(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale
  Copyright © 2017 Yann Régis-Gianas

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

Require Export Utf8.
Require Export List.
Export ListNotations.
Require Export Equalities.
Require Export Bool.

(* Option monad. It will be used to handle error cases. *)
Definition ret {A : Type} (x : A) := Some x.
Definition bind {A B : Type} (a : option A) (f : A → option B) :=
    match a with
    | None => None
    | Some x => f x
    end.

Notation "a >>= f" := (bind a f) (at level 40, left associativity).

(** Coplist: possibly infinite lists. *)
CoInductive coplist (A : Type) : Type :=
| CSingle : A -> coplist A
| CCons : A -> coplist A -> coplist A.

Definition coplist_id_force (A : Type) (l : coplist A)
  : coplist A :=
  match l with
  | CCons _ c l' => CCons _ c l'
  | CSingle _ c => CSingle _ c
  end.

Lemma coplist_id_force_eq:
  ∀ A l, coplist_id_force A l = l.
Proof.
  intros A l. destruct l ; reflexivity.
Qed.

Definition first {A : Type} (l : coplist A) : A :=
match l with
| CSingle _ x => x
| CCons _ x _ => x
end.

Definition next {A : Type} (l : coplist A) : option (coplist A) :=
match l with
| CSingle _ _ => None
| CCons _ _ xs => Some xs
end.

Fixpoint skipn {A : Type} (n : nat) (l : coplist A) :=
  match n with
  | 0 => Some l
  | S n' => next l >>= λ l',
            skipn n' l'
  end.

CoInductive coplist_in {A : Type} : A -> coplist A -> Prop :=
| CoplistInHead   : forall x l, coplist_in x (CCons _ x l)
| CoplistInSingle : forall x, coplist_in x (CSingle _ x)
| CoplistInTail   : forall x y l, coplist_in x l -> coplist_in x (CCons _ y l).

Fixpoint take {A} (n : nat) (l : coplist A) : list A * option (coplist A) :=
  match n with
    | O => (nil, Some l)
    | S k =>
      match next l with
      | None =>
        (first l :: nil, None)
      | Some l =>
        let r := take k l in
        (first l :: fst r, snd r)
      end
  end.

CoFixpoint map {A B : Type} f (s : coplist A) : coplist B :=
  match s with
    | CSingle _ x => CSingle _ (f x)
    | CCons _ x xs => CCons _ (f x) (map f xs)
  end.

(** Applying a partial function n times, or up to n times *)

Fixpoint napply {A : Type} (f : A → option A) (n : nat) (a : A) : option A :=
  match n with
  | 0 =>
    Some a
  | S n' =>
    f a >>= λ a',
    napply f n' a'
  end.

Fixpoint napply' {A : Type} (f : A → option A) (n : nat) (a : A) :
  A * nat :=
  match n with
  | 0 =>
    (a, 0)
  | S n' =>
    match f a with
    | Some a' =>
      napply' f n' a'
    | None =>
      (a, n)
    end
  end.

Property napply'_1:
  ∀ {A : Type} (f : A → option A) n a a' n',
    napply' f n a = (a', n') → n' ≤ n.
Proof.
  intros A f n.
  induction n as [ | n IH ] ; intros a a' n' H.
  - inversion H. auto.
  - simpl in H. destruct (f a) ; inversion H ; eauto.
Qed.

Property napply_napply'_Sn_iff:
  ∀ {A : Type} (f : A → option A) n a,
    napply f n a = None ↔
    ∃ n' a',
      napply' f n a = (a', S n') ∧ f a' = None.
Proof.
  intros A f n a. split.
  - revert a. induction n as [ | n IH ] ; intros a H.
    + simpl in H. discriminate H.
    + simpl in H.
      destruct (f a) as [a'' | ] eqn:Ha''.
      * simpl in H.
        simpl. rewrite Ha''. apply IH. auto.
      * simpl. rewrite Ha''. exists n, a. auto.
  - revert a. induction n as [ | n IH ] ; intros a [n' [a' [H₁ H₂]]].
    + simpl in H₁. discriminate H₁.
    + simpl in *.
      destruct (f a) as [a'' | ].
      * simpl. apply IH. exists n', a'. split ; auto.
      * reflexivity.
Qed.

Property napply'_prefix:
  ∀ {A : Type} (f : A → option A) n n' a a',
    napply' f n a = (a', n') →
    napply' f (n - n') a = (a', 0).
Proof.
  intros A f n n' a a' H.
  revert a a' n' H.
  induction n as [ | n IH ] ; intros a a' n' H.
  - simpl in H. inversion H. clear H. subst a' n'.
    simpl. auto.
  - simpl in H. destruct (f a) as [a'' | ] eqn:Ha''.
    + pose proof (napply'_1 _ _ _ _ _ H) as Hn'n.
      replace (S n - n') with (S (n - n')) ; auto with arith.
      simpl. rewrite Ha''. apply IH. auto.
    + inversion H. subst a' n'. simpl. replace (n - n) with 0 ; auto with arith.
Qed.

Property napply_napply'_0_iff:
  ∀ {A : Type} (f : A → option A) n a a',
    napply f n a = Some a' ↔ napply' f n a = (a', 0).
Proof.
  intros A f n a a'. split.
  - revert a a'. induction n as [ | n IH ] ; intros a a' H.
    + simpl in *. inversion H. auto.
    + simpl in *. destruct (f a) as [ai | ] ; try discriminate H.
      simpl in H. apply IH. auto.
  - revert a a'. induction n as [ | n IH ] ; intros a a' H.
    + simpl in *. inversion H. auto.
    + simpl in *. destruct (f a) as [ai | ] ; try discriminate H.
      simpl. apply IH. auto.
Qed.

Require Import Omega.

Property napply'_lt_id:
  ∀ {A : Type} (f : A → option A) n n' a a',
    napply' f n a = (a', n') → n - n' = 0 →
    a = a'.
Proof.
  intros A f [ | n ] n' a a' H H'.
  - simpl in H. inversion H. auto.
  - simpl in H.
    destruct (f a) as [a'' | ] eqn:Ha''.
    + pose proof (napply'_1 f n a'' a' n' H) as H''.
      exfalso. clear H Ha'' a a' a'' f A.
      omega.
    + inversion H. clear H. subst a' n'. auto.
Qed.