(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Yann Régis-Gianas

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

(** General definitions for difference languages.

    In this module, we introduce basic definitions about difference
    languages for deterministic programming languages.

    A difference language between two deterministic languages L1 and
    L2 is a record of type [difference_language L1 L2] made of a
    syntax for [difference]s and an [interpret]ation function.

    While the syntax for differences is left as abstract as possible,
    the semantics of differences uses a fixed (but general)
    interpretation domain : the relations between reduction traces.

    A difference term δ written in a [difference_language L1 L2] is an
    sound difference between P1 and P2 if its interpretation relates
    the reduction traces of P1 and P2.

*)

Require Import ProgrammingLanguage.
Require Import ReductionTrace.
Require Import ReductionTracesRelation.

(** A difference language is defined by a syntax for differences and a
    semantic that interprets differences as relations between
    configurations. *)
Record difference_language (L1 L2 : language) : Type :=
{
  (** The type for the terms of the difference language. *)
  difference : Type;

  (** The interpretation function. *)
  interpret : difference -> trace_relation L1 L2;
}.

Arguments difference {L1 L2} _.
Arguments interpret {L1 L2} _.

(** A difference δ is a sound difference between S and T, if the
    interpretation of δ relates the trace of S and the trace of T. *)
Definition sound_program_difference (L1 L2 : language) (D : difference_language L1 L2)
 (δ : difference D) (pL1 : prog L1) (pL2 : prog L2) : Prop :=
  interpret D δ (program_trace L1 pL1) (program_trace L2 pL2).


