(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
 *)

(**
  Big-step semantics of the Imp programming language, not used by the oracles
  nor the papers, but shows equivalence with the small-step semantics.
*)


Require Import Common.
Require Import Imp.Lang.
(* XXX: nsteps is defined there *)
Require Import ProgrammingLanguage.

(* TODO: For now, depends on Imp.Lang, but it should probably be the other way around? *)
Module Type ImpBigStep
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT).

  Reserved Notation "S ⊢ c ⇓ S'" (at level 50, no associativity).
  Inductive big_step : store → cmd → store → Prop :=
    | j_skip: ∀ {s},
      s ⊢ Skip ⇓ s
    | j_seq: ∀ {s₀ s₂ c₁ c₂} (s₁ : store),
      s₀ ⊢ c₁ ⇓ s₁ →
      s₁ ⊢ c₂ ⇓ s₂ →
      s₀ ⊢ (Seq c₁ c₂) ⇓ s₂
    | j_assign: ∀ {s s' x e} v,
      eval_exp s e = Some v →
      M.add x v s = s' →
      s ⊢ (Assign x e) ⇓ s'
    | j_if_then: ∀ {s s' b c₁ c₂},
      eval_bexp s b = Some true →
      s ⊢ c₁ ⇓ s' →
      s ⊢ ITE b c₁ c₂ ⇓ s'
    | j_if_else: ∀ {s s' b c₁ c₂},
      eval_bexp s b = Some false →
      s ⊢ c₂ ⇓ s' →
      s ⊢ ITE b c₁ c₂ ⇓ s'
    | j_while_iter: ∀ {s s' b c} s₁,
      eval_bexp s b = Some true →
      s ⊢ c ⇓ s₁ →
      s₁ ⊢ While b c ⇓ s' →
      s ⊢ While b c ⇓ s'
    | j_while_end: ∀ {s b c},
      eval_bexp s b = Some false →
      s ⊢ While b c ⇓ s
    | j_assert_true: ∀ {s b},
      eval_bexp s b = Some true →
      s ⊢ Assert b ⇓ s
  where "S ⊢ c ⇓ S'" := (big_step S c S').

  Hint Constructors big_step.

  Reserved Notation "S ⊢ c ⇒ '∞'" (at level 50, no associativity).
  Inductive big_step_div : store → cmd → Prop :=
    | j_seq₁: ∀ {s₀ c₁ c₂},
      s₀ ⊢ c₁ ⇒ ∞  →
      s₀ ⊢ (Seq c₁ c₂) ⇒ ∞
    | j_seq₂: ∀ {s₀ c₁ c₂} (s₁ : store),
      s₀ ⊢ c₁ ⇓ s₁ →
      s₁ ⊢ c₂ ⇒ ∞ →
      s₀ ⊢ (Seq c₁ c₂) ⇒ ∞
    | j_if_then_inf: ∀ {s b c₁ c₂},
      eval_bexp s b = Some true →
      s ⊢ c₁ ⇒ ∞ →
      s ⊢ ITE b c₁ c₂ ⇒ ∞
    | j_if_else_inf: ∀ {s b c₁ c₂},
      eval_bexp s b = Some false →
      s ⊢ c₂ ⇒ ∞ →
      s ⊢ ITE b c₁ c₂ ⇒ ∞
    | j_while₁: ∀ {s b c},
      eval_bexp s b = Some true →
      s ⊢ c ⇒ ∞ →
      s ⊢ While b c ⇒ ∞
    | j_while₂: ∀ {s b c} s₁,
      eval_bexp s b = Some true →
      s ⊢ c ⇓ s₁ →
      s₁ ⊢ While b c ⇒ ∞ →
      s ⊢ While b c ⇒ ∞
  where "S ⊢ c ⇒ '∞'" := (big_step_div S c).

  Hint Constructors big_step_div.

  Theorem big_implies_small:
    ∀ c s s',
    s ⊢ c ⇓ s' →
    ∃ n, @nsteps imp_language tt n ([c], s) = Some ([], s').
  Proof.
    intros c s s' Hbig.
    induction Hbig as [
                       s
                     | s s' c₁ c₂ s₁ Hc₁ IH₁ Hc₂ IH₂
                     | s s' x e v Hv Hs'
                     | s s' b c₁ c₂ Hb Hc₁ IH
                     | s s' b c₁ c₂ Hb Hc₂ IH
                     | s s' b c s₁ Hb Hc IH₁ H' IH₂
                     | s b c Hb
                     | s b Hb ].
    - exists 1. simpl. auto.
    - destruct IH₁ as (n₁, H₁), IH₂ as (n₂, H₂).
      exists (S (n₁ + n₂)). unfold nsteps. simpl.
      replace ([c₁ ; c₂]) with ([c₁] ++ [c₂]) ; auto.
      apply (nsteps_compose s₁) ; auto.
    - exists 1. unfold nsteps. simpl. rewrite Hv, <- Hs'. reflexivity.
    - destruct IH as (n, H'). exists (S n).
      unfold nsteps. simpl. rewrite Hb. simpl. apply H'.
    - destruct IH as (n, H'). exists (S n).
      unfold nsteps. simpl. rewrite Hb. simpl. apply H'.
    - destruct IH₁ as (n₁, H₁), IH₂ as (n₂, H₂).
      exists (S (n₁ + n₂)). unfold nsteps. simpl.
      rewrite Hb. simpl.
      replace ([c ; While b c]) with ([c] ++ [While b c]).
      + apply (nsteps_compose s₁) ; auto.
      + auto.
    - exists 1. unfold nsteps. simpl. rewrite Hb. reflexivity.
    - exists 1. unfold nsteps. simpl. rewrite Hb. reflexivity.
  Qed.

  Lemma seq_assoc:
    ∀ s c₁ c₂ c₃ s',
    s ⊢ Seq (Seq c₁ c₂) c₃ ⇓ s' ↔
    s ⊢ Seq c₁ (Seq c₂ c₃) ⇓ s'.
  Proof.
    intros s c₁ c₂ c₃ s'. split.
    + intro H. inversion H as [ | s_ s'_ c₁₂ c₃_ s₁ H₁₂ H₃ | | | | | | ]. subst.
      inversion H₁₂. eauto.
    + intro H. inversion H as [ | s_ s'_ c₁_ c₂₃ s₁ H₁ H₂₃ | | | | | | ]. subst.
      inversion H₂₃. eauto.
  Qed.

  Inductive big_step_cont (S : store) : store → cont → Prop :=
    | Halted :
      big_step_cont S S []
    | OneBigStep : ∀ s c s' k,
      s ⊢ c ⇓ s' →
      big_step_cont S s' k →
      big_step_cont S s (c::k).

  Inductive big_step_cont_div : store → cont → Prop :=
    | Diverges : ∀ s c k,
      s ⊢ c ⇒ ∞ →
      big_step_cont_div s (c::k)
    | OneBigStep' : ∀ s c s' k,
      s ⊢ c ⇓ s' →
      big_step_cont_div s' k →
      big_step_cont_div s (c::k).

  Hint Constructors big_step_cont.
  Hint Constructors big_step_cont_div.

  Theorem big_step_determinist:
    ∀ c s s' s'',
    s ⊢ c ⇓ s' →
    s ⊢ c ⇓ s'' →
    s' = s''.
  Proof.
    intros c s s' s'' H.
    revert s''.
    induction H as [
                    s
                  | s s' c₁ c₂ s₁ Hc₁ IH₁ Hc₂ IH₂
                  | s s' x e v Hv Hs'
                  | s s' b c₁ c₂ Hb Hc₁ IH₁
                  | s s' b c₁ c₂ Hb Hc₂ IH₂
                  | s s' b c s₁ Hb Hc IH₁ H' IH₂
                  | s b c Hb
                  | s b Hb ] ;
      intros s'' Hbig ;
      inversion Hbig as [
                         s_
                       | s_ s'_ c₁_ c₂_ s₁_ Hc₁' Hc₂'
                       | s_ s'_ x_ e_ v_ Hv' Hs''
                       | s_ s'_ b_ c₁_ c₂_ Hb' Hc₁'
                       | s_ s'_ b_ c₁_ c₂_ Hb' Hc₂'
                       | s_ s'_ b_ c_ s₁_ Hb' Hc' H''
                       | s_ b_ c_ Hb'
                       | s_ b_ Hb' ] ;
      subst s_ ; try subst c₁_ c₂_ ; try subst s'_ ; try subst b_ ;
        try subst s₁_ ; try subst s'' ; try subst c_ ; try congruence ; auto.
    - pose proof (IH₁ s₁_ Hc₁') ; auto. subst s₁_. auto.
    - pose proof (IH₁ s₁_ Hc') ; auto. subst s₁_. auto.
  Qed.

  Theorem big_div_implies_small_div_helper:
    ∀ c s,
    s ⊢ c ⇒ ∞ →
    ∃ c' k' s', step tt ([c], s) = Some (c'::k', s') ∧ big_step_cont_div s' (c'::k').
  Proof.
    intro c.
    destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ] ; intros s Hdiv.
    * inversion Hdiv.
    * exists c₁, [c₂], s. simpl. split ; auto. simpl.
      inversion Hdiv as [ s_ c₁_ c₂_ | s_ c₁_ c₂_ | | | | ] ; subst s_ c₁_ c₂_ ;
      eauto.
    * inversion Hdiv.
    * inversion Hdiv as [ | | s_ b_ c₁_ c₂_ Hb H₁ | s_ b_ c₁_ c₂_ Hb H₂ | | ] ;
      subst s_ b_ c₁_ c₂_.
      + exists c₁, [], s. simpl. rewrite Hb. simpl. split ; auto.
      + exists c₂, [], s. simpl. rewrite Hb. simpl. split ; auto.
    * inversion Hdiv as [ | | | | s_ b_ c_ Hb Hc | s_ b_ c_ s₁ Hb H₁ H ] ;
      subst s_ b_ c_ ; exists c, [While b c], s ; simpl ; rewrite Hb ;
      simpl ; split ; eauto.
    * inversion Hdiv.
  Qed.

  Theorem big_implies_small_helper:
    ∀ c s s'',
    s ⊢ c ⇓ s'' →
    ∃ k' s', step tt ([c], s) = Some (k', s') ∧ big_step_cont s'' s' k'.
  Proof.
    intro c.
    destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ] ; intros s s' Hbig ;
    inversion Hbig as [
        s_
      | s_ s''_ c₁_ c₂_ s₁ Hc₁ Hc₂
      | s_ s''_ x_ e_ v Hv Hs''
      | s_ s''_ b_ c₁_ c₂_ Hb Hc₁
      | s_ s''_ b_ c₁_ c₂_ Hb Hc₂
      | s_ s''_ b_ c_ s₁ Hb Hc Hw
      | s_ b_ c_ Hb
      | s_ b_ Hb ] ;
    subst s_ ; try subst s''_ ; try subst b_ ; try subst c₁_ c₂_ ; try subst s' ;
    try subst x_ e_ ; try subst c_.
    * exists [], s. split ; auto.
    * exists (c₁::[c₂]), s. split ; eauto.
    * exists [], (M.add x v s). split ; auto.
      simpl. rewrite Hv. simpl. auto.
    * exists [c₁], s. split ; eauto.
      simpl. rewrite Hb. simpl. auto.
    * exists [c₂], s. split ; eauto.
      simpl. rewrite Hb. simpl. auto.
    * exists [c; While b c], s. split ; eauto.
      simpl. rewrite Hb. simpl. auto.
    * exists [], s. split ; auto.
      simpl. rewrite Hb. simpl. auto.
    * exists [], s. split ; auto.
      simpl. rewrite Hb. simpl. auto.
  Qed.

  Theorem big_div_implies_small_div_helper':
    ∀ k s,
    big_step_cont_div s k →
    ∃ c' k' s', step tt (k, s) = Some (c'::k', s') ∧ big_step_cont_div s' (c'::k').
  Proof.
    intros k s Hdiv.
    induction Hdiv as [ s c k Hdiv | s c s' k Hs' Hdiv IH ].
    * destruct (big_div_implies_small_div_helper c s Hdiv) as [c' [k' [s' [H₁ H₂]]]].
      exists c', (k' ++ k), s'. split.
      + apply (step_app _ _ k) in H₁.
        apply H₁.
      + clear H₁. clear Hdiv. rewrite app_comm_cons.
        remember (c'::k') as K eqn:HK. clear HK.
        induction H₂.
        - simpl. apply Diverges. auto.
        - simpl. eauto.
    * pose proof (big_implies_small_helper c s s' Hs') as H'.
      destruct H' as [k' [s'' [H₁ H₂]]].
      destruct k' as [ | c' k' ].
      + inversion H₂. subst. clear H₂.
        destruct k as [ | c' k' ].
        - exfalso. destruct IH as [c' [k' [ s' [Hfalse _]]]]. discriminate Hfalse.
        - exists c', k', s''. split ; auto.
          apply (step_app _ _ (c'::k')) in H₁. apply H₁.
      + exists c', (k' ++ k), s''. split ; auto.
        - apply (step_app _ _ k) in H₁. apply H₁.
        - rewrite app_comm_cons. 
          remember (c'::k') as K eqn:HK. clear HK.
          clear IH Hs' c' H₁ k' s c.
          induction H₂.
          { simpl. auto. }
          { simpl. eauto. }
  Qed.

  Theorem big_cont_div_implies_small_div:
    ∀ k s,
    big_step_cont_div s k →
    infinite_small_step_exec (k, s).
  Proof.
    cofix H.
    intros k s Hbig.
    destruct (big_div_implies_small_div_helper' k s Hbig) as [c' [k' [s' [Hstep Hbig']]]].
    apply (InfStep _ _ Hstep) ; auto.
  Qed.

  Theorem big_div_implies_small_div:
    ∀ c s,
    s ⊢ c ⇒ ∞ →
    infinite_small_step_exec ([c], s).
  Proof.
    intros c s Hbig.
    apply big_cont_div_implies_small_div.
    apply Diverges. auto.
  Qed.

  Lemma small_implies_big_helper:
    ∀ S s',
    finite_small_step_exec s' S →
    big_step_cont s' (snd S) (fst S).
  Proof.
    intros S s' H.
    induction H as [ | [k s] [k₁ s₁] HS' H IH ] ;
    simpl in * ; eauto. destruct k ; eauto ; try discriminate HS'.
    destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ].
    + inversion_clear HS'. simpl. destruct k₁ ; eauto.
    + inversion HS'. clear HS'. subst k₁ s₁.
      inversion IH. subst.
      inversion H4. subst. eauto.
    + remember (eval_exp s e) as v eqn:Hv.
      destruct v as [ v | ] ; try discriminate HS'.
      simpl in HS'. inversion HS'. clear HS'. subst k₁ s₁.
      simpl. destruct k ; eauto.
    + remember (eval_bexp s b) as cond eqn:Hcond.
      destruct cond as [cond | ] ; try discriminate HS'.
      destruct cond ; simpl in * ; inversion HS' ; clear HS' ; subst k₁ s₁ ;
      inversion IH ; subst ; eauto.
    + remember (eval_bexp s b) as cond eqn:Hcond.
      destruct cond as [cond | ] ; try discriminate HS'.
      destruct cond ; simpl in * ; inversion HS' ; clear HS' ; subst k₁ s₁ ;
      inversion IH ; subst ; eauto. inversion H4. subst. eauto.
    + destruct (eval_bexp s b) as [[ | ] | ] eqn:Hcond ; try discriminate HS'.
      simpl in *. inversion HS'. clear HS'. subst k₁ s₁.
      inversion IH ; subst ; eauto.
  Qed.

  Lemma small_implies_big:
    ∀ c s s',
    (∃ n, @nsteps imp_language tt n ([c], s) = Some ([], s')) →
    s ⊢ c ⇓ s'.
  Proof.
    intros c s s' H.
    rewrite finite_small_step_exec_nsteps in H.
    pose proof (small_implies_big_helper _ _ H) as H'.
    simpl in H'.
    inversion H' as [ | s_ c_ s'_ k H₁ H₂ ].
    subst s_ c_ k. inversion H₂. subst s'_. auto.
  Qed.

  Theorem small_equiv_big:
    ∀ c s s',
    (∃ n, @nsteps imp_language tt n ([c], s) = Some ([], s')) ↔ s ⊢ c ⇓ s'.
  Proof.
    intros c s s'. split.
    + apply small_implies_big.
    + apply big_implies_small.
  Qed.
End ImpBigStep.

Module Make (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT)
  : ImpBigStep Var_as_DT Lang.
  Include (ImpBigStep Var_as_DT Lang).
End Make.