(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
 *)

(* Various definitions and lemmas on pairs of states with disjoint sets of variables *)

Require Import Common.
Require Import Imp.Lang.

Module Make
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT).

  Definition store_is_union (s s₁ s₂ : store) : Prop :=
    (∀ x v, M.MapsTo x v s₁ → M.MapsTo x v s)
    ∧ (∀ x v, M.MapsTo x v s₂ → M.MapsTo x v s)
    ∧ (∀ x v, M.MapsTo x v s → M.MapsTo x v s₁ ∨ M.MapsTo x v s₂).

  Lemma store_is_union_sym:
    ∀ s s₁ s₂,
      store_is_union s s₁ s₂ → store_is_union s s₂ s₁.
  Proof.
    intros s s₁ s₂ [H₁ [H₂ H₃]].
    repeat split ; auto.
    intros x v H. destruct (H₃ x v H) ; auto.
  Qed.

  Hint Resolve store_is_union_sym.

  Lemma equal_store_is_union:
    ∀ s₁ s₂ s s₁' s₂',
      M.Equal s₁' s₁ →
      M.Equal s₂' s₂ →
      store_is_union s s₁ s₂ →
      store_is_union s s₁' s₂'.
  Proof.
    intros s₁ s₂ s s₁' s₂' H₁ H₂ H.
    unfold store_is_union in *.
    destruct H as [Hs₁ [Hs₂ H']].
    rewrite MFacts.Equal_mapsto_iff in H₁, H₂.
    split ; [ | split ].
    - intros x v H. apply Hs₁, H₁, H.
    - intros x v H. apply Hs₂, H₂, H.
    - intros x v H. destruct (H' x v H) as [H'' | H''].
      + left. apply H₁ in H''. auto.
      + right. apply H₂ in H''. auto.
  Qed.

  Lemma store_is_union_add:
    ∀ s₁ s₂ s x v,
      store_is_union s s₁ s₂ →
      (¬ M.In x s₂) →
      store_is_union (M.add x v s) (M.add x v s₁) s₂.
  Proof.
    intros s₁ s₂ s x v [H₁ [H₂ H₃]] Hxs₂.
    repeat split ; auto.
    - intros y v' Hv'.
      rewrite MFacts.add_mapsto_iff in Hv'.
      destruct Hv' as [[Hy Hv'] | [Hy Hv']].
      + subst y v'. apply M.add_1 ; auto.
      + apply M.add_2 ; auto.
    - intros y v' Hv'.
      destruct (Var_as_DT.eq_dec x y) as [Hy | Hy].
      + subst y. exfalso. apply Hxs₂. exists v'. apply Hv'.
      + apply M.add_2 ; auto.
    - intros y v' Hv'.
      rewrite MFacts.add_mapsto_iff in Hv'.
      destruct Hv' as [[Hy Hv'] | [Hy Hv']].
      + subst y v'. left. apply M.add_1 ; auto.
      + destruct (H₃ y v' Hv') ; auto.
        left. apply M.add_2 ; auto.
  Qed.

  Hint Resolve store_is_union_add.

  Definition In_var_cont (x : varid) (k : cont) : Prop :=
    ∃ c, In c k ∧ In_var x c.

  Definition cont_classified (k : cont) (V : varid → Prop) : Prop :=
    ∀ x, In_var_cont x k → V x.

  Definition store_classified (s : store) (V : varid → Prop) : Prop :=
    ∀ x, M.In x s → V x.

  Property cont_classified_1:
    ∀ c k V,
    cont_classified (c::k) V →
    ∀ x, In_var x c → V x.
  Proof.
    intros c k V H x Hx.
    apply (H x).
    exists c. simpl. auto.
  Qed.

  Property cont_classified_2:
    ∀ c k V,
      cont_classified (c::k) V →
      cont_classified k V.
  Proof.
    intros c k V H x Hx.
    apply H. destruct Hx as [c' Hc']. exists c'. destruct Hc'. split ; auto.
    right. auto.
  Qed.

  Definition disjoint_states (S₁ S₂ : state) : Prop :=
    let '(k₁, s₁) := S₁ in
    let '(k₂, s₂) := S₂ in
    ∃ (V₁ V₂ : varid → Prop),
      cont_classified k₁ V₁
      ∧ cont_classified k₂ V₂
      ∧ store_classified s₁ V₁
      ∧ store_classified s₂ V₂
      ∧ (∀ x, V₁ x → V₂ x → Logic.False).

  Lemma disjoint_states_sym:
    ∀ S₁ S₂,
      disjoint_states S₁ S₂ → disjoint_states S₂ S₁.
  Proof.
    intros [k₁ s₁] [k₂ s₂] H.
    destruct H as [V₁ [V₂ [H₁ [H₂ [H₃ [H₄ H₅]]]]]].
    exists V₂, V₁.
    repeat split ; auto.
    intros x HV₂ HV₁. apply (H₅ x HV₁ HV₂).
  Qed.

  Hint Resolve disjoint_states_sym.

  Lemma step_disjoint_states:
    ∀ S₁ S₁' S₂,
      disjoint_states S₁ S₂ →
      step tt S₁ = Some S₁' →
      disjoint_states S₁' S₂.
  Proof.
    intros [k₁ s₁] [k₁' s₁'] [k₂ s₂].
    intros H₁ H₂.
    destruct H₁ as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ H₁]]]]]].
    destruct k₁ as [ | c k₁ ] ;
      try discriminate H₂.
    simpl in H₂.
    destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ].
    - inversion H₂. subst s₁' k₁'.
      exists V₁, V₂.
      repeat split ; eauto using cont_classified_2.
    - inversion H₂. clear H₂. subst k₁' s₁'.
      exists V₁, V₂.
      repeat split ; auto.
      intros x Hx. apply Hk₁.
      destruct Hx as [c₀ [H₁' H₂']].
      simpl in H₁'.
      destruct H₁' as [H₁' | [H₁' | H₁']].
      + subst c₀. exists (Seq c₁ c₂). split ; simpl ; auto.
      + subst c₀. exists (Seq c₁ c₂). split ; simpl ; auto.
      + exists c₀. split ; simpl ; auto.
    - destruct (eval_exp s₁ e) as [v | ] eqn:Hv ;
        try discriminate H₂.
      simpl in H₂. inversion H₂. clear H₂. subst s₁' k₁'.
      exists V₁, V₂.
      repeat split ; eauto using cont_classified_2.
      intros x' Hx'.
      rewrite MFacts.add_in_iff in Hx'.
      destruct Hx' as [Hx' | Hx'].
      * subst x'. apply Hk₁.
        exists (Assign x e). split ; simpl ; auto.
      * apply Hs₁. auto.
    - destruct (eval_bexp s₁ b) as [v | ] ;
        try discriminate H₂.
      simpl in H₂. inversion H₂. clear H₂.
      subst s₁' k₁'.
      exists V₁, V₂.
      repeat split ; auto.
      intros x Hx.
      apply Hk₁.
      destruct Hx as [c [H₁' H₂']].
      simpl in H₁'.
      destruct H₁' as [H₁' | H₁'].
      + exists (ITE b c₁ c₂). split ; simpl ; auto.
        destruct v ; subst c ; auto.
      + exists c. split ; simpl ; auto.
    - destruct (eval_bexp s₁ b) as [v | ] ;
        try discriminate H₂.
      destruct v.
      + simpl in H₂. inversion H₂. clear H₂.
        subst s₁' k₁'.
        exists V₁, V₂.
        repeat split ; auto.
        intros x Hx.
        apply Hk₁.
        destruct Hx as [c' [H₁' H₂']].
        simpl in H₁'.
        destruct H₁' as [H₁' | [H₁' | H₁']].
        * subst c'.
          exists (While b c). split ; simpl ; auto.
        * subst c'.
          exists (While b c). split ; simpl ; auto.
        * exists c'. split ; simpl ; auto.
      + simpl in H₂. inversion H₂. clear H₂.
        subst s₁' k₁'.
        exists V₁, V₂.
        repeat split ; eauto using cont_classified_2.
    - destruct (eval_bexp s₁ b) as [[ | ] | ] ;
        try discriminate H₂.
      inversion H₂. clear H₂. subst k₁' s₁'.
      exists V₁, V₂.
      repeat split ; eauto using cont_classified_2.
  Qed.

  Lemma step_2_disjoint_states:
    ∀ S₁ S₁' S₂ S₂',
      disjoint_states S₁ S₂ →
      step tt S₁ = Some S₁' →
      step tt S₂ = Some S₂' →
      disjoint_states S₁' S₂'.
  Proof.
    intros S₁ S₁' S₂ S₂' Hdisj H₁ H₂.
    pose proof (step_disjoint_states S₁ S₁' S₂ Hdisj H₁) as Hdisj'.
    eauto using step_disjoint_states.
  Qed.

  Lemma disjoint_eval_exp:
    ∀ (e : exp) (s : store) (V₁ V₂ : varid → Prop) (x : varid) v,
      (∀ y, In_exp_var y e → V₁ y) →
      V₂ x →
      (∀ x, V₁ x → V₂ x → Logic.False) →
      eval_exp (M.add x v s) e = eval_exp s e.
  Proof.
    intros e s V₁ V₂ x v He Hx Hdisj.
    induction e as [
                    y
                  | n
                  | e₁ IH₁ e₂ IH₂
                  | e₁ IH₁ e₂ IH₂
                  | e₁ IH₁ e₂ IH₂
                  | e₁ IH₁ e₂ IH₂
                  | e₁ IH₁ e₂ IH₂ ] ;
      simpl ; auto.
    - rewrite MFacts.add_neq_o ; auto. intro Hfalse.
      subst y. apply (Hdisj x) ; auto. apply He. simpl. auto.
    - rewrite IH₁, IH₂ ; auto ;
        intros z Hz ; apply He ; simpl ; auto.
    - rewrite IH₁, IH₂ ; auto ;
        intros z Hz ; apply He ; simpl ; auto.
    - rewrite IH₁, IH₂ ; auto ;
        intros z Hz ; apply He ; simpl ; auto.
    - rewrite IH₁, IH₂ ; auto ;
        intros z Hz ; apply He ; simpl ; auto.
    - rewrite IH₁, IH₂ ; auto ;
        intros z Hz ; apply He ; simpl ; auto.
  Qed.

  Lemma disjoint_eval_bexp:
    ∀ (b : bexp) (s : store) (V₁ V₂ : varid → Prop) (x : varid) v,
      (∀ y, In_bexp_var y b → V₁ y) →
      V₂ x →
      (∀ x, V₁ x → V₂ x → Logic.False) →
      eval_bexp (M.add x v s) b = eval_bexp s b.
  Proof.
    intros b s V₁ V₂ x v Hb Hx Hdisj.
    induction b as [ |
                     | e₁ e₂ | e₁ e₂ | e₁ e₂
                     | b₁ IH₁ b₂ IH₂
                     | b₁ IH₁ b₂ IH₂
                     | b₁ IH₁ b₂ IH₂
                     | b IH ] ;
      simpl ; auto.
    - rewrite (disjoint_eval_exp e₁ s V₁ V₂),
      (disjoint_eval_exp e₂ s V₁ V₂) ; auto ;
        intros z Hz ; apply Hb ; simpl ; auto.
    - rewrite (disjoint_eval_exp e₁ s V₁ V₂),
      (disjoint_eval_exp e₂ s V₁ V₂) ; auto ;
        intros z Hz ; apply Hb ; simpl ; auto.
    - rewrite (disjoint_eval_exp e₁ s V₁ V₂),
      (disjoint_eval_exp e₂ s V₁ V₂) ; auto ;
        intros z Hz ; apply Hb ; simpl ; auto.
    - rewrite IH₁, IH₂ ; auto ;
        intros z Hz ; apply Hb ; simpl ; auto.
    - rewrite IH₁, IH₂ ; auto ;
        intros z Hz ; apply Hb ; simpl ; auto.
    - rewrite IH₁, IH₂ ; auto ;
        intros z Hz ; apply Hb ; simpl ; auto.
    - rewrite IH ; auto ;
        intros z Hz ; apply Hb ; simpl ; auto.
  Qed.

  Lemma disjoint_union_eval_exp_same:
    ∀ (e : exp) (s₁ s₂ s : store) (V₁ V₂ : varid → Prop),
      store_is_union s s₁ s₂ →
      (∀ x, In_exp_var x e → V₁ x) →
      (∀ x, M.In x s₁ → V₁ x) →
      (∀ y, M.In y s₂ → V₂ y) →
      (∀ x, V₁ x → V₂ x → Logic.False) →
      eval_exp s₁ e = eval_exp s e.
  Proof.
    intros e s₁ s₂ s V₁ V₂ [H₁ [H₂ H₃]] He Hs₁ Hs₂ Hdisj.
    induction e as [
                    x
                  | n
                  | e₁ IH₁ e₂ IH₂
                  | e₁ IH₁ e₂ IH₂
                  | e₁ IH₁ e₂ IH₂
                  | e₁ IH₁ e₂ IH₂
                  | e₁ IH₁ e₂ IH₂ ].
    - simpl.
      destruct (M.find x s₁) as [v₁ | ] eqn:Hv₁.
      + rewrite <- MFacts.find_mapsto_iff in Hv₁.
        apply H₁ in Hv₁.
        rewrite MFacts.find_mapsto_iff in Hv₁. rewrite Hv₁.
        auto.
      + destruct (M.find x s) as [v₁' | ] eqn:Hv₁' ; auto.
        exfalso.
        rewrite <- MFacts.find_mapsto_iff in Hv₁'.
        destruct (H₃ x v₁' Hv₁') as [H | H].
        * rewrite MFacts.find_mapsto_iff in H.
          rewrite H in Hv₁. discriminate Hv₁.
        * apply (Hdisj x) ; auto.
          { apply He. simpl. auto. }
          { apply Hs₂. exists v₁'. auto. }
    - simpl. auto.
    - simpl. rewrite IH₁, IH₂ ; auto.
      + intros x Hx. apply He. simpl. auto.
      + intros x Hx. apply He. simpl. auto.
    - simpl. rewrite IH₁, IH₂ ; auto.
      + intros x Hx. apply He. simpl. auto.
      + intros x Hx. apply He. simpl. auto.
    - simpl. rewrite IH₁, IH₂ ; auto.
      + intros x Hx. apply He. simpl. auto.
      + intros x Hx. apply He. simpl. auto.
    - simpl. rewrite IH₁, IH₂ ; auto.
      + intros x Hx. apply He. simpl. auto.
      + intros x Hx. apply He. simpl. auto.
    - simpl. rewrite IH₁, IH₂ ; auto.
      + intros x Hx. apply He. simpl. auto.
      + intros x Hx. apply He. simpl. auto.
  Qed.

  Lemma disjoint_union_eval_bexp_same:
    ∀ (b : bexp) (s₁ s₂ s : store) (V₁ V₂ : varid → Prop),
      store_is_union s s₁ s₂ →
      (∀ x, In_bexp_var x b → V₁ x) →
      (∀ x, M.In x s₁ → V₁ x) →
      (∀ y, M.In y s₂ → V₂ y) →
      (∀ x, V₁ x → V₂ x → Logic.False) →
      eval_bexp s₁ b = eval_bexp s b.
  Proof.
    intros b s₁ s₂ s V₁ V₂ H₁ He Hs₁ Hs₂ Hdisj.
    induction b as [ |
                     | e₁ e₂ | e₁ e₂ | e₁ e₂
                     | b₁ IH₁ b₂ IH₂
                     | b₁ IH₁ b₂ IH₂
                     | b₁ IH₁ b₂ IH₂
                     | b IH ] ;
      simpl ; auto.
    - rewrite (disjoint_union_eval_exp_same e₁ s₁ s₂ s V₁ V₂),
      (disjoint_union_eval_exp_same e₂ s₁ s₂ s V₁ V₂) ;
        auto ; intros x H ; apply He ; simpl ; auto.
    - rewrite (disjoint_union_eval_exp_same e₁ s₁ s₂ s V₁ V₂),
      (disjoint_union_eval_exp_same e₂ s₁ s₂ s V₁ V₂) ;
        auto ; intros x H ; apply He ; simpl ; auto.
    - rewrite (disjoint_union_eval_exp_same e₁ s₁ s₂ s V₁ V₂),
      (disjoint_union_eval_exp_same e₂ s₁ s₂ s V₁ V₂) ;
        auto ; intros x H ; apply He ; simpl ; auto.
    - rewrite IH₁, IH₂ ;
        auto ; intros x H ; apply He ; simpl ; auto.
    - rewrite IH₁, IH₂ ;
        auto ; intros x H ; apply He ; simpl ; auto.
    - rewrite IH₁, IH₂ ;
        auto ; intros x H ; apply He ; simpl ; auto.
    - rewrite IH ;
        auto ; intros x H ; apply He ; simpl ; auto.
  Qed.

  Lemma step_disjoint_states_union:
    ∀ S₁ S₁' s s' k' S₂,
      disjoint_states S₁ S₂ →
      store_is_union s (snd S₁) (snd S₂) →
      step tt S₁ = Some S₁' →
      step tt (fst S₁, s) = Some (k', s') →
      k' = fst S₁' ∧ store_is_union s' (snd S₁') (snd S₂). 
  Proof.
    intros [k₁ s₁] [k₁' s₁'] s s' k' [k₂ s₂] Hdisj Hs HS₁' Hk's'.
    simpl fst in *. simpl snd in *.
    destruct k₁ as [ | c k₁ ] ; try discriminate HS₁'.
    destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ].
    - simpl in HS₁'. inversion HS₁'. clear HS₁'. subst k₁' s₁'.
      simpl in Hk's'. inversion Hk's'. subst k' s'.
      split ; auto.
    - simpl in HS₁'. inversion HS₁'. clear HS₁'. subst k₁' s₁'.
      simpl in Hk's'. inversion Hk's'. clear Hk's'. subst k' s'.
      split ; auto.
    - simpl in HS₁', Hk's'.
      destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
      rewrite <- (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂) in Hk's' ;
        auto.
      + destruct (eval_exp s₁ e) as [v₁ | ] eqn:Hv₁ ;
          try discriminate HS₁'.
        inversion HS₁'. clear HS₁'. subst k₁' s₁'.
        inversion Hk's'. clear Hk's'. subst k' s'.
        split ; auto.
        apply store_is_union_add ; auto.
        intro HFalse. apply (Hdisj x) ; auto.
        apply Hk₁. exists (Assign x e). simpl. auto.
      + intros y Hy. apply Hk₁. exists (Assign x e). simpl. auto.
    - simpl in HS₁', Hk's'.
      destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
      rewrite <- (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) in Hk's' ;
        auto.
      + destruct (eval_bexp s₁ b) as [v₁ | ] eqn:Hv₁ ;
          try discriminate HS₁'.
        simpl in HS₁'.
        inversion HS₁'. clear HS₁'. subst k₁' s₁'.
        inversion Hk's'. clear Hk's'. subst k' s'.
        split ; auto.
      + intros x Hx. apply Hk₁. exists (ITE b c₁ c₂). simpl. auto.
    - simpl in HS₁', Hk's'.
      destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
      rewrite <- (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) in Hk's' ;
        auto.
      + destruct (eval_bexp s₁ b) as [v₁ | ] eqn:Hv₁ ;
          try discriminate HS₁'.
        simpl in HS₁'.
        destruct v₁ ;
          inversion HS₁' ; clear HS₁' ; subst k₁' s₁' ;
          inversion Hk's' ; clear Hk's' ; subst k' s' ;
            split ; auto.
      + intros x Hx. apply Hk₁. exists (While b c). simpl. auto.
    - simpl in HS₁', Hk's'.
      destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
      rewrite <- (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) in Hk's' ;
        auto.
      + destruct (eval_bexp s₁ b) as [[ | ] | ] eqn:Hv₁ ;
          try discriminate HS₁'.
        inversion HS₁'. clear HS₁'. subst k₁' s₁'.
        inversion Hk's'. clear Hk's'. subst k' s'.
        split ; auto.
      + intros x Hx. apply Hk₁. exists (Assert b). simpl. auto.
  Qed.

  Lemma step_disjoint_states_union_same_crashes:
    ∀ S₁ s S₂,
      disjoint_states S₁ S₂ →
      store_is_union s (snd S₁) (snd S₂) →
      step tt S₁ = None <-> step tt (fst S₁, s) = None.
  Proof.
    intros [k₁ s₁] s [k₂ s₂] Hdisj Hs.
    simpl fst in *. simpl snd in *.
    split ; intro H.
    - destruct k₁ as [ | c k₁ ] ; auto.
      destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ] ;
        try discriminate H ; simpl in H.
      + simpl.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
        rewrite <- (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂) ; auto.
        * destruct (eval_exp s₁ e) ; auto ; discriminate H.
        * intros y Hy. apply Hk₁. exists (Assign x e). simpl. auto.
      + simpl.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
        rewrite <- (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) ; auto.
        * destruct (eval_bexp s₁ b) ; auto ; discriminate H.
        * intros y Hy. apply Hk₁. exists (ITE b c₁ c₂). simpl. auto.
      + simpl.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
        rewrite <- (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) ; auto.
        * destruct (eval_bexp s₁ b) ; auto ; discriminate H.
        * intros y Hy. apply Hk₁. exists (While b c). simpl. auto.
      + simpl.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
        rewrite <- (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) ; auto.
        * destruct (eval_bexp s₁ b) as [[ | ] | ] ; auto ; discriminate H.
        * intros y Hy. apply Hk₁. exists (Assert b). simpl. auto.
    - destruct k₁ as [ | c k₁ ] ; auto.
      destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ] ;
        try discriminate H ; simpl in H.
      + simpl.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
        rewrite (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂) ; auto.
        * destruct (eval_exp s e) ; auto ; discriminate H.
        * intros y Hy. apply Hk₁. exists (Assign x e). simpl. auto.
      + simpl.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
        rewrite (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) ; auto.
        * destruct (eval_bexp s b) ; auto ; discriminate H.
        * intros y Hy. apply Hk₁. exists (ITE b c₁ c₂). simpl. auto.
      + simpl.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
        rewrite (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) ; auto.
        * destruct (eval_bexp s b) ; auto ; discriminate H.
        * intros y Hy. apply Hk₁. exists (While b c). simpl. auto.
      + simpl.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [H's₁ [H's₂ Hdisj]]]]]].
        rewrite (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) ; auto.
        * destruct (eval_bexp s b) as [[ | ] | ] ; auto ; discriminate H.
        * intros y Hy. apply Hk₁. exists (Assert b). simpl. auto.
  Qed.

  Lemma step_disjoint_union_states:
    ∀ S₁ S₁' s s' k' S₂,
      disjoint_states S₁ S₂ →
      store_is_union s (snd S₁) (snd S₂) →
      step tt S₁ = Some S₁' →
      step tt (fst S₁, s) = Some (k', s') →
      k' = fst S₁' ∧ store_is_union s' (snd S₁') (snd S₂)
      ∧ disjoint_states S₁' S₂.
  Proof.
    intros S₁ S₁' s s' k' S₂ Hdisj Hunion HS₁' HS₂'.
    destruct (step_disjoint_states_union S₁ S₁' s s' k' S₂ Hdisj Hunion HS₁' HS₂')
      as [H₁ H₂].
    split ; auto. split ; auto.
    eauto using step_disjoint_states.
  Qed.
End Make.