(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

(** Partial correctness Hoare logic on Imp *)

Require Import Common.
Require Imp.SmallStep.
Require Imp.BigStep.
Require Imp.Lang.

Module Type ImpHoare (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : Lang.ImpProgrammingLanguage Var_as_DT)
       (Import BigStep : BigStep.ImpBigStep Var_as_DT Lang).

  (* Regular Hoare triplets *)
  Definition assertion : Type := store → Prop.

  Definition assertion_subst x e (P : assertion) : assertion :=
    λ (S : store), ∃ v,
    eval_exp S e = Some v
    ∧ P (M.add x v S).
  Notation "P [ x ↦ e ]" := (assertion_subst x e P) (at level 10).

  Definition assertion_of_bexp (b : bexp) : assertion :=
    λ S, eval_bexp S b = Some true.

  Definition assertion_implies (P Q : assertion) : Prop :=
    ∀ S, P S → Q S.

  Notation "P ⇒ Q" := (assertion_implies P Q) (at level 80).

  Property assertion_implies_refl:
    ∀ P, P ⇒ P.
  Proof.
    intro P.
    intros S H. auto.
  Qed.

  Hint Resolve assertion_implies_refl.

  Property assertion_implies_trans:
    ∀ P P' P'', P ⇒ P' → P' ⇒ P'' → P ⇒ P''.
  Proof.
    intros P P' P'' H₁ H₂.
    intros S H. auto.
  Qed.

  Hint Resolve assertion_implies_trans.

  Inductive hoare_proof : assertion → cmd → assertion → Type :=
  | HoareSkip : ∀ P,
    hoare_proof P Skip P
  | HoareAssign : ∀ P x e,
    hoare_proof (P [x ↦ e]) (Assign x e) P
  | HoareAssert : ∀ P b,
    P ⇒ assertion_of_bexp b →
    hoare_proof P (Assert b) P
  | HoareSeq : ∀ P Q R c₁ c₂,
    hoare_proof P c₁ Q →
    hoare_proof Q c₂ R →
    hoare_proof P (Seq c₁ c₂) R
  | HoareITE : ∀ P Q b c₁ c₂,
    hoare_proof (λ S, P S ∧ assertion_of_bexp b S) c₁ Q →
    hoare_proof (λ S, P S ∧ assertion_of_bexp (NOT b) S) c₂ Q →
    P ⇒ (λ S, eval_bexp S b ≠ None) → (* No crash! *)
    hoare_proof P (ITE b c₁ c₂) Q
  | HoareWhile : ∀ P b c,
    hoare_proof (λ S, P S ∧ assertion_of_bexp b S) c P →
    P ⇒ (λ S, eval_bexp S b ≠ None) → (* No crash! *)
    hoare_proof P (While b c) (λ S, P S ∧ assertion_of_bexp (NOT b) S)
  | HoareSub: ∀ P P' Q Q' c,
    hoare_proof P c Q →
    P' ⇒ P →
    Q ⇒ Q' →
    hoare_proof P' c Q'.

  Lemma hoare_soundness:
    ∀ P Q c, hoare_proof P c Q →
    ∀ S S', P S → S ⊢ c ⇓ S' → Q S'.
  Proof.
    intros P Q c H.
    induction H as [
                    P
                  | P x e
                  | P b HPb
                  | P Q R c₁ c₂ H₁ IH₁ H₂ IH₂
                  | P Q b c₁ c₂ H₁ IH₁ H₂ IH₂
                  | P b c H IH
                  | P P' Q Q' c H IH HP'P HQQ' ] ;
      intros S S' HP Hc.
    - inversion Hc. subst. auto.
    - inversion Hc as [ | | s_ s' x_ e_ v Hv Hs' | | | | | ].
      subst s_ x_ e_ s'. subst S'.
      destruct HP as [v' [Hv' HP]].
      rewrite Hv in Hv'. inversion Hv'. clear Hv'. subst v'.
      apply HP.
    - inversion Hc. subst. auto.
    - inversion Hc. subst. eauto.
    - inversion Hc
        as [ | |
             | S_ S'_ b_ c₁_ c₂_ Hb H IH
             | S_ S'_ b_ c₁_ c₂_ Hb H IH
             | | | ] ; subst.
      + apply (IH₁ S) ; auto.
      + apply (IH₂ S) ; auto. split ; auto.
        unfold assertion_of_bexp. simpl. rewrite Hb.
        auto.
    - remember (While b c) as c₀ eqn:Hc₀.
      induction Hc as [ | | | |
                        | S S' b_ c_ S₁ Hb H₁ IH₁ H₂ IH₂
                        | S b_ c_ Hb | ] ;
        try discriminate Hc₀ ; inversion Hc₀ ; clear Hc₀ ; subst b_ c_.
      + apply IH₂ ; auto. apply (IH S) ; auto.
      + split ; auto. unfold assertion_of_bexp.
        simpl. rewrite Hb. auto.
    - apply HQQ'. apply (IH S) ; auto.
  Qed.
End ImpHoare.

Module Make (Var_as_DT : UsualDecidableTypeOrig)
       (Lang : Lang.ImpProgrammingLanguage Var_as_DT)
       (BigStep : BigStep.ImpBigStep Var_as_DT Lang)
  : ImpHoare Var_as_DT Lang BigStep.

  Include ImpHoare Var_as_DT Lang BigStep.
End Make.
