(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

Require Import Common.
Require Import Imp.SmallStep.
Require Import Imp.Syntax.

Require ProgrammingLanguage.

Module Type ImpProgrammingLanguage (Var_as_DT : UsualDecidableTypeOrig).
  Module SmallStep := SmallStep.Imp Var_as_DT.
  Export SmallStep.

  Property genv_eq_dec:
    ∀ (x y : unit), { x = y } + { x ≠ y }.
  Proof.
    intros [ ] [ ]. left. reflexivity.
  Qed.

  Lemma state_final_no_step:
    ∀ genv S v,
      state_final S = Some v →
      step genv S = None.
  Proof.
    intros genv [k₁ s₁] v H.
    simpl in H. destruct k₁ ; try discriminate H.
    reflexivity.
  Qed.

  Definition imp_language : ProgrammingLanguage.language :=
    {|
      ProgrammingLanguage.return_type := unit;
      ProgrammingLanguage.genv_type := unit;
      ProgrammingLanguage.state := state;
      ProgrammingLanguage.step := step;
      ProgrammingLanguage.state_final := state_final;
      ProgrammingLanguage.state_final_no_step := state_final_no_step;

      (* TODO: maybe change that *)
      ProgrammingLanguage.prog := cmd;
      ProgrammingLanguage.initial_state := λ c, (tt, (c::nil, M.empty Z));

      ProgrammingLanguage.genv_eq := eq;
      ProgrammingLanguage.genv_eq_refl := @eq_refl unit;
      ProgrammingLanguage.genv_eq_sym := @eq_sym unit;
      ProgrammingLanguage.genv_eq_trans := @eq_trans unit;
      ProgrammingLanguage.genv_eq_dec := genv_eq_dec;

      ProgrammingLanguage.state_eq := state_eq;
      ProgrammingLanguage.state_eq_dec := state_eq_dec;
      ProgrammingLanguage.state_eq_sym := state_eq_sym;
      ProgrammingLanguage.state_eq_refl := state_eq_refl;
      ProgrammingLanguage.state_eq_trans := state_eq_trans;

      ProgrammingLanguage.state_final_eq := state_final_eq;
      ProgrammingLanguage.step_eq := step_eq;
      ProgrammingLanguage.step_eq' := step_eq';
    |}.

  Definition υ : ProgrammingLanguage.genv_type imp_language := tt.

  (* Additional theorems *)

  Inductive finite_small_step_exec s : state → Prop :=
    | Halted : finite_small_step_exec s ([], s)
    | Step : ∀ S S',
      step υ S = Some S' →
      finite_small_step_exec s S' →
      finite_small_step_exec s S.

  Hint Constructors finite_small_step_exec.

  CoInductive infinite_small_step_exec : state → Prop :=
    | InfStep : ∀ S S',
      step tt S = Some S' →
      infinite_small_step_exec S' →
      infinite_small_step_exec S.

  Lemma finite_small_step_exec_nsteps:
    ∀ S s',
      (∃ n, ProgrammingLanguage.nsteps υ n S = Some ([], s'))
        ↔ finite_small_step_exec s' S.
  Proof.
    unfold ProgrammingLanguage.nsteps.
    intros S s' ; split.
    - intros (n, H). revert S s' H.
      induction n as [ | n IH ] ; intros (k, s) s' H.
      + inversion H. subst. auto.
      + unfold napply in *.
        destruct (ProgrammingLanguage.step imp_language _ (k, s)) as [ S' | ] eqn:H'.
        * apply (Step _ _ S') ; auto.
        * simpl in H. discriminate H.
    - intro H. induction H ; try (exists 0 ; tauto) ;
      destruct IHfinite_small_step_exec as (n, IH) ;
      exists (Datatypes.S n) ; simpl ; auto ;
      rewrite H ; simpl ; auto.
  Qed.

  Import ProgrammingLanguage.

  Lemma nsteps_imp_app_1:
    ∀ n s s' k₁ k₂,
      nsteps υ n (k₁, s) = Some ([], s') →
      nsteps υ n (k₁ ++ k₂, s) = Some (k₂, s').
  Proof.
    intro n.
    unfold nsteps.
    induction n as [ | n IH ] ;
      intros s s' k₁ k₂ H.
    - inversion H. clear H. subst k₁ s'.
      reflexivity.
    - simpl in *.
      destruct k₁ as [ | c₁ k₁ ] ;
        try discriminate H.
      destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ] ;
        simpl in * ; auto.
      + apply (IH _ _ (c₁₁::c₁₂::k₁)). auto.
      + destruct (eval_exp s e₁) as [v₁ | ] ;
          try discriminate H.
        simpl in *. auto.
      + destruct (eval_bexp s b₁) as [[ | ] | ] ;
          try discriminate H ;
          simpl in *.
        * apply (IH _ _ (c₁₁::k₁)) ; auto.
        * apply (IH _ _ (c₁₂::k₁)) ; auto.
      + destruct (eval_bexp s b₁) as [[ | ] | ] ;
          try discriminate H ;
          simpl ; auto.
        apply (IH _ _ (c₁ :: While b₁ c₁ :: k₁)) ; auto.
      + destruct (eval_bexp s b₁) as [[ | ] | ] ;
          try discriminate H ;
          simpl ; auto.
  Qed.

  Lemma nsteps_compose:
    ∀ {n₁ n₂ k₁ k₂ k₃ s₀ s₂} s₁,
    nsteps υ n₁ (k₁, s₀) = Some ([], s₁) →
    nsteps υ n₂ (k₂, s₁) = Some (k₃, s₂) →
    nsteps υ (n₁ + n₂) (k₁ ++ k₂, s₀) = Some (k₃, s₂).
  Proof.
    intros n₁ n₂ k₁ k₂ k₃ s₀ s₂ s₁ H₁ H₂.
    rewrite <- H₂.
    apply nsteps_plus_1.
    apply nsteps_imp_app_1. auto.
  Qed.
End ImpProgrammingLanguage.

Module Make (Var_as_DT : UsualDecidableTypeOrig) : (ImpProgrammingLanguage Var_as_DT).
  Include ImpProgrammingLanguage Var_as_DT.
End Make.