(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

Require Import Common.
Require Import Imp.Lang.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.
Require Import OracleHelpers.

Module Make
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT)
       (Import OracleHelpers : ImpOracleHelpers Var_as_DT Lang).

  Inductive either (A B : Type) :=
  | Left : A → either A B
  | Right : B → either A B.
  Arguments Left {A B} a.
  Arguments Right {A B} b.

  Record mod := {
    c_left :
      cmd;
    c_right :
      cmd;
    left_fun :
      store → either store nat;
    right_fun :
      store → either store nat;
    left_equiv_right:
      ∀ s₁ s₂,
        M.Equal s₁ s₂ →
        (∃ b₁ b₂, left_fun s₁ = Right b₁ ∧ right_fun s₂ = Right b₂)
        ∨ (∃ s₁' s₂', left_fun s₁ = Left s₁' ∧ right_fun s₂ = Left s₂'
                      ∧ M.Equal s₁' s₂');
    left_fun_spec_1 :
      ∀ s s', left_fun s = Left s' →
              ∃ n, nsteps υ n ([c_left], s) = Some ([], s');
    left_fun_spec_2 :
      ∀ s n, left_fun s = Right n →
             ∀ k, nsteps υ n (c_left::k, s) = None;
    right_fun_spec_1 :
      ∀ s s', right_fun s = Left s' →
              ∃ n, nsteps υ n ([c_right], s) = Some ([], s');
    right_fun_spec_2 :
      ∀ s n, right_fun s = Right n →
             ∀ k, nsteps υ n (c_right::k, s) = None
  }.

  Inductive _cont_mod : Type := .

  Definition _genv := unit.

  Inductive _state : Type :=
  | Stuck : state → state → _state
  | State : state → state → list (cont_mod mod _cont_mod) → _state.

  Definition _step (_ : _genv) (S : _state) : option _state :=
    match S with
    | Stuck S₁ S₂ =>
      None
    | State S₁ S₂ km =>
      match step_helper S₁ S₂ km with
      | None =>
        None
      | Some StuckStep =>
        None
      | Some (GenericStep S₁' S₂' km') =>
        Some (State S₁' S₂' km')
      | Some (SpecialCmdStep m km') =>
        match left_fun m (snd S₁), right_fun m (snd S₂) with
        | Right b₁, Right b₂ =>
          let (S₁', n₁) := nsteps' υ b₁ S₁ in
          let (S₂', n₂) := nsteps' υ b₂ S₂ in
          match (b₁ - n₁), (b₂ - n₂) with
          | O, O =>
            None
          | _, _ =>
            Some (Stuck S₁' S₂')
          end
        | Left s₁', Left s₂' =>
          Some (State (tl (fst S₁), s₁') (tl (fst S₂), s₂') km')
        | _, _ =>
          None
        end
      | Some (SpecialContStep _ km') =>
        None
      end
    end.

  Definition _left_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | Stuck S₁ _ | State S₁ _ _ =>
      S₁
    end.

  Definition _right_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | Stuck _ S₂ | State _ S₂ _ =>
      S₂
    end.

  Inductive mod_invariant : cmd_mod mod → cmd → cmd → Prop :=
  | mod_id :
      ∀ c,
        mod_invariant Id c c
  | mod_rec_seq :
      ∀ m₁ m₂ c₁₁ c₁₂ c₂₁ c₂₂,
        mod_invariant m₁ c₁₁ c₂₁ →
        mod_invariant m₂ c₁₂ c₂₂ →
        mod_invariant (RecMod m₁ m₂) (Seq c₁₁ c₁₂) (Seq c₂₁ c₂₂)
  | mod_rec_ite :
      ∀ b m₁ m₂ c₁₁ c₁₂ c₂₁ c₂₂,
        mod_invariant m₁ c₁₁ c₂₁ →
        mod_invariant m₂ c₁₂ c₂₂ →
        mod_invariant (RecMod m₁ m₂) (ITE b c₁₁ c₁₂) (ITE b c₂₁ c₂₂)
  | mod_rec_while :
      ∀ b m m' c₁ c₂,
        mod_invariant m c₁ c₂ →
        mod_invariant (RecMod m m') (While b c₁) (While b c₂)
  | mod_special :
      ∀ m,
        mod_invariant (LeafMod m) (c_left m) (c_right m).

  Inductive km_invariant : list (cont_mod mod _cont_mod) → cont → cont → Prop :=
  | empty_cont_mod : km_invariant [] [] []
  | cmd_cont : ∀ c c' m km k₁ k₂,
      mod_invariant m c c' →
      km_invariant km k₁ k₂ →
      km_invariant (CmdMod m::km) (c::k₁) (c'::k₂).

  Definition _invariant (_ : _genv) (S : _state) : Prop :=
    match S with
    | Stuck S₁ S₂ => step υ S₁ = None ∧ step υ S₂ = None
    | State (k₁, s₁) (k₂, s₂) km =>
      km_invariant km k₁ k₂ ∧ M.Equal s₁ s₂
    end.

  Lemma _invariant_1:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      _invariant genv os'.
  Proof.
    intros [] [S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] os' Hinv Hstep ;
      try discriminate Hstep.
    unfold _step in Hstep.
    unfold _invariant in *.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ m km' k₁' k₂' Hm Hkm' ] ;
      subst km k₁ k₂ ;
      try discriminate Hstep.
    unfold step_helper in Hstep.
    destruct m as [ m | | m₁ m₂ ].
    * simpl snd in Hstep.
      inversion Hm as [ | | | | m_ ]. subst m_ c₁ c₂.
      simpl in Hstep.
      destruct (left_equiv_right m _ _ Hs₁s₂)
        as [[b₁ [b₂ [H₁ H₂]]] | [s₁' [s₂' [Hs₁' [Hs₂' Hs₁'s₂']]]]].
      { rewrite H₁, H₂ in Hstep.
        pose proof (left_fun_spec_2  m s₁ b₁ H₁ k₁') as H'₁.
        pose proof (right_fun_spec_2 m s₂ b₂ H₂ k₂') as H'₂.
        unfold nsteps in *.
        rewrite napply_napply'_Sn_iff in H'₁, H'₂.
        destruct H'₁ as [n₁ [S₁' [HS₁' H'S₁']]],
                 H'₂ as [n₂ [S₂' [HS₂' H'S₂']]].
        unfold nsteps' in *.
        simpl ProgrammingLanguage.step in *.
        simpl ProgrammingLanguage.state in *.
        unfold state, store, cont in *.
        rewrite HS₁', HS₂' in Hstep.
        destruct (b₁ - S n₁), (b₂ - S n₂) ;
          inversion Hstep ; clear Hstep ; subst os' ; auto. }
      { rewrite Hs₁', Hs₂' in Hstep. inversion Hstep. clear Hstep.
        subst os'.
        split ; auto. }
    * inversion Hm as [ c | | | | ]. subst c₁ c₂.
      clear Hm.
      destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ] ;
        try discriminate Hstep ;
        simpl in Hstep.
      { inversion Hstep ; split ;
          auto using cmd_cont. }
      { inversion Hstep ; split ;
          auto using cmd_cont, mod_id. }
      { rewrite (store_equal_eval_exp e s₁ s₂) in Hstep ; auto.
        destruct (eval_exp s₂ e) as [v₁ | ] eqn:Hv₁ ;
          inversion Hstep ; clear Hstep ; subst os' ; split ;
            auto using cmd_cont, mod_id, MFacts.add_m ;
            simpl ; try rewrite (store_equal_eval_exp _ s₁ s₂) ;
              try rewrite Hv₁ ; auto. }
      { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
        destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
          inversion Hstep ; clear Hstep ; subst os' ; split ;
            auto using cmd_cont, mod_id, MFacts.add_m ;
            simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂) ;
              try rewrite Hv₁ ; auto. }
      { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
        destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
          inversion Hstep ; clear Hstep ; subst os' ; split ;
            auto using cmd_cont, mod_id, MFacts.add_m ;
            simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂) ;
              try rewrite Hv₁ ; auto. }
      { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
        destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
          inversion Hstep ; clear Hstep ; subst os' ; split ;
            auto using cmd_cont, mod_id, MFacts.add_m ;
            simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂) ;
              try rewrite Hv₁ ; auto. }
    * inversion Hm as [ | m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                        | b m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                        | b m₁_ m₂_ c₁_ c₂_ Hm₁
                        | ] ;
        subst m₁_ m₂_ c₁ c₂.
      { simpl in Hstep.
        inversion Hstep. clear Hstep. subst os'.
        auto using cmd_cont. }
      { simpl in Hstep.
        rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
        destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv ;
          try discriminate Hstep ;
          simpl in Hstep ; inversion Hstep ; clear Hstep ; subst os' ;
            split ; auto using cmd_cont ;
              simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂) ;
                try rewrite Hv ; auto. }
      { simpl in Hstep.
        rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
        destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv ;
          try discriminate Hstep ;
          inversion Hstep ; clear Hstep ; subst os' ;
            split ; auto using cmd_cont ;
              simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂) ;
                try rewrite Hv ; auto. } 
  Qed.

  Lemma _prediction_soundness :
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (_left_state  os))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (_right_state os))
                       (Some (_right_state os'))
        ∧ n₁ + n₂ > 0.
  Proof.
    intros [ ] [S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] os' Hinv Hos' ;
      try discriminate Hos'.
    unfold _step in Hos'.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ m km' k₁' k₂' Hm Hkm' ] ;
      subst km k₁ k₂ ; try discriminate Hos'.
    destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H.
    + exists 1, 1. unfold nsteps.
      inversion Hos'. clear Hos'. subst os'.
      destruct (generic_step_bistep _ _ _ _ _ _ H) as [H₁ H₂].
      repeat split ; unfold _left_state, _right_state, nsteps ; auto ;
        simpl (ProgrammingLanguage.step imp_language).
      * unfold bind, napply. rewrite H₁. simpl. apply opt_state_eq_refl.
      * unfold bind, napply. rewrite H₂. simpl. apply opt_state_eq_refl.
    + apply step_helper_special_cmd_step_inv in H.
      destruct H as [H'm Hkm'']. inversion H'm as [H''m]. clear H'm.
      subst km'' m.
      simpl fst in *. simpl snd in *.
      simpl _left_state. simpl _right_state.
      inversion Hm as [ | | | | m ]. subst m' c₁ c₂.
      simpl in Hos'.
      destruct (left_equiv_right m _ _ Hs₁s₂)
        as [[b₁ [b₂ [H₁ H₂]]] | [s₁' [s₂' [Hs₁' [Hs₂' Hs₁'s₂']]]]].
      * rewrite H₁, H₂ in *.
        pose proof (left_fun_spec_2  _ _ _ H₁ k₁') as H'₁.
        pose proof (right_fun_spec_2 _ _ _ H₂ k₂') as H'₂.
        unfold nsteps in *.
        rewrite napply_napply'_Sn_iff in H'₁, H'₂.
        destruct H'₁ as [n₁ [S₁' [HS₁' H'S₁']]],
                 H'₂ as [n₂ [S₂' [HS₂' H'S₂']]].
        unfold nsteps' in Hos'.
        simpl ProgrammingLanguage.step in *. simpl ProgrammingLanguage.state in *.
        unfold state, store, cont in *.
        exists (b₁ - S n₁), (b₂ - S n₂).
        rewrite HS₁', HS₂' in Hos'.
        apply napply'_prefix in HS₁'. apply napply'_prefix in HS₂'.
        rewrite <- napply_napply'_0_iff in HS₁', HS₂'.
        rewrite HS₁', HS₂'.
        destruct (b₁ - S n₁) as [ | n₁' ], (b₂ - S n₂) as [ | n₂' ] ;
          inversion Hos' ; clear Hos' ; subst os' ;
            repeat split ; eauto using opt_state_eq_refl with arith.
      * rewrite Hs₁', Hs₂' in Hos'.
        inversion Hos'. clear Hos'. subst os'.
        destruct (left_fun_spec_1  _ _ _ Hs₁') as [n₁ H₁].
        destruct (right_fun_spec_1 _ _ _ Hs₂') as [n₂ H₂].
        exists n₁, n₂.
        replace (c_left m :: k₁') with ([c_left m] ++ k₁') ; auto.
        rewrite (nsteps_imp_app_1 _ _ _ _ _ H₁).
        replace (c_right m :: k₂') with ([c_right m] ++ k₂') ; auto.
        rewrite (nsteps_imp_app_1 _ _ _ _ _ H₂).
        simpl.
        repeat (split ; auto using opt_state_eq_refl).
        destruct n₁ as [ | n₁ ].
        { discriminate H₁. }
        auto with arith.
    + discriminate Hos'.
    + discriminate Hos'.
  Qed.

  Lemma _prediction_completeness :
    ∀ genv os,
      _invariant genv os →
      _step genv os = None →
      step υ (_left_state  os) = None
      ∧ step υ (_right_state os) = None.
  Proof.
    intros [ ] [S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] Hinv HNone ;
      auto.
    unfold _step in HNone.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ m km' k₁' k₂' Hm Hkm' ] ;
      subst km k₁ k₂ ; auto.
    destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H.
    - discriminate HNone.
    - apply step_helper_special_cmd_step_inv in H.
      destruct H as [H'm Hkm'']. inversion H'm as [H''m]. clear H'm.
      subst km'' m.
      simpl fst in *. simpl snd in *.
      simpl _left_state. simpl _right_state.
      inversion Hm as [ | | | | m ]. subst m' c₁ c₂.
      simpl in HNone.
      destruct (left_equiv_right m _ _ Hs₁s₂)
        as [[b₁ [b₂ [H₁ H₂]]] | [s₁' [s₂' [Hs₁' [Hs₂' Hs₁'s₂']]]]].
      * rewrite H₁, H₂ in *.
        pose proof (left_fun_spec_2  _ _ _ H₁ k₁') as H'₁.
        pose proof (right_fun_spec_2 _ _ _ H₂ k₂') as H'₂.
        unfold nsteps in *.
        rewrite napply_napply'_Sn_iff in H'₁, H'₂.
        destruct H'₁ as [n₁ [S₁' [HS₁' H'S₁']]],
                 H'₂ as [n₂ [S₂' [HS₂' H'S₂']]].
        unfold nsteps' in HNone.
        simpl ProgrammingLanguage.step in *. simpl ProgrammingLanguage.state in *.
        unfold state, store, cont in *.
        rewrite HS₁', HS₂' in HNone.
        apply napply'_prefix in HS₁'. apply napply'_prefix in HS₂'.
        rewrite <- napply_napply'_0_iff in HS₁', HS₂'.
        destruct (b₁ - S n₁) as [ | n₁' ], (b₂ - S n₂) as [ | n₂' ] ;
          try discriminate HNone.
        inversion HS₁'. inversion HS₂'. subst S₁' S₂'. auto.
      * rewrite Hs₁', Hs₂' in HNone.
        discriminate HNone.
    - destruct (generic_step_bicrash _ _ _ H) as [H₁ H₂].
      auto.
    - (* Thanks to the invariant, we can't crash step_helper: *)
      exfalso.
      simpl in H.
      destruct m as [ m | | m₁ m₂ ].
      * discriminate H.
      * inversion Hm as [ c | | | | ].
        subst c₁ c₂.
        destruct c ; try discriminate H.
        { rewrite (store_equal_eval_exp e s₁ s₂) in H ; auto.
          destruct (eval_exp s₂ e) ; try discriminate H. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H. }
      * inversion Hm as [ | m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                          | b m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                          | b m₁_ m₂_ c₁_ c₂_ Hm₁
                          | ] ;
          subst m₁_ m₂_ c₁ c₂.
        { discriminate H. }
        { simpl in H.
          rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ;
            discriminate H. }
        { simpl in H.
          rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          simpl in H.
          destruct (eval_bexp s₂ b) as [[|]|] ;
            discriminate H. }
  Qed.

  Definition abstract_equiv_oracle : oracle_language imp_language imp_language :=
    {|
      oracle_state := _state;
      oracle_genv := unit;
      oracle_step := _step;
      left_state := _left_state;
      right_state := _right_state;
      left_genv := λ _, υ;
      right_genv := λ _, υ;
      invariant := _invariant;
      invariant_1 := _invariant_1;
      prediction_soundness := _prediction_soundness;
      prediction_completeness := _prediction_completeness
    |}.
End Make.