(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

Require Import Common.
Require Import Imp.Lang.
Require Import Imp.BigStep.
Require Import Imp.RHL.
Require Import Imp.DisjointStates.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.
Require Import OracleHelpers.

Require Import Omega.

(* Core RHL with self-composition

  Self-composition ([CoreSCSelfComp] rule) simply consists in
  proving our postcondition on the sequential composition of the
  two programs under scrutiny using regular Hoare Logic, thus
  not imposing any condition on the relation between the two
  programs' structure.

  In our case, it means that when the head proof term begins with
  a [CoreSCSelfComp], we will only predict steps of the left
  sub-program until it ends (if it does) before switching to the
  right sub-program. Intuitively, this is possible to achieve by
  unfolding the head proof term until we get a [HoareSeq] rule.

  However, there is a tricky case we have to consider: we have
  seen in [Imp.Oracles.CoreRHL] that a few rules might apply to
  “fictitious” [Skip] sub-commands. That is also the case of
  [CoreSCSelfComp].

  There is a last trick with the [CoreSCSelfComp] rule in our
  setting: much like [CoreSCSub], it does not represent progress
  in the computation of either program.

*)

Module Make (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT).

  Module BigStep := Imp.BigStep.Make Var_as_DT Lang.
  Module RHL := Imp.RHL.Make Var_as_DT Lang BigStep.
  Import RHL.
  Import Hoare.
  Import DisjointStates.

  Definition _genv : Type := assertion. (* Post-condition *)

  (* This might lead to some duplication, but it's probably
     one of the sanest options… *)
  Inductive proof_term :=
  | HoareProof:
      ∀ {P Q c₁ c₂},
        core_rhl_sc_proof P c₁ c₂ Q →
        proof_term
  | HoareProofL:
      ∀ {P Q c₁},
        core_rhl_sc_proof P c₁ Skip Q →
        proof_term
  | HoareProofR:
      ∀ {P Q c₂},
        core_rhl_sc_proof P Skip c₂ Q →
        proof_term
  | SingleHoareProofL:
      ∀ {P Q c},
        hoare_proof P c Q →
        proof_term
  | SingleHoareProofR:
      ∀ {P Q c},
        hoare_proof P c Q →
        proof_term.

  Inductive _state :=
  | State :
      state →
      state →
      store →
      assertion →
      list proof_term →
      _state.

  Fixpoint proof_terms_consistent (l : list proof_term) (P Q : assertion) : Prop :=
    match l with
    | [] => P ⇒ Q
    | (@HoareProofL P' Q' _ _)::l'
    | (@HoareProofR P' Q' _ _)::l'
    | (@SingleHoareProofL P' Q' _ _)::l'
    | (@SingleHoareProofR P' Q' _ _)::l'
    | (@HoareProof P' Q' _ _ _)::l' =>
      P ⇒ P' ∧ proof_terms_consistent l' Q' Q
    end.

  Property proof_terms_consistent_1:
    ∀ l P P' Q,
      P ⇒ P' →
      proof_terms_consistent l P' Q →
      proof_terms_consistent l P Q.
  Proof.
    intro l.
    induction l as [
                  | [ P'' Q'' c₁ c₂ H
                    | P'' Q'' c H
                    | P'' Q'' c H
                    | P'' Q'' c H
                    | P'' Q'' c H] l ] ;
      intros P P' Q H₁ H₂.
    - simpl in *. eauto.
    - simpl in *. destruct H₂ as [H₂ H₃].
      split ; eauto.
    - simpl in *. destruct H₂ as [H₂ H₃].
      split ; eauto.
    - simpl in *. destruct H₂ as [H₂ H₃].
      split ; eauto.
    - simpl in *. destruct H₂ as [H₂ H₃].
      split ; eauto.
    - simpl in *. destruct H₂ as [H₂ H₃].
      split ; eauto.
  Qed.

  Inductive proof_terms_match : cont → cont → list proof_term → Prop :=
  | proof_terms_match_nil :
      proof_terms_match nil nil nil
  | proof_terms_match_cons :
      ∀ c₁ k₁ c₂ k₂ l P Q proof,
        proof_terms_match k₁ k₂ l →
        proof_terms_match (c₁::k₁) (c₂::k₂) (@HoareProof P Q c₁ c₂ proof::l)
  | proof_terms_match_cons_l :
      ∀ c₁ k₁ k₂ l P Q proof,
        proof_terms_match k₁ k₂ l →
        proof_terms_match (c₁::k₁) k₂ (@HoareProofL P Q c₁ proof::l)
  | proof_terms_match_cons_r :
      ∀ k₁ c₂ k₂ l P Q proof,
        proof_terms_match k₁ k₂ l →
        proof_terms_match k₁ (c₂::k₂) (@HoareProofR P Q c₂ proof::l)
  | proof_terms_match_cons_l' :
      ∀ c₁ k₁ k₂ l P Q proof,
        proof_terms_match k₁ k₂ l →
        proof_terms_match (c₁::k₁) k₂ (@SingleHoareProofL P Q c₁ proof::l)
  | proof_terms_match_cons_r' :
      ∀ k₁ c₂ k₂ l P Q proof,
        proof_terms_match k₁ k₂ l →
        proof_terms_match k₁ (c₂::k₂) (@SingleHoareProofR P Q c₂ proof::l).

  Hint Constructors proof_terms_match.

  Definition _invariant (Q : _genv) (os : _state) :=
    match os with
    | State S₁ S₂ s P l =>
      P s
      ∧ store_is_union s (snd S₁) (snd S₂)
      ∧ disjoint_states S₁ S₂
      ∧ proof_terms_consistent l P Q
      ∧ proof_terms_match (fst S₁) (fst S₂) l
    end.

  Fixpoint _split_hoare_seq {P Q c}
           (* XXX: ideally hoare_proof P (Seq c₁ c₂) Q,
              but dependent types are a nightmare  *)
           (H : hoare_proof P c Q)
           (l : list proof_term) :
    option (list proof_term) :=
    match H with
    | HoareSeq _ _ _ _ _ H₁ H₂ =>
      Some (SingleHoareProofL H₁ :: SingleHoareProofR H₂ :: l)
    | HoareSub _ _ _ _ _ H' _ _ =>
      _split_hoare_seq H' l
    | _ =>
      None
    end.

  Lemma _split_hoare_seq_1:
    ∀ P Q Q' c₁ c₂ (H : hoare_proof P (Seq c₁ c₂) Q) l,
      proof_terms_consistent l Q Q' →
    ∃ P' Q'' R' (H₁ : hoare_proof P' c₁ Q'') (H₂ : hoare_proof Q'' c₂ R'),
      _split_hoare_seq H l = Some (SingleHoareProofL H₁ :: SingleHoareProofR H₂ :: l)
      ∧ P ⇒ P' ∧ proof_terms_consistent l R' Q'.
  Proof.
    intros P Q Q' c₁ c₂ H l.
    remember (Seq c₁ c₂) as c eqn:Hc.
    induction H as [ | | | P Q R c₁' c₂' H₁ _ H₂ _ | | | P P' Q Q'' c H' IH H₁ H₂ ] ;
      intros Hl ;
      try discriminate Hc ; try subst c.
    - inversion Hc. clear Hc. subst c₁' c₂'. simpl. exists P, Q, R, H₁, H₂.
      repeat (split ; auto).
    - destruct (IH eq_refl) as [P'' [Q''' [R' [H'₁ [H'₂ [H [HPP' Hl']]]]]]] ;
        eauto using proof_terms_consistent_1.
      exists P'', Q''', R', H'₁, H'₂. split ; auto. split ; auto.
      eauto using assertion_implies_trans.
  Qed.

  (* TODO: refactor *)
  Fixpoint _step_helper_single_left {P Q c}
           (S₁ S₂ : state)
           (s : store)
           (H : hoare_proof P c Q)
           (l' : list proof_term)
    : option _state :=
    match H with
    | HoareSkip P =>
      step υ S₁ >>= λ S₁',
      Some (State S₁' S₂ s P l')
    | HoareAssign P x e =>
      step υ S₁ >>= λ S₁',
      eval_exp s e >>= λ v,
      Some (State S₁' S₂ (M.add x v s) P l')
    | HoareAssert P b H =>
      step υ S₁ >>= λ S₁',
      Some (State S₁' S₂ s P l')
    | HoareSeq P Q R c₁ c₂ H₁ H₂ =>
      step υ S₁ >>= λ S₁',
      Some (State S₁' S₂ s P (SingleHoareProofL H₁ :: SingleHoareProofL H₂ :: l'))
    | HoareITE P Q b c₁ c₂ H₁ H₂ _ =>
      step υ S₁ >>= λ S₁',
      eval_bexp s b >>= λ v,
      if v then
        Some (State S₁' S₂ s
                    (λ S, P S ∧ assertion_of_bexp b S)
                    ((SingleHoareProofL H₁)::l'))
      else
        Some (State S₁' S₂ s
                    (λ S, P S ∧ assertion_of_bexp (NOT b) S)
                    ((SingleHoareProofL H₂)::l'))
    | HoareWhile P b c H' Hnc =>
      step υ S₁ >>= λ S₁',
      eval_bexp s b >>= λ v,
      if v then
        Some (State S₁' S₂ s
                    (λ S, P S ∧ assertion_of_bexp b S)
                    ((SingleHoareProofL H')::(SingleHoareProofL (HoareWhile P b c H' Hnc))::l'))
      else
        Some (State S₁' S₂ s
                    (λ S, P S ∧ assertion_of_bexp (NOT b) S)
                    l')
    | HoareSub P P' Q Q' c H' H₁ H₂ =>
      _step_helper_single_left S₁ S₂ s H' l'
    end.

  (* TODO: refactor *)
  Fixpoint _step_helper_single_right {P Q c}
           (S₁ S₂ : state)
           (s : store)
           (H : hoare_proof P c Q)
           (l' : list proof_term)
    : option _state :=
    match H with
    | HoareSkip P =>
      step υ S₂ >>= λ S₂',
      Some (State S₁ S₂' s P l')
    | HoareAssign P x e =>
      step υ S₂ >>= λ S₂',
      eval_exp s e >>= λ v,
      Some (State S₁ S₂' (M.add x v s) P l')
    | HoareAssert P b H =>
      step υ S₂ >>= λ S₂',
      Some (State S₁ S₂' s P l')
    | HoareSeq P Q R c₁ c₂ H₁ H₂ =>
      step υ S₂ >>= λ S₂',
      Some (State S₁ S₂' s P (SingleHoareProofR H₁ :: SingleHoareProofR H₂ :: l'))
    | HoareITE P Q b c₁ c₂ H₁ H₂ _ =>
      step υ S₂ >>= λ S₂',
      eval_bexp s b >>= λ v,
      if v then
        Some (State S₁ S₂' s
                    (λ S, P S ∧ assertion_of_bexp b S)
                    ((SingleHoareProofR H₁)::l'))
      else
        Some (State S₁ S₂' s
                    (λ S, P S ∧ assertion_of_bexp (NOT b) S)
                    ((SingleHoareProofR H₂)::l'))
    | HoareWhile P b c H' Hnc =>
      step υ S₂ >>= λ S₂',
      eval_bexp s b >>= λ v,
      if v then
        Some (State S₁ S₂' s
                    (λ S, P S ∧ assertion_of_bexp b S)
                    ((SingleHoareProofR H')::(SingleHoareProofR (HoareWhile P b c H' Hnc))::l'))
      else
        Some (State S₁ S₂' s
                    (λ S, P S ∧ assertion_of_bexp (NOT b) S)
                    l')
    | HoareSub P P' Q Q' c H' H₁ H₂ =>
      _step_helper_single_right S₁ S₂ s H' l'
    end.


  Fixpoint _step_helper {P Q c₁ c₂}
           (S₁ S₂ : state)
           (s : store)
           (H : core_rhl_sc_proof P c₁ c₂ Q) (l' : list proof_term)
    : option _state :=
    match H with
    | CoreSCSkip P =>
      step υ S₁ >>= λ S₁',
      step υ S₂ >>= λ S₂',
      Some (State S₁' S₂' s P l')
    | CoreSCAssign P x e₁ y e₂ =>
      step υ S₁ >>= λ S₁',
      step υ S₂ >>= λ S₂',
      eval_exp s e₂ >>= λ v₁,
      let s' := M.add y v₁ s in
      eval_exp s e₁ >>= λ v₂,
      Some (State S₁' S₂' (M.add x v₂ s') P l')
    | CoreSCAssert P b₁ b₂ Hb =>
      step υ S₁ >>= λ S₁',
      step υ S₂ >>= λ S₂',
      Some (State S₁' S₂' s P l')
    | CoreSCSeq P Q R c₁₁ c₂₁ c₁₂ c₂₂ H₁ H₂ =>
      step υ S₁ >>= λ S₁',
      step υ S₂ >>= λ S₂',
      Some (State S₁' S₂' s P ((HoareProof H₁)::(HoareProof H₂)::l'))
    | CoreSCITE P Q b₁ b₂ c₁₁ c₂₁ c₁₂ c₂₂ H₁ H₂ Hb _ =>
      step υ S₁ >>= λ S₁',
      step υ S₂ >>= λ S₂',
      eval_bexp s b₁ >>= λ v,
      if v then
        Some (State S₁' S₂' s
                    (λ S, P S ∧ assertion_of_bexp b₁ S)
                    ((HoareProof H₁)::l'))
      else
        Some (State S₁' S₂' s
                    (λ S, P S ∧ assertion_of_bexp (NOT b₁) S)
                    ((HoareProof H₂)::l'))
    | CoreSCWhile P b₁ b₂ c₁ c₂ H Hb Hnc =>
      step υ S₁ >>= λ S₁',
      step υ S₂ >>= λ S₂',
      eval_bexp s b₁ >>= λ v,
      if v then
        Some (State S₁' S₂' s
                    (λ S, P S ∧ assertion_of_bexp b₁ S)
                    ((HoareProof H)::(HoareProof (CoreSCWhile P b₁ b₂ c₁ c₂ H Hb Hnc))::l'))
      else
        Some (State S₁' S₂' s
                    (λ S, P S ∧ assertion_of_bexp (NOT b₁) S)
                    l')
    | CoreSCSub P P' Q Q' c₁ c₂ H HP'P HQQ' =>
      _step_helper S₁ S₂ s H l'
    (* Executing only one of the programs: *)
    | CoreSCAssignLeft P Q x e _ H' =>
      step υ S₁ >>= λ S₁',
      eval_exp s e >>= λ v,
      Some (State S₁' S₂ (M.add x v s) P (HoareProofR H' :: l'))
    | CoreSCAssignRight P _ x e _ H' =>
      step υ S₂ >>= λ S₂',
      eval_exp s e >>= λ v,
      Some (State S₁ S₂' (M.add x v s) P (HoareProofL H' :: l'))
    | CoreSCAssertLeft P _ b _ H' _ =>
      step υ S₁ >>= λ S₁',
      Some (State S₁' S₂ s P (HoareProofR H' :: l'))
    | CoreSCAssertRight P _ b _ H' _ =>
      step υ S₂ >>= λ S₂',
      Some (State S₁ S₂' s P (HoareProofL H' :: l'))
    | CoreSCSeqSkipLeft P _ _ _ _ _ H₁ H₂ =>
      step υ S₁ >>= λ S₁',
      Some (State S₁' S₂ s P (HoareProofL H₁ :: HoareProof H₂ :: l'))
    | CoreSCSeqSkipRight P Q R _ _ _ H₁ H₂ =>
      step υ S₂ >>= λ S₂',
      Some (State S₁ S₂' s P (HoareProofR H₁ :: HoareProof H₂ :: l'))
    | CoreSCITELeft P Q b _ _ _ H₁ H₂ _ =>
      step υ S₁ >>= λ S₁',
      eval_bexp s b >>= λ v,
      if v then
        Some (State S₁' S₂ s
                    (λ S, P S ∧ assertion_of_bexp b S)
                    ((HoareProof H₁)::l'))
      else
        Some (State S₁' S₂ s
                    (λ S, P S ∧ assertion_of_bexp (NOT b) S)
                    ((HoareProof H₂)::l'))
    | CoreSCITERight P Q b _ _ _ H₁ H₂ _ =>
      step υ S₂ >>= λ S₂',
      eval_bexp s b >>= λ v,
      if v then
        Some (State S₁ S₂' s
                    (λ S, P S ∧ assertion_of_bexp b S)
                    ((HoareProof H₁)::l'))
      else
        Some (State S₁ S₂' s
                    (λ S, P S ∧ assertion_of_bexp (NOT b) S)
                    ((HoareProof H₂)::l'))
    | CoreSCSelfComp P Q c₁ c₂ H =>
      match _split_hoare_seq H l' with
      | Some (SingleHoareProofL H'::l') =>
        _step_helper_single_left S₁ S₂ s H' l'
      | _ =>
        None
      end
    end.

  Fixpoint _step_helper_left {P Q c₁ c₂}
           (S₁ S₂ : state)
           (s : store)
           (* XXX: ideally core_rhl_sc_proof P c₁ Skip Q,
              but dependent types are a nightmare  *)
           (H : core_rhl_sc_proof P c₁ c₂ Q)
           (l' : list proof_term)
    : option _state :=
    match H with
    | CoreSCSkip P =>
      step υ S₁ >>= λ S₁',
      Some (State S₁' S₂ s P l')
    | CoreSCSub P P' Q Q' c₁ Skip H HP'P HQQ' =>
      _step_helper_left S₁ S₂ s H l'
    (* Executing only one of the programs: *)
    | CoreSCAssignLeft P Q x e Skip H' =>
      step υ S₁ >>= λ S₁',
      eval_exp s e >>= λ v,
      (* XXX: H' is dropped completely as we know it only refers to “imaginary” stuff *)
      Some (State S₁' S₂ (M.add x v s) Q l')
    | CoreSCAssertLeft P Q b Skip H' _ =>
      step υ S₁ >>= λ S₁',
      (* XXX: H' is dropped completely as we know it only refers to “imaginary” stuff *)
      Some (State S₁' S₂ s Q l')
    | CoreSCSeqSkipLeft P _ _ _ _ Skip H₁ H₂ =>
      step υ S₁ >>= λ S₁',
      Some (State S₁' S₂ s P (HoareProofL H₁ :: HoareProofL H₂ :: l'))
    | CoreSCITELeft P Q b _ _ Skip H₁ H₂ _ =>
      step υ S₁ >>= λ S₁',
      eval_bexp s b >>= λ v,
      if v then
        Some (State S₁' S₂ s
                    (λ S, P S ∧ assertion_of_bexp b S)
                    ((HoareProofL H₁)::l'))
      else
        Some (State S₁' S₂ s
                    (λ S, P S ∧ assertion_of_bexp (NOT b) S)
                    ((HoareProofL H₂)::l'))
    | CoreSCSelfComp P Q c₁ c₂ H =>
      match _split_hoare_seq H l' with
      | Some (SingleHoareProofL H'::SingleHoareProofR _::l') =>
        _step_helper_single_left S₁ S₂ s H' l'
      | _ =>
        None
      end
    | _ =>
      None
    end.

  Fixpoint _step_helper_right {P Q c₁ c₂}
           (S₁ S₂ : state)
           (s : store)
           (* XXX: ideally core_rhl_sc_proof P Skip c₂ Q,
              but dependent types are a nightmare  *)
           (H : core_rhl_sc_proof P c₁ c₂ Q)
           (l' : list proof_term)
    : option _state :=
    match H with
    | CoreSCSkip P =>
      step υ S₂ >>= λ S₂',
      Some (State S₁ S₂' s P l')
    | CoreSCSub P P' Q Q' Skip c₂ H HP'P HQQ' =>
      _step_helper_right S₁ S₂ s H l'
    (* Executing only one of the programs: *)
    | CoreSCAssignRight P Q x e Skip H' =>
      step υ S₂ >>= λ S₂',
      eval_exp s e >>= λ v,
      (* XXX: H' is dropped completely as we know it only refers to “imaginary” stuff *)
      Some (State S₁ S₂' (M.add x v s) Q l')
    | CoreSCAssertRight P Q b Skip H' _ =>
      step υ S₂ >>= λ S₂',
      (* XXX: H' is dropped completely as we know it only refers to “imaginary” stuff *)
      Some (State S₁ S₂' s Q l')
    | CoreSCSeqSkipRight P _ _ Skip _ _ H₁ H₂ =>
      step υ S₂ >>= λ S₂',
      Some (State S₁ S₂' s P (HoareProofR H₁ :: HoareProofR H₂ :: l'))
    | CoreSCITERight P Q b _ _ Skip H₁ H₂ _ =>
      step υ S₂ >>= λ S₂',
      eval_bexp s b >>= λ v,
      if v then
        Some (State S₁ S₂' s
                    (λ S, P S ∧ assertion_of_bexp b S)
                    ((HoareProofR H₁)::l'))
      else
        Some (State S₁ S₂' s
                    (λ S, P S ∧ assertion_of_bexp (NOT b) S)
                    ((HoareProofR H₂)::l'))
    | CoreSCSelfComp P Q c₁ c₂ H =>
      match _split_hoare_seq H l' with
      | Some (SingleHoareProofL _::SingleHoareProofR H'::l') =>
        _step_helper_single_right S₁ S₂ s H' l'
      | _ =>
        None
      end
    | _ =>
      None
    end.

  Definition _step (_ : _genv) (S : _state) : option _state :=
    match S with
    | State S₁ S₂ s P [] =>
      (* Final state *)
      None
    | State S₁ S₂ s P (HoareProof H::l') =>
      _step_helper S₁ S₂ s H l'
    | State S₁ S₂ s P (HoareProofL H::l') =>
      _step_helper_left S₁ S₂ s H l'
    | State S₁ S₂ s P (HoareProofR H::l') =>
      _step_helper_right S₁ S₂ s H l'
    | State S₁ S₂ s P (SingleHoareProofL H::l') =>
      _step_helper_single_left S₁ S₂ s H l'
    | State S₁ S₂ s P (SingleHoareProofR H::l') =>
      _step_helper_single_right S₁ S₂ s H l'
    end.

  Definition _left_state (os : _state) : ProgrammingLanguage.state imp_language :=
    match os with
    | State S₁ _ _ _ _ =>
      S₁
    end.

  Definition _right_state (os : _state) : ProgrammingLanguage.state imp_language :=
    match os with
    | State _ S₂ _ _ _ =>
      S₂
    end.

  Lemma _step_helper_single_left_soundness:
    ∀ {P Q} c₁ k₁' s₁ k₂ s₂ s l' (Hc : hoare_proof P c₁ Q) os',
      disjoint_states (c₁ :: k₁', s₁) (k₂, s₂) →
      store_is_union s s₁ s₂ →
      P s →
      _step_helper_single_left (c₁ :: k₁', s₁) (k₂, s₂) s Hc l' = Some os' →
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (c₁ :: k₁', s₁))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (k₂, s₂))
                       (Some (_right_state os'))
        ∧ (n₁ + n₂ > 0).
  Proof.
    intros P Q c₁ k₁' s₁ k₂ s₂ s l' Hc os' Hdisj Hs₁s₂ HP Hos'.
    induction Hc as [ P | P x e | P b H
                      | P Q R c₁ c₂ H₁ _ H₂ _
                      | P Q b c₁ c₂ H₁ _ H₂ _ Hnc
                      | P b c H _ Hnc
                      | P P' Q Q' c H IH ].
    - inversion Hos'. clear Hos'. subst os'.
      exists 1, 0.
      simpl. repeat (split ; auto using opt_state_eq_refl).
    - exists 1, 0. unfold nsteps.
      destruct HP as [v [Hv HP]].
      simpl in *.
      rewrite Hv in Hos'.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_exp_var z e → V₁ z) as He.
      { intros z Hz. apply H'k₁. exists (Assign x e). simpl. auto. }
      simpl in *.
      rewrite (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂) in * ; auto.
      rewrite Hv in *. simpl in *.
      inversion Hos'. clear Hos'. subst os'.
      simpl.
      repeat (split ; auto using opt_state_eq_refl).
    - exists 1, 0. unfold nsteps. simpl in *.
      apply H in HP. unfold assertion_of_bexp in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₁ z) as He.
      { intros z Hz. apply H'k₁. exists (Assert b). simpl. auto. }
      simpl in *.
      rewrite (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) in * ; auto.
      rewrite HP in *. simpl in *.
      inversion Hos'. clear Hos'. subst os'.
      repeat (split ; auto using opt_state_eq_refl).
    - exists 1, 0. unfold nsteps.
      inversion Hos'. clear Hos'. subst os'.
      repeat (split ; auto using opt_state_eq_refl).
    - exists 1, 0. unfold nsteps. simpl in *.
      apply Hnc in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₁ z) as He.
      { intros z Hz. apply H'k₁. exists (ITE b c₁ c₂). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) in * ; auto.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ;
        try contradiction.
      destruct v ; inversion Hos' ; clear Hos' ; subst os' ;
        repeat (split ; auto using opt_state_eq_refl).
    - exists 1, 0. unfold nsteps. simpl in *.
      apply Hnc in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₁ z) as He.
      { intros z Hz. apply H'k₁. exists (While b c). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) in * ; auto.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ;
        try contradiction.
      destruct v ; inversion Hos' ; clear Hos' ; subst os' ;
        repeat (split ; auto using opt_state_eq_refl).
    - apply IH ; auto.
  Qed.

  Lemma _step_helper_single_left_no_crash:
    ∀ {P Q} c₁ k₁' s₁ k₂ s₂ s l' (Hc : hoare_proof P c₁ Q),
      disjoint_states (c₁ :: k₁', s₁) (k₂, s₂) →
      store_is_union s s₁ s₂ →
      P s →
      ∃ os', _step_helper_single_left (c₁ :: k₁', s₁) (k₂, s₂) s Hc l' = Some os'.
  Proof.
    intros P Q c₁ k₁' s₁ k₂ s₂ s l' Hc Hdisj Hs₁s₂ HP.
    induction Hc as [ P | P x e | P b H
                      | P Q R c₁ c₂ H₁ _ H₂ _
                      | P Q b c₁ c₂ H₁ _ H₂ _ Hnc
                      | P b c H _ Hnc
                      | P P' Q Q' c H IH ] ;
      simpl ; eauto.
    - destruct HP as [v [Hv HP]].
      simpl in *.
      rewrite Hv.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_exp_var z e → V₁ z) as He.
      { intros z Hz. apply H'k₁. exists (Assign x e). simpl. auto. }
      simpl in *.
      rewrite (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂) in * ; auto.
      rewrite Hv in *. simpl in *.
      eauto.
    - apply H in HP. unfold assertion_of_bexp in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₁ z) as He.
      { intros z Hz. apply H'k₁. exists (Assert b). simpl. auto. }
      simpl in *.
      rewrite (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) in * ; auto.
      rewrite HP in *. simpl in *.
      eauto.
    - apply Hnc in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₁ z) as He.
      { intros z Hz. apply H'k₁. exists (ITE b c₁ c₂). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) in * ; auto.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ;
        try contradiction.
      destruct v ; simpl ; eauto.
    - apply Hnc in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₁ z) as He.
      { intros z Hz. apply H'k₁. exists (While b c). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) in * ; auto.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ;
        try contradiction.
      destruct v ; simpl ; eauto.
  Qed.

  Lemma _step_helper_single_right_soundness:
    ∀ {P Q} k₁ s₁ c₂ k₂' s₂ s l' (Hc : hoare_proof P c₂ Q) os',
      disjoint_states (k₁, s₁) (c₂ :: k₂', s₂) →
      store_is_union s s₁ s₂ →
      P s →
      _step_helper_single_right (k₁, s₁) (c₂ :: k₂', s₂) s Hc l' = Some os' →
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (k₁, s₁))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (c₂ :: k₂', s₂))
                       (Some (_right_state os'))
        ∧ (n₁ + n₂ > 0).
  Proof.
    intros P Q k₁ s₁ c₂ k₂' s₂ s l' Hc os' Hdisj Hs₁s₂ HP Hos'.
    induction Hc as [ P | P x e | P b H
                      | P Q R c₁ c₂ H₁ _ H₂ _
                      | P Q b c₁ c₂ H₁ _ H₂ _ Hnc
                      | P b c H _ Hnc
                      | P P' Q Q' c H IH ].
    - exists 0, 1. inversion Hos'. clear Hos'. subst os'.
      simpl. repeat (split ; auto using opt_state_eq_refl).
    - exists 0, 1. unfold nsteps.
      simpl in *. destruct HP as [v [Hv HP]].
      rewrite Hv in *.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_exp_var z e → V₂ z) as He.
      { intros z Hz. apply H'k₂. exists (Assign x e). simpl. auto. }
      rewrite (disjoint_union_eval_exp_same e s₂ s₁ s V₂ V₁) in * ; eauto.
      rewrite Hv in *. simpl in *.
      inversion Hos'. clear Hos'. subst os'.
      repeat (split ; auto using opt_state_eq_refl).
    - exists 0, 1. unfold nsteps. simpl in *.
      apply H in HP. unfold assertion_of_bexp in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₂ z) as He.
      { intros z Hz. apply H'k₂. exists (Assert b). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b s₂ s₁ s V₂ V₁) in * ; eauto.
      rewrite HP in *. simpl in *.
      inversion Hos'. clear Hos'. subst os'.
      repeat (split ; auto using opt_state_eq_refl).
    - exists 0, 1. unfold nsteps. simpl in *.
      inversion Hos'. clear Hos'. subst os'.
      repeat (split ; auto using opt_state_eq_refl).
    - exists 0, 1. unfold nsteps. simpl in *.
      apply Hnc in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₂ z) as He.
      { intros z Hz. apply H'k₂. exists (ITE b c₁ c₂). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b s₂ s₁ s V₂ V₁) in * ; eauto.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ;
        try contradiction.
      destruct v ; inversion Hos' ; clear Hos' ; subst os' ;
        repeat (split ; auto using opt_state_eq_refl).
    - exists 0, 1. unfold nsteps. simpl in *.
      apply Hnc in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₂ z) as He.
      { intros z Hz. apply H'k₂. exists (While b c). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b s₂ s₁ s V₂ V₁) in * ; eauto.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ;
        try contradiction.
      destruct v ; inversion Hos' ; clear Hos' ; subst os' ;
        repeat (split ; auto using opt_state_eq_refl).
    - apply IH ; auto.
  Qed.

  Lemma _step_helper_single_right_no_crash:
    ∀ {P Q} k₁ s₁ c₂ k₂' s₂ s l' (Hc : hoare_proof P c₂ Q),
      disjoint_states (k₁, s₁) (c₂ :: k₂', s₂) →
      store_is_union s s₁ s₂ →
      P s →
      ∃ os', _step_helper_single_right (k₁, s₁) (c₂ :: k₂', s₂) s Hc l' = Some os'.
  Proof.
    intros P Q k₁ s₁ c₂ k₂' s₂ s l' Hc Hdisj Hs₁s₂ HP.
    induction Hc as [ P | P x e | P b H
                      | P Q R c₁ c₂ H₁ _ H₂ _
                      | P Q b c₁ c₂ H₁ _ H₂ _ Hnc
                      | P b c H _ Hnc
                      | P P' Q Q' c H IH ] ;
      simpl ; eauto.
    - simpl in *. destruct HP as [v [Hv HP]].
      rewrite Hv in *.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_exp_var z e → V₂ z) as He.
      { intros z Hz. apply H'k₂. exists (Assign x e). simpl. auto. }
      rewrite (disjoint_union_eval_exp_same e s₂ s₁ s V₂ V₁) in * ; eauto.
      rewrite Hv in *. simpl in *.
      eauto.
    - apply H in HP. unfold assertion_of_bexp in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₂ z) as He.
      { intros z Hz. apply H'k₂. exists (Assert b). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b s₂ s₁ s V₂ V₁) in * ; eauto.
      rewrite HP in *. simpl in *.
      eauto.
    - apply Hnc in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₂ z) as He.
      { intros z Hz. apply H'k₂. exists (ITE b c₁ c₂). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b s₂ s₁ s V₂ V₁) in * ; eauto.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ;
        try contradiction.
      destruct v ; simpl ; eauto.
    - apply Hnc in HP.
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b → V₂ z) as He.
      { intros z Hz. apply H'k₂. exists (While b c). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b s₂ s₁ s V₂ V₁) in * ; eauto.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ;
        try contradiction.
      destruct v ; simpl ; eauto.
  Qed.

  Lemma hoare_close_skip:
    ∀ P Q s (H : hoare_proof P Skip Q),
      P s → Q s.
  Proof.
    intros P Q s H.
    remember Skip as c eqn:Hc.
    induction H ;
      try discriminate Hc ; try subst c ; auto.
  Qed.

  Lemma hoare_close_skip_skip:
    ∀ P Q s (H : hoare_proof P (Seq Skip Skip) Q),
      P s → Q s.
  Proof.
    intros P Q s H.
    remember (Seq Skip Skip) as c eqn:Hc.
    induction H ;
      try discriminate Hc ; try subst c ; auto.
    inversion Hc. clear Hc. subst c₁ c₂.
    intro HP. apply (hoare_close_skip Q R) ; auto.
    apply (hoare_close_skip P Q) ; auto.
  Qed.

  Lemma _prediction_soundness:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (_left_state  os))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (_right_state os))
                       (Some (_right_state os'))
        ∧ n₁ + n₂ > 0.
  Proof.
    intros genv [[k₁ s₁] [k₂ s₂] s P₀ l] os' [HP₀ [Hs [Hdisj [H₁l H₂l]]]] Hos'.
    simpl fst in *. simpl snd in *.
    inversion H₂l as [
                    | c₁ k₁' c₂ k₂' l' P Q Hc H₂l'
                    | c₁ k₁'    k₂' l' P Q Hc H₂l'
                    |    k₁' c₂ k₂' l' P Q Hc H₂l'
                    | c₁ k₁'    k₂' l' P Q Hc H₂l'
                    |    k₁' c₂ k₂' l' P Q Hc H₂l' ] ;
      subst k₁ k₂ l ; clear H₂l ;
        try (inversion Hos' ; clear Hos' ; subst os').
    - destruct H₁l as [HP₀P H₁l].
      simpl.
      induction Hc as [
                       P
                     | P x e y e'
                     | P b₁ b₂ Hb₁b₂
                     | P Q R c₁₁ c₂₁ c₁₂ c₂₂ H₁ _ H₂ _
                     | P Q b₁ b₂ c₁₁ c₂₁ c₁₂ c₂₂ H₁ _ H₂ _ Hb₁b₂ Hnc
                     | P b₁ b₂ c₁ c₂ H₁ _ Hb₁b₂ Hnc
                     | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'
                     | P Q x e c₂ H₂ _
                     | P Q x e c₁ H₁ _
                     | P Q b c₂ H₂ _ Hb
                     | P Q b c₁ H₁ _ Hb
                     | P Q R c₁₁ c₁₂ c₂ H₁ _ H₂ _
                     | P Q R c₁ c₂₁ c₂₂ H₁ _ H₂ _
                     | P Q b c₁₁ c₁₂ c₂ H₁ _ H₂ _ Hnc
                     | P Q b c₂₁ c₂₂ c₁ H₁ _ H₂ _ Hnc
                     | P Q c₁ c₂ H ].
      + inversion Hos'. clear Hos'. subst os'.
        exists 1, 1.
        repeat split ; auto using opt_state_eq_refl.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v₁ [Hv₁ [v₂ [Hv₂ HP]]]].
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_exp_var z e → V₁ z) as He.
        { intros z Hz. apply H'k₁. exists (Assign x e). simpl. auto. }
        assert (∀ z, In_exp_var z e' → V₂ z) as He'.
        { intros z Hz. apply H'k₂. exists (Assign y e'). simpl. auto. }
        assert (V₂ y) as Hy.
        { apply H'k₂. exists (Assign y e'). simpl. auto. }
        exists 1, 1. unfold nsteps. simpl in *.
        rewrite (disjoint_eval_exp e s V₁ V₂ y v₁) in Hv₂ ; auto.
        rewrite Hv₂ in *. simpl in *. rewrite Hv₁ in *. simpl in *.
        rewrite (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂),
                (disjoint_union_eval_exp_same e' s₂ s₁ s V₂ V₁) in * ; eauto.
        rewrite Hv₁, Hv₂ in *. simpl in *.
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct (Hb₁b₂ s HP₀) as [Hb₁ Hb₂].
        unfold assertion_of_bexp in Hb₁, Hb₂.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_bexp_var z b₁ → V₁ z) as H'b₁.
        { intros z Hz. apply Hk₁. exists (Assert b₁). simpl. auto. }
        assert (∀ z, In_bexp_var z b₂ → V₂ z) as H'b₂.
        { intros z Hz. apply Hk₂. exists (Assert b₂). simpl. auto. }
        exists 1, 1. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_bexp_same
                   b₂ _ _ _ V₂ V₁
                   (store_is_union_sym _ _ _ Hs)),
                 (disjoint_union_eval_bexp_same
                   b₁ _ _ _ V₁ V₂ Hs) in * ; eauto.
        rewrite Hb₁, Hb₂ in *.
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      + exists 1, 1.
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct (Hnc s HP₀) as [Hnc' _].
        apply (Hb₁b₂ s) in HP₀.
        destruct (eval_bexp s b₁) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc' ; auto ; fail).
        clear Hnc'.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
        { intros x Hx. apply Hk₁. exists (ITE b₁ c₁₁ c₁₂).
          simpl. auto. }
        assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
        { intros x Hx. apply Hk₂. exists (ITE b₂ c₂₁ c₂₂).
          simpl. auto. }
        exists 1, 1. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_bexp_same
                   b₁ _ _ _ V₁ V₂
                   Hs),
                (disjoint_union_eval_bexp_same
                   b₂ _ _ _ V₂ V₁
                   (store_is_union_sym _ _ _ Hs)) in * ; eauto.
        rewrite <- HP₀, Hv in *. simpl in *.
        destruct v ; inversion Hos' ; clear Hos' ; subst os' ;
        repeat split ; auto using opt_state_eq_refl.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct (Hnc s HP₀) as [Hnc' _].
        apply (Hb₁b₂ s) in HP₀.
        destruct (eval_bexp s b₁) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc' ; auto ; fail).
        clear Hnc'.
                destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
        { intros x Hx. apply Hk₁. exists (While b₁ c₁).
          simpl. auto. }
        assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
        { intros x Hx. apply Hk₂. exists (While b₂ c₂).
          simpl. auto. }
        exists 1, 1. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_bexp_same
                   b₁ _ _ _ V₁ V₂
                   Hs),
                (disjoint_union_eval_bexp_same
                   b₂ _ _ _ V₂ V₁
                   (store_is_union_sym _ _ _ Hs)) in * ; eauto.
        rewrite <- HP₀, Hv in *. simpl in *.
        destruct v ; inversion Hos' ; clear Hos' ; subst os' ;
        repeat split ; auto using opt_state_eq_refl.
      + apply IH₁ ; eauto using assertion_implies_trans.
        clear IH₁.
        eauto using proof_terms_consistent_1.
      (* Executing only one of the two programs *)
      (* Assign *)
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_exp_var z e → V₁ z) as He.
        { intros z Hz. apply H'k₁. exists (Assign x e). simpl. auto. }
        exists 1, 0. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂) in * ; eauto.
        rewrite Hv in *. simpl in *.
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_exp_var z e → V₂ z) as He.
        { intros z Hz. apply H'k₂. exists (Assign x e). simpl. auto. }
        exists 0, 1. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_exp_same e s₂ s₁ s V₂ V₁), Hv in * ; eauto.
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      (* Assert *)
      + apply HP₀P in HP₀. clear HP₀P.
        unfold assertion_of_bexp in Hb.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_bexp_var z b → V₁ z) as H'b.
        { intros z Hz. apply Hk₁. exists (Assert b). simpl. auto. }
        exists 1, 0. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₁ V₂ Hs), Hb in * ; eauto.
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      + apply HP₀P in HP₀. clear HP₀P.
        unfold assertion_of_bexp in Hb.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_bexp_var z b → V₂ z) as H'b.
        { intros z Hz. apply Hk₂. exists (Assert b). simpl. auto. }
        exists 0, 1. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₂ V₁ (store_is_union_sym _ _ _ Hs)), Hb in * ; eauto.
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      (* Seq *)
      + exists 1, 0. inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      + exists 0, 1. inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      (* ITE *)
      + apply HP₀P in HP₀. clear HP₀P.
        pose proof (Hnc s HP₀) as H'nc.
        simpl in *.
        clear Hnc. rename H'nc into Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc ; auto ; fail).
        clear Hnc.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b → V₁ x) as Hb₁.
        { intros x Hx. apply Hk₁. exists (ITE b c₁₁ c₁₂).
          simpl. auto. }
        exists 1, 0. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₁ V₂
                   Hs), Hv in * ; eauto.
        destruct v ; inversion Hos' ; clear Hos' ; subst os' ;
        repeat split ; auto using opt_state_eq_refl.
      + apply HP₀P in HP₀. clear HP₀P.
        pose proof (Hnc s HP₀) as H'nc.
        simpl in *.
        clear Hnc. rename H'nc into Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc ; auto ; fail).
        clear Hnc.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b → V₂ x) as Hb₁.
        { intros x Hx. apply Hk₂. exists (ITE b c₂₁ c₂₂).
          simpl. auto. }
        exists 0, 1. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₂ V₁
                   (store_is_union_sym _ _ _ Hs)), Hv in * ; eauto.
        destruct v ; inversion Hos' ; clear Hos' ; subst os' ;
        repeat split ; auto using opt_state_eq_refl.
      (* Self-composition *)
      + simpl in *.
        destruct (_split_hoare_seq_1 P Q genv c₁ c₂ H l')
          as [P' [Q' [R' [H₁ [H₂ [H' [HPP' H'']]]]]]] ; auto.
        rewrite H' in *.
        destruct (_step_helper_single_left_soundness
                    c₁ k₁' s₁ (c₂::k₂') s₂ s
                    (SingleHoareProofR H₂ :: l') H₁ os' Hdisj Hs)
          as [n₁ [n₂ [H₁' [H₂' H₃']]]] ; auto.
        exists n₁, n₂. repeat (split ; auto).
    (* This subtree only involves the left program *)
    - destruct H₁l as [HP₀P H₁l].
      simpl in *. remember Skip as c eqn:HSkip.
      induction Hc as [
                       P
                     | | | | |
                     | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'
                     | P Q x e c₂ H₂ _
                     | P Q x e c₁ H₁ _
                     | P Q b c₂ H₂ IH₂ Hb
                     | P Q b c₁ H₁ IH₁ Hb
                     | P Q R c₁₁ c₁₂ c₂ H₁ _ H₂ _
                     | P Q R c₁ c₂₁ c₂₂ H₁ _ H₂ _
                     | P Q b c₁₁ c₁₂ c₂ H₁ _ H₂ _ Hnc
                     | P Q b c₂₁ c₂₂ c₁ H₁ _ H₂ _ Hnc
                     | P Q c₁ c₂ H ] ;
        try discriminate HSkip ; try subst c₂ ;
          try (inversion Hos' ; clear Hos' ; subst os' ; exists 1, 0 ;
               repeat split ; eauto using opt_state_eq_refl).
      + apply IH₁ ; eauto using assertion_implies_trans.
        eauto using proof_terms_consistent_1.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_exp_var z e → V₁ z) as He.
        { intros z Hz. apply H'k₁. exists (Assign x e). simpl. auto. }
        exists 1, 0. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂), Hv in * ; eauto.
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      + apply HP₀P in HP₀. clear HP₀P.
        unfold assertion_of_bexp in Hb.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_bexp_var z b → V₁ z) as H'b.
        { intros z Hz. apply Hk₁. exists (Assert b). simpl. auto. }
        exists 1, 0. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₁ V₂ Hs), Hb in * ; eauto.
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      + apply HP₀P in HP₀. clear HP₀P.
        pose proof (Hnc s HP₀) as H'nc.
        simpl in *.
        clear Hnc. rename H'nc into Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc ; auto ; fail).
        clear Hnc.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b → V₁ x) as Hb₁.
        { intros x Hx. apply Hk₁. exists (ITE b c₁₁ c₁₂).
          simpl. auto. }
        exists 1, 0. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₁ V₂
                   Hs), Hv in * ; eauto.
        destruct v ; inversion Hos' ; clear Hos' ; subst os' ;
        repeat split ; auto using opt_state_eq_refl.
      (* Self-composition *)
      + simpl in *.
        destruct (_split_hoare_seq_1 P Q genv c₁ Skip H l')
          as [P' [Q' [R' [H₁ [H₂ [H' [HPP' H'']]]]]]] ; auto.
        rewrite H' in *.
        destruct (_step_helper_single_left_soundness
                    c₁ k₁' s₁ k₂' s₂ s
                    l' H₁ os' Hdisj Hs)
          as [n₁ [n₂ [H₁' [H₂' H₃']]]] ; auto.
        eauto.
    (* This subtree only involves the right program *)
    - destruct H₁l as [HP₀P H₁l].
      simpl in *. remember Skip as c eqn:HSkip.
      induction Hc as [
                       P
                     | | | | |
                     | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'
                     | P Q x e c₂ H₂ _
                     | P Q x e c₁ H₁ _
                     | P Q b c₂ H₂ _ Hb
                     | P Q b c₁ H₁ _ Hb
                     | P Q R c₁₁ c₁₂ c₂ H₁ _ H₂ _
                     | P Q R c₁ c₂₁ c₂₂ H₁ _ H₂ _
                     | P Q b c₁₁ c₁₂ c₂ H₁ _ H₂ _ Hnc
                     | P Q b c₂₁ c₂₂ c₁ H₁ _ H₂ _ Hnc
                     | P Q c₁ c₂ H ] ;
        try discriminate HSkip ; try subst c₁ ;
          try (inversion Hos' ; clear Hos' ; subst os' ; exists 0, 1 ;
               repeat split ; eauto using opt_state_eq_refl).
      + apply IH₁ ; eauto using assertion_implies_trans.
        eauto using proof_terms_consistent_1.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_exp_var z e → V₂ z) as He.
        { intros z Hz. apply H'k₂. exists (Assign x e). simpl. auto. }
        exists 0, 1. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_exp_same e s₂ s₁ s V₂ V₁), Hv in * ; eauto.
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      + apply HP₀P in HP₀. clear HP₀P.
        unfold assertion_of_bexp in Hb.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_bexp_var z b → V₂ z) as H'b.
        { intros z Hz. apply Hk₂. exists (Assert b). simpl. auto. }
        exists 0, 1. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₂ V₁ (store_is_union_sym _ _ _ Hs)), Hb in * ; eauto.
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; auto using opt_state_eq_refl.
      + apply HP₀P in HP₀. clear HP₀P.
        pose proof (Hnc s HP₀) as H'nc.
        simpl in *.
        clear Hnc. rename H'nc into Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc ; auto ; fail).
        clear Hnc.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b → V₂ x) as Hb₁.
        { intros x Hx. apply Hk₂. exists (ITE b c₂₁ c₂₂).
          simpl. auto. }
        exists 0, 1. unfold nsteps. simpl in *.
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₂ V₁
                   (store_is_union_sym _ _ _ Hs)), Hv in * ; eauto.
        destruct v ; inversion Hos' ; clear Hos' ; subst os' ;
        repeat split ; auto using opt_state_eq_refl.
      (* Self-composition *)
      + simpl in *.
        destruct (_split_hoare_seq_1 P Q genv Skip c₂ H l')
          as [P' [Q' [R' [H₁ [H₂ [H' [HPP' H'']]]]]]] ; auto.
        rewrite H' in *.
        destruct (_step_helper_single_right_soundness
                    k₁' s₁ c₂ k₂' s₂ s
                    l' H₂ os' Hdisj Hs)
          as [n₁ [n₂ [H₁' [H₂' H₃']]]] ; auto.
        * apply (hoare_close_skip P') ; auto.
        * eauto.
    - simpl in *.
      destruct (_step_helper_single_left_soundness
                  c₁ k₁' s₁ k₂' s₂ s
                  l' Hc os' Hdisj Hs)
        as [n₁ [n₂ [H₁' [H₂' H₃']]]] ; auto.
      * inversion H₁l. auto.
      * eauto.
    - simpl in *.
      destruct (_step_helper_single_right_soundness
                  k₁' s₁ c₂ k₂' s₂ s
                  l' Hc os' Hdisj Hs)
        as [n₁ [n₂ [H₁' [H₂' H₃']]]] ; auto.
      * inversion H₁l. auto.
      * eauto.
  Qed.

  Lemma close_skip_skip:
    ∀ P Q s (H : core_rhl_sc_proof P Skip Skip Q),
      P s → Q s.
  Proof.
    intros P Q s H.
    remember Skip as c  eqn:Hc  in H at 1.
    remember Skip as c' eqn:Hc' in H at 1.
    induction H ;
      try discriminate Hc ; try discriminate Hc' ;
        try subst c₁ ; try subst c₂ ; try clear Hc ; try clear Hc' ;
          auto.
    intro HP.
    apply (hoare_close_skip_skip P) ; auto.
  Qed.

  Lemma _step_helper_single_left_invariant:
    ∀ genv P Q c₁ k₁ s₁ k₂ s₂ s l' (Hc : hoare_proof P c₁ Q) os',
      _step_helper_single_left (c₁ :: k₁, s₁) (k₂, s₂) s Hc l' = Some os' →
      P s →
      disjoint_states (c₁::k₁, s₁) (k₂, s₂) →
      store_is_union s s₁ s₂ →
      proof_terms_consistent l' Q genv →
      proof_terms_match k₁ k₂ l' →
      _invariant genv os'.
  Proof.
    intros genv P Q c₁ k₁ s₁ k₂ s₂ s l' Hc.
    induction Hc as [ P | P x e | P b H'
                      | P Q R c₁ c₂ H₁ _ H₂ _
                      | P Q b c₁ c₂ H₁ _ H₂ _ Hnc
                      | P b c H' _ Hnc
                      | P P' Q Q'' c H' IH ] ;
      intros os' H HP Hdisj Hs₁s₂ Hl' Hk₁k₂.
    - simpl in H. inversion H. clear H. subst os'.
      simpl. repeat (split ; auto).
      pose proof (step_disjoint_states _ (k₁, s₁) _ Hdisj) as H.
      apply H. auto.
    - destruct HP as [v [Hv HP]].
      assert (eval_exp s₁ e = eval_exp s e) as He.
      { destruct Hdisj as [V₁ [V₂ [Hk₁' [Hk₂' [Hs₁ [Hs₂ Hdisj']]]]]].
        apply (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂) ; eauto.
        intros z Hz. apply Hk₁'. exists (Assign x e). simpl. auto. }
      assert (step υ (Assign x e :: k₁, s₁) = Some (k₁, M.add x v s₁)) as HS₁'.
      { simpl. rewrite He, Hv. reflexivity. }
      destruct (step_disjoint_union_states _ _ s (M.add x v s) k₁ _ Hdisj Hs₁s₂ HS₁')
        as [Hk₁'' [Hs₁'s₂ Hdisj']].
      { simpl. rewrite Hv. reflexivity. }
      simpl in H. rewrite He, Hv in H. inversion H. clear H. subst os'.
      repeat (split ; auto).
    - pose proof (H' s HP) as Hv.
      unfold assertion_of_bexp in Hv.
      assert (eval_bexp s₁ b = eval_bexp s b) as Hb.
      { destruct Hdisj as [V₁ [V₂ [Hk₁' [Hk₂' [Hs₁ [Hs₂ Hdisj']]]]]].
        apply (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) ; eauto.
        intros z Hz. apply Hk₁'. exists (Assert b). simpl. auto. }
      assert (step υ (Assert b :: k₁, s₁) = Some (k₁, s₁)) as HS₁'.
      { simpl. rewrite Hb, Hv. reflexivity. }
      destruct (step_disjoint_union_states _ _ _ s k₁ _ Hdisj Hs₁s₂ HS₁')
        as [Hk₁'' [Hs₁'s₂ Hdisj']].
      { simpl. rewrite Hv. reflexivity. }
      simpl in H. rewrite Hb, Hv in H. inversion H. clear H. subst os'.
      repeat (split ; auto).
    - simpl in H. inversion H. clear H. subst os'.
      simpl. repeat (split ; auto).
      pose proof (step_disjoint_states _ (c₁::c₂::k₁, s₁) _ Hdisj) as H.
      apply H. auto.
    - pose proof (Hnc s HP) as Hnc'. simpl in H. clear Hnc. simpl in Hnc'.
      assert (eval_bexp s₁ b = eval_bexp s b) as Hb.
      { destruct Hdisj as [V₁ [V₂ [Hk₁' [Hk₂' [Hs₁ [Hs₂ Hdisj']]]]]].
        apply (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) ; eauto.
        intros z Hz. apply Hk₁'. exists (ITE b c₁ c₂). simpl. auto. }
      rewrite Hb in H.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ; try contradiction.
      assert (step υ (ITE b c₁ c₂ :: k₁, s₁) = Some ((if v then c₁ else c₂) :: k₁, s₁))
        as HS₁'.
      { simpl. rewrite Hb. reflexivity. }
      destruct (step_disjoint_union_states
                  _ _ _ s ((if v then c₁ else c₂)::k₁) _ Hdisj Hs₁s₂ HS₁')
        as [Hk₁'' [Hs₁'s₂ Hdisj']].
      { simpl. rewrite Hv. simpl. reflexivity. }
      simpl in H.
      destruct v ; inversion H ; clear H ; subst os' ; simpl in * ;
        repeat (split ; auto).
      unfold assertion_of_bexp. simpl. rewrite Hv. reflexivity.
    - pose proof (Hnc s HP) as Hnc'. simpl in H. simpl in Hnc'.
      assert (eval_bexp s₁ b = eval_bexp s b) as Hb.
      { destruct Hdisj as [V₁ [V₂ [Hk₁' [Hk₂' [Hs₁ [Hs₂ Hdisj']]]]]].
        apply (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂) ; eauto.
        intros z Hz. apply Hk₁'. exists (While b c). simpl. auto. }
      rewrite Hb in H.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ; try contradiction.
      simpl in H.
      destruct v.
      + inversion H. clear H. subst os'.
        assert (step υ (While b c :: k₁, s₁) = Some (c :: While b c :: k₁, s₁))
          as HS₁'.
        { simpl. rewrite Hb. reflexivity. }
        destruct (step_disjoint_union_states
                    _ _ _ s (c :: While b c :: k₁) _ Hdisj Hs₁s₂ HS₁')
          as [Hk₁'' [Hs₁'s₂ Hdisj']].
        { simpl. rewrite Hv. simpl. reflexivity. }
        simpl. repeat (split ; auto).
      + inversion H. clear H. subst os'.
        assert (step υ (While b c :: k₁, s₁) = Some (k₁, s₁))
          as HS₁'.
        { simpl. rewrite Hb. reflexivity. }
        destruct (step_disjoint_union_states
                    _ _ _ s k₁ _ Hdisj Hs₁s₂ HS₁')
          as [Hk₁'' [Hs₁'s₂ Hdisj']].
        { simpl. rewrite Hv. simpl. reflexivity. }
        simpl. repeat (split ; auto).
        unfold assertion_of_bexp. simpl. rewrite Hv. reflexivity.
    - apply IH ; auto.
      eauto using proof_terms_consistent_1.
  Qed.

  Lemma _step_helper_single_right_invariant:
    ∀ genv P Q k₁ s₁ c₂ k₂ s₂ s l' (Hc : hoare_proof P c₂ Q) os',
      _step_helper_single_right (k₁, s₁) (c₂ :: k₂, s₂) s Hc l' = Some os' →
      P s →
      disjoint_states (k₁, s₁) (c₂::k₂, s₂) →
      store_is_union s s₁ s₂ →
      proof_terms_consistent l' Q genv →
      proof_terms_match k₁ k₂ l' →
      _invariant genv os'.
  Proof.
    intros genv P Q k₁ s₁ c₂ k₂ s₂ s l' Hc.
    induction Hc as [ P | P x e | P b H'
                      | P Q R c₁ c₂ H₁ _ H₂ _
                      | P Q b c₁ c₂ H₁ _ H₂ _ Hnc
                      | P b c H' _ Hnc
                      | P P' Q Q'' c H' IH ] ;
      intros os' H HP Hdisj Hs₁s₂ Hl' Hk₁k₂.
    - simpl in H. inversion H. clear H. subst os'.
      simpl. repeat (split ; auto).
      apply disjoint_states_sym in Hdisj.
      pose proof (step_disjoint_states _ (k₂, s₂) _ Hdisj) as H.
      apply disjoint_states_sym in H ; auto.
    - destruct HP as [v [Hv HP]].
      assert (eval_exp s₂ e = eval_exp s e) as He.
      { destruct Hdisj as [V₁ [V₂ [Hk₁' [Hk₂' [Hs₁ [Hs₂ Hdisj']]]]]].
        apply (disjoint_union_eval_exp_same e s₂ s₁ s V₂ V₁) ; eauto.
        intros z Hz. apply Hk₂'. exists (Assign x e). simpl. auto. }
      assert (step υ (Assign x e :: k₂, s₂) = Some (k₂, M.add x v s₂)) as HS₂'.
      { simpl. rewrite He, Hv. reflexivity. }
      destruct (step_disjoint_union_states
                  _ _ s (M.add x v s) k₂ _ (disjoint_states_sym _ _ Hdisj)
                  (store_is_union_sym _ _ _ Hs₁s₂) HS₂')
        as [Hk₂'' [Hs₂'s₁ Hdisj']].
      { simpl. rewrite Hv. reflexivity. }
      simpl in H. rewrite He, Hv in H. inversion H. clear H. subst os'.
      repeat (split ; auto).
    - pose proof (H' s HP) as Hv.
      unfold assertion_of_bexp in Hv.
      assert (eval_bexp s₂ b = eval_bexp s b) as Hb.
      { destruct Hdisj as [V₁ [V₂ [Hk₁' [Hk₂' [Hs₁ [Hs₂ Hdisj']]]]]].
        apply (disjoint_union_eval_bexp_same b s₂ s₁ s V₂ V₁) ; eauto.
        intros z Hz. apply Hk₂'. exists (Assert b). simpl. auto. }
      assert (step υ (Assert b :: k₂, s₂) = Some (k₂, s₂)) as HS₂'.
      { simpl. rewrite Hb, Hv. reflexivity. }
      destruct (step_disjoint_union_states
                  _ _ _ s k₂ _ (disjoint_states_sym _ _ Hdisj)
                  (store_is_union_sym _ _ _ Hs₁s₂) HS₂')
        as [Hk₂'' [Hs₂'s₁ Hdisj']].
      { simpl. rewrite Hv. reflexivity. }
      simpl in H. rewrite Hb, Hv in H. inversion H. clear H. subst os'.
      repeat (split ; auto).
    - simpl in H. inversion H. clear H. subst os'.
      simpl. repeat (split ; auto).
      apply disjoint_states_sym in Hdisj.
      pose proof (step_disjoint_states _ (c₁::c₂::k₂, s₂) _ Hdisj) as H.
      apply disjoint_states_sym in H ; auto.
    - pose proof (Hnc s HP) as Hnc'. simpl in H. clear Hnc. simpl in Hnc'.
      assert (eval_bexp s₂ b = eval_bexp s b) as Hb.
      { destruct Hdisj as [V₁ [V₂ [Hk₁' [Hk₂' [Hs₁ [Hs₂ Hdisj']]]]]].
        apply (disjoint_union_eval_bexp_same b s₂ s₁ s V₂ V₁) ; eauto.
        intros z Hz. apply Hk₂'. exists (ITE b c₁ c₂). simpl. auto. }
      rewrite Hb in H.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ; try contradiction.
      assert (step υ (ITE b c₁ c₂ :: k₂, s₂) = Some ((if v then c₁ else c₂) :: k₂, s₂))
        as HS₂'.
      { simpl. rewrite Hb. reflexivity. }
      destruct (step_disjoint_union_states
                  _ _ _ s
                  ((if v then c₁ else c₂)::k₂) _
                  (disjoint_states_sym _ _ Hdisj)
                  (store_is_union_sym _ _ _ Hs₁s₂) HS₂')
        as [Hk₂'' [Hs₂'s₁ Hdisj']].
      { simpl. rewrite Hv. simpl. reflexivity. }
      simpl in H.
      apply disjoint_states_sym in Hdisj'.
      destruct v ; inversion H ; clear H ; subst os' ; simpl in * ;
        repeat (split ; auto).
      unfold assertion_of_bexp. simpl. rewrite Hv. reflexivity.
    - pose proof (Hnc s HP) as Hnc'. simpl in H. simpl in Hnc'.
      assert (eval_bexp s₂ b = eval_bexp s b) as Hb.
      { destruct Hdisj as [V₁ [V₂ [Hk₁' [Hk₂' [Hs₁ [Hs₂ Hdisj']]]]]].
        apply (disjoint_union_eval_bexp_same b s₂ s₁ s V₂ V₁) ; eauto.
        intros z Hz. apply Hk₂'. exists (While b c). simpl. auto. }
      rewrite Hb in H.
      destruct (eval_bexp s b) as [v | ] eqn:Hv ; try contradiction.
      simpl in H.
      destruct v.
      + inversion H. clear H. subst os'.
        assert (step υ (While b c :: k₂, s₂) = Some (c :: While b c :: k₂, s₂))
          as HS₂'.
        { simpl. rewrite Hb. reflexivity. }
        destruct (step_disjoint_union_states
                    _ _ _ s (c :: While b c :: k₂) _
                    (disjoint_states_sym _ _ Hdisj)
                    (store_is_union_sym _ _ _ Hs₁s₂) HS₂')
          as [Hk₂'' [Hs₂'s₁ Hdisj']].
        { simpl. rewrite Hv. simpl. reflexivity. }
        apply disjoint_states_sym in Hdisj'.
        simpl. repeat (split ; auto).
      + inversion H. clear H. subst os'.
        assert (step υ (While b c :: k₂, s₂) = Some (k₂, s₂))
          as HS₂'.
        { simpl. rewrite Hb. reflexivity. }
        destruct (step_disjoint_union_states
                    _ _ _ s k₂ _
                    (disjoint_states_sym _ _ Hdisj)
                    (store_is_union_sym _ _ _ Hs₁s₂) HS₂')
          as [Hk₂'' [Hs₂'s₁ Hdisj']].
        { simpl. rewrite Hv. simpl. reflexivity. }
        apply disjoint_states_sym in Hdisj'.
        simpl. repeat (split ; auto).
        unfold assertion_of_bexp. simpl. rewrite Hv. reflexivity.
    - apply IH ; auto.
      eauto using proof_terms_consistent_1.
  Qed.

  Lemma _invariant_1:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      _invariant genv os'.
  Proof.
    intros or [[k₁ s₁] [k₂ s₂] s P₀ l] os' Hinv H.
    destruct Hinv as [HP₀ [Hs₁s₂ [Hdisj [H₁l H₂l]]]].
    simpl fst in *. simpl snd in *.
    inversion H₂l as [
                    | c₁ k₁' c₂ k₂' l' P Q Hc H₂l'
                    | c₁ k₁' k₂' l' P Q Hc H₂l'
                    | k₁' c₂ k₂' l' P Q Hc H₂l'
                    | c₁ k₁' k₂' l' P Q Hc H₂l'
                    | k₁' c₂ k₂' l' P Q Hc H₂l' ] ;
      subst k₁ k₂ l ; clear H₂l.
    - simpl in H. inversion H.
    - unfold _step in H.
      simpl in H₁l.
      destruct H₁l as [HP₀P H₁l].
      induction Hc as [
                       P
                     | P x e y e'
                     | P b₁ b₂ Hb₁b₂
                     | P Q R c₁₁ c₂₁ c₁₂ c₂₂ H₁ _ H₂ _
                     | P Q b₁ b₂ c₁₁ c₂₁ c₁₂ c₂₂ H₁ _ H₂ _ Hb₁b₂ Hnc
                     | P b₁ b₂ c₁ c₂ H₁ _ Hb₁b₂ Hnc
                     | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'
                     | P Q x e c₂ H₂ _
                     | P Q x e c₁ H₁ _
                     | P Q b c₂ H₂ _ Hb
                     | P Q b c₁ H₁ _ Hb
                     | P Q R c₁₁ c₁₂ c₂ H₁ _ H₂ _
                     | P Q R c₁ c₂₁ c₂₂ H₁ _ H₂ _
                     | P Q b c₁₁ c₁₂ c₂ H₁ _ H₂ _ Hnc
                     | P Q b c₂₁ c₂₂ c₁ H₁ _ H₂ _ Hnc
                     | P Q c₁ c₂ H' ].
      + simpl in H. inversion H. clear H. subst os'.
        simpl.
        apply HP₀P in HP₀.
        repeat (split ; eauto using step_2_disjoint_states).
        apply (step_2_disjoint_states _ (k₁', s₁) _ (k₂', s₂) Hdisj) ;
          reflexivity.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v₁ [Hv₁ HP]].
        destruct HP  as [v₂ [Hv₂ HP]].
        unfold _step_helper in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct (step υ (_, s₂)) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        destruct S₁' as [k₁'' s₁'], S₂' as [k₂'' s₂'].
        destruct (step_disjoint_union_states
                    _ _ s (M.add y v₁ s) k₂' _
                    (disjoint_states_sym _ _ Hdisj)
                    (store_is_union_sym _ _ _ Hs₁s₂) HS₂') as [Hk₂'' [Hs₂'s₁ Hdisj']].
        { simpl. rewrite Hv₁. simpl. reflexivity. }
        simpl in Hk₂''. subst k₂''.
        simpl fst in *. simpl snd in *.
        destruct (step_disjoint_union_states
                    _ _ (M.add y v₁ s) (M.add x v₂ (M.add y v₁ s)) k₁' _
                    (disjoint_states_sym _ _ Hdisj')
                    (store_is_union_sym _ _ _ Hs₂'s₁) HS₁') as [Hk₁'' [Hs₁'s₂' Hdisj'']].
        { simpl. rewrite Hv₂. simpl. reflexivity. }
        simpl in Hk₁''. subst k₁''.
        simpl fst in *. simpl snd in *.
        rewrite Hv₁ in H.
        destruct Hdisj as [V₁ [V₂ [Hk₁' [Hk₂' [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z, In_exp_var z e → V₁ z) as He.
        { intros z Hz. apply Hk₁'. exists (Assign x e). simpl. auto. }
        assert (V₂ y) as Hy.
        { apply Hk₂'. exists (Assign y e'). simpl. auto. }
        rewrite <- (disjoint_eval_exp e s V₁ V₂ y v₁) in H ; eauto.
        rewrite Hv₂ in H. inversion H. clear H. subst os'.
        simpl.
        repeat (split ; auto).
      + apply HP₀P in HP₀.
        unfold _step_helper in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct (step υ (_, s₂)) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        destruct S₁' as [k₁'' s₁'], S₂' as [k₂'' s₂'].
        simpl in H. inversion H. clear H. subst os'.
        pose proof (step_2_disjoint_states _ _ _ _ Hdisj HS₁' HS₂') as Hdisj'.
        simpl in HS₁', HS₂'.
        destruct (eval_bexp s₁ b₁) as [[|]|], (eval_bexp s₂ b₂) as [[|]|] ;
          try discriminate HS₁' ; try discriminate HS₂'.
        inversion HS₁' as [H'S₁']. clear HS₁'. subst k₁'' s₁'.
        inversion HS₂' as [H'S₂']. clear HS₂'. subst k₂'' s₂'.
        repeat (split ; auto).
      + unfold _step_helper in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct (step υ (_, s₂)) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        destruct S₁' as [k₁'' s₁'], S₂' as [k₂'' s₂'].
        inversion H. clear H. subst os'.
        pose proof (step_2_disjoint_states _ _ _ _ Hdisj HS₁' HS₂') as Hdisj'.
        inversion HS₁' as [H'S₁']. clear HS₁'. subst k₁'' s₁'.
        inversion HS₂' as [H'S₂']. clear HS₂'. subst k₂'' s₂'.
        simpl.
        repeat (split ; auto).
      + unfold _step_helper in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct (step υ (_, s₂)) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        destruct S₁' as [k₁'' s₁'], S₂' as [k₂'' s₂'].
        pose proof (step_2_disjoint_states _ _ _ _ Hdisj HS₁' HS₂') as Hdisj'.
        apply HP₀P in HP₀.
        simpl in H.
        pose proof (Hb₁b₂ s HP₀) as H'b₁b₂. clear Hb₁b₂. simpl in H'b₁b₂.
        destruct (Hnc s HP₀) as [Hnc' _]. clear Hnc.
        destruct (eval_bexp s b₁) as [v₁ | ] eqn:Hv₁ ;
          try contradiction. clear Hnc'.
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        simpl in HS₁', HS₂'.
        assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
        { intros x Hx. apply H'k₁. exists (ITE b₁ c₁₁ c₁₂). simpl. auto. }
        assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
        { intros x Hx. apply H'k₂. exists (ITE b₂ c₂₁ c₂₂). simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same b₂ s₂ s₁ s V₂ V₁), <- H'b₁b₂ in HS₂' ;
          eauto.
        rewrite (disjoint_union_eval_bexp_same b₁ s₁ s₂ s V₁ V₂), Hv₁ in HS₁' ;
          eauto.
        simpl in HS₁', HS₂'.
        destruct v₁.
        * inversion HS₁'. inversion HS₂'. inversion H. clear HS₁' HS₂' H.
          subst k₁'' k₂'' s₁' s₂' os'.
          simpl.
          repeat (split ; eauto).
        * inversion HS₁'. inversion HS₂'. inversion H. clear HS₁' HS₂' H.
          subst k₁'' k₂'' s₁' s₂' os'.
          simpl.
          repeat (split ; eauto).
          unfold assertion_of_bexp. simpl. rewrite Hv₁. simpl. reflexivity.
      + unfold _step_helper in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct (step υ (_, s₂)) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        destruct S₁' as [k₁'' s₁'], S₂' as [k₂'' s₂'].
        pose proof (step_2_disjoint_states _ _ _ _ Hdisj HS₁' HS₂') as Hdisj'.
        apply HP₀P in HP₀.
        pose proof (Hb₁b₂ s HP₀) as H'b₁b₂. simpl in H'b₁b₂.
        destruct (Hnc s HP₀) as [Hnc' _].
        destruct (eval_bexp s b₁) as [v₁ | ] eqn:Hv₁ ;
          try contradiction. clear Hnc'.
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        simpl in HS₁', HS₂'.
        assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
        { intros x Hx. apply H'k₁. exists (While b₁ c₁). simpl. auto. }
        assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
        { intros x Hx. apply H'k₂. exists (While b₂ c₂). simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same b₂ s₂ s₁ s V₂ V₁), <- H'b₁b₂ in HS₂' ;
          eauto.
        rewrite (disjoint_union_eval_bexp_same b₁ s₁ s₂ s V₁ V₂), Hv₁ in HS₁' ;
          eauto.
        simpl in HS₁', HS₂'.
        destruct v₁.
        * inversion HS₁'. inversion HS₂'. inversion H. clear HS₁' HS₂' H.
          subst k₁'' k₂'' s₁' s₂' os'.
          simpl.
          repeat (split ; eauto).
        * inversion HS₁'. inversion HS₂'. inversion H. clear HS₁' HS₂' H.
          subst k₁'' k₂'' s₁' s₂' os'.
          simpl.
          repeat (split ; eauto).
          unfold assertion_of_bexp. simpl. rewrite Hv₁. simpl. reflexivity.
      + apply IH₁ ; clear IH₁ ; eauto using proof_terms_consistent_1.
      (* Executing only one of the two programs *)
      (* Assign *)
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        unfold _step_helper in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct S₁' as [k₁'' s₁'].
        destruct (step_disjoint_union_states
                    _ _ s (M.add x v s) k₁' _
                    Hdisj
                    Hs₁s₂ HS₁') as [Hk₁'' [Hs₁'s₂' Hdisj'']].
        { simpl. rewrite Hv. simpl. reflexivity. }
        simpl in Hk₁''. subst k₁''.
        simpl fst in *. simpl snd in *.
        rewrite Hv in H. inversion H. clear H. subst os'.
        simpl.
        repeat (split ; auto).
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        unfold _step_helper in H.
        destruct (step υ _) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        unfold bind in H.
        destruct S₂' as [k₂'' s₂'].
        destruct (step_disjoint_union_states
                    _ _ s (M.add x v s) k₂' _
                    (disjoint_states_sym _ _ Hdisj)
                    (store_is_union_sym _ _ _ Hs₁s₂) HS₂') as [Hk₂'' [Hs₁'s₂' Hdisj'']].
        { simpl. rewrite Hv. simpl. reflexivity. }
        simpl in Hk₂''. subst k₂''.
        simpl fst in *. simpl snd in *.
        rewrite Hv in H. inversion H. clear H. subst os'.
        simpl.
        apply disjoint_states_sym in Hdisj''.
        repeat (split ; auto).
      (* Assert *)
      + apply HP₀P in HP₀.
        unfold _step_helper in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct S₁' as [k₁'' s₁'].
        simpl in H. inversion H. clear H. subst os'.
        pose proof (step_disjoint_states _ _ _ Hdisj HS₁') as Hdisj'.
        simpl in HS₁'.
        destruct (eval_bexp s₁ b) as [[|]|] ;
          try discriminate HS₁'.
        inversion HS₁' as [H'S₁']. clear HS₁'. subst k₁'' s₁'.
        simpl.
        repeat (split ; auto).
      + apply HP₀P in HP₀.
        unfold _step_helper in H.
        destruct (step υ _) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        unfold bind in H.
        destruct S₂' as [k₂'' s₂'].
        simpl in H. inversion H. clear H. subst os'.
        pose proof (step_disjoint_states _ _ _ (disjoint_states_sym _ _ Hdisj) HS₂')
          as Hdisj'.
        simpl in HS₂'.
        destruct (eval_bexp s₂ b) as [[|]|] ;
          try discriminate HS₂'.
        inversion HS₂' as [H'S₂']. clear HS₂'. subst k₂'' s₂'.
        simpl.
        apply disjoint_states_sym in Hdisj'.
        repeat (split ; auto).
      (* Seq *)
      + unfold _step_helper in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct S₁' as [k₁'' s₁'].
        inversion H. clear H. subst os'.
        pose proof (step_disjoint_states _ _ _ Hdisj HS₁') as Hdisj'.
        inversion HS₁' as [H'S₁']. clear HS₁'. subst k₁'' s₁'.
        simpl.
        repeat (split ; auto).
      + unfold _step_helper in H.
        destruct (step υ _) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        unfold bind in H.
        destruct S₂' as [k₂'' s₂'].
        inversion H. clear H. subst os'.
        pose proof (step_disjoint_states _ _ _ (disjoint_states_sym _ _ Hdisj) HS₂')
          as Hdisj'.
        apply disjoint_states_sym in Hdisj'.
        inversion HS₂' as [H'S₂']. clear HS₂'. subst k₂'' s₂'.
        simpl.
        repeat (split ; auto).
      (* ITE *)
      + unfold _step_helper in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct S₁' as [k₁'' s₁'].
        pose proof (step_disjoint_states _ _ _ Hdisj HS₁') as Hdisj'.
        apply HP₀P in HP₀.
        pose proof (Hnc s HP₀) as Hnc'. clear Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try contradiction. clear Hnc'.
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        simpl in HS₁'.
        assert (∀ x, In_bexp_var x b → V₁ x) as Hb₁.
        { intros x Hx. apply H'k₁. exists (ITE b c₁₁ c₁₂). simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂), Hv in HS₁' ;
          eauto.
        simpl in HS₁'. inversion HS₁'. clear HS₁'. subst k₁'' s₁'.
        destruct v.
        * inversion H. clear H.
          subst os'.
          simpl.
          repeat (split ; eauto).
        * inversion H. clear H.
          subst os'.
          simpl.
          repeat (split ; eauto).
          unfold assertion_of_bexp. simpl. rewrite Hv. simpl. reflexivity.
      + unfold _step_helper in H.
        destruct (step υ _) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        unfold bind in H.
        destruct S₂' as [k₂'' s₂'].
        pose proof (step_disjoint_states _ _ _ (disjoint_states_sym _ _ Hdisj) HS₂')
          as Hdisj'.
        apply HP₀P in HP₀.
        pose proof (Hnc s HP₀) as Hnc'. clear Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try contradiction. clear Hnc'.
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        simpl in HS₂'.
        assert (∀ x, In_bexp_var x b → V₂ x) as Hb₁.
        { intros x Hx. apply H'k₂. exists (ITE b c₂₁ c₂₂). simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same b s₂ s₁ s V₂ V₁), Hv in HS₂' ;
          eauto.
        simpl in HS₂'. inversion HS₂'. clear HS₂'. subst k₂'' s₂'.
        apply disjoint_states_sym in Hdisj'.
        destruct v.
        * inversion H. clear H.
          subst os'.
          simpl.
          repeat (split ; eauto).
        * inversion H. clear H.
          subst os'.
          simpl.
          repeat (split ; eauto).
          unfold assertion_of_bexp. simpl. rewrite Hv. simpl. reflexivity.
      (* Self-comp *)
      + simpl in H.
        destruct (_split_hoare_seq_1 P Q or _ _ H' l')
          as [P' [Q' [R' [H₁ [H₂ [H'' [HPP' H''']]]]]]] ; auto.
        rewrite H'' in H.
        apply (_step_helper_single_left_invariant or P' Q' _ _ _ _ _ _ _ _ _ H) ; auto.
        simpl. split ; auto.
    (* This subtree only involves the left program *)
    - unfold _step in H.
      simpl in H₁l.
      destruct H₁l as [HP₀P H₁l].
      remember Skip as c eqn:HSkip.
      induction Hc as [
                       P
                     | | | | |
                     | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'
                     | P Q x e c₂ H₂ _
                     | P Q x e c₁ H₁ _
                     | P Q b c₂ H₂ IH₂ Hb
                     | P Q b c₁ H₁ IH₁ Hb
                     | P Q R c₁₁ c₁₂ c₂ H₁ _ H₂ _
                     | P Q R c₁ c₂₁ c₂₂ H₁ _ H₂ _
                     | P Q b c₁₁ c₁₂ c₂ H₁ _ H₂ _ Hnc
                     | P Q b c₂₁ c₂₂ c₁ H₁ _ H₂ _ Hnc
                     | P Q c₁ c₂ H' ] ;
        try discriminate HSkip ; try subst c₂ ; try clear HSkip.
      + simpl in H. inversion H. clear H. subst os'.
        simpl.
        apply HP₀P in HP₀.
        repeat (split ; eauto).
        apply (step_disjoint_states _ (k₁', s₁) _  Hdisj) ;
          reflexivity.
      + apply IH₁ ; clear IH₁ ; eauto using proof_terms_consistent_1.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        unfold _step_helper_left in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct S₁' as [k₁'' s₁'].
        destruct (step_disjoint_union_states
                    _ _ s (M.add x v s) k₁' _
                    Hdisj
                    Hs₁s₂ HS₁') as [Hk₁'' [Hs₁'s₂' Hdisj'']].
        { simpl. rewrite Hv. simpl. reflexivity. }
        simpl in Hk₁''. subst k₁''.
        simpl fst in *. simpl snd in *.
        rewrite Hv in H. inversion H. clear H. subst os'.
        simpl.
        repeat (split ; auto).
        apply (close_skip_skip _ _ _ H₂) ; auto.
      + apply HP₀P in HP₀.
        unfold _step_helper_left in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct S₁' as [k₁'' s₁'].
        simpl in H. inversion H. clear H. subst os'.
        pose proof (step_disjoint_states _ _ _ Hdisj HS₁') as Hdisj'.
        simpl in HS₁'.
        destruct (eval_bexp s₁ b) as [[|]|] ;
          try discriminate HS₁'.
        inversion HS₁' as [H'S₁']. clear HS₁'. subst k₁'' s₁'.
        simpl.
        repeat (split ; auto).
        apply (close_skip_skip _ _ _ H₂) ; auto.
      + unfold _step_helper_left in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct S₁' as [k₁'' s₁'].
        inversion H. clear H. subst os'.
        pose proof (step_disjoint_states _ _ _ Hdisj HS₁') as Hdisj'.
        inversion HS₁' as [H'S₁']. clear HS₁'. subst k₁'' s₁'.
        simpl.
        repeat (split ; auto).
      + unfold _step_helper_left in H.
        destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
          try discriminate H.
        unfold bind in H.
        destruct S₁' as [k₁'' s₁'].
        pose proof (step_disjoint_states _ _ _ Hdisj HS₁') as Hdisj'.
        apply HP₀P in HP₀.
        pose proof (Hnc s HP₀) as Hnc'. clear Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try contradiction. clear Hnc'.
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        simpl in HS₁'.
        assert (∀ x, In_bexp_var x b → V₁ x) as Hb₁.
        { intros x Hx. apply H'k₁. exists (ITE b c₁₁ c₁₂). simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same b s₁ s₂ s V₁ V₂), Hv in HS₁' ;
          eauto.
        simpl in HS₁'. inversion HS₁'. clear HS₁'. subst k₁'' s₁'.
        destruct v.
        * inversion H. clear H.
          subst os'.
          simpl.
          repeat (split ; eauto).
        * inversion H. clear H.
          subst os'.
          simpl.
          repeat (split ; eauto).
          unfold assertion_of_bexp. simpl. rewrite Hv. simpl. reflexivity.
      (* Self-comp *)
      + simpl in H.
        destruct (_split_hoare_seq_1 P Q or _ _ H' l')
          as [P' [Q' [R' [H₁ [H₂ [H'' [HPP' H''']]]]]]] ; auto.
        rewrite H'' in H.
        apply (_step_helper_single_left_invariant _ P' Q' _ _ _ _ _ _ _ _ _ H) ; auto.
        apply (proof_terms_consistent_1 _ _ R') ; auto.
        intro s'. apply (hoare_close_skip _ _ s' H₂).
    (* This subtree only involves the right program *)
    - unfold _step in H.
      simpl in H₁l.
      destruct H₁l as [HP₀P H₁l].
      remember Skip as c eqn:HSkip.
      induction Hc as [
                       P
                     | | | | |
                     | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'
                     | P Q x e c₂ H₂ _
                     | P Q x e c₁ H₁ _
                     | P Q b c₂ H₂ IH₂ Hb
                     | P Q b c₁ H₁ IH₁ Hb
                     | P Q R c₁₁ c₁₂ c₂ H₁ _ H₂ _
                     | P Q R c₁ c₂₁ c₂₂ H₁ _ H₂ _
                     | P Q b c₁₁ c₁₂ c₂ H₁ _ H₂ _ Hnc
                     | P Q b c₂₁ c₂₂ c₁ H₁ _ H₂ _ Hnc
                     | P Q c₁ c₂ H' ] ;
        try discriminate HSkip ; try subst c₁ ; try clear HSkip.
      + simpl in H. inversion H. clear H. subst os'.
        apply HP₀P in HP₀.
        repeat (split ; eauto).
        apply disjoint_states_sym.
        apply (step_disjoint_states _ (k₂', s₂) _  (disjoint_states_sym _ _ Hdisj)) ;
          reflexivity.
      + apply IH₁ ; clear IH₁ ; eauto using proof_terms_consistent_1.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        unfold _step_helper_right in H.
        destruct (step υ _) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        unfold bind in H.
        destruct S₂' as [k₂'' s₂'].
        destruct (step_disjoint_union_states
                    _ _ s (M.add x v s) k₂' _
                    (disjoint_states_sym _ _ Hdisj)
                    (store_is_union_sym _ _ _ Hs₁s₂) HS₂')
          as [Hk₂'' [Hs₁'s₂' Hdisj'']].
        { simpl. rewrite Hv. simpl. reflexivity. }
        simpl in Hk₂''. subst k₂''.
        simpl fst in *. simpl snd in *.
        rewrite Hv in H. inversion H. clear H. subst os'.
        simpl.
        apply disjoint_states_sym in Hdisj''.
        repeat (split ; auto).
        apply (close_skip_skip _ _ _ H₁) ; auto.
      + apply HP₀P in HP₀.
        unfold _step_helper_right in H.
        destruct (step υ _) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        unfold bind in H.
        destruct S₂' as [k₂'' s₂'].
        simpl in H. inversion H. clear H. subst os'.
        pose proof (step_disjoint_states _ _ _ (disjoint_states_sym _ _ Hdisj) HS₂')
          as Hdisj'.
        simpl in HS₂'.
        destruct (eval_bexp s₂ b) as [[|]|] ;
          try discriminate HS₂'.
        inversion HS₂' as [H'S₂']. clear HS₂'. subst k₂'' s₂'.
        simpl.
        apply disjoint_states_sym in Hdisj'.
        repeat (split ; auto).
        apply (close_skip_skip _ _ _ H₁) ; auto.
      + unfold _step_helper_right in H.
        destruct (step υ _) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        unfold bind in H.
        destruct S₂' as [k₂'' s₂'].
        inversion H. clear H. subst os'.
        pose proof (step_disjoint_states _ _ _ (disjoint_states_sym _ _ Hdisj) HS₂')
          as Hdisj'.
        inversion HS₂' as [H'S₂']. clear HS₂'. subst k₂'' s₂'.
        simpl.
        apply disjoint_states_sym in Hdisj'.
        repeat (split ; auto).
      + unfold _step_helper_right in H.
        destruct (step υ _) as [S₂' | ] eqn:HS₂' ;
          try discriminate H.
        unfold bind in H.
        destruct S₂' as [k₂'' s₂'].
        pose proof (step_disjoint_states _ _ _ (disjoint_states_sym _ _ Hdisj) HS₂')
          as Hdisj'.
        apply HP₀P in HP₀.
        pose proof (Hnc s HP₀) as Hnc'. clear Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try contradiction. clear Hnc'.
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        simpl in HS₂'.
        assert (∀ x, In_bexp_var x b → V₂ x) as Hb₁.
        { intros x Hx. apply H'k₂. exists (ITE b c₂₁ c₂₂). simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same b s₂ s₁ s V₂ V₁), Hv in HS₂' ;
          eauto.
        simpl in HS₂'. inversion HS₂'. clear HS₂'. subst k₂'' s₂'.
        apply disjoint_states_sym in Hdisj'.
        destruct v.
        * inversion H. clear H.
          subst os'.
          simpl.
          repeat (split ; eauto).
        * inversion H. clear H.
          subst os'.
          simpl.
          repeat (split ; eauto).
          unfold assertion_of_bexp. simpl. rewrite Hv. simpl. reflexivity.
      (* Self-comp *)
      + simpl in H.
        destruct (_split_hoare_seq_1 P Q or _ _ H' l')
          as [P' [Q' [R' [H₁ [H₂ [H'' [HPP' H''']]]]]]] ; auto.
        rewrite H'' in H.
        apply (_step_helper_single_right_invariant or Q' R' _ _ _ _ _ _ _ _ _ H) ;
          auto.
        apply (hoare_close_skip _ _ s H₁). auto.
    (* This subtree only involves the left program, in a single hoare proof *)
    - unfold _step in H.
      simpl in H₁l.
      destruct H₁l as [HP₀P H₁l].
      apply (_step_helper_single_left_invariant or P Q _ _ _ _ _ _ _ _ _ H) ; auto.
    (* This subtree only involves the right program, in a single hoare proof *)
    - unfold _step in H.
      simpl in H₁l.
      destruct H₁l as [HP₀P H₁l].
      apply (_step_helper_single_right_invariant or P Q _ _ _ _ _ _ _ _ _ H) ; auto.
  Qed.

  Lemma no_crash :
    ∀ genv S₁ S₂ s P₀ H₀ l,
      _invariant genv (State S₁ S₂ s P₀ (H₀::l)) →
      ∃ os', _step genv (State S₁ S₂ s P₀ (H₀::l)) = Some os'.
  Proof.
    intros genv [k₁ s₁] [k₂ s₂] s P₀ l H₀ [HP₀ [Hs [Hdisj [H₁l H₂l]]]].
    simpl fst in *. simpl snd in *.
    inversion H₂l as [
                    | c₁ k₁' c₂ k₂' l' P Q Hc H₂l'
                    | c₁ k₁'    k₂' l' P Q Hc H₂l'
                    |    k₁' c₂ k₂' l' P Q Hc H₂l'
                    | c₁ k₁'    k₂' l' P Q Hc H₂l'
                    |    k₁' c₂ k₂' l' P Q Hc H₂l' ] ;
      subst k₁ k₂ l H₀ ; clear H₂l.
    - destruct H₁l as [HP₀P H₁l].
      simpl.
      induction Hc as [
                       P
                     | P x e y e'
                     | P b₁ b₂ Hb₁b₂
                     | P Q R c₁₁ c₂₁ c₁₂ c₂₂ H₁ _ H₂ _
                     | P Q b₁ b₂ c₁₁ c₂₁ c₁₂ c₂₂ H₁ _ H₂ _ Hb₁b₂ Hnc
                     | P b₁ b₂ c₁ c₂ H₁ _ Hb₁b₂ Hnc
                     | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'
                     | P Q x e c₂ H₂ _
                     | P Q x e c₁ H₁ _
                     | P Q b c₂ H₂ _ Hb
                     | P Q b c₁ H₁ _ Hb
                     | P Q R c₁₁ c₁₂ c₂ H₁ _ H₂ _
                     | P Q R c₁ c₂₁ c₂₂ H₁ _ H₂ _
                     | P Q b c₁₁ c₁₂ c₂ H₁ _ H₂ _ Hnc
                     | P Q b c₂₁ c₂₂ c₁ H₁ _ H₂ _ Hnc
                     | P Q c₁ c₂ H ] ;
        simpl ; eauto.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v₁ [Hv₁ [v₂ [Hv₂ HP]]]].
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_exp_var z e → V₁ z) as He.
        { intros z Hz. apply H'k₁. exists (Assign x e). simpl. auto. }
        assert (∀ z, In_exp_var z e' → V₂ z) as He'.
        { intros z Hz. apply H'k₂. exists (Assign y e'). simpl. auto. }
        assert (V₂ y) as Hy.
        { apply H'k₂. exists (Assign y e'). simpl. auto. }
        rewrite (disjoint_eval_exp e s V₁ V₂ y v₁) in Hv₂ ; auto.
        rewrite Hv₂ in *. simpl in *. rewrite Hv₁ in *. simpl in *.
        rewrite (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂),
                (disjoint_union_eval_exp_same e' s₂ s₁ s V₂ V₁) in * ; eauto.
        rewrite Hv₁, Hv₂ in *. simpl in *.
        eauto.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct (Hb₁b₂ s HP₀) as [Hb₁ Hb₂].
        unfold assertion_of_bexp in Hb₁, Hb₂.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_bexp_var z b₁ → V₁ z) as H'b₁.
        { intros z Hz. apply Hk₁. exists (Assert b₁). simpl. auto. }
        assert (∀ z, In_bexp_var z b₂ → V₂ z) as H'b₂.
        { intros z Hz. apply Hk₂. exists (Assert b₂). simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same
                   b₂ _ _ _ V₂ V₁
                   (store_is_union_sym _ _ _ Hs)),
                 (disjoint_union_eval_bexp_same
                   b₁ _ _ _ V₁ V₂ Hs) in * ; eauto.
        rewrite Hb₁, Hb₂ in *. simpl.
        eauto.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct (Hnc s HP₀) as [Hnc' _].
        apply (Hb₁b₂ s) in HP₀.
        destruct (eval_bexp s b₁) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc' ; auto ; fail).
        clear Hnc'.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
        { intros x Hx. apply Hk₁. exists (ITE b₁ c₁₁ c₁₂).
          simpl. auto. }
        assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
        { intros x Hx. apply Hk₂. exists (ITE b₂ c₂₁ c₂₂).
          simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same
                   b₁ _ _ _ V₁ V₂
                   Hs),
                (disjoint_union_eval_bexp_same
                   b₂ _ _ _ V₂ V₁
                   (store_is_union_sym _ _ _ Hs)) in * ; eauto.
        rewrite <- HP₀, Hv in *. simpl in *.
        destruct v ; eauto.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct (Hnc s HP₀) as [Hnc' _].
        apply (Hb₁b₂ s) in HP₀.
        destruct (eval_bexp s b₁) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc' ; auto ; fail).
        clear Hnc'.
                destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
        { intros x Hx. apply Hk₁. exists (While b₁ c₁).
          simpl. auto. }
        assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
        { intros x Hx. apply Hk₂. exists (While b₂ c₂).
          simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same
                   b₁ _ _ _ V₁ V₂
                   Hs),
                (disjoint_union_eval_bexp_same
                   b₂ _ _ _ V₂ V₁
                   (store_is_union_sym _ _ _ Hs)) in * ; eauto.
        rewrite <- HP₀, Hv in *. simpl in *.
        destruct v ; eauto.
      + apply IH₁ ; eauto using assertion_implies_trans.
        clear IH₁.
        eauto using proof_terms_consistent_1.
      (* Executing only one of the two programs *)
      (* Assign *)
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_exp_var z e → V₁ z) as He.
        { intros z Hz. apply H'k₁. exists (Assign x e). simpl. auto. }
        rewrite (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂) in * ; eauto.
        rewrite Hv in *. simpl in *. eauto.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_exp_var z e → V₂ z) as He.
        { intros z Hz. apply H'k₂. exists (Assign x e). simpl. auto. }
        rewrite (disjoint_union_eval_exp_same e s₂ s₁ s V₂ V₁), Hv in * ; eauto.
        simpl. eauto.
      (* Assert *)
      + apply HP₀P in HP₀. clear HP₀P.
        unfold assertion_of_bexp in Hb.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_bexp_var z b → V₁ z) as H'b.
        { intros z Hz. apply Hk₁. exists (Assert b). simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₁ V₂ Hs), Hb in * ; eauto.
        simpl. eauto.
      + apply HP₀P in HP₀. clear HP₀P.
        unfold assertion_of_bexp in Hb.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_bexp_var z b → V₂ z) as H'b.
        { intros z Hz. apply Hk₂. exists (Assert b). simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₂ V₁ (store_is_union_sym _ _ _ Hs)), Hb in * ; eauto.
        simpl.
        eauto.
      (* ITE *)
      + apply HP₀P in HP₀. clear HP₀P.
        pose proof (Hnc s HP₀) as H'nc.
        simpl in *.
        clear Hnc. rename H'nc into Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc ; auto ; fail).
        clear Hnc.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b → V₁ x) as Hb₁.
        { intros x Hx. apply Hk₁. exists (ITE b c₁₁ c₁₂).
          simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₁ V₂
                   Hs), Hv in * ; eauto.
        destruct v ; simpl ; eauto.
      + apply HP₀P in HP₀. clear HP₀P.
        pose proof (Hnc s HP₀) as H'nc.
        simpl in *.
        clear Hnc. rename H'nc into Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc ; auto ; fail).
        clear Hnc.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b → V₂ x) as Hb₁.
        { intros x Hx. apply Hk₂. exists (ITE b c₂₁ c₂₂).
          simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₂ V₁
                   (store_is_union_sym _ _ _ Hs)), Hv in * ; eauto.
        destruct v ; simpl ; eauto.
      (* Self-composition *)
      + simpl in *.
        destruct (_split_hoare_seq_1 P Q genv c₁ c₂ H l')
          as [P' [Q' [R' [H₁ [H₂ [H' [HPP' H'']]]]]]] ; auto.
        rewrite H' in *.
        apply _step_helper_single_left_no_crash ; auto.
    (* This subtree only involves the left program *)
    - destruct H₁l as [HP₀P H₁l].
      simpl in *. remember Skip as c eqn:HSkip.
      induction Hc as [
                       P
                     | | | | |
                     | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'
                     | P Q x e c₂ H₂ _
                     | P Q x e c₁ H₁ _
                     | P Q b c₂ H₂ IH₂ Hb
                     | P Q b c₁ H₁ IH₁ Hb
                     | P Q R c₁₁ c₁₂ c₂ H₁ _ H₂ _
                     | P Q R c₁ c₂₁ c₂₂ H₁ _ H₂ _
                     | P Q b c₁₁ c₁₂ c₂ H₁ _ H₂ _ Hnc
                     | P Q b c₂₁ c₂₂ c₁ H₁ _ H₂ _ Hnc
                     | P Q c₁ c₂ H ] ;
        try discriminate HSkip ; try subst c₂ ;
          simpl ; eauto.
      + apply IH₁ ; eauto using assertion_implies_trans.
        eauto using proof_terms_consistent_1.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_exp_var z e → V₁ z) as He.
        { intros z Hz. apply H'k₁. exists (Assign x e). simpl. auto. }
        rewrite (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂), Hv in * ; eauto.
        simpl. eauto.
      + apply HP₀P in HP₀. clear HP₀P.
        unfold assertion_of_bexp in Hb.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_bexp_var z b → V₁ z) as H'b.
        { intros z Hz. apply Hk₁. exists (Assert b). simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₁ V₂ Hs), Hb in * ; eauto.
        simpl. eauto.
      + apply HP₀P in HP₀. clear HP₀P.
        pose proof (Hnc s HP₀) as H'nc.
        simpl in *.
        clear Hnc. rename H'nc into Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc ; auto ; fail).
        clear Hnc.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b → V₁ x) as Hb₁.
        { intros x Hx. apply Hk₁. exists (ITE b c₁₁ c₁₂).
          simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₁ V₂
                   Hs), Hv in * ; eauto.
        destruct v ; simpl ; eauto.
      (* Self-composition *)
      + simpl in *.
        destruct (_split_hoare_seq_1 P Q genv c₁ Skip H l')
          as [P' [Q' [R' [H₁ [H₂ [H' [HPP' H'']]]]]]] ; auto.
        rewrite H' in *.
        apply _step_helper_single_left_no_crash ; auto.
    (* This subtree only involves the right program *)
    - destruct H₁l as [HP₀P H₁l].
      simpl in *. remember Skip as c eqn:HSkip.
      induction Hc as [
                       P
                     | | | | |
                     | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'
                     | P Q x e c₂ H₂ _
                     | P Q x e c₁ H₁ _
                     | P Q b c₂ H₂ _ Hb
                     | P Q b c₁ H₁ _ Hb
                     | P Q R c₁₁ c₁₂ c₂ H₁ _ H₂ _
                     | P Q R c₁ c₂₁ c₂₂ H₁ _ H₂ _
                     | P Q b c₁₁ c₁₂ c₂ H₁ _ H₂ _ Hnc
                     | P Q b c₂₁ c₂₂ c₁ H₁ _ H₂ _ Hnc
                     | P Q c₁ c₂ H ] ;
        try discriminate HSkip ; try subst c₁ ;
          simpl ; eauto.
      + apply IH₁ ; eauto using assertion_implies_trans.
        eauto using proof_terms_consistent_1.
      + apply HP₀P in HP₀. clear HP₀P.
        destruct HP₀ as [v [Hv HP]].
        destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_exp_var z e → V₂ z) as He.
        { intros z Hz. apply H'k₂. exists (Assign x e). simpl. auto. }
        rewrite (disjoint_union_eval_exp_same e s₂ s₁ s V₂ V₁), Hv in * ; eauto.
        simpl. eauto.
      + apply HP₀P in HP₀. clear HP₀P.
        unfold assertion_of_bexp in Hb.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ z : varid, In_bexp_var z b → V₂ z) as H'b.
        { intros z Hz. apply Hk₂. exists (Assert b). simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₂ V₁ (store_is_union_sym _ _ _ Hs)), Hb in * ; eauto.
        simpl. eauto.
      + apply HP₀P in HP₀. clear HP₀P.
        pose proof (Hnc s HP₀) as H'nc.
        simpl in *.
        clear Hnc. rename H'nc into Hnc.
        destruct (eval_bexp s b) as [v | ] eqn:Hv ;
          try (exfalso ; apply Hnc ; auto ; fail).
        clear Hnc.
        destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
        assert (∀ x, In_bexp_var x b → V₂ x) as Hb₁.
        { intros x Hx. apply Hk₂. exists (ITE b c₂₁ c₂₂).
          simpl. auto. }
        rewrite (disjoint_union_eval_bexp_same
                   b _ _ _ V₂ V₁
                   (store_is_union_sym _ _ _ Hs)), Hv in * ; eauto.
        destruct v ; simpl ; eauto.
      (* Self-composition *)
      + simpl in *.
        destruct (_split_hoare_seq_1 P Q genv Skip c₂ H l')
          as [P' [Q' [R' [H₁ [H₂ [H' [HPP' H'']]]]]]] ; auto.
        rewrite H' in *.
        apply _step_helper_single_right_no_crash ; auto.
        apply (hoare_close_skip P') ; auto.
    - simpl in *.
      apply _step_helper_single_left_no_crash ; auto.
      inversion H₁l. auto.
    - simpl in *.
      apply _step_helper_single_right_no_crash ; auto.
      inversion H₁l. auto.
  Qed.

  Lemma _prediction_completeness :
    ∀ genv os,
      _invariant genv os →
      _step genv os = None →
      step υ (_left_state  os) = None
      ∧ step υ (_right_state os) = None.
  Proof.
    intros genv [[k₁ s₁] [k₂ s₂] s P₀ l] Hinv HNone.
    pose proof Hinv as [HP₀ [Hs [Hdisj [H₁l H₂l]]]].
    simpl fst in *. simpl snd in *.
    inversion H₂l as [ | c₁ k₁' c₂ k₂' l' P Q Hc H₂l' | | | | ] ;
      subst k₁ k₂ l ; clear H₂l ; auto ;
    destruct (no_crash genv _ _ _ _ _ _ Hinv) as [os' Hos'] ;
    rewrite Hos' in HNone ; discriminate HNone.
  Qed.

  Definition core_rhl_oracle : oracle_language imp_language imp_language :=
    {|
      oracle_state := _state;
      oracle_genv := _genv;
      oracle_step := _step;
      left_state := _left_state;
      right_state := _right_state;
      left_genv := λ _, υ;
      right_genv := λ _, υ;
      invariant := _invariant;
      invariant_1 := _invariant_1;
      prediction_soundness := _prediction_soundness;
      prediction_completeness := _prediction_completeness
    |}.

  (* Prove core RHL soundness using the oracle *)
  Import BigStep.

  Lemma state_final_nocrash:
    ∀ genv os,
      _invariant genv os →
      let S₁ := _left_state  os in
      let S₂ := _right_state os in
      step υ S₁ = None →
      step υ S₂ = None →
      ∃ v₁ v₂, state_final S₁ = Some v₁ ∧ state_final S₂ = Some v₂.
  Proof.
    intros genv [[k₁ s₁] [k₂ s₂] s P₀ l] Hinv S₁ S₂ HS₁ HS₂.
    pose proof Hinv as [HP₀ [Hs [Hdisj [H₁l H₂l]]]].
    simpl fst in *. simpl snd in *.
    inversion H₂l as [ | c₁ k₁' c₂ k₂' l' P Q Hc H₂l' | | | | ] ;
      subst k₁ k₂ l ; clear H₂l ; auto.
    { exists υ, υ. simpl. auto. }
    - exfalso.
      destruct (no_crash genv _ _ _ _ _ _ Hinv) as [os' Hos'].
      destruct (_prediction_soundness genv _ os' Hinv) as [n₁ [n₂ [H₁ [H₂ H₃]]]] ; auto.
      simpl in H₁, H₂.
      destruct n₁ as [ | n₁ ], n₂ as [ | n₂].
      + omega.
      + unfold nsteps, napply in H₂.
        simpl ProgrammingLanguage.step in *.
        subst S₂. simpl in HS₂. rewrite HS₂ in H₂.
        simpl in H₂. unfold opt_state_eq in H₂.
        destruct H₂ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
      + unfold nsteps, napply in H₁.
        simpl ProgrammingLanguage.step in *.
        subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
        simpl in H₁. unfold opt_state_eq in H₁.
        destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
      + unfold nsteps, napply in H₁.
        simpl ProgrammingLanguage.step in *.
        subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
        simpl in H₁. unfold opt_state_eq in H₁.
        destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
    - exfalso.
      destruct (no_crash genv _ _ _ _ _ _ Hinv) as [os' Hos'].
      destruct (_prediction_soundness genv _ os' Hinv) as [n₁ [n₂ [H₁ [H₂ H₃]]]] ; auto.
      simpl in H₁, H₂.
      destruct n₁ as [ | n₁ ], n₂ as [ | n₂].
      + omega.
      + unfold nsteps, napply in H₂.
        simpl ProgrammingLanguage.step in *.
        subst S₂. simpl in HS₂. rewrite HS₂ in H₂.
        simpl in H₂. unfold opt_state_eq in H₂.
        destruct H₂ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
      + unfold nsteps, napply in H₁.
        simpl ProgrammingLanguage.step in *.
        subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
        simpl in H₁. unfold opt_state_eq in H₁.
        destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
      + unfold nsteps, napply in H₁.
        simpl ProgrammingLanguage.step in *.
        subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
        simpl in H₁. unfold opt_state_eq in H₁.
        destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
    - exfalso.
      destruct (no_crash genv _ _ _ _ _ _ Hinv) as [os' Hos'].
      destruct (_prediction_soundness genv _ os' Hinv) as [n₁ [n₂ [H₁ [H₂ H₃]]]] ; auto.
      simpl in H₁, H₂.
      destruct n₁ as [ | n₁ ], n₂ as [ | n₂].
      + omega.
      + unfold nsteps, napply in H₂.
        simpl ProgrammingLanguage.step in *.
        subst S₂. simpl in HS₂. rewrite HS₂ in H₂.
        simpl in H₂. unfold opt_state_eq in H₂.
        destruct H₂ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
      + unfold nsteps, napply in H₁.
        simpl ProgrammingLanguage.step in *.
        subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
        simpl in H₁. unfold opt_state_eq in H₁.
        destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
      + unfold nsteps, napply in H₁.
        simpl ProgrammingLanguage.step in *.
        subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
        simpl in H₁. unfold opt_state_eq in H₁.
        destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
    - exfalso.
      destruct (no_crash genv _ _ _ _ _ _ Hinv) as [os' Hos'].
      destruct (_prediction_soundness genv _ os' Hinv) as [n₁ [n₂ [H₁ [H₂ H₃]]]] ; auto.
      simpl in H₁, H₂.
      destruct n₁ as [ | n₁ ], n₂ as [ | n₂].
      + omega.
      + unfold nsteps, napply in H₂.
        simpl ProgrammingLanguage.step in *.
        subst S₂. simpl in HS₂. rewrite HS₂ in H₂.
        simpl in H₂. unfold opt_state_eq in H₂.
        destruct H₂ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
      + unfold nsteps, napply in H₁.
        simpl ProgrammingLanguage.step in *.
        subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
        simpl in H₁. unfold opt_state_eq in H₁.
        destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
      + unfold nsteps, napply in H₁.
        simpl ProgrammingLanguage.step in *.
        subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
        simpl in H₁. unfold opt_state_eq in H₁.
        destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
    - exfalso.
      destruct (no_crash genv _ _ _ _ _ _ Hinv) as [os' Hos'].
      destruct (_prediction_soundness genv _ os' Hinv) as [n₁ [n₂ [H₁ [H₂ H₃]]]] ; auto.
      simpl in H₁, H₂.
      destruct n₁ as [ | n₁ ], n₂ as [ | n₂].
      + omega.
      + unfold nsteps, napply in H₂.
        simpl ProgrammingLanguage.step in *.
        subst S₂. simpl in HS₂. rewrite HS₂ in H₂.
        simpl in H₂. unfold opt_state_eq in H₂.
        destruct H₂ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
      + unfold nsteps, napply in H₁.
        simpl ProgrammingLanguage.step in *.
        subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
        simpl in H₁. unfold opt_state_eq in H₁.
        destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
      + unfold nsteps, napply in H₁.
        simpl ProgrammingLanguage.step in *.
        subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
        simpl in H₁. unfold opt_state_eq in H₁.
        destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
  Qed.
  
  Lemma core_rhl_sc_soundness:
    ∀ c₁ s₁ c₂ s₂,
    terminates υ ([c₁], s₁) →
    terminates υ ([c₂], s₂) →
    ∀ s (P Q : assertion),
    store_is_union s s₁ s₂ →
    disjoint_states ([c₁], s₁) ([c₂], s₂) →
    P s →
    core_rhl_sc_proof P c₁ c₂ Q →
    ∃ s' s₁' s₂',
      s₁ ⊢ c₁ ⇓ s₁'
    ∧ s₂ ⊢ c₂ ⇓ s₂'
    ∧ store_is_union s' s₁' s₂' ∧ Q s'.
  Proof.
    intros c₁ s₁ c₂ s₂ H₁ H₂ s P Q Hs₁s₂ Hdisj HP₀ Hc₁c₂.
    pose (State ([c₁], s₁) ([c₂], s₂) s P [HoareProof Hc₁c₂]) as os.
    assert (invariant core_rhl_oracle Q os) as Hinv.
    { simpl. repeat (split ; eauto). }
    destruct (@oracle_nsteps_termination''
                _ _ core_rhl_oracle
                Q os)
      as [n [os' [Hos' [Hinv' H]]]] ; auto.
    simpl in H. destruct H as [H₁' H₂'].
    destruct (state_final_nocrash Q os' Hinv' H₁' H₂') as [v₁ [v₂ [H'₁ H'₂]]].
    clear H₁' H₂'.
    destruct os' as [S₁' S₂' s' P' l'].
    destruct S₁' as [[ | ] s₁'] ; try discriminate H'₁.
    destruct S₂' as [[ | ] s₂'] ; try discriminate H'₂.
    (* Recover small-step execution of c₁ and c₂ *)
    (* First: c₁! *)
    destruct (λ H, @oracle_nsteps_left _ _ core_rhl_oracle Q n os _ H Hos')
      as [n₁ Hn₁] ; auto.
    simpl in Hn₁.
    destruct Hn₁ as [ [_ HFalse] | [S₁' [S₁'' [HS₁' [HS₁'' H]]]] ] ;
      try discriminate HFalse.
    inversion HS₁''. clear HS₁''. subst S₁''. simpl in H.
    destruct S₁' as [k₁' s₁'']. simpl in H. destruct H as [Hk₁' Hs₁'].
    subst k₁'.
    (* Second: c₂! *)
    destruct (λ H, @oracle_nsteps_right _ _ core_rhl_oracle Q n os _ H Hos')
      as [n₂ Hn₂] ; auto.
    simpl in Hn₂.
    destruct Hn₂ as [ [_ HFalse] | [S₂' [S₂'' [HS₂' [HS₂'' H]]]] ] ;
      try discriminate HFalse.
    inversion HS₂''. clear HS₂''. subst S₂''. simpl in H.
    destruct S₂' as [k₂' s₂'']. simpl in H. destruct H as [Hk₂' Hs₂'].
    subst k₂'.
    (* Now, we have everything we need! *)
    clear H₁ H₂ Hs₁s₂ Hdisj HP₀ H'₁ H'₂ Hinv s Hc₁c₂ os Hos'.
    exists s', s₁'', s₂''.
    destruct Hinv' as [HP' [Hs' [Hdisj' [H₁ H₂]]]].
    simpl fst in *. simpl snd in *.
    split ; [ | split ; [ | split ] ] ; eauto using small_implies_big.
    + eauto using equal_store_is_union.
    + inversion H₂. clear H₂. subst l'. simpl in H₁. apply H₁. auto.
  Qed.
End Make.
