(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

Require Import Common.
Require Import Imp.Lang.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.
Require Import OracleHelpers.

Module Make
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT)
       (Import OracleHelpers : ImpOracleHelpers Var_as_DT Lang).
  Record mod := {
    condition :
      (* Condition to add in order to guard the buggy program *)
      bexp;
    c_then :
      (* Buggy sub-program that is to be guarded *)
      cmd;
    c_else :
      (* Alternative code to be executed in case of problematic input *)
      cmd;
    crash_bound:
      store → nat;
    crash_avoided_1:
      ∀ s, eval_bexp s condition ≠ Some true →
      ∀ k',
        nsteps υ (crash_bound s) (c_then::k', s) = None;
  }.

  Inductive _cont_mod : Type := .

  Definition _genv := unit.

  Inductive _state : Type :=
  | StuckLeft : state → state → _state
  | State : state → state → list (cont_mod mod _cont_mod) → _state.

  Definition _step (_ : _genv) (S : _state) : option _state :=
    match S with
    | StuckLeft S₁ S₂ =>
      step υ S₂ >>= λ S₂',
      Some (StuckLeft S₁ S₂')
    | State S₁ S₂ km =>
      match step_helper S₁ S₂ km with
      | None =>
        None
      | Some StuckStep =>
        None
      | Some (GenericStep S₁' S₂' km') =>
        Some (State S₁' S₂' km')
      | Some (SpecialCmdStep m km') =>
        match eval_bexp (snd S₂) (condition m) with
        | None =>
          let b := crash_bound m (snd S₁) in
          let (S₁', n') := nsteps' υ b S₁ in
          match (b - n') with
          | O =>
            None
          | _ =>
            Some (StuckLeft S₁' S₂)
          end
        | Some true =>
          step υ S₂ >>= λ S₂',
          Some (State S₁ S₂' (CmdMod Id::km'))
        | Some false =>
          step υ S₂ >>= λ S₂',
          let b := crash_bound m (snd S₁) in
          Some (StuckLeft (fst (nsteps' υ b S₁)) S₂')
        end
      | Some (SpecialContStep _ km') =>
        None
      end
    end.

  Definition _left_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | StuckLeft S₁ _ | State S₁ _ _ =>
      S₁
    end.

  Definition _right_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | StuckLeft _ S₂ | State _ S₂ _ =>
      S₂
    end.

  Inductive mod_invariant : cmd_mod mod → cmd → cmd → Prop :=
  | mod_id :
      ∀ c,
        mod_invariant Id c c
  | mod_rec_seq :
      ∀ m₁ m₂ c₁₁ c₁₂ c₂₁ c₂₂,
        mod_invariant m₁ c₁₁ c₂₁ →
        mod_invariant m₂ c₁₂ c₂₂ →
        mod_invariant (RecMod m₁ m₂) (Seq c₁₁ c₁₂) (Seq c₂₁ c₂₂)
  | mod_rec_ite :
      ∀ b m₁ m₂ c₁₁ c₁₂ c₂₁ c₂₂,
        mod_invariant m₁ c₁₁ c₂₁ →
        mod_invariant m₂ c₁₂ c₂₂ →
        mod_invariant (RecMod m₁ m₂) (ITE b c₁₁ c₁₂) (ITE b c₂₁ c₂₂)
  | mod_rec_while :
      ∀ b m m' c₁ c₂,
        mod_invariant m c₁ c₂ →
        mod_invariant (RecMod m m') (While b c₁) (While b c₂)
  | mod_special :
      ∀ m,
        mod_invariant (LeafMod m) (c_then m) (ITE (condition m) (c_then m) (c_else m)).

  Inductive km_invariant : list (cont_mod mod _cont_mod) → cont → cont → Prop :=
  | empty_cont_mod : km_invariant [] [] []
  | cmd_cont : ∀ c c' m km k₁ k₂,
      mod_invariant m c c' →
      km_invariant km k₁ k₂ →
      km_invariant (CmdMod m::km) (c::k₁) (c'::k₂).

  Definition _invariant (_ : _genv) (S : _state) : Prop :=
    match S with
    | StuckLeft S₁ S₂ => step υ S₁ = None
    | State (k₁, s₁) (k₂, s₂) km =>
      km_invariant km k₁ k₂ ∧ M.Equal s₁ s₂
    end.

  Lemma _invariant_1:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      _invariant genv os'.
  Proof.
    intros [] [ S₁ [k₂ s₂] | [k₁ s₁] [k₂ s₂] km ] os' Hinv Hstep.
    - unfold _step in Hstep.
      destruct (step υ (k₂, s₂)) as [ S₂' | ] ; inversion Hstep ; simpl ; auto.
    - unfold _step in Hstep.
      unfold _invariant in *.
      destruct Hinv as [Hkm Hs₁s₂].
      inversion Hkm as [
                      | c₁ c₂ m km' k₁' k₂' Hm Hkm' ] ;
        subst km k₁ k₂ ; try discriminate Hstep.
      unfold step_helper in Hstep.
      destruct m as [ m | | m₁ m₂ ].
      * simpl snd in Hstep.
        inversion Hm as [ | | | | m_ ]. subst m_ c₁ c₂.
        destruct (eval_bexp s₂ (condition m)) as [[ | ] | ] eqn:Hcond.
        { simpl in Hstep.
          rewrite Hcond in Hstep. simpl in Hstep.
          inversion Hstep. clear Hstep. subst os'.
          split ; auto.
          apply cmd_cont ; auto.
          apply mod_id. }
        { simpl in Hstep. rewrite Hcond in Hstep. simpl in Hstep.
          inversion Hstep. clear Hstep. subst os'.
          assert (eval_bexp s₁ (condition m) ≠ Some true) as Hfalse.
          { rewrite (store_equal_eval_bexp _ s₁ s₂) ; auto.
            rewrite Hcond. discriminate. }
          pose proof (crash_avoided_1 m s₁ Hfalse k₁').
          unfold nsteps in H.
          rewrite napply_napply'_Sn_iff in H.
          destruct H as [n' [S' [HS' H'S']]].
          unfold state, store, cont in *.
          unfold nsteps'.
          rewrite HS'. simpl. auto. }
        { pose proof (λ Hfalse, crash_avoided_1 m s₁ Hfalse k₁').
          unfold nsteps in H.
          rewrite napply_napply'_Sn_iff in H.
          destruct H as [n' [S' [HS' H'S']]].
          { rewrite (store_equal_eval_bexp _ s₁ s₂) ; auto.
            rewrite Hcond. discriminate. }
          unfold nsteps' in Hstep.
          unfold state, store, cont in *.
          rewrite HS' in Hstep.
          destruct (crash_bound m s₁ - S n') ; try discriminate.
          inversion Hstep. clear Hstep. subst os'.
          auto. }
      * inversion Hm as [ c | | | | ]. subst c₁ c₂.
        clear Hm.
        destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ] ;
          try discriminate Hstep ;
          simpl in Hstep.
        { inversion Hstep ; split ;
            auto using cmd_cont. }
        { inversion Hstep ; split ;
            auto using cmd_cont, mod_id. }
        { rewrite (store_equal_eval_exp e s₁ s₂) in Hstep ; auto.
          destruct (eval_exp s₂ e) as [v₁ | ] eqn:Hv₁ ;
            inversion Hstep ; clear Hstep ; subst os' ; repeat split ;
              auto using cmd_cont, mod_id, MFacts.add_m. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
            inversion Hstep ; clear Hstep ; subst os' ; repeat split ;
              auto using cmd_cont, mod_id, MFacts.add_m. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
            inversion Hstep ; clear Hstep ; subst os' ; repeat split ;
              auto using cmd_cont, mod_id, MFacts.add_m. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
            inversion Hstep ; clear Hstep ; subst os' ; repeat split ;
              auto using cmd_cont, mod_id, MFacts.add_m. }
      * inversion Hm as [ | m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                          | b m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                          | b m₁_ m₂_ c₁_ c₂_ Hm₁
                          | ] ;
          subst m₁_ m₂_ c₁ c₂.
        { simpl in Hstep.
          inversion Hstep. clear Hstep. subst os'.
          auto using cmd_cont. }
        { simpl in Hstep.
          rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv ;
            try discriminate Hstep ;
            simpl in Hstep ; inversion Hstep ; clear Hstep ; subst os' ;
              repeat split ; auto using cmd_cont. }
        { simpl in Hstep.
          rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv ;
            try discriminate Hstep ;
            inversion Hstep ; clear Hstep ; subst os' ;
              repeat split ; auto using cmd_cont. }
  Qed.

  Lemma _prediction_soundness :
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (_left_state  os))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (_right_state os))
                       (Some (_right_state os'))
        ∧ n₁ + n₂ > 0.
  Proof.
    intros [ ] [ S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] os' Hinv Hos'.
    { unfold _step in Hos'.
      destruct (step υ S₂) as [S₂' | ] eqn:HS₂' ;
        inversion Hos'. subst os'.
      exists 0, 1.
      unfold nsteps. simpl. rewrite HS₂'. simpl.
      repeat (split ; auto using opt_state_eq_refl). }
    unfold _step in Hos'.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ m km' k₁' k₂' Hm Hkm' ] ;
      subst km k₁ k₂ ; try discriminate Hos'.
    destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H.
    + inversion Hos'. clear Hos'. subst os'.
      exists 1, 1.
      destruct (generic_step_bistep _ _ _ _ _ _ H) as [H₁ H₂].
      repeat split ; unfold _left_state, _right_state, nsteps, napply ; auto ;
        simpl (ProgrammingLanguage.step imp_language).
      * unfold bind. rewrite H₁. simpl. apply opt_state_eq_refl.
      * unfold bind. rewrite H₂. simpl. apply opt_state_eq_refl.
    + apply step_helper_special_cmd_step_inv in H.
      destruct H as [H'm Hkm'']. inversion H'm as [H''m]. clear H'm.
      subst km'' m.
      simpl snd. simpl _left_state. simpl _right_state.
      unfold bind.
      inversion Hm as [ | | | | m ]. subst m' c₁ c₂.
      simpl in Hos'.
      destruct (eval_bexp s₂ (condition m)) as [[ | ] | ] eqn:Hcond ;
        simpl in Hos' ;
        try (inversion Hos' ; clear Hos' ; subst os').
      * exists 0, 1. unfold nsteps, napply. simpl. rewrite Hcond. simpl.
        repeat (split ; auto using opt_state_eq_refl).
      * pose proof (λ Hfalse, crash_avoided_1 m s₁ Hfalse k₁') as H.
        unfold nsteps in H. rewrite napply_napply'_Sn_iff in H.
        destruct H as [n' [S' [HS' H'S']]].
        { rewrite (store_equal_eval_bexp _ s₁ s₂) ; auto.
          rewrite Hcond. discriminate. }
        unfold nsteps'. simpl.
        unfold napply' in *.
        simpl ProgrammingLanguage.step in *.
        simpl ProgrammingLanguage.state in *.
        unfold state, store, cont in *.
        rewrite HS'.
        exists (crash_bound m s₁ - S n'), 1.
        simpl.
        apply napply'_prefix in HS'.
        rewrite <- napply_napply'_0_iff in HS'.
        unfold nsteps. simpl ProgrammingLanguage.step in *.
        simpl ProgrammingLanguage.state in *.
        unfold state, store, cont in *.
        rewrite HS'. simpl. rewrite Hcond. simpl.
        repeat split ; auto using opt_state_eq_refl.
        auto with arith.
      * pose proof (λ Hfalse, crash_avoided_1 m s₁ Hfalse k₁').
        unfold nsteps in H. rewrite napply_napply'_Sn_iff in H.
        destruct H as [n' [S' [HS' H'S']]].
        { rewrite (store_equal_eval_bexp _ s₁ s₂ ) ; auto.
          rewrite Hcond. discriminate. }
        unfold state, store, cont in *.
        unfold nsteps' in Hos'. rewrite HS' in Hos'.
        destruct (crash_bound m s₁ - S n') as [ | n'' ] eqn:Hn'' ;
          try discriminate Hos'.
        inversion Hos'. clear Hos'. subst os'.
        apply napply'_prefix in HS'.
        rewrite <- napply_napply'_0_iff in HS'.
        unfold nsteps. simpl.
        simpl ProgrammingLanguage.step in *.
        simpl ProgrammingLanguage.state in *.
        exists (crash_bound m s₁ - S n'), 0.
        rewrite HS'. repeat split ; auto using opt_state_eq_refl.
        rewrite Hn''. auto with arith.
    + discriminate Hos'.
    + discriminate Hos'.
  Qed.

  Lemma _prediction_completeness :
    ∀ genv os,
      _invariant genv os →
      _step genv os = None →
      step υ (_left_state  os) = None
      ∧ step υ (_right_state os) = None.
  Proof.
    intros [ ] [ S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] Hinv HNone.
    { simpl in Hinv. split ; auto. simpl in HNone. simpl.
      destruct (step υ S₂) ; auto ; discriminate. }
    unfold _step in HNone.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ m km' k₁' k₂' Hm Hkm' ] ;
      subst km k₁ k₂ ; auto.
    destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H.
    - discriminate HNone.
    - apply step_helper_special_cmd_step_inv in H.
      destruct H as [H'm Hkm'']. inversion H'm as [H''m]. clear H'm.
      subst m km''.
      inversion Hm as [ | | | | m ]. subst m' c₁ c₂.
      simpl in HNone.
      destruct (eval_bexp s₂ (condition m)) as [[ | ] | ] eqn:Hcond ;
        simpl in HNone ;
        try discriminate HNone.
      pose proof (λ Hfalse, crash_avoided_1 m s₁ Hfalse k₁') as H.
      unfold nsteps in H. rewrite napply_napply'_Sn_iff in H.
      destruct H as [n' [S' [HS' H'S']]].
      { rewrite (store_equal_eval_bexp _ s₁ s₂ ) ; auto.
        rewrite Hcond. discriminate. }
      unfold state, store, cont in *.
      unfold nsteps' in HNone. rewrite HS' in HNone.
      destruct (crash_bound m s₁ - S n') eqn:H ; try discriminate HNone.
      unfold _left_state, _right_state.
      split ; [ | simpl ; rewrite Hcond ; auto ].
      pose proof (napply'_lt_id (ProgrammingLanguage.step imp_language υ)
                                _ _ _ _ HS' H) as H''S'.
      subst S'.
      auto.
    - destruct (generic_step_bicrash _ _ _ H) as [H₁ H₂].
      unfold _left_state, _right_state. auto.
    - (* Thanks to the invariant, we can't crash step_helper: *)
      exfalso.
      simpl in H.
      destruct m as [ m | | m₁ m₂ ].
      * discriminate H.
      * inversion Hm as [ c | | | | ].
        subst c₁ c₂.
        destruct c ; try discriminate H.
        { rewrite (store_equal_eval_exp e s₁ s₂) in H ; auto.
          destruct (eval_exp s₂ e) ; try discriminate H. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H. }
      * inversion Hm as [ | m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                          | b m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                          | b m₁_ m₂_ c₁_ c₂_ Hm₁
                          | ] ;
          subst m₁_ m₂_ c₁ c₂.
        { discriminate H. }
        { simpl in H.
          rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ;
            discriminate H. }
        { simpl in H.
          rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          simpl in H.
          destruct (eval_bexp s₂ b) as [[|]|] ;
            discriminate H. }
  Qed.

  Definition crashfix_oracle : oracle_language imp_language imp_language :=
    {|
      oracle_state := _state;
      oracle_genv := unit;
      oracle_step := _step;
      left_state := _left_state;
      right_state := _right_state;
      left_genv := λ _, υ;
      right_genv := λ _, υ;
      invariant := _invariant;
      invariant_1 := _invariant_1;
      prediction_soundness := _prediction_soundness;
      prediction_completeness := _prediction_completeness
    |}.
End Make.