(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

(* Fix crash caused by off-by-one *)

Require Import Common.
Require Import Imp.Lang.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.
Require Import OracleHelpers.
Import ZArith.

Module Make
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT)
       (Import OracleHelpers : ImpOracleHelpers Var_as_DT Lang).
  Record modif := {
    expression :
      (* Expression used in the while loop condition *)
      exp;
    bound :
      Z;
    body :
      (* Loop body *)
      cmd;
    crash_bound:
      store → nat;
    crash_avoided_1:
      ∀ s, eval_exp s expression = Some bound →
      ∀ k',
        nsteps υ (crash_bound s) (body::k', s) = None;
  }.

  Inductive _cont_mod : Type := .

  Definition _genv := unit.

  Inductive _state : Type :=
  | StuckLeft : state → state → _state
  | State : state → state → list (cont_mod modif _cont_mod) → _state.

  Definition _step (_ : _genv) (S : _state) : option _state :=
    match S with
    | StuckLeft S₁ S₂ =>
      step υ S₂ >>= λ S₂',
      Some (StuckLeft S₁ S₂')
    | State S₁ S₂ km =>
      match step_helper S₁ S₂ km with
      | None =>
        None
      | Some StuckStep =>
        None
      | Some (GenericStep S₁' S₂' km') =>
        Some (State S₁' S₂' km')
      | Some (SpecialCmdStep m km') =>
        eval_exp (snd S₂) (expression m) >>= λ val,
        if Z.eqb val (bound m) then
          step υ S₂ >>= λ S₂',
          let b := crash_bound m (snd S₁) in
          Some (StuckLeft (fst (nsteps' υ (Datatypes.S b) S₁)) S₂')
        else
          step υ S₁ >>= λ S₁',
          step υ S₂ >>= λ S₂',
          match fst S₂ with | While b _ :: _ => Some b | _ => None end >>= λ b,
          eval_bexp (snd S₂) b >>= λ v,
          Some (State S₁' S₂'
                      (if v then
                         CmdMod Id :: CmdMod (LeafMod m) :: km'
                       else
                         km'))
      | Some (SpecialContStep _ km') =>
        None
      end
    end.

  Definition _left_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | StuckLeft S₁ _ | State S₁ _ _ =>
      S₁
    end.

  Definition _right_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | State _ S₂ _ | StuckLeft _ S₂ =>
      S₂
    end.

  Inductive mod_invariant : cmd_mod modif → cmd → cmd → Prop :=
  | mod_id :
      ∀ c,
        mod_invariant Id c c
  | mod_rec_seq :
      ∀ m₁ m₂ c₁₁ c₁₂ c₂₁ c₂₂,
        mod_invariant m₁ c₁₁ c₂₁ →
        mod_invariant m₂ c₁₂ c₂₂ →
        mod_invariant (RecMod m₁ m₂) (Seq c₁₁ c₁₂) (Seq c₂₁ c₂₂)
  | mod_rec_ite :
      ∀ b m₁ m₂ c₁₁ c₁₂ c₂₁ c₂₂,
        mod_invariant m₁ c₁₁ c₂₁ →
        mod_invariant m₂ c₁₂ c₂₂ →
        mod_invariant (RecMod m₁ m₂) (ITE b c₁₁ c₁₂) (ITE b c₂₁ c₂₂)
  | mod_rec_while :
      ∀ b m m' c₁ c₂,
        mod_invariant m c₁ c₂ →
        mod_invariant (RecMod m m') (While b c₁) (While b c₂)
  | mod_special :
      ∀ m,
        mod_invariant (LeafMod m)
                      (While (LE (expression m) (Int (bound m))) (body m))
                      (While (LT (expression m) (Int (bound m))) (body m))
  | mod_special' :
      ∀ m,
        mod_invariant (LeafMod m)
                      (While (LE (Int (bound m)) (expression m)) (body m))
                      (While (LT (Int (bound m)) (expression m)) (body m)).

  Inductive km_invariant : list (cont_mod modif _cont_mod) → cont → cont → Prop :=
  | empty_cont_mod : km_invariant [] [] []
  | cmd_cont : ∀ c c' m km k₁ k₂,
      mod_invariant m c c' →
      km_invariant km k₁ k₂ →
      km_invariant (CmdMod m::km) (c::k₁) (c'::k₂).

  Definition _invariant (_ : _genv) (S : _state) : Prop :=
    match S with
    | StuckLeft S₁ _ => step υ S₁ = None
    | State (k₁, s₁) (k₂, s₂) km =>
      km_invariant km k₁ k₂ ∧ M.Equal s₁ s₂
    end.

  Lemma _invariant_1:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      _invariant genv os'.
  Proof.
    intros [] [ S₁ [k₂ s₂] | [k₁ s₁] [k₂ s₂] km ] os' Hinv Hstep.
    - unfold _step in Hstep.
      destruct (step υ (k₂, s₂)) as [ S₂' | ] ; inversion Hstep ; simpl ; auto.
    - unfold _step in Hstep.
      unfold _invariant in *.
      destruct Hinv as [Hkm Hs₁s₂].
      inversion Hkm as [
                      | c₁ c₂ m km' k₁' k₂' Hm Hkm' ] ;
        subst km k₁ k₂ ; try discriminate Hstep.
      unfold step_helper in Hstep.
      destruct m as [ m | | m₁ m₂ ].
      * simpl snd in Hstep.
        inversion Hm as [ | | | | m_ | m_ ] ; subst m_ c₁ c₂ ;
          destruct (eval_exp s₂ (expression m)) as [ | ] eqn:Hv.
        { destruct (z =? bound m)%Z eqn:H.
          - simpl in Hstep. rewrite Hv, H in Hstep. simpl in Hstep.
            inversion Hstep. clear Hstep. subst os'.
            unfold nsteps'. simpl.
            rewrite (store_equal_eval_exp _ s₁ s₂) ; auto.
            rewrite Hv. simpl.
            rewrite Z.eqb_eq in H. subst z.
            rewrite Z.leb_refl.
            rewrite <- (store_equal_eval_exp _ s₁ s₂) in Hv ; auto.
            pose proof
                 (crash_avoided_1 m s₁ Hv
                                  (While (LE (expression m) (Int (bound m)))
                                         (body m) :: k₁')).
            unfold nsteps in H.
            rewrite napply_napply'_Sn_iff in H.
            destruct H as [n' [S' [HS' H'S']]].
            simpl ProgrammingLanguage.step in *.
            simpl ProgrammingLanguage.state in *.
            rewrite HS'.
            auto.
          - simpl in Hstep.
            rewrite (store_equal_eval_exp (expression m) s₁ s₂) in Hstep ; auto.
            rewrite Hv in Hstep. simpl in Hstep.
            destruct (z <? bound m)%Z eqn:H'.
            + assert (z <=? bound m = true)%Z as Hz.
              { rewrite Z.leb_le.
                rewrite Z.ltb_lt in H'. omega. }
              rewrite Hz, H in Hstep.
              inversion Hstep. clear Hstep. subst os'.
              split ; auto using cmd_cont, mod_id.
            + assert (z <=? bound m = false)%Z as Hz.
              { rewrite Z.leb_gt in *.
                rewrite Z.ltb_ge in *. rewrite Z.eqb_neq in *. omega. }
              rewrite Hz, H in Hstep. inversion Hstep. clear Hstep.
              subst os'.
              split ; auto. }
        { discriminate Hstep. }
        { destruct (z =? bound m)%Z eqn:H.
          - simpl in Hstep. rewrite Hv, H in Hstep. simpl in Hstep.
            inversion Hstep. clear Hstep. subst os'.
            unfold nsteps'. simpl.
            rewrite (store_equal_eval_exp _ s₁ s₂) ; auto.
            rewrite Hv. simpl.
            rewrite Z.eqb_eq in H. subst z.
            rewrite Z.leb_refl.
            rewrite <- (store_equal_eval_exp _ s₁ s₂) in Hv ; auto.
            pose proof
                 (crash_avoided_1 m s₁ Hv
                                  (While (LE (Int (bound m)) (expression m))
                                         (body m) :: k₁')).
            unfold nsteps in H.
            rewrite napply_napply'_Sn_iff in H.
            destruct H as [n' [S' [HS' H'S']]].
            simpl ProgrammingLanguage.step in *.
            simpl ProgrammingLanguage.state in *.
            rewrite HS'.
            auto.
          - simpl in Hstep.
            rewrite (store_equal_eval_exp (expression m) s₁ s₂) in Hstep ; auto.
            rewrite Hv in Hstep. simpl in Hstep.
            destruct (bound m <? z)%Z eqn:H'.
            + assert (bound m <=? z = true)%Z as Hz.
              { rewrite Z.leb_le.
                rewrite Z.ltb_lt in H'. omega. }
              rewrite Hz, H in Hstep.
              inversion Hstep. clear Hstep. subst os'.
              split ; auto using cmd_cont, mod_id.
            + assert (bound m <=? z = false)%Z as Hz.
              { rewrite Z.leb_gt in *.
                rewrite Z.ltb_ge in *. rewrite Z.eqb_neq in *. omega. }
              rewrite Hz, H in Hstep. inversion Hstep. clear Hstep.
              subst os'.
              split ; auto. }
        { discriminate Hstep. }
      * inversion Hm as [ c | | | | | ]. subst c₁ c₂.
        clear Hm.
        destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ] ;
          try discriminate Hstep ;
          simpl in Hstep.
        { inversion Hstep ; split ;
            auto using cmd_cont. }
        { inversion Hstep ; split ;
            auto using cmd_cont, mod_id. }
        { rewrite (store_equal_eval_exp e s₁ s₂) in Hstep ; auto.
          destruct (eval_exp s₂ e) as [v₁ | ] eqn:Hv₁ ;
            inversion Hstep ; clear Hstep ; subst os' ; repeat split ;
              auto using cmd_cont, mod_id, MFacts.add_m. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
            inversion Hstep ; clear Hstep ; subst os' ; repeat split ;
              auto using cmd_cont, mod_id, MFacts.add_m. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
            inversion Hstep ; clear Hstep ; subst os' ; repeat split ;
              auto using cmd_cont, mod_id, MFacts.add_m. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
            inversion Hstep ; clear Hstep ; subst os' ; repeat split ;
              auto using cmd_cont, mod_id, MFacts.add_m ;
              simpl ; rewrite (store_equal_eval_bexp _ s₁ s₂), Hv₁ ; auto. }
      * inversion Hm as [ | m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                          | b m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                          | b m₁_ m₂_ c₁_ c₂_ Hm₁
                          | | ] ;
          subst m₁_ m₂_ c₁ c₂.
        { simpl in Hstep.
          inversion Hstep. clear Hstep. subst os'.
          auto using cmd_cont. }
        { simpl in Hstep.
          rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv ;
            try discriminate Hstep ;
            simpl in Hstep ; inversion Hstep ; clear Hstep ; subst os' ;
              repeat split ; auto using cmd_cont. }
        { simpl in Hstep.
          rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv ;
            try discriminate Hstep ;
            inversion Hstep ; clear Hstep ; subst os' ;
              repeat split ; auto using cmd_cont. }
  Qed.

  Lemma _prediction_soundness :
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (_left_state  os))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (_right_state os))
                       (Some (_right_state os'))
        ∧ n₁ + n₂ > 0.
  Proof.
    intros [ ] [ S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] os' Hinv Hos'.
    { unfold _step in Hos'.
      destruct (step υ S₂) as [S₂' | ] eqn:HS₂' ;
        inversion Hos' ; clear Hos' ; subst os'.
      exists 0, 1. unfold nsteps. simpl. rewrite HS₂'. simpl.
      repeat (split ; auto using opt_state_eq_refl). }
    unfold _step in Hos'.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ m km' k₁' k₂' Hm Hkm' ] ;
      subst km k₁ k₂ ; try discriminate Hos'.
    destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H.
    - exists 1, 1.
      inversion Hos'. clear Hos'. subst os'.
      destruct (generic_step_bistep _ _ _ _ _ _ H) as [H₁ H₂].
      repeat split ; unfold _left_state, _right_state, nsteps, napply ; auto ;
        simpl (ProgrammingLanguage.step imp_language).
      + unfold bind. rewrite H₁. simpl. apply opt_state_eq_refl.
      + unfold bind. rewrite H₂. simpl. apply opt_state_eq_refl.
    - apply step_helper_special_cmd_step_inv in H.
      destruct H as [H'm Hkm'']. inversion H'm as [H''m]. clear H'm.
      subst km'' m.
      simpl snd. simpl _left_state. simpl _left_state. simpl _right_state.
      unfold bind.
      inversion Hm as [ | | | | m | m ] ; subst m' c₁ c₂ ;
        destruct (eval_exp s₂ (expression m)) as [v | ] eqn:Hv.
      + destruct (v =? bound m)%Z eqn:H.
        * simpl in Hos'. rewrite Hv in Hos'. simpl in Hos'.
          rewrite Z.eqb_eq in H. subst v.
          rewrite <- (store_equal_eval_exp _ s₁ s₂) in Hv ; auto.
          pose proof
               (crash_avoided_1 m s₁ Hv
                                (While (LE (expression m) (Int (bound m)))
                                       (body m) :: k₁')).
          unfold nsteps in H.
          rewrite napply_napply'_Sn_iff in H.
          destruct H as [n' [S' [HS' H'S']]].
          simpl in Hos'. rewrite Z.eqb_refl in Hos'.
          inversion Hos'. clear Hos'. subst os'.
          exists (1 + (crash_bound m s₁ - S n')), 1.
          unfold nsteps. simpl.
          rewrite <- (store_equal_eval_exp _ s₁ s₂) ; auto.
          rewrite Hv. simpl.
          rewrite Z.leb_refl, Z.ltb_irrefl.
          unfold nsteps'. simpl. rewrite Hv. simpl.
          rewrite Z.leb_refl.
          simpl ProgrammingLanguage.step in *.
          simpl ProgrammingLanguage.state in *.
          rewrite HS'.
          simpl.
          apply napply'_prefix in HS'.
          rewrite <- napply_napply'_0_iff in HS'.
          rewrite HS'. repeat split ; auto using opt_state_eq_refl.
          auto with arith.
        * simpl in Hos'.
          rewrite (store_equal_eval_exp _ s₁ s₂) in Hos' ; auto.
          rewrite Hv in Hos'. simpl in Hos'.
          destruct (v <? bound m)%Z eqn:H'v.
          { assert (v <=? bound m = true)%Z as H''v.
            { rewrite Z.leb_le. rewrite Z.eqb_neq in H.
              rewrite Z.ltb_lt in H'v. omega. }
            rewrite H''v in Hos'.
            exists 1, 1.
            rewrite H in Hos'.
            inversion Hos'. clear Hos'. subst os'.
            unfold nsteps. simpl.
            rewrite (store_equal_eval_exp _ s₁ s₂) ; auto.
            rewrite Hv. simpl. rewrite H'v, H''v.
            repeat split ; auto using opt_state_eq_refl. }
          { assert (v <=? bound m = false)%Z as H''v.
            { rewrite Z.leb_gt. rewrite Z.eqb_neq in H.
              rewrite Z.ltb_ge in H'v. omega. }
            rewrite H''v in Hos'.
            rewrite H in Hos'.
            inversion Hos'. clear Hos'. subst os'.
            exists 1, 1. unfold nsteps. simpl.
            rewrite (store_equal_eval_exp _ s₁ s₂) ; auto.
            rewrite Hv. simpl.
            rewrite H''v, H'v.
            repeat split ; auto using opt_state_eq_refl. }
      + simpl in Hos'. rewrite Hv in Hos'. discriminate Hos'.
      + destruct (v =? bound m)%Z eqn:H.
        * simpl in Hos'. rewrite Hv in Hos'. simpl in Hos'.
          rewrite Z.eqb_eq in H. subst v.
          rewrite <- (store_equal_eval_exp _ s₁ s₂) in Hv ; auto.
          pose proof
               (crash_avoided_1 m s₁ Hv
                                (While (LE (Int (bound m)) (expression m))
                                       (body m) :: k₁')).
          unfold nsteps in H.
          rewrite napply_napply'_Sn_iff in H.
          destruct H as [n' [S' [HS' H'S']]].
          rewrite Z.eqb_refl in Hos'.
          inversion Hos'. clear Hos'. subst os'.
          exists (1 + (crash_bound m s₁ - S n')), 1.
          unfold nsteps, nsteps'. simpl.
          rewrite <- (store_equal_eval_exp _ s₁ s₂) ; auto.
          rewrite Hv. simpl.
          rewrite Z.leb_refl, Z.ltb_irrefl.
          simpl ProgrammingLanguage.step in *.
          simpl ProgrammingLanguage.state in *.
          rewrite HS'.
          apply napply'_prefix in HS'.
          rewrite <- napply_napply'_0_iff in HS'.
          rewrite HS'.
          repeat split ; auto using opt_state_eq_refl.
          auto with arith.
        * simpl in Hos'. rewrite (store_equal_eval_exp _ s₁ s₂) in Hos' ; auto.
          rewrite Hv in Hos'. simpl in Hos'.
          destruct (bound m <? v)%Z eqn:H'v.
          { assert (bound m <=? v = true)%Z as H''v.
            { rewrite Z.leb_le. rewrite Z.eqb_neq in H.
              rewrite Z.ltb_lt in H'v. omega. }
            rewrite H''v, H in Hos'.
            inversion Hos'. clear Hos'. subst os'.
            exists 1, 1.
            unfold nsteps. simpl.
            rewrite (store_equal_eval_exp _ s₁ s₂) ; auto.
            rewrite Hv. simpl. rewrite H'v, H''v.
            repeat (split ; auto using opt_state_eq_refl). }
          { assert (bound m <=? v = false)%Z as H''v.
            { rewrite Z.leb_gt. rewrite Z.eqb_neq in H.
              rewrite Z.ltb_ge in H'v. omega. }
            rewrite H''v, H in Hos'.
            inversion Hos'. clear Hos'. subst os'.
            exists 1, 1. unfold nsteps.
            simpl. rewrite (store_equal_eval_exp _ s₁ s₂) ; auto.
            rewrite Hv. simpl. rewrite H'v, H''v.
            repeat (split ; auto using opt_state_eq_refl). }
      + simpl in Hos'. rewrite Hv in Hos'.
        discriminate Hos'.
    - discriminate Hos'.
    - discriminate Hos'.
  Qed.

  Lemma _prediction_completeness :
    ∀ genv os,
      _invariant genv os →
      _step genv os = None →
      step υ (_left_state  os) = None
      ∧ step υ (_right_state os) = None.
  Proof.
    intros [ ] [ S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] Hinv HNone.
    { simpl in *. split ; auto. destruct (step υ S₂) ; auto. discriminate. }
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ m km' k₁' k₂' Hm Hkm' ] ;
      subst km k₁ k₂ ; auto.
    unfold _step in HNone.
    destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H.
    - discriminate HNone.
    - apply step_helper_special_cmd_step_inv in H.
      destruct H as [H'm Hkm'']. inversion H'm as [H''m]. clear H'm.
      subst km'' m.
      simpl in HNone.
      inversion Hm as [ | | | | m | m ] ; subst m' c₁ c₂ ;
        destruct (eval_exp s₂ (expression m)) as [v | ] eqn:Hv.
      + destruct (v =? bound m)%Z eqn:H ;
          simpl in HNone ;
          try rewrite (store_equal_eval_exp _ s₁ s₂) in HNone ; auto ;
            rewrite Hv, H in HNone ; discriminate HNone.
      + simpl.
        rewrite (store_equal_eval_exp _ s₁ s₂) ; auto.
        rewrite Hv. simpl. auto.
      + destruct (v =? bound m)%Z eqn:H ;
          simpl in HNone ;
          try rewrite (store_equal_eval_exp _ s₁ s₂) in HNone ; auto ;
            rewrite Hv, H in HNone ; discriminate HNone.
      + simpl.
        rewrite (store_equal_eval_exp _ s₁ s₂) ; auto.
        rewrite Hv. simpl. auto.
    - destruct (generic_step_bicrash _ _ _ H) as [H₁ H₂].
      auto.
    - (* Thanks to the invariant, we can't crash step_helper: *)
      exfalso.
      simpl in H.
      destruct m as [ m | | m₁ m₂ ].
      + discriminate H.
      + inversion Hm as [ c | | | | | ].
        subst c₁ c₂.
        destruct c ; try discriminate H.
        * rewrite (store_equal_eval_exp e s₁ s₂) in H ; auto.
          destruct (eval_exp s₂ e) ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
      + inversion Hm as [ | m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                          | b m₁_ m₂_ c₁₁_ c₁₂_ c₂₁_ c₂₂_ Hm₁ Hm₂
                          | b m₁_ m₂_ c₁_ c₂_ Hm₁
                          | | ] ;
          subst m₁_ m₂_ c₁ c₂.
        * discriminate H.
        * simpl in H.
          rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ;
            discriminate H.
        * simpl in H.
          rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          simpl in H.
          destruct (eval_bexp s₂ b) as [[|]|] ;
            discriminate H.
  Qed.

  Definition crashfixwhile_oracle : oracle_language imp_language imp_language :=
    {|
      oracle_state := _state;
      oracle_genv := unit;
      oracle_step := _step;
      left_state := _left_state;
      right_state := _right_state;
      left_genv := λ _, υ;
      right_genv := λ _, υ;
      invariant := _invariant;
      invariant_1 := _invariant_1;
      prediction_soundness := _prediction_soundness;
      prediction_completeness := _prediction_completeness
    |}.
End Make.