(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

Require Import Common.
Require Import Imp.Lang.
Require Import Imp.BigStep.
Require Import Imp.RHL.
Require Import Imp.DisjointStates.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.
Require Import OracleHelpers.

Require Import Omega.

Module Make (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT).

  Module BigStep := Imp.BigStep.Make Var_as_DT Lang.
  Module RHL := Imp.RHL.Make Var_as_DT Lang BigStep.
  Import RHL.
  Import Hoare.
  Import DisjointStates.

  Definition _genv : Type := assertion. (* Post-condition *)

  Inductive proof_term :=
  | HoareProof:
      ∀ {P Q c₁ c₂},
        minimal_rhl_proof P c₁ c₂ Q →
        proof_term.

  Inductive _state :=
  | State :
      state →
      state →
      store →
      assertion →
      list proof_term →
      _state.

  Fixpoint proof_terms_consistent (l : list proof_term) (P Q : assertion) : Prop :=
    match l with
    | [] => P ⇒ Q
    | (@HoareProof P' Q' _ _ _)::l' =>
      P ⇒ P' ∧ proof_terms_consistent l' Q' Q
    end.

  Property proof_terms_consistent_1:
    ∀ l P P' Q,
      P ⇒ P' →
      proof_terms_consistent l P' Q →
      proof_terms_consistent l P Q.
  Proof.
    intro l.
    induction l as [ | [P'' Q'' c H] l ] ;
      intros P P' Q H₁ H₂.
    - simpl in *. eauto.
    - simpl in *. destruct H₂ as [H₂ H₃].
      split ; eauto.
  Qed.

  Inductive proof_terms_match : cont → cont → list proof_term → Prop :=
  | proof_terms_match_nil :
      proof_terms_match nil nil nil
  | proof_terms_match_cons :
      ∀ c₁ k₁ c₂ k₂ l P Q proof,
        proof_terms_match k₁ k₂ l →
        proof_terms_match (c₁::k₁) (c₂::k₂) (@HoareProof P Q c₁ c₂ proof::l).

  Hint Constructors proof_terms_match.

  Definition _invariant (Q : _genv) (os : _state) :=
    match os with
    | State S₁ S₂ s P l =>
      P s
      ∧ store_is_union s (snd S₁) (snd S₂)
      ∧ disjoint_states S₁ S₂
      ∧ proof_terms_consistent l P Q
      ∧ proof_terms_match (fst S₁) (fst S₂) l
    end.

  Fixpoint _step_helper {P Q c₁ c₂}
           (s : store)
           (H : minimal_rhl_proof P c₁ c₂ Q) (l' : list proof_term)
    : option (store * assertion * list proof_term) :=
    match H with
    | MinimalSkip P =>
      Some (s, P, l')
    | MinimalAssign P x e₁ y e₂ =>
      eval_exp s e₂ >>= λ v₁,
      let s' := M.add y v₁ s in
      eval_exp s e₁ >>= λ v₂,
      Some (M.add x v₂ s', P, l')
    | MinimalAssert P b₁ b₂ Hb =>
      (* XXX: If the precondition holds, it won't crash *)
      Some (s, P, l')
    | MinimalSeq P Q R c₁₁ c₂₁ c₁₂ c₂₂ H₁ H₂ =>
      Some (s, P, (HoareProof H₁)::(HoareProof H₂)::l')
    | MinimalITE P Q b₁ b₂ c₁₁ c₂₁ c₁₂ c₂₂ H₁ H₂ Hb _ =>
      eval_bexp s b₁ >>= λ v,
      if v then
        Some (s,
              (λ S, P S ∧ assertion_of_bexp b₁ S),
              (HoareProof H₁)::l')
      else
        Some (s,
              (λ S, P S ∧ assertion_of_bexp (NOT b₁) S),
              (HoareProof H₂)::l')
    | MinimalWhile P b₁ b₂ c₁ c₂ H Hb Hnc =>
      eval_bexp s b₁ >>= λ v,
      if v then
        Some (s,
              (λ S, P S ∧ assertion_of_bexp b₁ S),
              (HoareProof H)::(HoareProof (MinimalWhile P b₁ b₂ c₁ c₂ H Hb Hnc))::l')
      else
        Some (s,
              (λ S, P S ∧ assertion_of_bexp (NOT b₁) S),
              l')
    | MinimalSub P P' Q Q' c₁ c₂ H HP'P HQQ' =>
      _step_helper s H l'
    end.

  Definition _step (_ : _genv) (S : _state) : option _state :=
    match S with
    | State S₁ S₂ s P [] =>
      (* Final state *)
      None
    | State S₁ S₂ s P (HoareProof H::l') =>
      (* Minimal RHL relates two programs that run in lock-step,
         allowing us to factor the step operations…

         Furthermore, as Minimal RHL gives partial correctness,
         we don't have to handle a Crashed state. *)
      step υ S₁ >>= λ S₁',
      step υ S₂ >>= λ S₂',
      _step_helper s H l' >>= λ res,
      let '(s', P', l') := res in
      Some (State S₁' S₂' s' P' l')
    end.

  Definition _left_state (os : _state) : ProgrammingLanguage.state imp_language :=
    match os with
    | State S₁ _ _ _ _ =>
      S₁
    end.

  Definition _right_state (os : _state) : ProgrammingLanguage.state imp_language :=
    match os with
    | State _ S₂ _ _ _ =>
      S₂
    end.

  Lemma _prediction_soundness:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (_left_state  os))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (_right_state os))
                       (Some (_right_state os'))
        ∧ n₁ + n₂ > 0.
  Proof.
    intros genv [[k₁ s₁] [k₂ s₂] s P₀ l] os' [HP₀ [Hs [Hdisj [H₁l H₂l]]]] Hos'.
    simpl fst in *. simpl snd in *.
    inversion H₂l as [ | c₁ k₁' c₂ k₂' l' P Q Hc H₂l' ] ;
      subst k₁ k₂ l ; clear H₂l ; try discriminate Hos'.
    destruct H₁l as [HP₀P H₁l].
    simpl.
    induction Hc as [
                     P
                   | P x e y e'
                   | P b₁ b₂ Hb₁b₂
                   | P Q R c₁₁ c₂₁ c₁₂ c₂₂ H₁ IH₁ H₂ IH₂
                   | P Q b₁ b₂ c₁₁ c₂₁ c₁₂ c₂₂ H₁ IH₁ H₂ IH₂ Hb₁b₂ Hnc
                   | P b₁ b₂ c₁ c₂ H₁ IH₁ Hb₁b₂ Hnc
                   | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'].
    + simpl. inversion Hos'. clear Hos'. subst os'.
      exists 1, 1. unfold nsteps.
      simpl. repeat split ; auto using opt_state_eq_refl.
    + apply HP₀P in HP₀. clear HP₀P.
      destruct HP₀ as [v₁ [Hv₁ [v₂ [Hv₂ HP]]]].
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_exp_var z e → V₁ z) as He.
      { intros z Hz. apply H'k₁. exists (Assign x e). simpl. auto. }
      assert (∀ z, In_exp_var z e' → V₂ z) as He'.
      { intros z Hz. apply H'k₂. exists (Assign y e'). simpl. auto. }
      assert (V₂ y) as Hy.
      { apply H'k₂. exists (Assign y e'). simpl. auto. }
      exists 1, 1. unfold nsteps. simpl in *.
      rewrite (disjoint_eval_exp e s V₁ V₂ y v₁) in * ; auto.
      rewrite Hv₂ in *. simpl in *. rewrite Hv₁ in *. simpl in *.
      rewrite (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂),
              (disjoint_union_eval_exp_same e' s₂ s₁ s V₂ V₁) in * ; eauto.
      rewrite Hv₁, Hv₂ in *. simpl in *.
      inversion Hos'. clear Hos'. subst os'.
      simpl. repeat split ; auto using opt_state_eq_refl.
    + apply HP₀P in HP₀. clear HP₀P.
      destruct (Hb₁b₂ s HP₀) as [Hb₁ Hb₂].
      simpl in *. unfold assertion_of_bexp in Hb₁, Hb₂.
      destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b₁ → V₁ z) as H'b₁.
      { intros z Hz. apply Hk₁. exists (Assert b₁). simpl. auto. }
      assert (∀ z, In_bexp_var z b₂ → V₂ z) as H'b₂.
      { intros z Hz. apply Hk₂. exists (Assert b₂). simpl. auto. }
      exists 1, 1. unfold nsteps. simpl in *.
      rewrite (disjoint_union_eval_bexp_same
                 b₂ _ _ _ V₂ V₁
                 (store_is_union_sym _ _ _ Hs)),
              (disjoint_union_eval_bexp_same
                 b₁ _ _ _ V₁ V₂ Hs) in * ; eauto.
      rewrite Hb₁, Hb₂ in *. simpl in *.
      inversion Hos'. clear Hos'. subst os'.
      repeat split ; auto using opt_state_eq_refl.
    + clear IH₁ IH₂.
      exists 1, 1. unfold nsteps.
      simpl in *.
      inversion Hos'. clear Hos'. subst os'.
      simpl. repeat split ; auto using opt_state_eq_refl.
    + clear IH₁ IH₂.
      apply HP₀P in HP₀. clear HP₀P.
      destruct (Hnc s HP₀) as [Hnc' _].
      apply (Hb₁b₂ s) in HP₀.
      destruct (eval_bexp s b₁) as [v | ] eqn:Hv ;
        try (exfalso ; apply Hnc' ; auto ; fail).
      clear Hnc'.
      destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
      { intros x Hx. apply Hk₁. exists (ITE b₁ c₁₁ c₁₂).
        simpl. auto. }
      assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
      { intros x Hx. apply Hk₂. exists (ITE b₂ c₂₁ c₂₂).
        simpl. auto. }
      exists 1, 1. unfold nsteps. simpl in *.
      rewrite (disjoint_union_eval_bexp_same
                 b₁ _ _ _ V₁ V₂
                 Hs),
              (disjoint_union_eval_bexp_same
                 b₂ _ _ _ V₂ V₁
                 (store_is_union_sym _ _ _ Hs)) in * ; eauto.
      rewrite <- HP₀, Hv in *. simpl in *.
      destruct v ; inversion Hos' ; clear Hos' ; subst os' ; simpl ;
        repeat split ; auto using opt_state_eq_refl.
    + clear IH₁.
      apply HP₀P in HP₀. clear HP₀P.
      destruct (Hnc s HP₀) as [Hnc' _].
      apply (Hb₁b₂ s) in HP₀.
      destruct (eval_bexp s b₁) as [v | ] eqn:Hv ;
        try (exfalso ; apply Hnc' ; auto ; fail).
      clear Hnc'.
      destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
      { intros x Hx. apply Hk₁. exists (While b₁ c₁).
        simpl. auto. }
      assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
      { intros x Hx. apply Hk₂. exists (While b₂ c₂).
        simpl. auto. }
      exists 1, 1. unfold nsteps. simpl in *.
      rewrite (disjoint_union_eval_bexp_same
                 b₁ _ _ _ V₁ V₂
                 Hs),
              (disjoint_union_eval_bexp_same
                 b₂ _ _ _ V₂ V₁
                 (store_is_union_sym _ _ _ Hs)) in * ; eauto.
      rewrite <- HP₀, Hv in *. simpl in *.
      destruct v ; inversion Hos' ; clear Hos' ; subst os' ;
        simpl ;
        repeat split ; auto using opt_state_eq_refl.
    + apply IH₁ ; eauto using assertion_implies_trans.
      clear IH₁.
      eauto using proof_terms_consistent_1.
  Qed.

  Lemma _invariant_1:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      _invariant genv os'.
  Proof.
    intros or [[k₁ s₁] [k₂ s₂] s P₀ l] os' Hinv H.
    destruct Hinv as [HP₀ [Hs₁s₂ [Hdisj [H₁l H₂l]]]].
    simpl fst in *. simpl snd in *.
    inversion H₂l as [ | c₁ k₁' c₂ k₂' l' P Q Hc H₂l' ] ;
      subst k₁ k₂ l ; clear H₂l ; try discriminate H.
    unfold _step in H.
    destruct (step υ _) as [S₁' | ] eqn:HS₁' ;
      try discriminate H.
    unfold bind in H.
    destruct (step υ (_, s₂)) as [S₂' | ] eqn:HS₂' ;
      try discriminate H.
    destruct (_step_helper s Hc l') as [[[s' P''] l] | ] eqn:H' ;
      inversion H.
    clear H.
    subst os'.
    simpl in H₁l.
    destruct H₁l as [HP₀P H₁l].
    induction Hc as [
                     P
                   | P x e y e'
                   | P b₁ b₂ Hb₁b₂
                   | P Q R c₁₁ c₂₁ c₁₂ c₂₂ H₁ _ H₂ _
                   | P Q b₁ b₂ c₁₁ c₂₁ c₁₂ c₂₂ H₁ _ H₂ _ Hb₁b₂ Hnc
                   | P b₁ b₂ c₁ c₂ H₁ _ Hb₁b₂ Hnc
                   | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'].
    + inversion HS₁' as [H'S₁']. clear HS₁'. subst S₁'.
      inversion HS₂' as [H'S₂']. clear HS₂'. subst S₂'.
      apply HP₀P in HP₀.
      simpl in H'. inversion H'. clear H'. subst s' P'' l'.
      repeat (split ; eauto using step_2_disjoint_states).
    + apply HP₀P in HP₀. clear HP₀P.
      destruct HP₀ as [v₁ [Hv₁ HP]].
      destruct HP  as [v₂ [Hv₂ HP]].
      simpl in H'.
      destruct S₁' as [k₁'' s₁'], S₂' as [k₂'' s₂'].
      destruct (step_disjoint_union_states
                  _ _ s (M.add y v₁ s) k₂' _
                  (disjoint_states_sym _ _ Hdisj)
                  (store_is_union_sym _ _ _ Hs₁s₂) HS₂') as [Hk₂'' [Hs₂'s₁ Hdisj']].
      { simpl. rewrite Hv₁. simpl. reflexivity. }
      simpl in Hk₂''. subst k₂''.
      simpl fst in *. simpl snd in *.
      destruct (step_disjoint_union_states
                  _ _ (M.add y v₁ s) (M.add x v₂ (M.add y v₁ s)) k₁' _
                  (disjoint_states_sym _ _ Hdisj')
                  (store_is_union_sym _ _ _ Hs₂'s₁) HS₁') as [Hk₁'' [Hs₁'s₂' Hdisj'']].
      { simpl. rewrite Hv₂. simpl. reflexivity. }
      simpl in Hk₁''. subst k₁''.
      simpl fst in *. simpl snd in *.
      rewrite Hv₁ in H'. simpl in H'.
      destruct Hdisj as [V₁ [V₂ [Hk₁' [Hk₂' [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z, In_exp_var z e → V₁ z) as He.
      { intros z Hz. apply Hk₁'. exists (Assign x e). simpl. auto. }
      assert (V₂ y) as Hy.
      { apply Hk₂'. exists (Assign y e'). simpl. auto. }
      rewrite <- (disjoint_eval_exp e s V₁ V₂ y v₁) in H' ; eauto.
      rewrite Hv₂ in H'. inversion H'. clear H'. subst l' P'' s'.
      simpl.
      repeat (split ; auto).
    + apply HP₀P in HP₀.
      simpl in H'. inversion H'. clear H'. subst s' P'' l'.
      pose proof (step_2_disjoint_states _ _ _ _ Hdisj HS₁' HS₂') as Hdisj'.
      simpl in HS₁', HS₂'.
      destruct (eval_bexp s₁ b₁) as [[|]|], (eval_bexp s₂ b₂) as [[|]|] ;
        try discriminate HS₁' ; try discriminate HS₂'.
      inversion HS₁' as [H'S₁']. clear HS₁'. subst S₁'.
      inversion HS₂' as [H'S₂']. clear HS₂'. subst S₂'.
      repeat (split ; auto).
    + simpl in H'. inversion H'. clear H'. subst s' P'' l.
      pose proof (step_2_disjoint_states _ _ _ _ Hdisj HS₁' HS₂') as Hdisj'.
      inversion HS₁' as [H'S₁']. clear HS₁'. subst S₁'.
      inversion HS₂' as [H'S₂']. clear HS₂'. subst S₂'.
      simpl.
      repeat (split ; auto).
    + pose proof (step_2_disjoint_states _ _ _ _ Hdisj HS₁' HS₂') as Hdisj'.
      apply HP₀P in HP₀.
      simpl in H'.
      pose proof (Hb₁b₂ s HP₀) as H'b₁b₂. clear Hb₁b₂. simpl in H'b₁b₂.
      destruct (Hnc s HP₀) as [Hnc' _]. clear Hnc.
      destruct (eval_bexp s b₁) as [v₁ | ] eqn:Hv₁ ;
        try contradiction. clear Hnc'.
      simpl in H'.
      destruct S₁' as [k₁'' s₁'], S₂' as [k₂'' s₂'].
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      simpl in HS₁', HS₂'.
      assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
      { intros x Hx. apply H'k₁. exists (ITE b₁ c₁₁ c₁₂). simpl. auto. }
      assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
      { intros x Hx. apply H'k₂. exists (ITE b₂ c₂₁ c₂₂). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b₂ s₂ s₁ s V₂ V₁), <- H'b₁b₂ in HS₂' ;
        eauto.
      rewrite (disjoint_union_eval_bexp_same b₁ s₁ s₂ s V₁ V₂), Hv₁ in HS₁' ;
        eauto.
      simpl in HS₁', HS₂', H'.
      destruct v₁.
      * inversion HS₁'. inversion HS₂'. inversion H'. clear HS₁' HS₂' H'.
        subst k₁'' k₂'' s₁' s₂' s' P'' l.
        simpl.
        repeat (split ; eauto).
      * inversion HS₁'. inversion HS₂'. inversion H'. clear HS₁' HS₂' H'.
        subst k₁'' k₂'' s₁' s₂' s' P'' l.
        simpl.
        repeat (split ; eauto).
        unfold assertion_of_bexp. simpl. rewrite Hv₁. simpl. reflexivity.
    + pose proof (step_2_disjoint_states _ _ _ _ Hdisj HS₁' HS₂') as Hdisj'.
      apply HP₀P in HP₀.
      simpl in H'.
      pose proof (Hb₁b₂ s HP₀) as H'b₁b₂. simpl in H'b₁b₂.
      destruct (Hnc s HP₀) as [Hnc' _].
      destruct (eval_bexp s b₁) as [v₁ | ] eqn:Hv₁ ;
        try contradiction. clear Hnc'.
      simpl in H'.
      destruct S₁' as [k₁'' s₁'], S₂' as [k₂'' s₂'].
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      simpl in HS₁', HS₂'.
      assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
      { intros x Hx. apply H'k₁. exists (While b₁ c₁). simpl. auto. }
      assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
      { intros x Hx. apply H'k₂. exists (While b₂ c₂). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same b₂ s₂ s₁ s V₂ V₁), <- H'b₁b₂ in HS₂' ;
        eauto.
      rewrite (disjoint_union_eval_bexp_same b₁ s₁ s₂ s V₁ V₂), Hv₁ in HS₁' ;
        eauto.
      simpl in HS₁', HS₂', H'.
      destruct v₁.
      * inversion HS₁'. inversion HS₂'. inversion H'. clear HS₁' HS₂' H'.
        subst k₁'' k₂'' s₁' s₂' s' P'' l.
        simpl.
        repeat (split ; eauto).
      * inversion HS₁'. inversion HS₂'. inversion H'. clear HS₁' HS₂' H'.
        subst k₁'' k₂'' s₁' s₂' s' P'' l.
        simpl.
        repeat (split ; eauto).
        unfold assertion_of_bexp. simpl. rewrite Hv₁. simpl. reflexivity.
    + apply IH₁ ; clear IH₁ ; eauto using proof_terms_consistent_1.
  Qed.

  Lemma no_crash :
    ∀ genv S₁ S₂ s P₀ H₀ l,
      _invariant genv (State S₁ S₂ s P₀ (H₀::l)) →
      ∃ os', _step genv (State S₁ S₂ s P₀ (H₀::l)) = Some os'.
  Proof.
    intros genv [k₁ s₁] [k₂ s₂] s P₀ H₀ l [HP₀ [Hs [Hdisj [H₁l H₂l]]]].
    simpl fst in *. simpl snd in *.
    inversion H₂l as [ | c₁ k₁' c₂ k₂' l' P Q Hc H₂l' ] ;
      subst k₁ k₂ l H₀ ; clear H₂l.
    destruct H₁l as [HP₀P H₁l].
    simpl.
    induction Hc as [
                     P
                   | P x e y e'
                   | P b₁ b₂ Hb₁b₂
                   | P Q R c₁₁ c₂₁ c₁₂ c₂₂ H₁ IH₁ H₂ IH₂
                   | P Q b₁ b₂ c₁₁ c₂₁ c₁₂ c₂₂ H₁ IH₁ H₂ IH₂ Hb₁b₂ Hnc
                   | P b₁ b₂ c₁ c₂ H₁ IH₁ Hb₁b₂ Hnc
                   | P P' Q Q' c₁ c₂ H₁ IH₁ HP'P HQQ'].
    + simpl. eauto.
    + apply HP₀P in HP₀. clear HP₀P.
      destruct HP₀ as [v₁ [Hv₁ [v₂ [Hv₂ HP]]]].
      destruct Hdisj as [V₁ [V₂ [H'k₁ [H'k₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_exp_var z e → V₁ z) as He.
      { intros z Hz. apply H'k₁. exists (Assign x e). simpl. auto. }
      assert (∀ z, In_exp_var z e' → V₂ z) as He'.
      { intros z Hz. apply H'k₂. exists (Assign y e'). simpl. auto. }
      assert (V₂ y) as Hy.
      { apply H'k₂. exists (Assign y e'). simpl. auto. }
      simpl.
      rewrite (disjoint_eval_exp e s V₁ V₂ y v₁) in * ; auto.
      rewrite Hv₂ in *. simpl in *. rewrite Hv₁ in *. simpl in *.
      rewrite (disjoint_union_eval_exp_same e s₁ s₂ s V₁ V₂),
              (disjoint_union_eval_exp_same e' s₂ s₁ s V₂ V₁) in * ; eauto.
      rewrite Hv₁, Hv₂ in *. simpl in *.
      eauto.
    + apply HP₀P in HP₀. clear HP₀P.
      destruct (Hb₁b₂ s HP₀) as [Hb₁ Hb₂].
      simpl in *. unfold assertion_of_bexp in Hb₁, Hb₂.
      destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ z : varid, In_bexp_var z b₁ → V₁ z) as H'b₁.
      { intros z Hz. apply Hk₁. exists (Assert b₁). simpl. auto. }
      assert (∀ z, In_bexp_var z b₂ → V₂ z) as H'b₂.
      { intros z Hz. apply Hk₂. exists (Assert b₂). simpl. auto. }
      rewrite (disjoint_union_eval_bexp_same
                 b₂ _ _ _ V₂ V₁
                 (store_is_union_sym _ _ _ Hs)),
              (disjoint_union_eval_bexp_same
                 b₁ _ _ _ V₁ V₂ Hs) in * ; eauto.
      rewrite Hb₁, Hb₂ in *. simpl in *.
      eauto.
    + clear IH₁ IH₂.
      simpl in *.
      eauto.
    + clear IH₁ IH₂.
      apply HP₀P in HP₀. clear HP₀P.
      destruct (Hnc s HP₀) as [Hnc' _].
      apply (Hb₁b₂ s) in HP₀.
      destruct (eval_bexp s b₁) as [v | ] eqn:Hv ;
        try (exfalso ; apply Hnc' ; auto ; fail).
      clear Hnc'.
      destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
      { intros x Hx. apply Hk₁. exists (ITE b₁ c₁₁ c₁₂).
        simpl. auto. }
      assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
      { intros x Hx. apply Hk₂. exists (ITE b₂ c₂₁ c₂₂).
        simpl. auto. }
      simpl.
      rewrite (disjoint_union_eval_bexp_same
                 b₁ _ _ _ V₁ V₂
                 Hs),
              (disjoint_union_eval_bexp_same
                 b₂ _ _ _ V₂ V₁
                 (store_is_union_sym _ _ _ Hs)) in * ; eauto.
      rewrite <- HP₀, Hv in *. simpl in *.
      destruct v ; simpl ; eauto.
    + clear IH₁.
      apply HP₀P in HP₀. clear HP₀P.
      destruct (Hnc s HP₀) as [Hnc' _].
      apply (Hb₁b₂ s) in HP₀.
      destruct (eval_bexp s b₁) as [v | ] eqn:Hv ;
        try (exfalso ; apply Hnc' ; auto ; fail).
      clear Hnc'.
      destruct Hdisj as [V₁ [V₂ [Hk₁ [Hk₂ [Hs₁ [Hs₂ Hdisj]]]]]].
      assert (∀ x, In_bexp_var x b₁ → V₁ x) as Hb₁.
      { intros x Hx. apply Hk₁. exists (While b₁ c₁).
        simpl. auto. }
      assert (∀ x, In_bexp_var x b₂ → V₂ x) as Hb₂.
      { intros x Hx. apply Hk₂. exists (While b₂ c₂).
        simpl. auto. }
      simpl.
      rewrite (disjoint_union_eval_bexp_same
                 b₁ _ _ _ V₁ V₂
                 Hs),
              (disjoint_union_eval_bexp_same
                 b₂ _ _ _ V₂ V₁
                 (store_is_union_sym _ _ _ Hs)) in * ; eauto.
      rewrite <- HP₀, Hv in *. simpl in *.
      destruct v ; simpl ; eauto.
    + apply IH₁ ; eauto using assertion_implies_trans.
      clear IH₁.
      eauto using proof_terms_consistent_1.
  Qed.

  Lemma _prediction_completeness :
    ∀ genv os,
      _invariant genv os →
      _step genv os = None →
      step υ (_left_state  os) = None
      ∧ step υ (_right_state os) = None.
  Proof.
    intros genv [[k₁ s₁] [k₂ s₂] s P₀ l] Hinv HNone.
    pose proof Hinv as [HP₀ [Hs [Hdisj [H₁l H₂l]]]].
    simpl fst in *. simpl snd in *.
    inversion H₂l as [ | c₁ k₁' c₂ k₂' l' P Q Hc H₂l' ] ;
      subst k₁ k₂ l ; clear H₂l ; auto.
    destruct (no_crash genv _ _ _ _ _ _ Hinv) as [os' Hos'].
    rewrite Hos' in HNone. discriminate HNone.
  Qed.

  Definition minimal_rhl_oracle : oracle_language imp_language imp_language :=
    {|
      oracle_state := _state;
      oracle_genv := _genv;
      oracle_step := _step;
      left_state := _left_state;
      right_state := _right_state;
      left_genv := λ _, υ;
      right_genv := λ _, υ;
      invariant := _invariant;
      invariant_1 := _invariant_1;
      prediction_soundness := _prediction_soundness;
      prediction_completeness := _prediction_completeness
    |}.

  (* Prove minimal RHL soundness using the oracle *)
  Import BigStep.

  Lemma state_final_nocrash:
    ∀ genv os,
      _invariant genv os →
      let S₁ := _left_state  os in
      let S₂ := _right_state os in
      step υ S₁ = None →
      step υ S₂ = None →
      ∃ v₁ v₂, state_final S₁ = Some v₁ ∧ state_final S₂ = Some v₂.
  Proof.
    intros genv [[k₁ s₁] [k₂ s₂] s P₀ l] Hinv S₁ S₂ HS₁ HS₂.
    pose proof Hinv as [HP₀ [Hs [Hdisj [H₁l H₂l]]]].
    simpl fst in *. simpl snd in *.
    inversion H₂l as [ | c₁ k₁' c₂ k₂' l' P Q Hc H₂l' ] ;
      subst k₁ k₂ l ; clear H₂l ; auto.
    { exists υ, υ. simpl. auto. }
    exfalso.
    destruct (no_crash genv _ _ _ _ _ _ Hinv) as [os' Hos'].
    destruct (_prediction_soundness genv _ os' Hinv) as [n₁ [n₂ [H₁ [H₂ H₃]]]] ; auto.
    simpl in H₁, H₂.
    destruct n₁ as [ | n₁ ], n₂ as [ | n₂].
    - omega.
    - unfold nsteps, napply in H₂.
      simpl ProgrammingLanguage.step in *.
      subst S₂. simpl in HS₂. rewrite HS₂ in H₂.
      simpl in H₂. unfold opt_state_eq in H₂.
      destruct H₂ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
    - unfold nsteps, napply in H₁.
      simpl ProgrammingLanguage.step in *.
      subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
      simpl in H₁. unfold opt_state_eq in H₁.
      destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
    - unfold nsteps, napply in H₁.
      simpl ProgrammingLanguage.step in *.
      subst S₁. simpl in HS₁. rewrite HS₁ in H₁.
      simpl in H₁. unfold opt_state_eq in H₁.
      destruct H₁ as [[_ HFalse] | [S'₁ [S'₂ [HFalse _]]]] ; discriminate.
  Qed.

  Lemma minimal_rhl_soundness:
    ∀ c₁ s₁ c₂ s₂,
    terminates υ ([c₁], s₁) →
    terminates υ ([c₂], s₂) →
    ∀ s (P Q : assertion),
    store_is_union s s₁ s₂ →
    disjoint_states ([c₁], s₁) ([c₂], s₂) →
    P s →
    minimal_rhl_proof P c₁ c₂ Q →
    ∃ s' s₁' s₂',
      s₁ ⊢ c₁ ⇓ s₁'
    ∧ s₂ ⊢ c₂ ⇓ s₂'
    ∧ store_is_union s' s₁' s₂' ∧ Q s'.
  Proof.
    intros c₁ s₁ c₂ s₂ H₁ H₂ s P Q Hs₁s₂ Hdisj HP₀ Hc₁c₂.
    pose (State ([c₁], s₁) ([c₂], s₂) s P [HoareProof Hc₁c₂]) as os.
    assert (invariant minimal_rhl_oracle Q os) as Hinv.
    { simpl. repeat (split ; eauto). }
    destruct (@oracle_nsteps_termination''
                _ _ minimal_rhl_oracle
                Q os)
      as [n [os' [Hos' [Hinv' H]]]] ; auto.
    simpl in H. destruct H as [H₁' H₂'].
    destruct (state_final_nocrash Q os' Hinv' H₁' H₂') as [v₁ [v₂ [H'₁ H'₂]]].
    clear H₁' H₂'.
    destruct os' as [S₁' S₂' s' P' l'].
    destruct S₁' as [[ | ] s₁'] ; try discriminate H'₁.
    destruct S₂' as [[ | ] s₂'] ; try discriminate H'₂.
    (* Recover small-step execution of c₁ and c₂ *)
    (* First: c₁! *)
    destruct (λ H, @oracle_nsteps_left _ _ minimal_rhl_oracle Q n os _ H Hos')
      as [n₁ Hn₁] ; auto.
    simpl in Hn₁.
    destruct Hn₁ as [ [_ HFalse] | [S₁' [S₁'' [HS₁' [HS₁'' H]]]] ] ;
      try discriminate HFalse.
    inversion HS₁''. clear HS₁''. subst S₁''. simpl in H.
    destruct S₁' as [k₁' s₁'']. simpl in H. destruct H as [Hk₁' Hs₁'].
    subst k₁'.
    (* Second: c₂! *)
    destruct (λ H, @oracle_nsteps_right _ _ minimal_rhl_oracle Q n os _ H Hos')
      as [n₂ Hn₂] ; auto.
    simpl in Hn₂.
    destruct Hn₂ as [ [_ HFalse] | [S₂' [S₂'' [HS₂' [HS₂'' H]]]] ] ;
      try discriminate HFalse.
    inversion HS₂''. clear HS₂''. subst S₂''. simpl in H.
    destruct S₂' as [k₂' s₂'']. simpl in H. destruct H as [Hk₂' Hs₂'].
    subst k₂'.
    (* Now, we have everything we need! *)
    clear H₁ H₂ Hs₁s₂ Hdisj HP₀ H'₁ H'₂ Hinv s Hc₁c₂ os Hos'.
    exists s', s₁'', s₂''.
    destruct Hinv' as [HP' [Hs' [Hdisj' [H₁ H₂]]]].
    simpl fst in *. simpl snd in *.
    split ; [ | split ; [ | split ] ] ; eauto using small_implies_big.
    + eauto using equal_store_is_union.
    + inversion H₂. clear H₂. subst l'. simpl in H₁. apply H₁. auto.
  Qed.
End Make.
