(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

(** Helpers for writing oracle languages on Imp **)

Require Import Common.
Require Import Imp.Lang.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.
Require Omega.

Module Type ImpOracleHelpers
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT).
  Inductive cmd_mod (A : Type) :=
  | LeafMod : A → cmd_mod A
  | Id : cmd_mod A
  | RecMod : cmd_mod A → cmd_mod A → cmd_mod A.

  Arguments Id {A}.
  Arguments LeafMod {A} mod.
  Arguments RecMod {A} mod₁ mod₂.

  Definition local_identity {A : Type} (m : cmd_mod A) : Prop :=
    match m with
    | Id => Logic.True
    | RecMod _ _ => Logic.True
    | LeafMod _ => Logic.False
    end.

  Definition left_mod {A : Type} (m : cmd_mod A) : cmd_mod A :=
    match m with
    | Id => Id
    | LeafMod m =>
      (* XXX: This case does not really make sense, but this helper
         function is only called when “local_identity m” holds.
         This case is “handled” nonetheless to make the code simpler. *)
      LeafMod m
    | RecMod m₁ _ => m₁
    end.

  Definition right_mod {A : Type} (m : cmd_mod A) : cmd_mod A :=
    match m with
    | Id => Id
    | LeafMod m =>
      (* XXX: This case does not really make sense, but this helper
         function is only called when “local_identity m” holds.
         This case is “handled” nonetheless to make the code simpler. *)
      LeafMod m
    | RecMod _ m₂ => m₂
    end.

  Inductive cont_mod (A B : Type) :=
  | CmdMod : cmd_mod A → cont_mod A B
  | ContMod : B → cont_mod A B.

  Arguments CmdMod {A B} m.
  Arguments ContMod {A B} m.

  Inductive step_helper_res (A B : Type) :=
  | GenericStep :
      state →
      state →
      list (cont_mod A B) →
      step_helper_res A B
  | SpecialCmdStep :
      A →
      list (cont_mod A B) →
      step_helper_res A B
  | SpecialContStep :
      B →
      list (cont_mod A B) →
      step_helper_res A B
  | StuckStep :
      step_helper_res A B.

  Arguments GenericStep {A B} oS₁' oS₂' km'.
  Arguments SpecialContStep {A B} mod km'.
  Arguments SpecialCmdStep {A B} mod km'.
  Arguments StuckStep {A B}.

  Definition step_helper_id
            {A B : Type}
            (St₁ St₂ : state)
            (m : cmd_mod A)
            (km' : list (cont_mod A B)) :
    option (step_helper_res A B) :=
    match St₁, St₂ with
    | ([], _), ([], _) =>
      Some StuckStep
    | (Skip::_, _), (Skip::_, _)
    | (Assign _ _::_, _), (Assign _ _::_, _)
    | (Assert _::_, _), (Assert _::_, _) =>
      match step υ St₁, step υ St₂ with
      | None, None => Some StuckStep
      | Some S₁', Some S₂' => Some (GenericStep S₁' S₂' km')
      | _, _ => None
      end
    | (Seq c₁₁ c₁₂::k₁, s₁), (Seq c₂₁ c₂₂::k₂, s₂) =>
      Some (GenericStep
              (c₁₁::c₁₂::k₁, s₁)
              (c₂₁::c₂₂::k₂, s₂)
              (CmdMod (left_mod m)::CmdMod (right_mod m)::km'))
    | (ITE b₁ c₁₁ c₁₂::k₁, s₁), (ITE b₂ c₂₁ c₂₂::k₂, s₂) =>
      match eval_bexp s₁ b₁, eval_bexp s₂ b₂ with
      | Some true, Some true =>
        Some (GenericStep (c₁₁::k₁, s₁) (c₂₁::k₂, s₂) (CmdMod (left_mod m)::km'))
      | Some false, Some false =>
        Some (GenericStep (c₁₂::k₁, s₁) (c₂₂::k₂, s₂) (CmdMod (right_mod m)::km'))
      | None, None =>
        Some StuckStep
      | _, _ => None
      end
    | (While b₁ c₁::k₁, s₁), (While b₂ c₂::k₂, s₂) =>
      match eval_bexp s₁ b₁, eval_bexp s₂ b₂ with
      | Some true, Some true =>
        Some (GenericStep
                (c₁::While b₁ c₁::k₁, s₁)
                (c₂::While b₂ c₂::k₂, s₂)
                (CmdMod (left_mod m)::CmdMod m::km'))
      | Some false, Some false =>
        Some (GenericStep (k₁, s₁) (k₂, s₂) km')
      | None, None =>
        Some StuckStep
      | _, _ => None
      end
    | _, _ =>
      None
    end.

  Definition step_helper
             {A B : Type}
             (St₁ St₂ : state)
             (km : list (cont_mod A B))
    : option (step_helper_res A B) :=
    match km with
    | (ContMod m)::km' =>
      Some (SpecialContStep m km')
    | (CmdMod m)::km' =>
      match m with
      | Id
      | RecMod _ _ =>
        step_helper_id St₁ St₂ m km'
      | LeafMod m =>
        Some (SpecialCmdStep m km')
      end
    | [] =>
      match St₁, St₂ with
      | ([], _), ([], _) =>
        Some StuckStep
      | _, _ =>
        None
      end
    end.

  Lemma step_helper_special_cmd_step_inv:
    ∀ {A B : Type} S₁ S₂ (m : cont_mod A B) km m' km',
      step_helper S₁ S₂ (m::km) = Some (SpecialCmdStep m' km') →
      m = CmdMod (LeafMod m') ∧ km = km'.
  Proof.
    intros A B [k₁ s₁] [k₂ s₂] m km m' km' H.
    destruct m as [ [m | | m₁ m₂] | ].
    - simpl in H. inversion H as [[Hm Hkm]]. clear H.
      subst m' km'. auto.
    - simpl in H.
      destruct k₁ as [ | [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ] k₁],
               k₂ as [ | [ | c₂₁ c₂₂ | x₂ e₂ | b₂ c₂₁ c₂₂ | b₂ c₂ | b₂ ] k₂] ;
        try discriminate H ; simpl in H.
      + destruct (eval_exp s₁ e₁), (eval_exp s₂ e₂) ; discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
    - simpl in H.
      destruct k₁ as [ | [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ] k₁],
               k₂ as [ | [ | c₂₁ c₂₂ | x₂ e₂ | b₂ c₂₁ c₂₂ | b₂ c₂ | b₂ ] k₂] ;
        try discriminate H ; simpl in H.
      + destruct (eval_exp s₁ e₁), (eval_exp s₂ e₂) ; discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
    - discriminate H.
  Qed.

    Lemma step_helper_special_cont_step_inv:
    ∀ {A B : Type} S₁ S₂ (m : cont_mod A B) km m' km',
      step_helper S₁ S₂ (m::km) = Some (SpecialContStep m' km') →
      m = ContMod m' ∧ km = km'.
  Proof.
    intros A B [k₁ s₁] [k₂ s₂] m km m' km' H.
    destruct m as [ [m | | m₁ m₂] | ].
    - simpl in H. discriminate H.
    - simpl in H.
      destruct k₁ as [ | [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ] k₁],
               k₂ as [ | [ | c₂₁ c₂₂ | x₂ e₂ | b₂ c₂₁ c₂₂ | b₂ c₂ | b₂ ] k₂] ;
        try discriminate H ; simpl in H.
      + destruct (eval_exp s₁ e₁), (eval_exp s₂ e₂) ; discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
    - simpl in H.
      destruct k₁ as [ | [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ] k₁],
               k₂ as [ | [ | c₂₁ c₂₂ | x₂ e₂ | b₂ c₂₁ c₂₂ | b₂ c₂ | b₂ ] k₂] ;
        try discriminate H ; simpl in H.
      + destruct (eval_exp s₁ e₁), (eval_exp s₂ e₂) ; discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
      + destruct (eval_bexp s₁ b₁) as [[ | ] | ],
                 (eval_bexp s₂ b₂) as [[ | ] | ] ;
          discriminate H.
    - inversion H. subst. auto.
  Qed.

  Lemma generic_step_bistep:
    ∀ {A B : Type} S₁ S₂ S₁' S₂' (km km' : list (cont_mod A B)),
      step_helper S₁ S₂ km = Some (GenericStep S₁' S₂' km') →
      step υ S₁ = Some S₁' ∧ step υ S₂ = Some S₂'.
  Proof.
    intros A B [k₁ s₁] [k₂ s₂] S₁' S₂' [ | m km ] km' H.
    { simpl in H. destruct k₁, k₂ ; discriminate H. }
    destruct m as [ m | ] ; try discriminate H.
    destruct m as [ m | | m₁ m₂ ].
    { simpl in H. discriminate H. }
    - destruct k₁ as [ | c₁ k₁ ], k₂ as [ | c₂ k₂ ] ;
        try discriminate H ; try (destruct c₁ ; discriminate H).
      destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ],
               c₂ as [ | c₂₁ c₂₂ | x₂ e₂ | b₂ c₂₁ c₂₂ | b₂ c₂ | b₂ ] ;
        try discriminate H.
      + simpl in H. inversion H. clear H. subst S₁' S₂' km'.
        simpl. auto.
      + simpl in H. inversion H. clear H. subst S₁' S₂' km'.
        simpl. auto.
      + simpl in H.
        destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H.
        simpl in H. inversion H. clear H. subst S₁' S₂' km'.
        simpl. rewrite Hv₁, Hv₂. simpl. auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ;
        simpl in H ; inversion H ; clear H ; subst S₁' S₂' km' ;
        simpl ; rewrite Hv₁, Hv₂ ; simpl ; auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ;
        simpl in H ; inversion H ; clear H ; subst S₁' S₂' km' ;
        simpl ; rewrite Hv₁, Hv₂ ; simpl ; auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ;
        simpl in H ; inversion H ; clear H ; subst S₁' S₂' km' ;
        simpl ; rewrite Hv₁, Hv₂ ; simpl ; auto.
    - destruct k₁ as [ | c₁ k₁ ], k₂ as [ | c₂ k₂ ] ;
        try discriminate H ; try (destruct c₁ ; discriminate H).
      destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ],
               c₂ as [ | c₂₁ c₂₂ | x₂ e₂ | b₂ c₂₁ c₂₂ | b₂ c₂ | b₂ ] ;
        try discriminate H.
      + simpl in H. inversion H. clear H. subst S₁' S₂' km'.
        simpl. auto.
      + simpl in H. inversion H. clear H. subst S₁' S₂' km'.
        simpl. auto.
      + simpl in H.
        destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H.
        simpl in H. inversion H. clear H. subst S₁' S₂' km'.
        simpl. rewrite Hv₁, Hv₂. simpl. auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ;
        simpl in H ; inversion H ; clear H ; subst S₁' S₂' km' ;
        simpl ; rewrite Hv₁, Hv₂ ; simpl ; auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ;
        simpl in H ; inversion H ; clear H ; subst S₁' S₂' km' ;
        simpl ; rewrite Hv₁, Hv₂ ; simpl ; auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ;
        simpl in H ; inversion H ; clear H ; subst S₁' S₂' km' ;
        simpl ; rewrite Hv₁, Hv₂ ; simpl ; auto.
  Qed.

  Lemma generic_step_bicrash:
    ∀ {A B : Type} S₁ S₂ (km : list (cont_mod A B)),
      step_helper S₁ S₂ km = Some StuckStep →
      step υ S₁ = None ∧ step υ S₂ = None.
  Proof.
    intros A B [k₁ s₁] [k₂ s₂] [ | m km ] H.
    { simpl in H. destruct k₁, k₂ ; try discriminate H. simpl. auto. }
    destruct m as [ [m | | m₁ m₂] | ] ;
      try discriminate H.
    - destruct k₁ as [ | c₁ k₁ ], k₂ as [ | c₂ k₂ ] ;
        try discriminate H ; try (destruct c₁ ; discriminate H) ;
          simpl ; auto.
      destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ],
               c₂ as [ | c₂₁ c₂₂ | x₂ e₂ | b₂ c₂₁ c₂₂ | b₂ c₂ | b₂ ] ;
        try discriminate H ; simpl ; auto.
      + simpl in H.
        destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H ; simpl ; auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ; simpl ; auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ; simpl ; auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ; simpl ; auto.
    - destruct k₁ as [ | c₁ k₁ ], k₂ as [ | c₂ k₂ ] ;
        try discriminate H ; try (destruct c₁ ; discriminate H) ;
          simpl ; auto.
      destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ],
               c₂ as [ | c₂₁ c₂₂ | x₂ e₂ | b₂ c₂₁ c₂₂ | b₂ c₂ | b₂ ] ;
        try discriminate H ; simpl ; auto.
      + simpl in H.
        destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H ; simpl ; auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ; simpl ; auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ; simpl ; auto.
      + simpl in H.
        destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv₁,
                 (eval_bexp s₂ b₂) as [[ | ] | ] eqn:Hv₂ ;
          try discriminate H ; simpl ; auto.
  Qed.
End ImpOracleHelpers.

Module Make (Var_as_DT : UsualDecidableTypeOrig)
       (Lang : ImpProgrammingLanguage Var_as_DT) : (ImpOracleHelpers Var_as_DT Lang).
  Include ImpOracleHelpers Var_as_DT Lang.
End Make.