(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

Require Import Common.
Require Import Imp.Lang.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.
Require Import OracleHelpers.

Module Make
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT)
       (Import OracleHelpers : ImpOracleHelpers Var_as_DT Lang).
  Inductive mod :=
  | Refactor.

  Inductive _cont_mod := .

  Definition _genv := unit.

  Inductive _state : Type :=
  | Stuck : state → state → _state
  | State : state → state → list (cont_mod mod _cont_mod) → _state.

  (* Perform the same number of steps in both programs.
     The programs are assumed to have roughly the same crashing behaviors:
     if one crashes in at most the given [n] steps, so does the other
     (although it is not assumed to crash in the exact same number of
     steps) *)
  Definition perform_n_steps
             (n : nat) (S₁ S₂ : state)
             (km' : list (cont_mod mod _cont_mod)) : option _state :=
    (* Sanity check: if both programs are already stuck, return None *)
    match step υ S₁, step υ S₂ with
    | None, None =>
      None
    | _, _ =>
      match nsteps' υ n S₁, nsteps' υ n S₂ with
      | (S₁', 0), (S₂', 0) =>
        (* No crash in either program *)
        Some (State S₁' S₂' km')
      | (S₁', S _), (S₂', S _) =>
        (* Crash in both programs *)
        Some (Stuck S₁' S₂')
      | _, _ =>
        (* Programs have different crashing behaviors! *)
        None
      end
    end.

  Definition _step (_ : _genv) (S : _state) : option _state :=
    match S with
    | Stuck S₁ S₂ =>
      None
    | State S₁ S₂ km =>
      match step_helper S₁ S₂ km with
      | None =>
        None
      | Some StuckStep =>
        None
      | Some (GenericStep S₁' S₂' km') =>
        Some (State S₁' S₂' km')
      | Some (SpecialCmdStep _ km') =>
        perform_n_steps 2 S₁ S₂ (CmdMod Id::CmdMod Id::km')
      | Some (SpecialContStep _ km') =>
        None
      end
    end.

  Definition _left_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | Stuck S₁ _ | State S₁ _ _ =>
      S₁
    end.

  Definition _right_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | Stuck _ S₂ | State _ S₂ _ =>
      S₂
    end.

  Definition refactor_common_trailer c :=
    match c with
    | ITE b (Seq c₁₁ c₁₂) (Seq c₂₁ c₂₂) =>
      if cmd_eq_dec c₁₂ c₂₂ then
        ret (Seq (ITE b c₁₁ c₂₁) c₁₂)
      else
        None
    | _ => None
    end.

  Fixpoint apply_cmd_mod c m :=
    match m with
    | Id => Some c
    | RecMod m₁ m₂ =>
      match c with
      | Seq c₁ c₂ =>
        apply_cmd_mod c₁ m₁ >>= λ c₁',
        apply_cmd_mod c₂ m₂ >>= λ c₂',
        Some (Seq c₁' c₂')
      | ITE b c₁ c₂ =>
        apply_cmd_mod c₁ m₁ >>= λ c₁',
        apply_cmd_mod c₂ m₂ >>= λ c₂',
        Some (ITE b c₁' c₂')
      | While b c =>
        apply_cmd_mod c m₁ >>= λ c',
        Some (While b c')
      | Skip | Assign _ _ | Assert _ =>
        None
      end
    | LeafMod Refactor => refactor_common_trailer c
    end.

  Lemma apply_cmd_mod_id:
    ∀ c, apply_cmd_mod c Id = Some c.
  Proof.
    intro c. destruct c ; simpl ; auto.
  Qed.

  Inductive km_invariant : list (cont_mod mod _cont_mod) → cont → cont → Prop :=
  | empty_cont_mod : km_invariant [] [] []
  | cmd_cont : ∀ c c' m km k₁ k₂,
      apply_cmd_mod c m = Some c' →
      km_invariant km k₁ k₂ →
      km_invariant (CmdMod m::km) (c::k₁) (c'::k₂).

  Definition _invariant (_ : _genv) (S : _state) : Prop :=
    match S with
    | Stuck S₁ S₂ => step υ S₁ = None ∧ step υ S₂ = None
    | State (k₁, s₁) (k₂, s₂) km =>
      km_invariant km k₁ k₂ ∧ M.Equal s₁ s₂
    end.

  Lemma _invariant_1:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      _invariant genv os'.
  Proof.
    intros [] [S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] os' Hinv Hstep ;
      try discriminate Hstep.
    simpl in Hstep.
    unfold _invariant in *.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ m km' k₁' k₂' Hc₁c₂ Hkm' ] ;
      subst km k₁ k₂ ; try discriminate.
    simpl in Hstep. destruct m as [ m | | m₁ m₂ ].
    - destruct m as [ ].
      destruct c₁ as [ | | | b₁ [ | c₁₁ c | | | | ] [ | c₁₂ c' | | | | ] | | ] ;
        try discriminate Hc₁c₂.
      simpl in Hc₁c₂.
      destruct (cmd_eq_dec c c') as [Hc' | ] ; try discriminate Hc₁c₂.
      subst c'.
      inversion Hc₁c₂ as [H'].
      clear Hc₁c₂. subst c₂.
      unfold perform_n_steps, nsteps' in Hstep.
      simpl in Hstep.
      rewrite (store_equal_eval_bexp b₁ s₁ s₂) in Hstep ; auto.
      destruct (eval_bexp s₂ b₁) as [[ | ] | ] eqn:Hv ;
        inversion Hstep ; clear Hstep ; subst os'.
      + split ; auto. apply cmd_cont ; simpl ; auto.
        * destruct c₁₁ ; auto.
        * apply cmd_cont ; auto. destruct c ; auto.
      + split ; auto. apply cmd_cont ; simpl ; auto.
        * destruct c₁₂ ; auto.
        * apply cmd_cont ; auto. destruct c ; auto.
      + simpl. rewrite (store_equal_eval_bexp _ s₁ s₂) ; auto. rewrite Hv. auto.
    - rewrite apply_cmd_mod_id in Hc₁c₂. inversion Hc₁c₂ as [H'c₁c₂].
      clear Hc₁c₂. subst c₂.
      destruct c₁ as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ] ;
        try discriminate Hstep.
      + simpl in Hstep.
        inversion Hstep ; split ;
          auto using cmd_cont, apply_cmd_mod_id.
      + simpl in Hstep.
        inversion Hstep ; split ;
          auto using cmd_cont, apply_cmd_mod_id.
      + rewrite (store_equal_eval_exp e s₁ s₂) in Hstep ; auto.
        destruct (eval_exp s₂ e) as [v₁ | ] eqn:Hv₁ ;
          inversion Hstep ; clear Hstep ; subst os'; split ;
            auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m ;
            simpl ; try rewrite (store_equal_eval_exp _ s₁ s₂) ;
              try rewrite Hv₁ ; auto.
      + rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
        destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
          inversion Hstep ; clear Hstep ; subst os' ; split ;
            auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m ;
            simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂ ) ;
              try rewrite Hv₁ ; auto.
      + rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
        destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
          inversion Hstep ; clear Hstep ; subst os' ; split ;
            auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m ;
            simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂) ;
              try rewrite Hv₁ ; auto.
      + rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
        destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
          inversion Hstep ; clear Hstep ; subst os' ; split ;
            auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m ;
            simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂) ;
              try rewrite Hv₁ ; auto.
    - destruct c₁ as [ | c₁₁ c₁₂ | | b₁ c₁₁ c₁₂ | b₁ c₁ | ] ;
        try discriminate Hc₁c₂.
      + simpl in Hstep.
        simpl in Hc₁c₂.
        destruct (apply_cmd_mod c₁₁ m₁) as [c₁' | ] eqn:Hc₁' ;
          try discriminate Hc₁c₂.
        simpl in Hc₁c₂.
        destruct (apply_cmd_mod c₁₂ m₂) as [c₂' | ] eqn:Hc₂' ;
          try discriminate Hc₁c₂.
        inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
        inversion Hstep. clear Hstep. subst os'.
        auto using cmd_cont.
      + simpl in Hc₁c₂.
        destruct (apply_cmd_mod c₁₁ m₁) as [c₁' | ] eqn:Hc₁' ;
          try discriminate Hc₁c₂.
        simpl in Hc₁c₂.
        destruct (apply_cmd_mod c₁₂ m₂) as [c₂' | ] eqn:Hc₂' ;
          try discriminate Hc₁c₂.
        inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
        rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
        destruct (eval_bexp s₂ b₁) as [[ | ] | ] eqn:Hv ;
          simpl in Hstep ; inversion Hstep ; clear Hstep ; subst os' ;
            split ; auto using cmd_cont ;
              simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂) ;
                try rewrite Hv ; auto.
      + simpl in Hc₁c₂.
        destruct (apply_cmd_mod c₁ m₁) as [c₁' | ] eqn:Hc₁' ;
          try discriminate Hc₁c₂.
        inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
        rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
        destruct (eval_bexp s₂ b₁) as [[ | ] | ] eqn:Hv ;
          inversion Hstep ; clear Hstep ; subst os' ;
            split ; auto using cmd_cont ;
              simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂) ;
                try rewrite Hv ; auto.
  Qed.

  Lemma _prediction_soundness :
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (_left_state  os))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (_right_state os))
                       (Some (_right_state os'))
        ∧ n₁ + n₂ > 0.
  Proof.
    intros [ ] [S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] os' Hinv Hos' ;
      try discriminate Hos'.
    simpl in *.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ m km' k₁' k₂' Hc₁c₂ Hkm' ] ;
      subst km k₁ k₂ ; try discriminate Hos'.
    destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H.
    + exists 1, 1. inversion Hos'. clear Hos'. subst os'.
      destruct (generic_step_bistep _ _ _ _ _ _ H) as [H₁ H₂].
      repeat split ; unfold _left_state, _right_state, nsteps, napply ; auto ;
        simpl (ProgrammingLanguage.step imp_language).
      * rewrite H₁. simpl. apply opt_state_eq_refl.
      * rewrite H₂. simpl. apply opt_state_eq_refl.
    + apply step_helper_special_cmd_step_inv in H.
      destruct H as [Hm Hkm'']. inversion Hm as [H'm]. clear Hm.
      subst km'' m.
      destruct m'.
      destruct c₁ as [
                    | | | b₁ [ | c₁₁ c | | | | ] [ | c₁₂ c' | | | | ] | | ] ;
        try discriminate Hc₁c₂.
      simpl in Hc₁c₂.
      destruct (cmd_eq_dec c c') as [Hc' | ] ;
        try discriminate Hc₁c₂.
      subst c'.
      inversion Hc₁c₂ as [Hc₂]. clear Hc₁c₂. subst c₂.
      unfold perform_n_steps, nsteps' in Hos'. simpl in Hos'.
      rewrite (store_equal_eval_bexp b₁ s₁ s₂) in Hos' ; auto.
      destruct (eval_bexp s₂ b₁) as [[ | ] | ] eqn:Hv ; simpl in Hos' ;
        inversion Hos' ; clear Hos' ; subst os'.
      * exists 2, 2. unfold nsteps. simpl.
        rewrite (store_equal_eval_bexp b₁ s₁ s₂) ; auto.
        rewrite Hv. simpl. repeat split ; auto using opt_state_eq_refl.
        auto with arith.
      * exists 2, 2. unfold nsteps. simpl.
        rewrite (store_equal_eval_bexp b₁ s₁ s₂) ; auto.
        rewrite Hv. simpl. repeat split ; auto using opt_state_eq_refl.
        auto with arith.
      * exists 0, 1. unfold nsteps. simpl.
        repeat split ; auto using opt_state_eq_refl with arith.
    + discriminate Hos'.
      (* TODO: useful later
         destruct (generic_step_bicrash _ _ _ H) as [H₁ H₂].
         match goal with | |- ∃ _ _ _, Some(?x) = Some _ ∧ _ => exists x end.
         exists 0, 0. simpl.
         repeat split ; unfold _left_state, _right_state, nsteps ;
         auto using opt_state_eq_refl. *)
    + discriminate Hos'.
  Qed.

  Lemma _prediction_completeness :
    ∀ genv os,
      _invariant genv os →
      _step genv os = None →
      step υ (_left_state  os) = None
      ∧ step υ (_right_state os) = None.
  Proof.
    intros [ ] [S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] Hinv HNone.
    { simpl. auto. }
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ m km' k₁' k₂' Hc₁c₂ Hkm' ] ;
      subst km k₁ k₂.
    { simpl. auto. }
    unfold _step in HNone.
    destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H ;
      try discriminate HNone.
    - apply step_helper_special_cmd_step_inv in H.
      destruct H as [Hm Hkm'']. inversion Hm as [H'm]. clear Hm.
      subst km'' m.
      destruct m'.
      destruct c₁ as [
                    | | | b₁ [ | c₁₁ c | | | | ] [ | c₁₂ c' | | | | ] | | ] ;
        try discriminate Hc₁c₂.
      simpl in Hc₁c₂.
      destruct (cmd_eq_dec c c') as [Hc' | ] ;
        try discriminate Hc₁c₂.
      subst c'.
      inversion Hc₁c₂ as [Hc₂]. clear Hc₁c₂. subst c₂.
      unfold perform_n_steps, nsteps' in HNone. simpl in HNone.
      rewrite (store_equal_eval_bexp b₁ s₁ s₂) in HNone ; auto.
      destruct (eval_bexp s₂ b₁) as [[ | ] | ] eqn:Hv ; simpl in HNone ;
        try discriminate HNone.
    - apply generic_step_bicrash in H.
      simpl _left_state. simpl _right_state. auto.
    - simpl in H.
      destruct m as [ m | | m₁ m₂ ].
      + discriminate H.
      + destruct c₁ ; inversion Hc₁c₂ ; clear Hc₁c₂ ; subst c₂ ; try discriminate H.
        * rewrite (store_equal_eval_exp e s₁ s₂) in H ; auto.
          destruct (eval_exp s₂ e) ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
      + destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ] ;
          try discriminate Hc₁c₂.
        * simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁₁ m₁), (apply_cmd_mod c₁₂ m₂) ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂. discriminate H.
        * simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁₁ m₁), (apply_cmd_mod c₁₂ m₂) ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
          simpl in H.
          rewrite (store_equal_eval_bexp b₁ s₁ s₂) in H ; auto.
          simpl in H.
          destruct (eval_bexp s₂ b₁) as [[|]|] ;
            try discriminate H.
        * simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁ m₁) ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
          simpl in H.
          rewrite (store_equal_eval_bexp b₁ s₁ s₂) in H ; auto.
          simpl in H.
          destruct (eval_bexp s₂ b₁) as [[|]|] ;
            try discriminate H.
  Qed.

  Definition refactor_common_branch_trailer_oracle
    : oracle_language imp_language imp_language :=
    {|
      oracle_state := _state;
      oracle_genv := unit;
      oracle_step := _step;
      left_state := _left_state;
      right_state := _right_state;
      left_genv := λ _, υ;
      right_genv := λ _, υ;
      invariant := _invariant;
      invariant_1 := _invariant_1;
      prediction_soundness := _prediction_soundness;
      prediction_completeness := _prediction_completeness
    |}.
End Make.