(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

Require Import Common.
Require Import Imp.Lang.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.

Module Make
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT).

  Definition bijective {A B : Type} (f : A → B) : Prop :=
    ∃ f', ∀ x y, f x = y ↔ f' y = x.

  Definition bijection (A B : Type) :=
    { f : A → B | bijective f }.

  Definition _genv := bijection varid varid.

  Inductive _state : Type :=
  | State : state → state → _state.

  Definition _step (_ : _genv) (S : _state) : option _state :=
    let '(State S₁ S₂) := S in
    match step tt S₁, step tt S₂ with
    | Some S₁', Some S₂' =>
      Some (State S₁' S₂')
    | _, _ =>
      None
    end.

  Definition _left_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | State S₁ _ => S₁
    end.

  Definition _right_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | State _ S₂ => S₂
    end.

  Fixpoint rename_exp (f : varid → varid) e :=
    match e with
    | Int n => Int n
    | Var x => Var (f x)
    | Sum e₁ e₂ => Sum (rename_exp f e₁) (rename_exp f e₂)
    | Sub e₁ e₂ => Sub (rename_exp f e₁) (rename_exp f e₂)
    | Mult e₁ e₂ => Mult (rename_exp f e₁) (rename_exp f e₂)
    | Div e₁ e₂ => Div (rename_exp f e₁) (rename_exp f e₂)
    | Modulo e₁ e₂ => Modulo (rename_exp f e₁) (rename_exp f e₂)
    end.

  Fixpoint rename_bexp (f : varid → varid) b :=
    match b with
    | True | False => b
    | EQ e₁ e₂ => EQ (rename_exp f e₁) (rename_exp f e₂)
    | LT e₁ e₂ => LT (rename_exp f e₁) (rename_exp f e₂)
    | LE e₁ e₂ => LE (rename_exp f e₁) (rename_exp f e₂)
    | EQB b₁ b₂ => EQB (rename_bexp f b₁) (rename_bexp f b₂)
    | AND b₁ b₂ => AND (rename_bexp f b₁) (rename_bexp f b₂)
    | OR b₁ b₂ => OR (rename_bexp f b₁) (rename_bexp f b₂)
    | NOT b => NOT (rename_bexp f b)
    end.

  Fixpoint rename_cmd (f : varid → varid) c :=
    match c with
    | Skip => Skip
    | Assign x e => Assign (f x) (rename_exp f e)
    | Seq c₁ c₂ => Seq (rename_cmd f c₁) (rename_cmd f c₂)
    | ITE b c₁ c₂ => ITE (rename_bexp f b) (rename_cmd f c₁) (rename_cmd f c₂)
    | While b c' => While (rename_bexp f b) (rename_cmd f c')
    | Assert b => Assert (rename_bexp f b)
    end.

  Definition rename_cont (f : varid → varid) k := List.map (rename_cmd f) k.
  Definition rename_store_helper (f : varid → varid) x v (s : store) := M.add (f x) v s.
  Definition rename_store (f : varid → varid) s :=
    M.fold (rename_store_helper f) s (M.empty _).

  Definition _invariant (m : _genv) (S : _state) : Prop :=
    let '(State (k₁, s₁) (k₂, s₂)) := S in
    let f := proj1_sig m in
    rename_cont f k₁ = k₂ ∧ M.Equal (rename_store f s₁) s₂.

  Lemma rename_store_preserves_values:
    ∀ f, bijective f →
    ∀ x s, M.find (f x) (rename_store f s) = M.find x s.
  Proof.
    intros f [f' Hff'] x s.
    unfold rename_store.
    apply MProps.fold_rec.
    * intros m Hm. rewrite MFacts.empty_o.
      remember (M.find x m) as v eqn:Hv. destruct v ; auto.
      exfalso. apply (Hm x z). apply MFacts.find_mapsto_iff. auto.
    * intros y v s' s'' s''' Hyv. intros Hnys'' Hys''' IH.
      unfold rename_store_helper.
      destruct (Var_as_DT.eq_dec x y).
      + subst y. rewrite Hys'''.
        repeat (rewrite MFacts.add_eq_o ; auto).
      + rewrite MFacts.add_neq_o ; auto.
        - rewrite Hys'''. rewrite MFacts.add_neq_o ; auto.
        - intro Hfxfy. apply n.
          assert (f' (f x) = x) as Hx. rewrite <- Hff'. auto.
          assert (f' (f y) = y) as Hy. rewrite <- Hff'. auto.
          rewrite <- Hx, <- Hy. apply f_equal. auto.
  Qed.

  Lemma rename_eval_exp:
    ∀ f, bijective f →
    ∀ s₁ s₂, M.Equal (rename_store f s₁) s₂ →
    ∀ e, eval_exp s₁ e = eval_exp s₂ (rename_exp f e).
  Proof.
    intros f Hf.
    intros s₁ s₂ Hs₁s₂ e. revert s₁ s₂ Hs₁s₂.
    induction e as [ x | v'
                     | e₁ IH₁ e₂ IH₂
                     | e₁ IH₁ e₂ IH₂
                     | e₁ IH₁ e₂ IH₂
                     | e₁ IH₁ e₂ IH₂
                     | e₁ IH₁ e₂ IH₂ ] ;
      intros s₁ s₂ Hs₁s₂ ; simpl in * ;
        try rewrite (IH₁ s₁ s₂ Hs₁s₂), (IH₂ s₁ s₂ Hs₁s₂) ; auto.
    unfold M.Equal in *. rewrite <- Hs₁s₂.
    rewrite (rename_store_preserves_values f Hf). reflexivity.
  Qed.

  Lemma rename_eval_bexp:
    ∀ f, bijective f →
    ∀ s₁ s₂, M.Equal (rename_store f s₁) s₂ →
    ∀ b, eval_bexp s₁ b = eval_bexp s₂ (rename_bexp f b).
  Proof.
    intros f Hf.
    intros s₁ s₂ Hs₁s₂ b. revert s₁ s₂ Hs₁s₂.
    induction b as [ | | e₁ e₂ | e₁ e₂ | e₁ e₂
                     | b₁ IH₁ b₂ IH₂ | b₁ IH₁ b₂ IH₂ | b₁ IH₁ b₂ IH₂
                     | b₁ IH₁ ] ;
      intros s₁ s₂ Hs₁s₂ ; simpl in * ; auto ;
        repeat rewrite (rename_eval_exp f Hf _ _ Hs₁s₂) ;
        try rewrite (IH₁ s₁ s₂ Hs₁s₂) ;
        try rewrite (IH₂ s₁ s₂ Hs₁s₂) ;
        auto.
  Qed.

  Lemma rename_Equal:
    ∀ f, bijective f →
    ∀ s x, M.find x s = M.find (f x) (rename_store f s).
  Proof.
    intros f [f' Hff'] s. unfold rename_store.
    apply MProps.fold_rec.
    - intros m Hm x. rewrite MFacts.empty_o.
      assert (¬ M.In x m).
      { intro Hfalse. destruct Hfalse as (v, Hv). apply Hm in Hv. auto. }
      apply MFacts.not_find_in_iff. auto.
    - intros y v s' m m' Hy Hm Hm' IH. intro x.
      unfold rename_store_helper.
      destruct (Var_as_DT.eq_dec (f x) (f y)).
      + replace x with y in *.
        * rewrite MFacts.add_eq_o ; auto. rewrite Hm'. rewrite MFacts.add_eq_o ; auto.
        * replace x with (f' (f x)) ; [ | rewrite <- Hff' ; auto].
          replace y with (f' (f y)) ; [ | rewrite <- Hff' ; auto].
          apply f_equal. auto.
      + rewrite MFacts.add_neq_o ; auto. rewrite <- IH.
        rewrite Hm'. rewrite MFacts.add_neq_o ; auto.
        intro Hfalse. apply n. subst x. auto.
  Qed.

  Lemma rename_store_comm_add:
    ∀ f, bijective f →
    ∀ s x v,
    M.Equal (rename_store f (M.add x v s)) (M.add (f x) v (rename_store f s)).
  Proof.
    intros f [f' Hff']. intros s x v. unfold M.Equal.
    intro y. remember (f' y) as k eqn:Hk. symmetry in Hk. rewrite <- Hff' in Hk. subst y.
    rewrite <- (rename_Equal f) ; unfold bijective ; eauto.
    destruct (Var_as_DT.eq_dec x k).
    + subst k. repeat rewrite MFacts.add_eq_o ; auto.
    + repeat rewrite MFacts.add_neq_o ; auto.
      * rewrite <- (rename_Equal f) ; auto.
        unfold bijective. eauto.
      * intro Hfalse. apply n.
        replace x with (f' (f x)) ; [ | rewrite <- Hff' ; auto].
        replace k with (f' (f k)) ; [ | rewrite <- Hff' ; auto].
        apply f_equal. auto.
  Qed.

  Lemma _invariant_1:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      _invariant genv os'.
  Proof.
    intros [f Hf] [S₁ S₂] os' Hinv Hstep.
    simpl in Hstep.
    destruct (step tt S₁) as [S₁' | ] eqn:HS₁',
             (step tt S₂) as [S₂' | ] eqn:HS₂' ;
      inversion Hstep ; clear Hstep ; subst os' ; auto.
    unfold _invariant in *. simpl in *.
    destruct S₁ as [k₁ s₁], S₂ as [k₂ s₂], S₁' as [k₁' s₁'], S₂' as [k₂' s₂'].
    destruct Hinv as [Hk₁k₂ Hs₁s₂]. subst k₂.
    unfold step in *.
    destruct k₁ as [ | c₁ k₁ ] ; try discriminate HS₁'.
    destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ].
    - simpl in HS₂'.
      inversion HS₁'. inversion HS₂'. clear HS₁' HS₂'. subst k₁' k₂' s₁' s₂'.
      split ; auto.
    - simpl in HS₂'.
      inversion HS₁'. inversion HS₂'. clear HS₁' HS₂'. subst k₁' k₂' s₁' s₂'.
      split ; auto.
    - simpl in HS₂'.
      rewrite <- (rename_eval_exp f Hf _ _ Hs₁s₂) in HS₂'.
      destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁ ; try discriminate HS₁'.
      simpl in *.
      inversion HS₁'. inversion HS₂'. clear HS₁' HS₂'. subst k₁' k₂' s₁' s₂'.
      split ; auto. rewrite rename_store_comm_add ; auto.
      apply MFacts.add_m ; auto.
    - simpl in HS₂'.
      rewrite <- (rename_eval_bexp f Hf _ _ Hs₁s₂) in HS₂'.
      destruct (eval_bexp s₁ b₁) as [[ | ] | ] ; try discriminate HS₁' ;
        simpl in * ; inversion HS₁' ; inversion HS₂' ; clear HS₁' HS₂' ;
          subst k₁' k₂' s₁' s₂' ;
          split ; auto.
    - simpl in HS₂'.
      rewrite <- (rename_eval_bexp f Hf _ _ Hs₁s₂) in HS₂'.
      destruct (eval_bexp s₁ b₁) as [[ | ] | ] ; try discriminate HS₁' ;
        simpl in * ; inversion HS₁' ; inversion HS₂' ; clear HS₁' HS₂' ;
          subst k₁' k₂' s₁' s₂' ;
          split ; auto.
    - simpl in HS₂'.
      rewrite <- (rename_eval_bexp f Hf _ _ Hs₁s₂) in HS₂'.
      destruct (eval_bexp s₁ b₁) as [[ | ] | ] ; try discriminate HS₁' ;
        simpl in * ; inversion HS₁' ; inversion HS₂' ; clear HS₁' HS₂' ;
          subst k₁' k₂' s₁' s₂' ;
          split ; auto.
  Qed.

  Lemma _prediction_soundness :
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (_left_state  os))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (_right_state os))
                       (Some (_right_state os'))
        ∧ n₁ + n₂ > 0.
  Proof.
    intros [f Hf] [[k₁ s₁] [k₂ s₂]] os' [Hk₁k₂ Hs₁s₂] Hos'.
    simpl in *.
    subst k₂.
    destruct k₁ as [ | c₁ k₁ ] ; simpl ; try discriminate Hos'.
    destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ] ; simpl ;
      try (inversion Hos' ; clear Hos' ; subst os') ; exists 1, 1 ;
        unfold nsteps, napply ;
        simpl in * ; try (repeat split ; auto using opt_state_eq_refl ; fail).
    - rewrite <- (rename_eval_exp _ Hf _ _ Hs₁s₂) in *.
      destruct (eval_exp s₁ e₁) as [v | ] eqn:Hv ;
        inversion Hos' ; clear Hos' ; subst os' ;
          repeat split ; simpl ; auto using opt_state_eq_refl.
    - rewrite <- (rename_eval_bexp _ Hf _ _ Hs₁s₂) in *.
      destruct (eval_bexp s₁ b₁) as [v | ] eqn:Hv ;
        inversion Hos' ; clear Hos' ; subst os' ;
          repeat split ; simpl ; auto using opt_state_eq_refl.
    - rewrite <- (rename_eval_bexp _ Hf _ _ Hs₁s₂) in *.
      destruct (eval_bexp s₁ b₁) as [v | ] eqn:Hv ;
        inversion Hos' ; clear Hos' ; subst os' ;
          repeat split ; simpl ; auto using opt_state_eq_refl.
    - rewrite <- (rename_eval_bexp _ Hf _ _ Hs₁s₂) in *.
      destruct (eval_bexp s₁ b₁) as [[ | ] | ] eqn:Hv ;
        inversion Hos' ; clear Hos' ; subst os' ;
          repeat split ; simpl ; auto using opt_state_eq_refl.
  Qed.

  Lemma _prediction_completeness :
    ∀ genv os,
      _invariant genv os →
      _step genv os = None →
      step tt (_left_state  os) = None
      ∧ step tt (_right_state os) = None.
  Proof.
    intros [f Hf] [[k₁ s₁] [k₂ s₂]] [Hk₁k₂ Hs₁s₂] HNone.
    simpl in *.
    subst k₂.
    destruct k₁ as [ | c₁ k₁ ] ; simpl ; auto.
    destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ] ; simpl ;
      try discriminate HNone.
    - simpl in HNone.
      rewrite <- (rename_eval_exp _ Hf _ _ Hs₁s₂) in *.
      destruct (eval_exp s₁ e₁) ; auto ; discriminate HNone.
    - simpl in HNone.
      rewrite <- (rename_eval_bexp _ Hf _ _ Hs₁s₂) in *.
      destruct (eval_bexp s₁ b₁) ; auto ; discriminate HNone.
    - simpl in HNone.
      rewrite <- (rename_eval_bexp _ Hf _ _ Hs₁s₂) in *.
      destruct (eval_bexp s₁ b₁) ; auto ; discriminate HNone.
    - simpl in HNone.
      rewrite <- (rename_eval_bexp _ Hf _ _ Hs₁s₂) in *.
      destruct (eval_bexp s₁ b₁) as [[ | ] | ] ; auto ; discriminate HNone.
  Qed.

  Definition renaming_oracle : oracle_language imp_language imp_language :=
    {|
      oracle_state := _state;
      oracle_genv := _genv;
      oracle_step := _step;
      left_state := _left_state;
      right_state := _right_state;
      left_genv := λ _, υ;
      right_genv := λ _, υ;
      invariant := _invariant;
      invariant_1 := _invariant_1;
      prediction_soundness := _prediction_soundness;
      prediction_completeness := _prediction_completeness
    |}.
End Make.
