(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

Require Import Common.
Require Import Imp.Lang.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.
Require Import OracleHelpers.

Module Make
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT)
       (Import OracleHelpers : ImpOracleHelpers Var_as_DT Lang).
  Inductive mod :=
  | UnfoldSeq
  | FoldSeq.

  Inductive _cont_mod :=
  | ExtraUnfoldLeft
  | ExtraUnfoldRight.

  Definition _genv := unit.

  Inductive _state : Type :=
  | State : state → state → list (cont_mod mod _cont_mod) → _state.

  Definition perform_n_m_steps
             (n m : nat) (S₁ S₂ : state)
             (km' : list (cont_mod mod _cont_mod)) : option _state :=
    match nsteps' υ n S₁, nsteps' υ m S₂ with
    | (S₁', _), (S₂', _) => Some (State S₁' S₂' km')
    end.

  Definition _step (_ : _genv) (S : _state) : option _state :=
    match S with
    | State S₁ S₂ km =>
      match step_helper S₁ S₂ km with
      | None =>
        None
      | Some StuckStep =>
        None
      | Some (GenericStep S₁' S₂' km') =>
        Some (State S₁' S₂' km')
      | Some (SpecialCmdStep m km') =>
        match m with
        | UnfoldSeq =>
          perform_n_m_steps 2 1 S₁ S₂ (CmdMod Id::ContMod ExtraUnfoldRight::km')
        | FoldSeq =>
          perform_n_m_steps 1 2 S₁ S₂ (CmdMod Id::ContMod ExtraUnfoldLeft::km')
        end
      | Some (SpecialContStep m km') =>
        match m with
        | ExtraUnfoldLeft =>
          step υ S₁ >>= λ S₁',
          Some (State S₁' S₂ (CmdMod Id::CmdMod Id::km'))
        | ExtraUnfoldRight =>
          step υ S₂ >>= λ S₂',
          Some (State S₁ S₂' (CmdMod Id::CmdMod Id::km'))
        end
      end
    end.

  Definition _left_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | State S₁ _ _ =>
      S₁
    end.

  Definition _right_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | State _ S₂ _ =>
      S₂
    end.

  Definition unfold_seq c :=
    match c with
    | Seq (Seq c₁ c₂) c₃ =>
      ret (Seq c₁ (Seq c₂ c₃))
    | _ => None
    end.

  Definition fold_seq c :=
    match c with
    | Seq c₁ (Seq c₂ c₃) =>
      ret (Seq (Seq c₁ c₂) c₃)
    | _ => None
    end.

  Fixpoint apply_cmd_mod c m {struct m} :=
    match m with
    | Id => Some c
    | RecMod m₁ m₂ =>
      match c with
      | Seq c₁ c₂ =>
        apply_cmd_mod c₁ m₁ >>= λ c₁',
        apply_cmd_mod c₂ m₂ >>= λ c₂',
        Some (Seq c₁' c₂')
      | ITE b c₁ c₂ =>
        apply_cmd_mod c₁ m₁ >>= λ c₁',
        apply_cmd_mod c₂ m₂ >>= λ c₂',
        Some (ITE b c₁' c₂')
      | While b c =>
        apply_cmd_mod c m₁ >>= λ c',
        Some (While b c')
      | Skip | Assign _ _ | Assert _ =>
        None
      end
    | LeafMod UnfoldSeq => unfold_seq c
    | LeafMod FoldSeq => fold_seq c
    end.

  Lemma apply_cmd_mod_id:
    ∀ c, apply_cmd_mod c Id = Some c.
  Proof.
    intro c. destruct c ; simpl ; auto.
  Qed.

  Inductive km_invariant : list (cont_mod mod _cont_mod) → cont → cont → Prop :=
  | empty_cont_mod : km_invariant [] [] []
  | extra_unfold_left_inv :
      ∀ c₁ c₂ km k₁ k₂,
        km_invariant km k₁ k₂ →
        km_invariant (ContMod ExtraUnfoldLeft::km)
                     ((Seq c₁ c₂)::k₁)
                     (c₁::c₂::k₂)
  | extra_unfold_right_inv :
      ∀ c₁ c₂ km k₁ k₂,
        km_invariant km k₁ k₂ →
        km_invariant (ContMod ExtraUnfoldRight::km)
                     (c₁::c₂::k₁)
                     ((Seq c₁ c₂)::k₂)
  | cmd_cont : ∀ c c' m km k₁ k₂,
      apply_cmd_mod c m = Some c' →
      km_invariant km k₁ k₂ →
      km_invariant (CmdMod m::km) (c::k₁) (c'::k₂).

  Definition leaf_mod_of_cmds (c₁ c₂ : cmd) : option mod :=
    match unfold_seq c₁ >>= λ c₁',
          if beq_cmd c₁' c₂ then
            Some UnfoldSeq
          else
            None
    with
    | Some m => Some m
    | None =>
      fold_seq c₁ >>= λ c₁',
      if beq_cmd c₁' c₂ then
        Some FoldSeq
      else
        None
    end.

  Definition _invariant (_ : _genv) (S : _state) : Prop :=
    match S with
    | State (k₁, s₁) (k₂, s₂) km =>
      km_invariant km k₁ k₂ ∧ M.Equal s₁ s₂
    end.

  Lemma _invariant_1:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      _invariant genv os'.
  Proof.
    intros [] [ [k₁ s₁] [k₂ s₂] km ] os' Hinv Hstep.
    - simpl in Hstep.
      unfold _invariant in *.
      destruct Hinv as [Hkm Hs₁s₂].
      inversion Hkm as [
                      | c₁ c₂ km' k₁' k₂' Hkm'
                      | c₁ c₂ km' k₁' k₂' Hkm'
                      | c₁ c₂ m km' k₁' k₂' Hc₁c₂ Hkm' ] ;
        subst km k₁ k₂ ; try discriminate Hstep.
      + simpl in Hstep.
        inversion Hstep. clear Hstep. subst os'.
        split ; auto using cmd_cont.
      + simpl in Hstep.
        inversion Hstep. clear Hstep. subst os'.
        split ; auto using cmd_cont.
      + simpl in Hstep. destruct m as [ m | | m₁ m₂ ].
        * destruct m as [ | ].
          { destruct c₁ as [ | [ | c₁₁ c₁₂ | | | | ] c₁₃ | | | | ] ;
              try discriminate Hc₁c₂.
            inversion Hc₁c₂ as [H'].
            clear Hc₁c₂. subst c₂.
            unfold perform_n_m_steps in Hstep.
            simpl in Hstep. inversion Hstep. clear Hstep. subst os'.
            split ; auto using cmd_cont, extra_unfold_right_inv. }
          { destruct c₁ as [ | c₁₁ [ | c₁₂ c₁₃ | | | | ] | | | | ] ;
              try discriminate Hc₁c₂.
            inversion Hc₁c₂ as [H'].
            clear Hc₁c₂. subst c₂.
            unfold perform_n_m_steps in Hstep.
            simpl in Hstep. inversion Hstep. clear Hstep. subst os'.
            split ; auto using cmd_cont, extra_unfold_left_inv. }
        * rewrite apply_cmd_mod_id in Hc₁c₂. inversion Hc₁c₂ as [H'c₁c₂].
          clear Hc₁c₂. subst c₂.
          destruct c₁ as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ] ;
            try discriminate Hstep.
          { simpl in Hstep.
            inversion Hstep ; split ;
              auto using cmd_cont, apply_cmd_mod_id. }
          { simpl in Hstep.
            inversion Hstep ; split ;
              auto using cmd_cont, apply_cmd_mod_id. }
          { rewrite (store_equal_eval_exp e s₁ s₂) in Hstep ; auto.
            destruct (eval_exp s₂ e) as [v₁ | ] eqn:Hv₁ ;
              inversion Hstep ; clear Hstep ; subst os' ;
                repeat split ;
                auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m. }
          { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
            destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
              inversion Hstep ; clear Hstep ; subst os' ;
                repeat split ;
                auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m. }
          { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
            destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
              inversion Hstep ; clear Hstep ; subst os' ;
                repeat split ;
                auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m ;
                simpl ; try rewrite (store_equal_eval_bexp b s₁ s₂) ; auto ;
                  rewrite Hv₁ ; reflexivity. }
          { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
            destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
              inversion Hstep ; clear Hstep ; subst os' ;
                repeat split ;
                auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m ;
                simpl ; try rewrite (store_equal_eval_bexp b s₁ s₂) ; auto ;
                  rewrite Hv₁ ; reflexivity. }
        *  destruct c₁ as [ | c₁₁ c₁₂ | | b₁ c₁₁ c₁₂ | b₁ c₁ | ] ;
             try discriminate Hc₁c₂.
           { simpl in Hc₁c₂.
             destruct (apply_cmd_mod c₁₁ m₁) as [c₁' | ] eqn:Hc₁' ;
               try discriminate Hc₁c₂.
             simpl in Hc₁c₂.
             destruct (apply_cmd_mod c₁₂ m₂) as [c₂' | ] eqn:Hc₂' ;
               try discriminate Hc₁c₂.
             inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
             inversion Hstep. clear Hstep. subst os'.
             auto using cmd_cont. }
           { simpl in Hc₁c₂.
             destruct (apply_cmd_mod c₁₁ m₁) as [c₁' | ] eqn:Hc₁' ;
               try discriminate Hc₁c₂.
             simpl in Hc₁c₂.
             destruct (apply_cmd_mod c₁₂ m₂) as [c₂' | ] eqn:Hc₂' ;
               try discriminate Hc₁c₂.
             inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
             rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
             destruct (eval_bexp s₂ b₁) as [[ | ] | ] eqn:Hv ;
               try discriminate Hstep ;
               simpl in Hstep ; inversion Hstep ; clear Hstep ; subst os' ;
               repeat split ; auto using cmd_cont ;
                 simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂) ; auto ;
                   rewrite Hv ; reflexivity. }
           { simpl in Hc₁c₂.
             destruct (apply_cmd_mod c₁ m₁) as [c₁' | ] eqn:Hc₁' ;
               try discriminate Hc₁c₂.
             inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
             rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
             destruct (eval_bexp s₂ b₁) as [[ | ] | ] eqn:Hv ;
               try discriminate Hstep ;
               inversion Hstep ; clear Hstep ; subst os' ;
                 repeat split ; auto using cmd_cont ;
                   simpl ; try rewrite (store_equal_eval_bexp _ s₁ s₂) ; auto ;
                     try rewrite Hv ; reflexivity. }
  Qed.

  Lemma _prediction_soundness :
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (_left_state  os))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (_right_state os))
                       (Some (_right_state os'))
        ∧ n₁ + n₂ > 0.
  Proof.
    intros [ ] [[k₁ s₁] [k₂ s₂] km ] os' Hinv Hos'.
    simpl in *.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ km' k₁' k₂' Hkm'
                    | c₁ c₂ km' k₁' k₂' Hkm'
                    | c₁ c₂ m km' k₁' k₂' Hc₁c₂ Hkm' ] ;
      subst km k₁ k₂ ; try discriminate Hos'.
    - simpl in Hos'. inversion Hos'. clear Hos'. subst os'.
      exists 1, 0. simpl.
      repeat split ; auto using opt_state_eq_refl with arith.
    - simpl in Hos'. inversion Hos'. clear Hos'. subst os'.
      exists 0, 1. simpl.
      repeat split ; auto using opt_state_eq_refl with arith.
    - destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H ;
        try (inversion Hos' ; clear Hos' ; subst os').
      + exists 1, 1.
        destruct (generic_step_bistep _ _ _ _ _ _ H) as [H₁ H₂].
        repeat split ; unfold _left_state, _right_state, nsteps, napply ; auto ;
          simpl (ProgrammingLanguage.step imp_language).
        * rewrite H₁. simpl. apply opt_state_eq_refl.
        * rewrite H₂. simpl. apply opt_state_eq_refl.
      + apply step_helper_special_cmd_step_inv in H.
        destruct H as [Hm Hkm'']. inversion Hm as [H'm]. clear Hm.
        subst km'' m.
        destruct m'.
        * destruct c₁ as [ | [ | c₁₁ c₁₂ | | | | ] c₁₃ | | | | ] ;
            try discriminate Hc₁c₂.
          simpl in Hc₁c₂. inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
          unfold perform_n_m_steps in Hos'. simpl in Hos'.
          inversion Hos'. clear Hos'. subst os'.
          exists 2, 1. simpl.
          repeat split ; auto using opt_state_eq_refl with arith.
        * destruct c₁ as [ | c₁₁ [ | c₁₂ c₁₃ | | | | ] | | | | ] ;
            try discriminate Hc₁c₂.
          simpl in Hc₁c₂. inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
          unfold perform_n_m_steps in Hos'. simpl in Hos'.
          inversion Hos'. clear Hos'. subst os'.
          exists 1, 2. simpl.
          repeat split ; auto using opt_state_eq_refl with arith.
      + apply step_helper_special_cont_step_inv in H.
        destruct H as [H _]. discriminate H.
      + apply step_helper_special_cont_step_inv in H.
        destruct H as [H _]. discriminate H.
  Qed.

  Lemma _prediction_completeness :
    ∀ genv os,
      _invariant genv os →
      _step genv os = None →
      step υ (_left_state  os) = None
      ∧ step υ (_right_state os) = None.
  Proof.
    intros [ ] [[k₁ s₁] [k₂ s₂] km ] Hinv HNone.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | c₁ c₂ km' k₁' k₂' Hkm'
                    | c₁ c₂ km' k₁' k₂' Hkm'
                    | c₁ c₂ m km' k₁' k₂' Hc₁c₂ Hkm' ] ;
      subst km k₁ k₂ ; try discriminate HNone.
    { simpl. auto. }
    unfold _step in HNone.
    destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H ;
      try discriminate.
    - apply step_helper_special_cmd_step_inv in H.
      destruct H as [Hm Hkm'']. inversion Hm as [H'm]. clear Hm.
      subst km'' m.
      destruct m'.
      + destruct c₁ as [ | [ | c₁₁ c₁₂ | | | | ] c₁₃ | | | | ] ;
          try discriminate Hc₁c₂.
        simpl in Hc₁c₂. inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
        discriminate HNone.
      + destruct c₁ as [ | c₁₁ [ | c₁₂ c₁₃ | | | | ] | | | | ] ;
          try discriminate Hc₁c₂.
        simpl in Hc₁c₂. inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
        discriminate HNone.
    - simpl in H.
      destruct m as [ m | | m₁ m₂ ].
      + discriminate H.
      + destruct c₁ ; inversion Hc₁c₂ ; clear Hc₁c₂ ; subst c₂ ; try discriminate H.
        * rewrite (store_equal_eval_exp e s₁ s₂) in H ; auto.
          destruct (eval_exp s₂ e) ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
      + destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ] ;
          try discriminate Hc₁c₂.
        * simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁₁ m₁), (apply_cmd_mod c₁₂ m₂) ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂. discriminate H.
        * simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁₁ m₁), (apply_cmd_mod c₁₂ m₂) ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
          simpl in H.
          rewrite (store_equal_eval_bexp b₁ s₁ s₂) in H ; auto.
          simpl in H.
          destruct (eval_bexp s₂ b₁) as [[|]|] ;
            try discriminate H.
        * simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁ m₁) ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
          simpl in H.
          rewrite (store_equal_eval_bexp b₁ s₁ s₂) in H ; auto.
          simpl in H.
          destruct (eval_bexp s₂ b₁) as [[|]|] ;
            try discriminate H.
    - apply step_helper_special_cont_step_inv in H.
      destruct H as [H _]. discriminate.
    - apply generic_step_bicrash in H.
      simpl _left_state. simpl _right_state. auto.
    - simpl in H.
      destruct m as [ m | | m₁ m₂ ].
      + discriminate H.
      + destruct c₁ ; inversion Hc₁c₂ ; clear Hc₁c₂ ; subst c₂ ; try discriminate H.
        * rewrite (store_equal_eval_exp e s₁ s₂) in H ; auto.
          destruct (eval_exp s₂ e) ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
        * rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
          destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H.
      + destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ] ;
          try discriminate Hc₁c₂.
        * simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁₁ m₁), (apply_cmd_mod c₁₂ m₂) ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂. discriminate H.
        * simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁₁ m₁), (apply_cmd_mod c₁₂ m₂) ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
          simpl in H.
          rewrite (store_equal_eval_bexp b₁ s₁ s₂) in H ; auto.
          simpl in H.
          destruct (eval_bexp s₂ b₁) as [[|]|] ;
            try discriminate H.
        * simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁ m₁) ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
          simpl in H.
          rewrite (store_equal_eval_bexp b₁ s₁ s₂) in H ; auto.
          simpl in H.
          destruct (eval_bexp s₂ b₁) as [[|]|] ;
            try discriminate H.
  Qed.

  Definition seqassoc_oracle : oracle_language imp_language imp_language :=
    {|
      oracle_state := _state;
      oracle_genv := unit;
      oracle_step := _step;
      left_state := _left_state;
      right_state := _right_state;
      left_genv := λ _, υ;
      right_genv := λ _, υ;
      invariant := _invariant;
      invariant_1 := _invariant_1;
      prediction_soundness := _prediction_soundness;
      prediction_completeness := _prediction_completeness
    |}.
End Make.