(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

Require Import Common.
Require Import Imp.Lang.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.
Require Import OracleHelpers.

Module Make
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT)
       (Import OracleHelpers : ImpOracleHelpers Var_as_DT Lang).
  Inductive mod :=
  | Swap.

  Inductive _cont_mod :=
  | SwapCont.

  Definition _genv := unit.

  Inductive _state : Type :=
  | Stuck : state → state → _state
  | State : state → state → list (cont_mod mod _cont_mod) → _state.

  (* Perform the same number of steps in both programs.
     The programs are assumed to have roughly the same crashing behaviors:
     if one crashes in at most the given [n] steps, so does the other
     (although it is not assumed to crash in the exact same number of
     steps) *)
  Definition perform_n_steps
             (n : nat) (S₁ S₂ : state)
             (km' : list (cont_mod mod _cont_mod)) : option _state :=
    (* Sanity check: if both programs are already stuck, return None *)
    match step υ S₁, step υ S₂ with
    | None, None =>
      None
    | _, _ =>
      match nsteps' υ n S₁, nsteps' υ n S₂ with
      | (S₁', 0), (S₂', 0) =>
        (* No crash in either program *)
        Some (State S₁' S₂' km')
      | (S₁', S _), (S₂', S _) =>
        (* Crash in both programs *)
        Some (Stuck S₁' S₂')
      | _, _ =>
        (* Programs have different crashing behaviors! *)
        None
      end
    end.

  Definition _step (_ : _genv) (S : _state) : option _state :=
    match S with
    | Stuck S₁ S₂ =>
      None
    | State S₁ S₂ km =>
      match step_helper S₁ S₂ km with
      | None =>
        None
      | Some StuckStep =>
        None
      | Some (GenericStep S₁' S₂' km') =>
        Some (State S₁' S₂' km')
      | Some (SpecialCmdStep _ km') =>
        perform_n_steps 1 S₁ S₂ ((ContMod SwapCont)::km')
      | Some (SpecialContStep _ km') =>
        perform_n_steps 2 S₁ S₂ km'
      end
    end.

  Definition _left_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | Stuck S₁ _ | State S₁ _ _ =>
      S₁
    end.

  Definition _right_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | Stuck _ S₂ | State _ S₂ _ =>
      S₂
    end.

  Definition swap_assignments c :=
    match c with
    | Seq (Assign x e) (Assign y e') =>
      ret (Seq (Assign y e') (Assign x e))
    | _ => None
    end.

  Fixpoint apply_cmd_mod c m :=
    match m with
    | Id => Some c
    | RecMod m₁ m₂ =>
      match c with
      | Seq c₁ c₂ =>
        apply_cmd_mod c₁ m₁ >>= λ c₁',
        apply_cmd_mod c₂ m₂ >>= λ c₂',
        Some (Seq c₁' c₂')
      | ITE b c₁ c₂ =>
        apply_cmd_mod c₁ m₁ >>= λ c₁',
        apply_cmd_mod c₂ m₂ >>= λ c₂',
        Some (ITE b c₁' c₂')
      | While b c =>
        apply_cmd_mod c m₁ >>= λ c',
        Some (While b c')
      | Skip | Assign _ _ | Assert _ =>
        None
      end
    | LeafMod Swap => swap_assignments c
    end.

  Lemma apply_cmd_mod_id:
    ∀ c, apply_cmd_mod c Id = Some c.
  Proof.
    intro c. destruct c ; simpl ; auto.
  Qed.

  Definition assign_indep' c :=
    match c with
    | Seq (Assign x e) (Assign y e') =>
      if In_exp_var_dec x e' then
        false
      else
        if In_exp_var_dec y e then
          false
        else
          negb (MFacts.eqb x y)
    | _ => false
    end.

  Fixpoint assign_indep m c :=
    match m with
    | Id => true
    | LeafMod Swap =>
      assign_indep' c
    | RecMod m₁ m₂ =>
      match c with
      | Seq c₁ c₂ | ITE _ c₁ c₂ => assign_indep m₁ c₁ && assign_indep m₂ c₂
      | While _ c₁ => assign_indep m₁ c₁
      | Skip | Assign _ _ | Assert _ => false
      end
    end.

  Inductive km_invariant : list (cont_mod mod _cont_mod) → cont → cont → Prop :=
  | empty_cont_mod : km_invariant [] [] []
  | cont_mod_inv :
      ∀ x₁ e₁ x₂ e₂ km k₁ k₂,
        x₁ ≠ x₂ →
        (¬ In_exp_var x₁ e₂) →
        (¬ In_exp_var x₂ e₁) →
        km_invariant km k₁ k₂ →
        km_invariant (ContMod SwapCont::km)
                     (Assign x₁ e₁::Assign x₂ e₂::k₁)
                     (Assign x₂ e₂::Assign x₁ e₁::k₂)
  | cmd_cont : ∀ c c' m km k₁ k₂,
      assign_indep m c = true →
      apply_cmd_mod c m = Some c' →
      km_invariant km k₁ k₂ →
      km_invariant (CmdMod m::km) (c::k₁) (c'::k₂).

  Definition _invariant (_ : _genv) (S : _state) : Prop :=
    match S with
    | Stuck S₁ S₂ => step υ S₁ = None ∧ step υ S₂ = None
    | State (k₁, s₁) (k₂, s₂) km =>
      km_invariant km k₁ k₂ ∧ M.Equal s₁ s₂
    end.

  Lemma eval_exp_add_neq:
    ∀ s x v e,
      (¬ In_exp_var x e) →
      eval_exp s e = eval_exp (M.add x v s) e.
  Proof.
    intros s x v e.
    induction e as
        [ y | n | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂
          | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ ] ;
      intros Hxe.
    - simpl in Hxe. simpl. rewrite MFacts.add_neq_o ; auto.
    - reflexivity.
    - simpl. simpl in Hxe.
      rewrite IH₁, IH₂ ; auto.
    - simpl. simpl in Hxe.
      rewrite IH₁, IH₂ ; auto.
    - simpl. simpl in Hxe.
      rewrite IH₁, IH₂ ; auto.
    - simpl. simpl in Hxe.
      rewrite IH₁, IH₂ ; auto.
    - simpl. simpl in Hxe.
      rewrite IH₁, IH₂ ; auto.
  Qed.

  Lemma eval_exp_add_neq':
    ∀ s₁ s₂ x v e,
      M.Equal s₁ s₂ →
      (¬ In_exp_var x e) →
      eval_exp s₁ e = eval_exp (M.add x v s₂) e.
  Proof.
    intros s₁ s₂ x v e Hs₁s₂ Hxe.
    rewrite <- (eval_exp_add_neq s₂ x v) ; auto.
    apply store_equal_eval_exp ; auto.
  Qed.

  Lemma map_swap_add_neq:
    ∀ {X : Type} s₁ s₂ x y (v₁ v₂ : X),
      M.Equal s₁ s₂ →
      x ≠ y →
      M.Equal (M.add x v₁ (M.add y v₂ s₁)) (M.add y v₂ (M.add x v₁ s₂)).
  Proof.
    intros X s₁ s₂ x y v₁ v₂ H Hxy.
    intro z.
    destruct (Var_as_DT.eq_dec x z) as [Hz | Hz].
    - subst z. rewrite MFacts.add_eq_o ; auto.
      rewrite MFacts.add_neq_o ; auto.
      rewrite MFacts.add_eq_o ; auto.
    - rewrite MFacts.add_neq_o ; auto.
      destruct (Var_as_DT.eq_dec y z) as [H'z | H'z].
      + subst z.
        repeat (rewrite MFacts.add_eq_o ; auto).
      + repeat (rewrite MFacts.add_neq_o ; auto).
  Qed.

  Lemma _invariant_1:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      _invariant genv os'.
  Proof.
    intros [] [ | [k₁ s₁] [k₂ s₂] km ] os' Hinv Hstep ;
      try discriminate Hstep.
    simpl in Hstep.
    unfold _invariant in *.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | x₁ e₁ x₂ e₂ km' k₁' k₂' Hx₁x₂ Hx₁e₂ Hx₂e₁ Hkm'
                    | c₁ c₂ m km' k₁' k₂' Hindep Hc₁c₂ Hkm' ] ;
      subst km k₁ k₂ ; try discriminate Hstep.
    + simpl in Hstep.
      unfold perform_n_steps, nsteps' in Hstep.
      simpl in Hstep.
      destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁.
      * simpl in Hstep.
        rewrite (eval_exp_add_neq' s₂ s₁ x₁ v₁) in Hstep ;
          auto using MFacts.Equal_sym.
        destruct (eval_exp (M.add x₁ v₁ s₁) e₂) as [v₂ | ] eqn:Hv₂.
        { simpl in Hstep.
          rewrite <- (eval_exp_add_neq' s₁ s₂ x₂ v₂) in Hstep ;
            auto.
          rewrite Hv₁ in Hstep. simpl in Hstep.
          inversion Hstep as [H']. clear Hstep. subst os'.
          split ; auto.
          apply map_swap_add_neq ; auto. }
        { simpl in Hstep.
          inversion Hstep. clear Hstep. subst os'.
          simpl. rewrite Hv₂. simpl.
          rewrite (eval_exp_add_neq' s₂ s₁ x₁ v₁) ; auto using MFacts.Equal_sym.
          rewrite Hv₂. auto. }
      * simpl in Hstep.
        destruct (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂.
        { simpl in Hstep.
          rewrite <- (eval_exp_add_neq' s₁ s₂ x₂ v₂) in Hstep ;
            auto.
          rewrite Hv₁ in Hstep. simpl in Hstep.
          inversion Hstep. clear Hstep. subst os'.
          simpl.
          rewrite <- (eval_exp_add_neq' s₁ s₂ x₂ v₂) ;
            auto.
          rewrite Hv₁. auto. }
        { simpl in Hstep. inversion Hstep. }
    + simpl in Hstep. destruct m as [ m | | m₁ m₂ ].
      * destruct m as [ ].
        destruct c₁ as [ | [ | | x e | | | ] [ | | y e' | | | ] | | | | ] ;
          try discriminate Hc₁c₂.
        inversion Hc₁c₂ as [H'].
        clear Hc₁c₂. subst c₂.
        simpl in Hindep.
        destruct (In_exp_var_dec x e') as [ | Hxe'] ;
          try discriminate Hindep.
        destruct (In_exp_var_dec y e) as [ | Hye] ;
          try discriminate Hindep.
        rewrite negb_true_iff in Hindep.
        unfold MFacts.eqb in Hindep.
        destruct (MFacts.eq_dec x y) as [ | Hxy] ;
          try discriminate Hindep.
        clear Hindep.
        unfold perform_n_steps in Hstep.
        simpl in Hstep. inversion Hstep. clear Hstep. subst os'.
        eauto using cont_mod_inv.
      * rewrite apply_cmd_mod_id in Hc₁c₂. inversion Hc₁c₂ as [H'c₁c₂].
        clear Hc₁c₂ Hindep. subst c₂.
        destruct c₁ as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b ] ;
          try discriminate Hstep.
        { simpl in Hstep.
          inversion Hstep ; split ;
            auto using cmd_cont, apply_cmd_mod_id. }
        { simpl in Hstep.
          inversion Hstep ; split ;
            auto using cmd_cont, apply_cmd_mod_id. }
        { rewrite (store_equal_eval_exp e s₁ s₂) in Hstep ; auto.
          destruct (eval_exp s₂ e) as [v₁ | ] eqn:Hv₁ ;
            simpl in Hstep ;
            inversion Hstep ; clear Hstep ; subst os' ; split ;
              auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m ;
              simpl ; try rewrite (store_equal_eval_exp e s₁ s₂) ;
                try rewrite Hv₁ ; auto. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
            inversion Hstep ; clear Hstep ; subst os' ; split ;
              auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m ;
              simpl ; try rewrite (store_equal_eval_bexp b s₁ s₂) ;
                try rewrite Hv₁ ; auto. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
            inversion Hstep ; clear Hstep ; subst os' ; split ;
              auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m ;
              simpl ; try rewrite (store_equal_eval_bexp b s₁ s₂) ;
                try rewrite Hv₁ ; auto. }
        { rewrite (store_equal_eval_bexp b s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv₁ ;
            inversion Hstep ; clear Hstep ; subst os' ; split ;
              auto using cmd_cont, apply_cmd_mod_id, MFacts.add_m ;
              simpl ; try rewrite (store_equal_eval_bexp b s₁ s₂) ;
                try rewrite Hv₁ ; auto. }
      * destruct c₁ as [ | c₁₁ c₁₂ | | b₁ c₁₁ c₁₂ | b₁ c₁ | ] ;
          try discriminate Hc₁c₂.
        { simpl in Hstep.
          simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁₁ m₁) as [c₁' | ] eqn:Hc₁' ;
            try discriminate Hc₁c₂.
          simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁₂ m₂) as [c₂' | ] eqn:Hc₂' ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
          inversion Hstep. clear Hstep. subst os'.
          simpl in Hindep. rewrite andb_true_iff in Hindep.
          destruct Hindep as [H₁ H₂].
          auto using cmd_cont. }
        { simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁₁ m₁) as [c₁' | ] eqn:Hc₁' ;
            try discriminate Hc₁c₂.
          simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁₂ m₂) as [c₂' | ] eqn:Hc₂' ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
          rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
          simpl in Hindep. rewrite andb_true_iff in Hindep.
          destruct Hindep as [H₁ H₂].
          destruct (eval_bexp s₂ b₁) as [[ | ] | ] eqn:Hv ;
            try discriminate Hstep ;
            simpl in Hstep ; inversion Hstep ; clear Hstep ; subst os' ;
              split ; auto using cmd_cont ;
                simpl ; try rewrite (store_equal_eval_bexp b₁ s₁ s₂) ;
                  try rewrite Hv ; auto. }
        { simpl in Hc₁c₂.
          destruct (apply_cmd_mod c₁ m₁) as [c₁' | ] eqn:Hc₁' ;
            try discriminate Hc₁c₂.
          inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
          rewrite (store_equal_eval_bexp _ s₁ s₂) in Hstep ; auto.
          destruct (eval_bexp s₂ b₁) as [[ | ] | ] eqn:Hv ;
            try discriminate Hstep ;
            inversion Hstep ; clear Hstep ; subst os' ;
              split ; auto using cmd_cont ;
                simpl ; try rewrite (store_equal_eval_bexp b₁ s₁ s₂) ;
                  try rewrite Hv ; auto. }
  Qed.

  Lemma _prediction_soundness :
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →      
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (_left_state  os))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (_right_state os))
                       (Some (_right_state os'))
        ∧ n₁ + n₂ > 0.
  Proof.
    intros [ ] [ S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] os' Hinv Hos' ;
      try discriminate Hos'.
    simpl in *.
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | x₁ e₁ x₂ e₂ km' k₁' k₂' Hx₁x₂ Hx₁e₂ Hx₂e₁
                    | c₁ c₂ m km' k₁' k₂' Hindep Hc₁c₂ Hkm' ] ;
      subst km k₁ k₂ ;
      try discriminate Hos'.
    - simpl in Hos'.
      unfold perform_n_steps in *.
      unfold nsteps' in *.
      simpl in Hos'.
      destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁.
      + simpl in Hos'.
        rewrite <- (eval_exp_add_neq' s₂ s₁ x₁ v₁) in Hos' ;
          auto using MFacts.Equal_sym.
        destruct (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂.
        * simpl in Hos'.
          rewrite <- (eval_exp_add_neq' s₁ s₂ x₂ v₂) in Hos' ; auto.
          rewrite Hv₁ in Hos'.
          simpl in Hos'.
          inversion Hos'. clear Hos'. subst os'.
          exists 2, 2. unfold nsteps. simpl.
          rewrite Hv₁, Hv₂. simpl.
          rewrite <- (eval_exp_add_neq' s₁ s₂ x₂ v₂) ; auto.
          rewrite <- (eval_exp_add_neq' s₂ s₁ x₁ v₁) ; auto using MFacts.Equal_sym.
          rewrite Hv₁, Hv₂. simpl.
          repeat split ; auto using opt_state_eq_refl with arith.
        * simpl in Hos'.
          inversion Hos'. clear Hos'. subst os'.
          exists 1, 0. unfold nsteps. simpl.
          rewrite Hv₁. simpl. split ; auto using opt_state_eq_refl.
      + simpl in Hos'.
        destruct (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂.
        * simpl in Hos'.
          rewrite <- (eval_exp_add_neq' s₁ s₂ x₂ v₂) in Hos' ; auto.
          rewrite Hv₁ in Hos'.
          simpl in Hos'.
          inversion Hos'. clear Hos'. subst os'.
          exists 0, 1. unfold nsteps. simpl.
          rewrite Hv₂. simpl.
          split ; auto using opt_state_eq_refl.
        * simpl in Hos'. discriminate Hos'.
    - destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H.
      + exists 1, 1.
        destruct (generic_step_bistep _ _ _ _ _ _ H) as [H₁ H₂].
        inversion Hos'. clear Hos'. subst os'.
        repeat split ; unfold _left_state, _right_state, nsteps, napply ; auto ;
          simpl (ProgrammingLanguage.step imp_language).
        * rewrite H₁. simpl. apply opt_state_eq_refl.
        * rewrite H₂. simpl. apply opt_state_eq_refl.
      + apply step_helper_special_cmd_step_inv in H.
        destruct H as [Hm Hkm'']. inversion Hm as [H'm]. clear Hm.
        subst km'' m.
        destruct m'.
        destruct c₁ as [
                      | [ | | x₁ e₁ | | | ]
                        [ | | x₂ e₂ | | | ]
                      | | | | ] ; try discriminate Hindep.
        simpl in Hc₁c₂. inversion Hc₁c₂ as [Hc₂]. clear Hc₁c₂. subst c₂.
        unfold perform_n_steps. simpl.
        unfold perform_n_steps in Hos'. simpl in Hos'.
        inversion Hos'. clear Hos'. subst os'.
        exists 1, 1.
        unfold nsteps, napply.
        simpl. repeat split ; auto using opt_state_eq_refl.
      + apply step_helper_special_cont_step_inv in H.
        destruct H as [H _]. discriminate H.
      + discriminate Hos'.
      + (* Thanks to the invariant, we can't crash step_helper: *)
        discriminate Hos'.
  Qed.

  Lemma _prediction_completeness :
    ∀ genv os,
      _invariant genv os →
      _step genv os = None →
      step υ (_left_state  os) = None
      ∧ step υ (_right_state os) = None.
  Proof.
    intros [ ] [ S₁ S₂ | [k₁ s₁] [k₂ s₂] km ] Hinv HNone.
    { unfold _invariant in Hinv. simpl. auto. }
    destruct Hinv as [Hkm Hs₁s₂].
    inversion Hkm as [
                    | x₁ e₁ x₂ e₂ km' k₁' k₂' Hx₁x₂ Hx₁e₂ Hx₂e₁ Hkm'
                    | c₁ c₂ m km' k₁' k₂' Hindep Hc₁c₂ Hkm' ] ;
      subst km k₁ k₂.
    { simpl. auto. }
    - simpl. simpl in HNone.
      unfold perform_n_steps, nsteps' in HNone.
      simpl in HNone.
      destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:H₁,
               (eval_exp s₂ e₂) as [v₂ | ] eqn:H₂ ;
        simpl in * ; auto ;
        try rewrite <- (eval_exp_add_neq' s₁ s₂ x₂ v₂), H₁ in HNone ; auto ;
        try rewrite <- (eval_exp_add_neq' s₂ s₁ x₁ v₁), H₂ in HNone ;
        auto using MFacts.Equal_sym ;
        simpl in HNone ; discriminate.
    - unfold _step in HNone.
      destruct (step_helper _ _) as [ [S₁' S₂' km'' | m' km'' | [] | ] | ] eqn:H ;
        try discriminate.
      + apply step_helper_special_cmd_step_inv in H.
        destruct H as [Hm Hkm''].
        inversion Hm. clear Hm.
        subst m km''.
        destruct m'.
        destruct c₁ as [
                      | [ | | x₁ e₁ | | | ]
                        [ | | x₂ e₂ | | | ]
                      | | | | ] ;
          try discriminate Hindep.
        simpl in HNone, Hc₁c₂. inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
        simpl in HNone. discriminate HNone.
      + apply step_helper_special_cont_step_inv in H.
        destruct H as [Hm Hl]. discriminate.
      + apply generic_step_bicrash in H.
        simpl _left_state. simpl _right_state. auto.
      + destruct m as [ [] | | m₁ m₂ ] ; simpl in H.
        * simpl in H. discriminate H.
        * destruct c₁ ; inversion Hc₁c₂ ; clear Hc₁c₂ ; subst c₂ ; try discriminate H.
          { rewrite (store_equal_eval_exp e s₁ s₂) in H ; auto.
            destruct (eval_exp s₂ e) ; try discriminate H. }
          { rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
            destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H. }
          { rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
            destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H. }
          { rewrite (store_equal_eval_bexp b s₁ s₂) in H ; auto.
            destruct (eval_bexp s₂ b) as [[|]|] ; try discriminate H. }
        * destruct c₁ as [ | c₁₁ c₁₂ | x₁ e₁ | b₁ c₁₁ c₁₂ | b₁ c₁ | b₁ ] ;
            try discriminate Hc₁c₂.
          { simpl in Hc₁c₂.
            destruct (apply_cmd_mod c₁₁ m₁), (apply_cmd_mod c₁₂ m₂) ;
              try discriminate Hc₁c₂.
            inversion Hc₁c₂. clear Hc₁c₂. subst c₂. discriminate H. }
          { simpl in Hc₁c₂.
            destruct (apply_cmd_mod c₁₁ m₁), (apply_cmd_mod c₁₂ m₂) ;
              try discriminate Hc₁c₂.
            inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
            simpl in H.
            rewrite (store_equal_eval_bexp b₁ s₁ s₂) in H ; auto.
            simpl in H.
            destruct (eval_bexp s₂ b₁) as [[|]|] ;
              try discriminate H. }
          { simpl in Hc₁c₂.
            destruct (apply_cmd_mod c₁ m₁) ;
              try discriminate Hc₁c₂.
            inversion Hc₁c₂. clear Hc₁c₂. subst c₂.
            simpl in H.
            rewrite (store_equal_eval_bexp b₁ s₁ s₂) in H ; auto.
            simpl in H.
            destruct (eval_bexp s₂ b₁) as [[|]|] ;
              try discriminate H. }
  Qed.

  Definition swapassign_oracle : oracle_language imp_language imp_language :=
    {|
      oracle_state := _state;
      oracle_genv := unit;
      oracle_step := _step;
      left_state := _left_state;
      right_state := _right_state;
      left_genv := λ _, υ;
      right_genv := λ _, υ;
      invariant := _invariant;
      invariant_1 := _invariant_1;
      prediction_soundness := _prediction_soundness;
      prediction_completeness := _prediction_completeness
    |}.
End Make.