(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

Require Import Common.
Require Import Imp.Lang.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.

Module Make
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT).

  Definition _genv := list varid.

  Inductive _state : Type :=
  | State : state → state → _state.

  Definition _step (_ : _genv) (S : _state) : option _state :=
    match S with
    | State S₁ S₂ =>
      step υ S₁ >>= λ S₁',
      step υ S₂ >>= λ S₂',
      Some (State S₁' S₂')
    end.

  Definition _left_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | State S₁ _ => S₁
    end.

  Definition _right_state (S : _state) : ProgrammingLanguage.state imp_language :=
    match S with
    | State _ S₂ => S₂
    end.

  Inductive same_control_cmd (modified_vars : list varid) : cmd → cmd → Prop :=
  | same_skip :
      same_control_cmd modified_vars Skip Skip
  | same_seq : ∀ c₁₁ c₂₁ c₁₂ c₂₂,
      same_control_cmd modified_vars c₁₁ c₂₁ →
      same_control_cmd modified_vars c₁₂ c₂₂ →
      same_control_cmd modified_vars (Seq c₁₁ c₁₂) (Seq c₂₁ c₂₂)
  | same_ITE : ∀ b c₁₁ c₂₁ c₁₂ c₂₂,
      same_control_cmd modified_vars c₁₁ c₂₁ →
      same_control_cmd modified_vars c₁₂ c₂₂ →
      (∀ x, In_bexp_var x b → ¬ In x modified_vars) →
      same_control_cmd modified_vars (ITE b c₁₁ c₁₂) (ITE b c₂₁ c₂₂)
  | same_While : ∀ b c₁ c₂,
      same_control_cmd modified_vars c₁ c₂ →
      (∀ x, In_bexp_var x b → ¬ In x modified_vars) →
      same_control_cmd modified_vars (While b c₁) (While b c₂)
  | same_assign : ∀ x e,
      (∀ x, In_exp_var x e → ¬ In x modified_vars) →
      same_control_cmd modified_vars (Assign x e) (Assign x e)
  | close_assign : ∀ x e₁ e₂,
      (* Same set of vars and divisors in e₁ and e₂ *)
      (∀ x, In_exp_var x e₁ <-> In_exp_var x e₂) →
      (∀ d, In_exp_div d e₁ <-> In_exp_div d e₂) →
      (* No modified var in divisors of e₁ nor e₂ *)
      (∀ d, In_exp_div d e₁ → ∀ x, In_exp_var x d → ¬ In x modified_vars) →
      (∀ d, In_exp_div d e₂ → ∀ x, In_exp_var x d → ¬ In x modified_vars) →
      In x modified_vars →
      same_control_cmd modified_vars (Assign x e₁) (Assign x e₂)
  | same_assert : ∀ b,
      (∀ x, In_bexp_var x b → ¬ In x modified_vars) →
      same_control_cmd modified_vars (Assert b) (Assert b).

  Definition same_control_cont (modified_vars : list varid) (k₁ k₂ : cont) :=
    Forall2 (same_control_cmd modified_vars) k₁ k₂.

  Definition same_control_store (modified_vars : list varid) (s₁ s₂ : store) : Prop :=
    (∀ x, M.In x s₁ <-> M.In x s₂)
    ∧ ∀ x, (¬ In x modified_vars) → M.find x s₁ = M.find x s₂.

  Lemma same_control_store_eval_exp:
    ∀ m s₁ s₂ e,
      same_control_store m s₁ s₂ →
      (∀ x, In_exp_var x e → ¬ In x m) →
      eval_exp s₁ e = eval_exp s₂ e.
  Proof.
    intros m s₁ s₂ e H H'.
    induction e as [ x | n | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂
                     | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ ] ;
      simpl ; auto.
    - apply H. apply H'. simpl. auto.
    - rewrite IH₁, IH₂ ; auto ;
      intros x Hx ; apply H' ; simpl ; auto.
    - rewrite IH₁, IH₂ ; auto ;
      intros x Hx ; apply H' ; simpl ; auto.
    - rewrite IH₁, IH₂ ; auto ;
      intros x Hx ; apply H' ; simpl ; auto.
    - rewrite IH₁, IH₂ ; auto ;
      intros x Hx ; apply H' ; simpl ; auto.
    - rewrite IH₁, IH₂ ; auto ;
      intros x Hx ; apply H' ; simpl ; auto.
  Qed.

  Lemma same_control_store_eval_bexp:
    ∀ m s₁ s₂ b,
      same_control_store m s₁ s₂ →
      (∀ x, In_bexp_var x b → ¬ In x m) →
      eval_bexp s₁ b = eval_bexp s₂ b.
  Proof.
    intros m s₁ s₂ b H H'.
    induction b as [ | | e₁ e₂ | e₁ e₂ | e₁ e₂
                     | b₁ IH₁ b₂ IH₂ | b₁ IH₁ b₂ IH₂
                     | b₁ IH₁ b₂ IH₂ | b IH ] ;
      simpl ; auto.
    - rewrite (same_control_store_eval_exp m s₁ s₂ e₁),
              (same_control_store_eval_exp m s₁ s₂ e₂) ;
      auto ; intros x Hx ; apply H' ; simpl ; auto.
    - rewrite (same_control_store_eval_exp m s₁ s₂ e₁),
      (same_control_store_eval_exp m s₁ s₂ e₂) ;
        auto ; intros x Hx ; apply H' ; simpl ; auto.
    - rewrite (same_control_store_eval_exp m s₁ s₂ e₁),
      (same_control_store_eval_exp m s₁ s₂ e₂) ;
      auto ; intros x Hx ; apply H' ; simpl ; auto.
    - rewrite IH₁, IH₂ ;
      auto ; intros x Hx ; apply H' ; simpl ; auto.
    - rewrite IH₁, IH₂ ;
      auto ; intros x Hx ; apply H' ; simpl ; auto.
    - rewrite IH₁, IH₂ ;
      auto ; intros x Hx ; apply H' ; simpl ; auto.
    - rewrite IH ;
      auto ; intros x Hx ; apply H' ; simpl ; auto.
  Qed.

  Definition _invariant (m : _genv) (S : _state) : Prop :=
    match S with
    | State (k₁, s₁) (k₂, s₂) =>
      same_control_cont m k₁ k₂
      ∧ same_control_store m s₁ s₂
    end.

  Lemma same_control_exp_crash:
    ∀ m s₁ s₂ e e',
    same_control_store m s₁ s₂ →
    (∀ x, In_exp_var x e <-> In_exp_var x e') →
    (∀ d, In_exp_div d e <-> In_exp_div d e') →
    (∀ d, In_exp_div d e → ∀ x, In_exp_var x d → ¬ In x m) →
    (∀ d, In_exp_div d e' → ∀ x, In_exp_var x d → ¬ In x m) →
    eval_exp s₁ e = None ↔ eval_exp s₂ e' = None.
  Proof.
    intros m s₁ s₂ e e' H Hvar Hdiv H₁ H₂.
    split ; intro H'.
    - assert (∀ x, In_exp_var x e → In_exp_var x e') as Hvar'.
      { apply Hvar. }
      clear Hvar.
      assert (∀ d, In_exp_div d e → In_exp_div d e') as Hdiv'.
      { apply Hdiv. }
      clear Hdiv.
      induction e as [ x | n | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂
                       | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ ].
      + pose proof (Hvar' x eq_refl) as Hx.
        clear Hvar'. simpl in H'.
        clear Hdiv' H₁ H₂.
        destruct H as [H _].
        induction e' as [ y | n | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂
                          | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ ].
        * simpl in Hx. subst. simpl. rewrite <- MFacts.not_find_in_iff in *.
          intro Hfalse. apply H'. apply H. auto.
        * simpl in Hx. exfalso. apply Hx.
        * simpl in Hx. destruct Hx as [Hx' | Hx'].
          { simpl. rewrite IH₁ ; auto. }
          { simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁) ; auto. }
        * simpl in Hx. destruct Hx as [Hx' | Hx'].
          { simpl. rewrite IH₁ ; auto. }
          { simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁) ; auto. }
        * simpl in Hx. destruct Hx as [Hx' | Hx'].
          { simpl. rewrite IH₁ ; auto. }
          { simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁) ; auto. }
        * simpl in Hx. destruct Hx as [Hx' | Hx'].
          { simpl. rewrite IH₁ ; auto. }
          { simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁) ; auto. }
        * simpl in Hx. destruct Hx as [Hx' | Hx'].
          { simpl. rewrite IH₁ ; auto. }
          { simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁) ; auto. }
      + discriminate H'.
      + simpl in H'.
        destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₁ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H'.
        * apply IH₂ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
      + simpl in H'.
        destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₁ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H'.
        * apply IH₂ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
      + simpl in H'.
        destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₁ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H'.
        * apply IH₂ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
      + simpl in H'.
        destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₁ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H'.
        * simpl in H'.
          rewrite (same_control_store_eval_exp m s₁ s₂ e₂) in Hv₂ ; auto ;
          [ | apply H₂ ; apply Hdiv' ; simpl ; auto ].
          destruct v₂ ; try discriminate H'.
          clear H' Hvar' IH₁ IH₂.
          assert (In_exp_div e₂ e') as Hdiv.
          { apply Hdiv'. simpl. auto. }
          clear Hdiv' e₁ v₁ Hv₁ H₁ H₂.
          induction e' as [ y | n | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂
                            | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂ ].
          { simpl in Hdiv. exfalso. apply Hdiv. }
          { simpl in Hdiv. exfalso. apply Hdiv. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | [Hdiv | Hdiv]].
            * subst e₂'. simpl. rewrite Hv₂. simpl.
              destruct (eval_exp s₂ e₁') ; auto.
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | [Hdiv | Hdiv]].
            * subst e₂'. simpl. rewrite Hv₂. simpl.
              destruct (eval_exp s₂ e₁') ; auto.
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁') ; auto. }
        * apply IH₂ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
      + simpl in H'.
        destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₁ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H'.
        * simpl in H'.
          rewrite (same_control_store_eval_exp m s₁ s₂ e₂) in Hv₂ ; auto ;
          [ | apply H₂ ; apply Hdiv' ; simpl ; auto ].
          destruct v₂ ; try discriminate H'.
          clear H' Hvar' IH₁ IH₂.
          assert (In_exp_div e₂ e') as Hdiv.
          { apply Hdiv'. simpl. auto. }
          clear Hdiv' e₁ v₁ Hv₁ H₁ H₂.
          induction e' as [ y | n | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂
                            | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂ ].
          { simpl in Hdiv. exfalso. apply Hdiv. }
          { simpl in Hdiv. exfalso. apply Hdiv. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | [Hdiv | Hdiv]].
            * subst e₂'. simpl. rewrite Hv₂. simpl.
              destruct (eval_exp s₂ e₁') ; auto.
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | [Hdiv | Hdiv]].
            * subst e₂'. simpl. rewrite Hv₂. simpl.
              destruct (eval_exp s₂ e₁') ; auto.
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₂ e₁') ; auto. }
        * apply IH₂ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₁. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
    - assert (∀ x, In_exp_var x e' → In_exp_var x e) as Hvar'.
      { apply Hvar. }
      clear Hvar.
      assert (∀ d, In_exp_div d e' → In_exp_div d e) as Hdiv'.
      { apply Hdiv. }
      clear Hdiv.
      induction e' as [ x | n | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂
                        | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ ].
      + pose proof (Hvar' x eq_refl) as Hx.
        clear Hvar'. simpl in H'.
        clear Hdiv' H₁ H₂.
        destruct H as [H _].
        induction e as [ y | n | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂
                         | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂ ].
        * simpl in Hx. subst. simpl. rewrite <- MFacts.not_find_in_iff in *.
          intro Hfalse. apply H'. apply H. auto.
        * simpl in Hx. exfalso. apply Hx.
        * simpl in Hx. destruct Hx as [Hx' | Hx'].
          { simpl. rewrite IH₁ ; auto. }
          { simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁) ; auto. }
        * simpl in Hx. destruct Hx as [Hx' | Hx'].
          { simpl. rewrite IH₁ ; auto. }
          { simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁) ; auto. }
        * simpl in Hx. destruct Hx as [Hx' | Hx'].
          { simpl. rewrite IH₁ ; auto. }
          { simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁) ; auto. }
        * simpl in Hx. destruct Hx as [Hx' | Hx'].
          { simpl. rewrite IH₁ ; auto. }
          { simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁) ; auto. }
        * simpl in Hx. destruct Hx as [Hx' | Hx'].
          { simpl. rewrite IH₁ ; auto. }
          { simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁) ; auto. }
      + discriminate H'.
      + simpl in H'.
        destruct (eval_exp s₂ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H'.
        * apply IH₂ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
      + simpl in H'.
        destruct (eval_exp s₂ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H'.
        * apply IH₂ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
      + simpl in H'.
        destruct (eval_exp s₂ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H'.
        * apply IH₂ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
      + simpl in H'.
        destruct (eval_exp s₂ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H'.
        * simpl in H'.
          rewrite <- (same_control_store_eval_exp m s₁ s₂ e₂) in Hv₂ ; auto ;
          [ | apply H₁ ; apply Hdiv' ; simpl ; auto ].
          destruct v₂ ; try discriminate H'.
          clear H' Hvar' IH₁ IH₂.
          assert (In_exp_div e₂ e) as Hdiv.
          { apply Hdiv'. simpl. auto. }
          clear Hdiv' e₁ v₁ Hv₁ H₁ H₂.
          induction e as [ y | n | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂
                           | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂ ].
          { simpl in Hdiv. exfalso. apply Hdiv. }
          { simpl in Hdiv. exfalso. apply Hdiv. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | [Hdiv | Hdiv]].
            * subst e₂'. simpl. rewrite Hv₂. simpl.
              destruct (eval_exp s₁ e₁') ; auto.
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | [Hdiv | Hdiv]].
            * subst e₂'. simpl. rewrite Hv₂. simpl.
              destruct (eval_exp s₁ e₁') ; auto.
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁') ; auto. }
        * apply IH₂ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
      + simpl in H'.
        destruct (eval_exp s₂ e₁) as [v₁ | ] eqn:Hv₁,
                 (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
          try discriminate H'.
        * simpl in H'.
          rewrite <- (same_control_store_eval_exp m s₁ s₂ e₂) in Hv₂ ; auto ;
          [ | apply H₁ ; apply Hdiv' ; simpl ; auto ].
          destruct v₂ ; try discriminate H'.
          clear H' Hvar' IH₁ IH₂.
          assert (In_exp_div e₂ e) as Hdiv.
          { apply Hdiv'. simpl. auto. }
          clear Hdiv' e₁ v₁ Hv₁ H₁ H₂.
          induction e as [ y | n | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂
                           | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂ | e₁' IH₁ e₂' IH₂ ].
          { simpl in Hdiv. exfalso. apply Hdiv. }
          { simpl in Hdiv. exfalso. apply Hdiv. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | Hdiv].
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | [Hdiv | Hdiv]].
            * subst e₂'. simpl. rewrite Hv₂. simpl.
              destruct (eval_exp s₁ e₁') ; auto.
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁') ; auto. }
          { simpl in Hdiv. destruct Hdiv as [Hdiv | [Hdiv | Hdiv]].
            * subst e₂'. simpl. rewrite Hv₂. simpl.
              destruct (eval_exp s₁ e₁') ; auto.
            * simpl. rewrite IH₁ ; auto.
            * simpl. rewrite IH₂ ; auto. destruct (eval_exp s₁ e₁') ; auto. }
        * apply IH₂ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
        * apply IH₁ ; auto.
          { intros d Hd. apply H₂. simpl. auto. }
          { intros x Hx. apply Hvar'. simpl. auto. }
          { intros d Hd. apply Hdiv'. simpl. auto. }
  Qed.

  Lemma same_control_store_add_stable:
    ∀ m s₁ s₂, same_control_store m s₁ s₂ →
    ∀ x v, same_control_store m (M.add x v s₁) (M.add x v s₂).
  Proof.
    intros m s₁ s₂ H x v.
    unfold same_control_store in *.
    destruct H as [H₁ H₂].
    split.
    - intro y. rewrite ! MFacts.add_in_iff.
      split ; intros [Hxy | Hin] ; auto.
      * right. apply H₁. auto.
      * right. apply H₁. auto.
    - intros y Hy. rewrite ! MFacts.add_o.
      destruct (MFacts.eq_dec x y) ; auto.
  Qed.

  Lemma same_control_store_add_stable':
    ∀ m s₁ s₂, same_control_store m s₁ s₂ →
    ∀ x v v',
    In x m →
    same_control_store m (M.add x v s₁) (M.add x v' s₂).
  Proof.
    intros m s₁ s₂ H x v v' Hx.
    unfold same_control_store in *.
    destruct H as [H₁ H₂].
    split.
    - intro y. rewrite ! MFacts.add_in_iff.
      split ; intros [Hxy | Hin] ; auto.
      * right. apply H₁. auto.
      * right. apply H₁. auto.
    - intros y Hy. rewrite ! MFacts.add_o.
      destruct (MFacts.eq_dec x y) ; auto.
      subst x. exfalso. apply Hy. apply Hx.
  Qed.

  Lemma _invariant_1:
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      _invariant genv os'.
  Proof.
    intros modified_vars [ [k₁ s₁] [k₂ s₂] ] os' Hinv Hstep.
    destruct Hinv as [H₁ H₂].
    inversion H₁ as [ | c₁_ c₂_ k₁' k₂' Hc₁c₂ H₁' ] ;
      subst k₁ k₂ ; try discriminate Hstep.
    inversion Hc₁c₂ as [ | c₁₁ c₂₁ c₁₂ c₂₂ Hc₁ Hc₂
                         | b c₁₁ c₂₁ c₁₂ c₂₂ Hc₁ Hc Hb
                         | b c₁ c₂ Hc Hb
                         | x e He
                         | x e₁ e₂ Hvars Hdivs H₁divs H₂divs Hx
                         | b Hb ] ;
      subst c₁_ c₂_.
    - simpl in Hstep. inversion Hstep. clear Hstep. subst os'.
      simpl. auto.
    - simpl in Hstep. inversion Hstep. clear Hstep. subst os'.
      simpl. split ; auto.
      apply Forall2_cons ; auto.
    - simpl in Hstep.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂) in Hstep ; auto.
      destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv ;
        inversion Hstep ; clear Hstep ; subst os'.
      + simpl. split ; auto.
        apply Forall2_cons ; auto.
      + simpl. split ; auto.
        apply Forall2_cons ; auto.
    - simpl in Hstep.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂) in Hstep ; auto.
      destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv ;
        inversion Hstep ; clear Hstep ; subst os'.
      + simpl. split ; auto.
        apply Forall2_cons ; auto.
      + simpl. split ; auto.
    - simpl in Hstep.
      rewrite (same_control_store_eval_exp modified_vars s₁ s₂) in Hstep ; auto.
      destruct (eval_exp s₂ e) as [v | ] eqn:Hv ;
        inversion Hstep ; clear Hstep ; subst os' ; simpl ; auto.
      split ; auto.
      apply same_control_store_add_stable ; auto.
    - simpl in Hstep.
      destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
               (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
        inversion Hstep ; clear Hstep ; subst os' ; simpl ; auto.
      split ; auto.
      apply same_control_store_add_stable' ; auto.
    - simpl in Hstep.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂) in Hstep ; auto.
      destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv ;
        inversion Hstep ; clear Hstep ; subst os' ; simpl ; auto.
  Qed.

  Lemma _prediction_soundness :
    ∀ genv os os',
      _invariant genv os →
      _step genv os = Some os' →
      ∃ n₁ n₂,
        opt_state_eq (nsteps υ n₁ (_left_state  os))
                       (Some (_left_state  os'))
        ∧ opt_state_eq (nsteps υ n₂ (_right_state os))
                       (Some (_right_state os'))
        ∧ n₁ + n₂ > 0.
  Proof.
    intros modified_vars [ [k₁ s₁] [k₂ s₂] ] os' Hinv Hos'.
    destruct Hinv as [H₁ H₂].
    inversion H₁ as [ | c₁_ c₂_ k₁' k₂' Hc₁c₂ H₁' ] ; subst k₁ k₂ ;
      try discriminate Hos'.
    exists 1, 1. unfold nsteps.
    inversion Hc₁c₂ as [ | c₁₁ c₂₁ c₁₂ c₂₂ Hc₁ Hc₂
                         | b c₁₁ c₂₁ c₁₂ c₂₂ Hc₁ Hc Hb
                         | b c₁ c₂ Hc Hb
                         | x e He
                         | x e₁ e₂ Hvars Hdivs H₁divs H₂divs Hx
                         | b Hb ] ;
      subst c₁_ c₂_ ;
      try (inversion Hos' ; clear Hos' ; subst os') ;
      try (repeat split ; auto using opt_state_eq_refl with arith ; fail).
    - simpl in Hos'.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂) in Hos' ; auto.
      destruct (eval_bexp s₂ b) as [v | ] eqn:Hv ; simpl in Hos' ;
        inversion Hos' ; clear Hos' ; subst os'.
      simpl.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂), Hv ; auto.
      repeat split ; auto using opt_state_eq_refl.
    - simpl in Hos'.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂) in Hos' ; auto.
      destruct (eval_bexp s₂ b) as [v | ] eqn:Hv ; simpl in Hos' ;
        inversion Hos' ; clear Hos' ; subst os'.
      simpl.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂), Hv ; auto.
      simpl.
      repeat split ; auto using opt_state_eq_refl.
    - simpl in Hos'.
      rewrite (same_control_store_eval_exp modified_vars s₁ s₂) in Hos' ; auto.
      destruct (eval_exp s₂ e) as [v | ] eqn:Hv ; simpl in Hos' ;
        inversion Hos' ; clear Hos' ; subst os'.
      simpl.
      rewrite (same_control_store_eval_exp modified_vars s₁ s₂), Hv ; auto.
      simpl.
      repeat split ; auto using opt_state_eq_refl.
    - simpl in Hos'.
      destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
               (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
        simpl in Hos' ;
        inversion Hos' ; clear Hos' ; subst os'.
      simpl.
      rewrite Hv₁, Hv₂. simpl.
      repeat split ; auto using opt_state_eq_refl.
    - simpl in Hos'.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂) in Hos' ; auto.
      destruct (eval_bexp s₂ b) as [[ | ] | ] eqn:Hv ;
        inversion Hos' ; clear Hos' ; subst os'.
      unfold napply. simpl.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂) ; auto.
      rewrite Hv. simpl.
      repeat split ; auto using opt_state_eq_refl.
  Qed.

  Lemma _prediction_completeness :
    ∀ genv os,
      _invariant genv os →
      _step genv os = None →
      step υ (_left_state  os) = None
      ∧ step υ (_right_state os) = None.
  Proof.
    intros modified_vars [ [k₁ s₁] [k₂ s₂] ] Hinv HNone.
    destruct Hinv as [H₁ H₂].
    inversion H₁ as [ | c₁_ c₂_ k₁' k₂' Hc₁c₂ H₁' ] ; subst k₁ k₂ ;
      auto.
    inversion Hc₁c₂ as [ | c₁₁ c₂₁ c₁₂ c₂₂ Hc₁ Hc₂
                         | b c₁₁ c₂₁ c₁₂ c₂₂ Hc₁ Hc Hb
                         | b c₁ c₂ Hc Hb
                         | x e He
                         | x e₁ e₂ Hvars Hdivs H₁divs H₂divs Hx
                         | b Hb ] ;
      subst c₁_ c₂_.
    - discriminate HNone.
    - discriminate HNone.
    - simpl. simpl in HNone.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂) in * ; auto.
      destruct (eval_bexp s₂ b) ; auto ; discriminate HNone.
    - simpl. simpl in HNone.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂) in * ; auto.
      destruct (eval_bexp s₂ b) ; auto ; discriminate HNone.
    - simpl. simpl in HNone.
      rewrite (same_control_store_eval_exp modified_vars s₁ s₂) in * ; auto.
      destruct (eval_exp s₂ e) ; auto ; discriminate HNone.
    - simpl. simpl in HNone.
      destruct (eval_exp s₁ e₁) as [v₁ | ] eqn:Hv₁,
               (eval_exp s₂ e₂) as [v₂ | ] eqn:Hv₂ ;
        auto ; try discriminate HNone.
      + rewrite <- (same_control_exp_crash modified_vars s₁ s₂) in Hv₂ ;
          auto.
        rewrite Hv₂ in Hv₁. discriminate.
      + rewrite    (same_control_exp_crash modified_vars s₁ s₂) in Hv₁ ;
          auto.
        rewrite Hv₂ in Hv₁. discriminate.
    - simpl. simpl in HNone.
      rewrite (same_control_store_eval_bexp modified_vars s₁ s₂) in * ; auto.
      destruct (eval_bexp s₂ b) as [[|]|] ; auto ; discriminate HNone.
  Qed.

  Definition valuechange_oracle : oracle_language imp_language imp_language :=
    {|
      oracle_state := _state;
      oracle_genv := _genv;
      oracle_step := _step;
      left_state := _left_state;
      right_state := _right_state;
      left_genv := λ _, υ;
      right_genv := λ _, υ;
      invariant := _invariant;
      invariant_1 := _invariant_1;
      prediction_soundness := _prediction_soundness;
      prediction_completeness := _prediction_completeness
    |}.
End Make.
