(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

(** Relational Hoare Logic on Imp

    This module defines several variants of Relational Hoare Logic as described
    by Gilles Barthe, Juan Manuel Crespo and César Kunz in
    _Product programs and relational program logics_

    Soundness proofs for those RHL variants are available in the following modules:
    - [OracleLanguages.Imp.Oracles.MinimalRHL] for [minimal_rhl_proof]
    - [Imp.Oracles.CoreRHL] for [core_rhl_proof]
    - [Imp.Oracles.CoreRHLSC] for [core_rhl_sc_proof]

 *)

Require Import Common.
Require Import Imp.Lang.
Require Import Imp.BigStep.
Require Import Imp.Hoare.
Require Import Imp.DisjointStates.

Module Make
       (Var_as_DT : UsualDecidableTypeOrig)
       (Import Lang : ImpProgrammingLanguage Var_as_DT)
       (BigStep : ImpBigStep Var_as_DT Lang).
  
  Module DisjointStates := Imp.DisjointStates.Make Var_as_DT Lang.
  Module Hoare := Imp.Hoare.Make Var_as_DT Lang BigStep.

  Import DisjointStates.
  Import Hoare.

  (* Minimal RHL logic, that is, lock-step execution *)
  Inductive minimal_rhl_proof :
    assertion → cmd → cmd → assertion → Type :=
  | MinimalSkip : ∀ P,
    minimal_rhl_proof P Skip Skip P
  | MinimalAssign : ∀ P x e y e',
    minimal_rhl_proof (P [x ↦ e] [y ↦ e']) (Assign x e) (Assign y e') P
  | MinimalAssert : ∀ P b₁ b₂,
    P ⇒ (λ S, (assertion_of_bexp b₁) S ∧ (assertion_of_bexp b₂) S) →
    minimal_rhl_proof P (Assert b₁) (Assert b₂) P
  | MinimalSeq : ∀ P Q R c₁₁ c₂₁ c₁₂ c₂₂,
    minimal_rhl_proof P c₁₁ c₂₁ Q →
    minimal_rhl_proof Q c₁₂ c₂₂ R →
    minimal_rhl_proof P (Seq c₁₁ c₁₂) (Seq c₂₁ c₂₂) R
  | MinimalITE : ∀ P Q b₁ b₂ c₁₁ c₂₁ c₁₂ c₂₂,
    minimal_rhl_proof (λ S, P S ∧ assertion_of_bexp b₁ S)
                      c₁₁ c₂₁ Q →
    minimal_rhl_proof (λ S, P S ∧ assertion_of_bexp (NOT b₁) S)
                      c₁₂ c₂₂ Q →
    P ⇒ (λ S, eval_bexp S b₁ = eval_bexp S b₂) →
    P ⇒ (λ S, eval_bexp S b₁ ≠ None ∧ eval_bexp S b₂ ≠ None) → (* No crash! *)
    minimal_rhl_proof P (ITE b₁ c₁₁ c₁₂) (ITE b₂ c₂₁ c₂₂) Q
  | MinimalWhile : ∀ P b₁ b₂ c₁ c₂,
    minimal_rhl_proof (λ S, P S ∧ assertion_of_bexp b₁ S)
                      c₁ c₂ P →
    P ⇒ (λ S, eval_bexp S b₁ = eval_bexp S b₂) →
    P ⇒ (λ S, eval_bexp S b₁ ≠ None ∧ eval_bexp S b₂ ≠ None) → (* No crash! *)
    minimal_rhl_proof P
                      (While b₁ c₁) (While b₂ c₂)
                      (λ S, P S ∧ assertion_of_bexp (NOT b₁) S)
  | MinimalSub: ∀ P P' Q Q' c₁ c₂,
    minimal_rhl_proof P c₁ c₂ Q →
    P' ⇒ P →
    Q ⇒ Q' →
    minimal_rhl_proof P' c₁ c₂ Q'.

  (* Core RHL logic *)
  Inductive core_rhl_proof :
    assertion → cmd → cmd → assertion → Type :=
  | CoreSkip : ∀ P,
    core_rhl_proof P Skip Skip P
  | CoreAssign : ∀ P x e y e',
    core_rhl_proof (P [x ↦ e] [y ↦ e']) (Assign x e) (Assign y e') P
  | CoreAssert : ∀ P b₁ b₂,
    P ⇒ (λ S, (assertion_of_bexp b₁) S ∧ (assertion_of_bexp b₂) S) →
    core_rhl_proof P (Assert b₁) (Assert b₂) P
  | CoreSeq : ∀ P Q R c₁₁ c₂₁ c₁₂ c₂₂,
    core_rhl_proof P c₁₁ c₂₁ Q →
    core_rhl_proof Q c₁₂ c₂₂ R →
    core_rhl_proof P (Seq c₁₁ c₁₂) (Seq c₂₁ c₂₂) R
  | CoreITE : ∀ P Q b₁ b₂ c₁₁ c₂₁ c₁₂ c₂₂,
    core_rhl_proof (λ S, P S ∧ assertion_of_bexp b₁ S)
                   c₁₁ c₂₁ Q →
    core_rhl_proof (λ S, P S ∧ assertion_of_bexp (NOT b₁) S)
                   c₁₂ c₂₂ Q →
    P ⇒ (λ S, eval_bexp S b₁ = eval_bexp S b₂) →
    P ⇒ (λ S, eval_bexp S b₁ ≠ None ∧ eval_bexp S b₂ ≠ None) → (* No crash! *)
    core_rhl_proof P (ITE b₁ c₁₁ c₁₂) (ITE b₂ c₂₁ c₂₂) Q
  | CoreWhile : ∀ P b₁ b₂ c₁ c₂,
    core_rhl_proof (λ S, P S ∧ assertion_of_bexp b₁ S)
                   c₁ c₂ P →
    P ⇒ (λ S, eval_bexp S b₁ = eval_bexp S b₂) →
    P ⇒ (λ S, eval_bexp S b₁ ≠ None ∧ eval_bexp S b₂ ≠ None) → (* No crash! *)
    core_rhl_proof P
                   (While b₁ c₁) (While b₂ c₂)
                   (λ S, P S ∧ assertion_of_bexp (NOT b₁) S)
  | CoreSub: ∀ P P' Q Q' c₁ c₂,
    core_rhl_proof P c₁ c₂ Q →
    P' ⇒ P →
    Q ⇒ Q' →
    core_rhl_proof P' c₁ c₂ Q'
  (* Core-specific rules *)
  (* Assignment *) (* XXX: The rules in the original paper are incorrect! *)
  | CoreAssignLeft : ∀ P Q x e c,
    core_rhl_proof P Skip c Q →
    core_rhl_proof (P [ x ↦ e ]) (Assign x e) c Q
  | CoreAssignRight : ∀ P Q x e c,
    core_rhl_proof P c Skip Q →
    core_rhl_proof (P [ x ↦ e ]) c (Assign x e) Q
  (* Assertion *)
  | CoreAssertLeft : ∀ P Q b c,
    core_rhl_proof P Skip c Q →
    P ⇒ (λ S, assertion_of_bexp b S) →
    core_rhl_proof P (Assert b) c Q
  | CoreAssertRight : ∀ P Q b c,
    core_rhl_proof P c Skip Q →
    P ⇒ (λ S, assertion_of_bexp b S) →
    core_rhl_proof P c (Assert b) Q
  (* Sequence *) (* XXX: Not present in the original paper! *)
  | CoreSeqSkipLeft : ∀ P Q R c₁₁ c₁₂ c₂,
      core_rhl_proof P c₁₁ Skip Q →
      core_rhl_proof Q c₁₂ c₂ R →
      core_rhl_proof P (Seq c₁₁ c₁₂) c₂ R
  | CoreSeqSkipRight : ∀ P Q R c₁ c₂₁ c₂₂,
      core_rhl_proof P Skip c₂₁ Q →
      core_rhl_proof Q c₁ c₂₂ R →
      core_rhl_proof P c₁ (Seq c₂₁ c₂₂) R
  (* Conditional *)
  | CoreITELeft : ∀ P Q b c₁ c₂ c,
    core_rhl_proof (λ S, P S ∧ assertion_of_bexp b S) c₁ c Q →
    core_rhl_proof (λ S, P S ∧ assertion_of_bexp (NOT b) S) c₂ c Q →
    P ⇒ (λ S, eval_bexp S b ≠ None) → (* No crash! *)
    core_rhl_proof P (ITE b c₁ c₂) c Q
  | CoreITERight : ∀ P Q b c₁ c₂ c,
    core_rhl_proof (λ S, P S ∧ assertion_of_bexp b S) c c₁ Q →
    core_rhl_proof (λ S, P S ∧ assertion_of_bexp (NOT b) S) c c₂ Q →
    P ⇒ (λ S, eval_bexp S b ≠ None) → (* No crash! *)
    core_rhl_proof P c (ITE b c₁ c₂) Q.

  (* Core RHL logic + self-composition *)
  (* XXX: It would be better not to have all this duplication,
     but doing otherwise is tricky… *)
  Inductive core_rhl_sc_proof :
    assertion → cmd → cmd → assertion → Type :=
  | CoreSCSkip : ∀ P,
    core_rhl_sc_proof P Skip Skip P
  | CoreSCAssign : ∀ P x e y e',
    core_rhl_sc_proof (P [x ↦ e] [y ↦ e']) (Assign x e) (Assign y e') P
  | CoreSCAssert : ∀ P b₁ b₂,
    P ⇒ (λ S, (assertion_of_bexp b₁) S ∧ (assertion_of_bexp b₂) S) →
    core_rhl_sc_proof P (Assert b₁) (Assert b₂) P
  | CoreSCSeq : ∀ P Q R c₁₁ c₂₁ c₁₂ c₂₂,
    core_rhl_sc_proof P c₁₁ c₂₁ Q →
    core_rhl_sc_proof Q c₁₂ c₂₂ R →
    core_rhl_sc_proof P (Seq c₁₁ c₁₂) (Seq c₂₁ c₂₂) R
  | CoreSCITE : ∀ P Q b₁ b₂ c₁₁ c₂₁ c₁₂ c₂₂,
    core_rhl_sc_proof (λ S, P S ∧ assertion_of_bexp b₁ S)
                      c₁₁ c₂₁ Q →
    core_rhl_sc_proof (λ S, P S ∧ assertion_of_bexp (NOT b₁) S)
                      c₁₂ c₂₂ Q →
    P ⇒ (λ S, eval_bexp S b₁ = eval_bexp S b₂) →
    P ⇒ (λ S, eval_bexp S b₁ ≠ None ∧ eval_bexp S b₂ ≠ None) → (* No crash! *)
    core_rhl_sc_proof P (ITE b₁ c₁₁ c₁₂) (ITE b₂ c₂₁ c₂₂) Q
  | CoreSCWhile : ∀ P b₁ b₂ c₁ c₂,
    core_rhl_sc_proof (λ S, P S ∧ assertion_of_bexp b₁ S)
                      c₁ c₂ P →
    P ⇒ (λ S, eval_bexp S b₁ = eval_bexp S b₂) →
    P ⇒ (λ S, eval_bexp S b₁ ≠ None ∧ eval_bexp S b₂ ≠ None) → (* No crash! *)
    core_rhl_sc_proof P
                      (While b₁ c₁) (While b₂ c₂)
                      (λ S, P S ∧ assertion_of_bexp (NOT b₁) S)
  | CoreSCSub: ∀ P P' Q Q' c₁ c₂,
    core_rhl_sc_proof P c₁ c₂ Q →
    P' ⇒ P →
    Q ⇒ Q' →
    core_rhl_sc_proof P' c₁ c₂ Q'
  (* Core-specific rules *)
  (* Assignment *) (* XXX: The rules in the original paper are incorrect! *)
  | CoreSCAssignLeft : ∀ P Q x e c,
    core_rhl_sc_proof P Skip c Q →
    core_rhl_sc_proof (P [ x ↦ e ]) (Assign x e) c Q
  | CoreSCAssignRight : ∀ P Q x e c,
    core_rhl_sc_proof P c Skip Q →
    core_rhl_sc_proof (P [ x ↦ e ]) c (Assign x e) Q
  (* Assertion *)
  | CoreSCAssertLeft : ∀ P Q b c,
    core_rhl_sc_proof P Skip c Q →
    P ⇒ (λ S, assertion_of_bexp b S) →
    core_rhl_sc_proof P (Assert b) c Q
  | CoreSCAssertRight : ∀ P Q b c,
    core_rhl_sc_proof P c Skip Q →
    P ⇒ (λ S, assertion_of_bexp b S) →
    core_rhl_sc_proof P c (Assert b) Q
  (* Sequence *) (* XXX: Not present in the original paper! *)
  | CoreSCSeqSkipLeft : ∀ P Q R c₁₁ c₁₂ c₂,
      core_rhl_sc_proof P c₁₁ Skip Q →
      core_rhl_sc_proof Q c₁₂ c₂ R →
      core_rhl_sc_proof P (Seq c₁₁ c₁₂) c₂ R
  | CoreSCSeqSkipRight : ∀ P Q R c₁ c₂₁ c₂₂,
      core_rhl_sc_proof P Skip c₂₁ Q →
      core_rhl_sc_proof Q c₁ c₂₂ R →
      core_rhl_sc_proof P c₁ (Seq c₂₁ c₂₂) R
  (* Conditional *)
  | CoreSCITELeft : ∀ P Q b c₁ c₂ c,
    core_rhl_sc_proof (λ S, P S ∧ assertion_of_bexp b S) c₁ c Q →
    core_rhl_sc_proof (λ S, P S ∧ assertion_of_bexp (NOT b) S) c₂ c Q →
    P ⇒ (λ S, eval_bexp S b ≠ None) → (* No crash! *)
    core_rhl_sc_proof P (ITE b c₁ c₂) c Q
  | CoreSCITERight : ∀ P Q b c₁ c₂ c,
    core_rhl_sc_proof (λ S, P S ∧ assertion_of_bexp b S) c c₁ Q →
    core_rhl_sc_proof (λ S, P S ∧ assertion_of_bexp (NOT b) S) c c₂ Q →
    P ⇒ (λ S, eval_bexp S b ≠ None) → (* No crash! *)
    core_rhl_sc_proof P c (ITE b c₁ c₂) Q
  | CoreSCSelfComp : ∀ P Q c₁ c₂,
      hoare_proof P (Seq c₁ c₂) Q →
      core_rhl_sc_proof P c₁ c₂ Q.
End Make.
