(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

(**
  Small-step semantics and change-indexed presentation of small-step
  semantics of the Imp imperative language.
*)

Require Import Common.
Require Import Imp.Syntax.
Require Import List.
Import ListNotations.

Module Imp (Var_as_DT : UsualDecidableTypeOrig).
  Module Syntax := Imp.Syntax.Imp Var_as_DT.
  Export Syntax.

  Definition cont := list cmd.
  Definition state := prod cont store.

  Definition state_eq (S₁ S₂ : state) : Prop :=
    let '(k₁, s₁) := S₁ in
    let '(k₂, s₂) := S₂ in
    k₁ = k₂ ∧ M.Equal s₁ s₂.

  Lemma state_eq_dec:
    ∀ S₁ S₂, {state_eq S₁ S₂} + {¬ state_eq S₁ S₂}.
  Proof.
    intros (k₁, s₁) (k₂, s₂).
    destruct (list_eq_dec (cmd_eq_dec) k₁ k₂).
    * subst k₂. destruct (store_Equal_dec s₁ s₂).
      + left. split ; auto.
      + right. intro Hfalse. apply n. destruct Hfalse as (_, Hfalse). auto.
    * right. intro Hfalse. destruct Hfalse as (H, _). apply n. apply H.
  Defined.

  Definition beq_state (S₁ S₂ : state) :=
    if state_eq_dec S₁ S₂ then true else false.

  Lemma state_eq_sym:
    ∀ S₁ S₂,
    state_eq S₁ S₂ → state_eq S₂ S₁.
  Proof.
    intros [k₁ s₁] [k₂ s₂] [Hk Hs] ; subst k₂ ; split ; intuition.
  Qed.

  Lemma state_eq_refl:
    ∀ S,
    state_eq S S.
  Proof.
    intros [k s]. unfold state_eq. intuition.
  Qed.

  Lemma state_eq_trans:
    ∀ S₁ S₂ S₃,
    state_eq S₁ S₂ →
    state_eq S₂ S₃ →
    state_eq S₁ S₃.
  Proof.
    intros [k₁ s₁] [k₂ s₂] [k₃ s₃] [Hk₁₂ Hs₁₂] [Hk₂₃ Hs₂₃].
    subst k₂ k₃. split ; intuition.
    intro x. rewrite Hs₁₂. auto.
  Qed.

  Definition state_final (S : state) :=
    match S with
      | ([], _) => Some tt
      | _ => None
    end.

  Lemma state_final_eq:
    ∀ S S',
      state_eq S S' →
      state_final S = state_final S'.
  Proof.
    intros [k s] [k' s'] [Hkk' Hss'].
    simpl. subst k'. destruct k ; reflexivity.
  Qed.

  Definition step (_ : unit) (S : state) : option state :=
    let '(k, s) := S in
    match k with
    | [] => None
    | Skip::k' => ret (k', s)
    | (Seq c₁ c₂)::k' => ret (c₁::c₂::k', s)
    | (Assign x e)::k' => eval_exp s e >>= λ v,
      ret (k', M.add x v s)
    | (ITE b c₁ c₂)::k' => eval_bexp s b >>= λ v,
      ret ((if v then c₁ else c₂)::k', s)
    | (While b c')::k' => eval_bexp s b >>= λ v,
      ret (if v then (c'::(While b c')::k', s) else (k', s))
    | (Assert b)::k' =>
      eval_bexp s b >>= λ v,
      if v then ret (k', s)
      else None
    end.

  Lemma step_app:
    ∀ k k' k₂ s s', step tt (k, s) = Some (k', s') →
    step tt (k ++ k₂, s) = Some (k' ++ k₂, s').
  Proof.
    intros k. induction k as [ | c k IH ] ; intros k' k₂ s s' H.
    * discriminate H.
    * simpl. simpl in H. destruct c.
      + inversion H. subst. reflexivity.
      + inversion H. subst. reflexivity.
      + destruct (eval_exp s e) ; inversion H. subst. reflexivity.
      + destruct (eval_bexp s b) as [[|] |] ; inversion H ; subst ; reflexivity.
      + destruct (eval_bexp s b) as [[|] |] ; inversion H ; subst ; reflexivity.
      + destruct (eval_bexp s b) as [[|] |] ; inversion H ; subst ; reflexivity.
  Qed.

  Lemma step_eq:
    ∀ genv₁ genv₂ S₁ S₂ S₁' S₂',
      genv₁ = genv₂ →
      state_eq S₁ S₂ →
      step genv₁ S₁ = Some S₁' →
      step genv₂ S₂ = Some S₂' →
      state_eq S₁' S₂'.
  Proof.
    intros [] [] [k₁ s₁] [k₂ s₂] [k₁' s₁'] [k₂' s₂'] _ [Hk₁k₂ Hs₁s₂] HS₁' HS₂'.
    subst k₂.
    destruct k₁ as [ | c k₁] ;
      try discriminate HS₁'.
    destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b] ;
      simpl in *.
    - inversion HS₁' as [[Hk₁k₁' Hs₁s₁']].
      inversion HS₂' as [[Hk₁k₂' Hs₂s₂']].
      clear HS₁' HS₂'.
      subst k₁' s₁' k₂' s₂'.
      split ; auto.
    - inversion HS₁' as [[Hk₁k₁' Hs₁s₁']].
      inversion HS₂' as [[Hk₁k₂' Hs₂s₂']].
      clear HS₁' HS₂'.
      subst k₁' s₁' k₂' s₂'.
      split ; auto.
    - rewrite (store_equal_eval_exp e s₁ s₂) in HS₁' ; auto.
      destruct (eval_exp s₂ e) as [v | ] ; try discriminate HS₁'.
      simpl in *.
      inversion HS₁' as [[Hk₁k₁' Hs₁s₁']].
      inversion HS₂' as [[Hk₁k₂' Hs₂s₂']].
      clear HS₁' HS₂'.
      subst k₁' s₁' k₂' s₂'.
      split ; auto. apply MFacts.add_m ; auto.
    - rewrite (store_equal_eval_bexp b s₁ s₂) in HS₁' ; auto.
      destruct (eval_bexp s₂ b) as [v | ] ; try discriminate HS₁'.
      simpl in *.
      inversion HS₁' as [[Hk₁k₁' Hs₁s₁']].
      inversion HS₂' as [[Hk₁k₂' Hs₂s₂']].
      clear HS₁' HS₂'.
      subst k₁' s₁' k₂' s₂'.
      split ; auto.
    - rewrite (store_equal_eval_bexp b s₁ s₂) in HS₁' ; auto.
      destruct (eval_bexp s₂ b) as [v | ] ; try discriminate HS₁'.
      simpl in *.
      destruct v ;
      inversion HS₁' as [[Hk₁k₁' Hs₁s₁']] ;
      inversion HS₂' as [[Hk₁k₂' Hs₂s₂']] ;
      clear HS₁' HS₂' ;
      subst k₁' s₁' k₂' s₂' ;
      split ; auto.
    - rewrite (store_equal_eval_bexp b s₁ s₂) in HS₁' ; auto.
      destruct (eval_bexp s₂ b) as [[ | ] | ] ; try discriminate HS₁'.
      simpl in *.
      inversion HS₁' as [[Hk₁k₁' Hs₁s₁']].
      inversion HS₂' as [[Hk₁k₂' Hs₂s₂']].
      clear HS₁' HS₂'.
      subst k₁' s₁' k₂' s₂'.
      split ; auto.
  Qed.

  Lemma step_eq':
    ∀ genv₁ genv₂ S₁ S₂,
      genv₁ = genv₂ →
      state_eq S₁ S₂ →
      step genv₁ S₁ = None ↔ step genv₂ S₂ = None.
  Proof.
    intros [] [] [k₁ s₁] [k₂ s₂] _ [Hk₁k₂ Hs₁s₂].
    subst k₂.
    destruct k₁ as [ | c k₁] ;
      try reflexivity.
    destruct c as [ | c₁ c₂ | x e | b c₁ c₂ | b c | b] ;
      simpl ; try (split ; discriminate).
    - rewrite (store_equal_eval_exp e s₁ s₂) ; auto.
      destruct (eval_exp s₂ e) as [v | ] ;
        simpl ; split ; try discriminate ; reflexivity.
    - rewrite (store_equal_eval_bexp b s₁ s₂) ; auto.
      destruct (eval_bexp s₂ b) as [v | ] ;
        simpl ; split ; try discriminate ; reflexivity.
    - rewrite (store_equal_eval_bexp b s₁ s₂) ; auto.
      destruct (eval_bexp s₂ b) as [v | ] ;
        simpl ; split ; try discriminate ; reflexivity.
    - rewrite (store_equal_eval_bexp b s₁ s₂) ; auto.
      destruct (eval_bexp s₂ b) as [[ | ] | ] ;
        simpl ; split ; try discriminate ; reflexivity.
  Qed.

  Inductive finite_small_step_exec s : state → Prop :=
    | Halted : finite_small_step_exec s ([], s)
    | Step : ∀ S S',
      step tt S = Some S' →
      finite_small_step_exec s S' →
      finite_small_step_exec s S.

  Hint Constructors finite_small_step_exec.

  CoInductive infinite_small_step_exec : state → Prop :=
    | InfStep : ∀ S S',
      step tt S = Some S' →
      infinite_small_step_exec S' →
      infinite_small_step_exec S.
End Imp.
