(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
 *)

(**
  Syntax of the Imp programming language.
*)

Require Import Common.
Require Export ZArith.
Require Export FMapWeakList.
Require Export FMapFacts.

Module Imp (Var_as_DT : UsualDecidableTypeOrig).
  (* Specialized maps that will be used later *)
  Module M := FMapWeakList.Make(Var_as_DT).
  Module MProps := FMapFacts.WProperties_fun Var_as_DT M.
  Module MFacts := MProps.F.

  (* AST *)
  Definition varid := Var_as_DT.t.

  Inductive exp :=
      | Var : varid → exp
      | Int : Z → exp
      | Sum : exp → exp → exp
      | Mult : exp → exp → exp
      | Sub : exp → exp → exp
      | Div : exp → exp → exp
      | Modulo : exp → exp → exp.

  Fixpoint In_exp_var (x : varid) (e : exp) : Prop :=
    match e with
      | Var y =>
        x = y
      | Int _ =>
        False
      | Sum e₁ e₂ | Mult e₁ e₂ | Sub e₁ e₂ | Div e₁ e₂ | Modulo e₁ e₂ =>
        In_exp_var x e₁ ∨ In_exp_var x e₂
    end.

  Fixpoint In_exp_div (div : exp) (e : exp) : Prop :=
    match e with
      | Var _ | Int _ =>
        False
      | Sum e₁ e₂ | Mult e₁ e₂ | Sub e₁ e₂ =>
        In_exp_div div e₁ ∨ In_exp_div div e₂
      | Div e₁ e₂ | Modulo e₁ e₂ =>
        e₂ = div ∨ In_exp_div div e₁ ∨ In_exp_div div e₂
    end.

  Lemma exp_eq_dec:
    ∀ (e₁ e₂ : exp), { e₁ = e₂ } + { e₁ ≠ e₂ }.
  Proof.
    decide equality.
    * apply Var_as_DT.eq_dec.
    * apply Z_eq_dec.
  Defined.

  Lemma In_exp_var_dec:
    ∀ x e, {In_exp_var x e} + {¬ In_exp_var x e}.
  Proof.
    intros x e.
    induction e as [ y
                   | n
                   | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂
                   | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂
                   | e₁ IH₁ e₂ IH₂ ] ; simpl ;
    auto using Var_as_DT.eq_dec.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; [ apply H₁ | apply H₂ ] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; [ apply H₁ | apply H₂ ] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; [ apply H₁ | apply H₂ ] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; [ apply H₁ | apply H₂ ] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; [ apply H₁ | apply H₂ ] ; auto.
  Defined.

  Lemma In_exp_div_dec:
    ∀ d e, {In_exp_div d e} + {¬ In_exp_div d e}.
  Proof.
    intros d e.
    induction e as [ y | n
                     | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂
                     | e₁ IH₁ e₂ IH₂ | e₁ IH₁ e₂ IH₂
                     | e₁ IH₁ e₂ IH₂ ] ; simpl ;
    auto using exp_eq_dec.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; [ apply H₁ | apply H₂ ] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; [ apply H₁ | apply H₂ ] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; [ apply H₁ | apply H₂ ] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      destruct (exp_eq_dec e₂ d) as [H₃ | H₃] ; auto.
      right.
      intro Hfalse.
      destruct Hfalse as [H | [H | H]] ;
        [ apply H₃ | apply H₁ | apply H₂ ] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      destruct (exp_eq_dec e₂ d) as [H₃ | H₃] ; auto.
      right.
      intro Hfalse.
      destruct Hfalse as [H | [H | H]] ;
        [ apply H₃ | apply H₁ | apply H₂ ] ; auto.
  Defined.

  Definition beq_exp (e₁ e₂ : exp) : bool :=
    if exp_eq_dec e₁ e₂ then true else false.

  Property beq_exp_reflx:
    ∀ e, beq_exp e e = true.
  Proof.
    intro e. unfold beq_exp.
    destruct (exp_eq_dec e e) ; auto.
  Qed.

  Inductive bexp :=
      | True | False
      | EQ : exp → exp → bexp
      | LT : exp → exp → bexp
      | LE : exp → exp → bexp
      | EQB : bexp → bexp → bexp
      | AND : bexp → bexp → bexp
      | OR : bexp → bexp → bexp
      | NOT : bexp → bexp.

  Fixpoint In_bexp_var (x : varid) (b : bexp) : Prop :=
    match b with
      | True | False =>
        Logic.False
      | EQ e₁ e₂ | LT e₁ e₂ | LE e₁ e₂ =>
        In_exp_var x e₁ ∨ In_exp_var x e₂
      | AND b₁ b₂ | OR b₁ b₂ | EQB b₁ b₂ =>
        In_bexp_var x b₁ ∨ In_bexp_var x b₂
      | NOT b =>
        In_bexp_var x b
    end.

  Fixpoint In_bexp_div (div : exp) (b : bexp) : Prop :=
    match b with
      | True | False =>
        Logic.False
      | EQ e₁ e₂ | LT e₁ e₂ | LE e₁ e₂ =>
        In_exp_div div e₁ ∨ In_exp_div div e₂
      | AND b₁ b₂ | OR b₁ b₂ | EQB b₁ b₂ =>
        In_bexp_div div b₁ ∨ In_bexp_div div b₂
      | NOT b =>
        In_bexp_div div b
    end.

  Lemma bexp_eq_dec:
    ∀ (b₁ b₂ : bexp), { b₁ = b₂ } + { b₁ ≠ b₂ }.
  Proof.
    decide equality ; apply exp_eq_dec.
  Defined.

  Lemma In_bexp_var_dec:
    ∀ x b, {In_bexp_var x b} + {¬ In_bexp_var x b}.
  Proof.
    intros x b.
    induction b as [ | | e₁ e₂ | e₁ e₂ | e₁ e₂
                     | b₁ IH₁ b₂ IH₂ | b₁ IH₁ b₂ IH₂
                     | b₁ IH₁ b₂ IH₂ | b IH ] ; simpl ;
    auto using In_exp_var_dec.
    - destruct (In_exp_var_dec x e₁) as [H₁ | H₁] ; auto.
      destruct (In_exp_var_dec x e₂) as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
    - destruct (In_exp_var_dec x e₁) as [H₁ | H₁] ; auto.
      destruct (In_exp_var_dec x e₂) as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
    - destruct (In_exp_var_dec x e₁) as [H₁ | H₁] ; auto.
      destruct (In_exp_var_dec x e₂) as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
  Defined.

  Lemma In_bexp_div_dec:
    ∀ e b, {In_bexp_div e b} + {¬ In_bexp_div e b}.
  Proof.
    intros e b.
    induction b as [ | | e₁ e₂ | e₁ e₂ | e₁ e₂
                     | b₁ IH₁ b₂ IH₂ | b₁ IH₁ b₂ IH₂
                     | b₁ IH₁ b₂ IH₂ | b IH ] ; simpl ;
    auto using In_exp_div_dec.
    - destruct (In_exp_div_dec e e₁) as [H₁ | H₁] ; auto.
      destruct (In_exp_div_dec e e₂) as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
    - destruct (In_exp_div_dec e e₁) as [H₁ | H₁] ; auto.
      destruct (In_exp_div_dec e e₂) as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
    - destruct (In_exp_div_dec e e₁) as [H₁ | H₁] ; auto.
      destruct (In_exp_div_dec e e₂) as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
    - destruct IH₁ as [H₁ | H₁] ; auto.
      destruct IH₂ as [H₂ | H₂] ; auto.
      right. intro Hfalse. destruct Hfalse as [H | H] ; auto.
  Defined.

  Definition beq_bexp (b₁ b₂ : bexp) : bool :=
    if bexp_eq_dec b₁ b₂ then true else false.

  Inductive cmd :=
      | Skip
      | Seq : cmd → cmd → cmd
      | Assign : varid → exp → cmd
      | ITE : bexp → cmd → cmd → cmd
      | While : bexp → cmd → cmd
      | Assert : bexp → cmd.

  Fixpoint In_var (x : varid) (c : cmd) :=
    match c with
      | Skip => Logic.False
      | Seq c₁ c₂ => In_var x c₁ ∨ In_var x c₂
      | Assign y e => x = y ∨ In_exp_var x e
      | ITE b c₁ c₂ => In_bexp_var x b ∨ In_var x c₁ ∨ In_var x c₂
      | While b c' => In_bexp_var x b ∨ In_var x c'
      | Assert b => In_bexp_var x b
    end.

  Lemma In_var_dec:
    ∀ x c, {In_var x c} + {¬ In_var x c}.
  Proof.
    intros x c.
    induction c as [ | c₁ IH₁ c₂ IH₂ | y e | b c₁ IH₁ c₂ IH₂ | b c IH | b ] ;
      simpl ;
      auto using In_exp_var_dec, In_bexp_var_dec ;
      try tauto.
    - destruct (Var_as_DT.eq_dec x y) as [H | H] ; auto.
      destruct (In_exp_var_dec x e) as [H₁ | H₂] ; [ left | right ] ; auto.
      intro HFalse. destruct HFalse ; auto.
    - destruct (In_bexp_var_dec x b) as [Hb | Hb] ; [ left ; auto | ].
      destruct IH₁ as [IH₁ | IH₁] ; auto.
      destruct IH₂ as [IH₂ | IH₂] ; auto.
      right. intro HFalse. intuition.
    - destruct (In_bexp_var_dec x b) as [Hb | Hb] ; [ left ; auto | ].
      destruct IH as [IH | IH] ; auto.
      right. intro HFalse. intuition.
  Qed.

  Lemma cmd_eq_dec:
    ∀ (c₁ c₂ : cmd), { c₁ = c₂ } + { c₁ ≠ c₂}.
  Proof.
    decide equality ; apply exp_eq_dec || apply bexp_eq_dec || apply Var_as_DT.eq_dec.
  Defined.

  Definition beq_cmd (c₁ c₂ : cmd) : bool :=
    if cmd_eq_dec c₁ c₂ then true else false.

  Property beq_cmd_reflx:
    ∀ c, beq_cmd c c = true.
  Proof.
    intro c. unfold beq_cmd.
    destruct (cmd_eq_dec c c) ; auto.
  Qed.

  Property beq_cmd_true_iff:
    ∀ c₁ c₂, beq_cmd c₁ c₂ = true ↔  c₁ = c₂.
  Proof.
    intros c₁ c₂. split.
    * intro H. unfold beq_cmd in H. destruct (cmd_eq_dec c₁ c₂) ; try discriminate H. auto.
    * intro H. subst c₂. apply beq_cmd_reflx.
  Qed.

  Property beq_cmd_false_iff:
    ∀ c₁ c₂, beq_cmd c₁ c₂ = false ↔  c₁ ≠ c₂.
  Proof.
    intros c₁ c₂. split ; intro H.
    * unfold beq_cmd in H. destruct (cmd_eq_dec c₁ c₂) ; try discriminate H. auto.
    * unfold beq_cmd. destruct (cmd_eq_dec c₁ c₂).
      + exfalso. apply H. auto.
      + reflexivity.
  Qed.

  Hint Rewrite beq_cmd_reflx : cmd.

  (* Semantics *)
  Definition store := M.t Z.

  Definition store_equal (s₁ s₂ : store) :=
    M.equal Z.eqb s₁ s₂.

  Lemma store_equal_true_iff:
    ∀ (s₁ s₂ : store), store_equal s₁ s₂ = true ↔ M.Equal s₁ s₂.
  Proof.
    intros s₁ s₂. unfold store_equal. rewrite <- MFacts.equal_iff.
    rewrite (MFacts.Equal_Equivb Z.eqb_eq).
    reflexivity.
  Qed.

  Lemma store_Equal_dec:
  (* XXX: might not be the most straightforward way to do that… *)
    ∀ (s₁ s₂ : store), {M.Equal s₁ s₂} + {¬ M.Equal s₁ s₂}.
  Proof.
    intros s₁ s₂.
    destruct (store_equal s₁ s₂) eqn:H.
    - rewrite store_equal_true_iff in H. left. exact H.
    - right. intro H'. rewrite <- store_equal_true_iff in H'.
      rewrite H in H'. discriminate H'.
  Defined.

  Fixpoint eval_exp (s : store) (e : exp) : option Z :=
    match e with
    | Int n => ret n
    | Var x => M.find x s
    | Sum e₁ e₂ => eval_exp s e₁ >>= λ x, eval_exp s e₂ >>= λ y, ret (x + y)%Z
    | Sub e₁ e₂ => eval_exp s e₁ >>= λ x, eval_exp s e₂ >>= λ y, ret (x - y)%Z
    | Mult e₁ e₂ => eval_exp s e₁ >>= λ x, eval_exp s e₂ >>= λ y, ret (x * y)%Z
    | Div e₁ e₂ => eval_exp s e₁ >>= λ x, eval_exp s e₂ >>= λ y,
      match y with
      | Z0 => None
      | _ => ret (Z.div x y)
      end
    | Modulo e₁ e₂ => eval_exp s e₁ >>= λ x, eval_exp s e₂ >>= λ y,
      match y with
      | Z0 => None
      | _ => ret (Z.modulo x y)
      end
    end.

  Lemma store_equal_eval_exp:
    ∀ e s₁ s₂, M.Equal s₁ s₂ → eval_exp s₁ e = eval_exp s₂ e.
  Proof.
    intros e s₁ s₂ H.
    induction e ; simpl ; auto ;
    rewrite IHe1, IHe2 ; auto.
  Qed.

  Fixpoint eval_bexp (s : store) (b : bexp) : option bool :=
    match b with
    | True => ret true
    | False => ret false
    | EQ e₁ e₂ => eval_exp s e₁ >>= λ x, eval_exp s e₂ >>= λ y, ret (Z.eqb x y)
    | LT e₁ e₂ => eval_exp s e₁ >>= λ x, eval_exp s e₂ >>= λ y, ret (Z.ltb x y)
    | LE e₁ e₂ => eval_exp s e₁ >>= λ x, eval_exp s e₂ >>= λ y, ret (Z.leb x y)
    | EQB b₁ b₂ => eval_bexp s b₁ >>= λ x, eval_bexp s b₂ >>= λ y, ret (eqb x y)
    | AND b₁ b₂ =>
      eval_bexp s b₁ >>= λ x,
      if x then
        eval_bexp s b₂
      else
        ret false
    | OR b₁ b₂ =>
      eval_bexp s b₁ >>= λ x,
      if x then
        ret true
      else
        eval_bexp s b₂
    | NOT b => eval_bexp s b >>= λ x, ret (negb x)
    end.

  Lemma store_equal_eval_bexp:
    ∀ b s₁ s₂, M.Equal s₁ s₂ → eval_bexp s₁ b = eval_bexp s₂ b.
  Proof.
    intros b s₁ s₂ H.
    induction b ; simpl ; auto ;
    repeat rewrite (store_equal_eval_exp _ s₁ s₂) ; auto.
    * rewrite IHb1, IHb2. auto.
    * rewrite IHb1, IHb2. auto.
    * rewrite IHb1, IHb2. auto.
    * rewrite IHb. auto.
  Qed.

  Module Notations.
    (* Expressions *)
    Notation "a + b" := (Sum a b)
      (at level 50, left associativity) : ast_scope.
    Notation "a - b" := (Sub a b)
      (at level 50, left associativity) : ast_scope.
    Notation "a / b" := (Div a b)
      (at level 40, left associativity) : ast_scope.
    Notation "a * b" := (Mult a b)
      (at level 40, left associativity) : ast_scope.
    Notation "a '%' b" := (Modulo a b)
      (at level 40, left associativity) : ast_scope.

    (* Boolean expressions *)
    Notation "a == b" := (EQ a b)
      (at level 70, no associativity) : ast_scope.
    Notation "a < b" := (LT a b)
      (at level 70, no associativity) : ast_scope.
    Notation "a ≤ b" := (LE a b)
      (at level 70, no associativity) : ast_scope.
    Notation "a || b" := (OR a b)
      (at level 50, left associativity) : ast_scope.
    Notation "a && b" := (AND a b)
      (at level 40, left associativity) : ast_scope.
    Notation "¬ b" := (NOT b)
      (at level 75, right associativity) : ast_scope.

    (* Commands *)
    Notation "x ← e" := (Assign x e)
                          (at level 80, no associativity) : ast_scope.
    Notation "{ x ; y ; .. ; z }" := (Seq x (Seq y .. (Seq z Skip) ..)) : ast_scope.
    Notation "'IF' b 'THEN' c₁ 'ELSE' c₂" := (ITE b c₁ c₂) (at level 200) : ast_scope.
    Notation "'WHILE' b 'DO' c" := (While b c) (at level 200) : ast_scope.

    Delimit Scope ast_scope with AST.
  End Notations.
End Imp.
