(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale
  Copyright © 2017 Yann Régis-Gianas

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

(** Basic definitions about correlating oracles.

    A correlating oracle is a program that realizes a relation between
    reduction traces in such a way that the verification that this
    relation holds between the reduction traces of two programs is
    reduced to the proof of a program, more precisely to the proof
    that an invariant holds during the evaluation of a Coq program.

    We are interested in defining languages of correlating oracles,
    each oracle language corresponding to a specific class of
    relations. For instance, the oracle language of "renaming"s
    contains all the oracles that relate pair of reduction traces
    [(tr₁, tr₂)] where the configurations of [tr₁] can be obtained by
    a renaming of the configurations of [tr₂].

    An oracle language is characterized by a class of invariants that
    is fulfilled by its inhabitants. The purpose of this invariant is
    twofold: first, it represents the static property that holds
    between the program configurations related by the oracle; second,
    it is a sufficient condition for the oracle to effectively realize
    a trace relation (i.e. the oracle must be “productive” and it must
    relate actual program configurations).
*)

Require Import Common.
Require Import ProgrammingLanguage.
Require Import ReductionTrace.
Require Import ReductionTracesRelation.

(** As with programming languages, we specify a record signature for
    oracle languages. This signature is parameterized by two programming
    languages, as an oracle relates a pair of programs written in
    two potentially different programming languages.

    Oracle languages are very similar to programming languages, except
    they are not required to provide a decidable equality on their states.

    Furthermore, oracle languages are equipped with projection functions
    to recover left and right program states ([left_state] and [right_state])
    from an oracle state, as well as as left and right global environments
    ([left_genv] and [right_genv]) from global oracle environments.

    Oracle languages also provide a class of invariants [invariant] as well
    as several proofs:
    1. Proof [invariant_1] that the invariant holds
    2. Proof [prediction_soundness] that if an oracle configuration for which
       [invariant] holds relates two program configurations, the oracle's
       reduction function [oracle_step] is defined on this oracle state and
       leads to an oracle state relating two program configurations reachable
       from the previous ones.
*)

Record oracle_language (L₁ L₂ : language) :=
  {
    (* Semantics *)
    oracle_state : Type;
    oracle_genv  : Type;
    oracle_step  : oracle_genv → oracle_state → option oracle_state;

    left_state :
      oracle_state → state L₁;
    right_state :
      oracle_state → state L₂;
    left_genv :
      oracle_genv → genv_type L₁;
    right_genv :
      oracle_genv → genv_type L₂;

    (* Another important aspect of oracle languages is that it
       exposes an invariant on the program states it relates *)
    invariant :
      oracle_genv → oracle_state → Prop ;

    (* Properties of well-defined oracles: *)
    invariant_1 :
      ∀ genv os os',
        invariant genv os →
        oracle_step genv os = Some os' →
        invariant genv os';

    (* Any prediction from a sound oracle should be productive and
       indeed correspond to an execution step of the correlated
       programs *)
    prediction_soundness :
      ∀ genv os os',
        invariant genv os →
        oracle_step genv os = Some os' →
        ∃ n₁ n₂,
          opt_state_eq (nsteps (left_genv  genv) n₁ (left_state os))
                       (Some (left_state  os'))
        ∧ opt_state_eq (nsteps (right_genv genv) n₂ (right_state os))
                       (Some (right_state os'))
        (* Productivity: *)
        ∧ n₁ + n₂ > 0;

    (* We may need a stronger notion of productivity: the oracle
       will eventually step in each program. But we probably don't
       want to enforce that on each oracle language, as it would
       prevent reasonning about some pairs of programs *)

    (* Sound oracles only get stuck when both the underlying
       programs are stuck *)
    prediction_completeness :
      ∀ genv os,
        invariant genv os →
        oracle_step genv os = None →
        step L₁ (left_genv  genv) (left_state  os) = None
        ∧ step L₂ (right_genv genv) (right_state os) = None;

  }.

Arguments oracle_genv {L₁ L₂} _.
Arguments oracle_state {L₁ L₂} _.
Arguments oracle_step {L₁ L₂} _ genv S.
Arguments left_state {L₁ L₂} _ S.
Arguments right_state {L₁ L₂} _ S.
Arguments left_genv {L₁ L₂} _ genv.
Arguments right_genv {L₁ L₂} _ genv.
Arguments invariant {L₁ L₂} _ genv os.

(**

   An oracle realizes a specific trace relation. This trace relation
   relates the execution traces of two specific programs.

   By contrast, an oracle language defines a trace relation relating
   every pair of traces for which an instance exists.

   In general, a trace T₁ is related to a trace T₂ by an oracle O if:

   A. T₁ starts with a configuration C₁ and T₂ starts with a
      configuration C₂ that are related by O.

   B. Once initialized, the oracle can be executed and at each step of
      its execution the state of the oracle can be projected as
      configurations appearing in order in T₁ and T₂.

   If the invariant of O holds on the initial configurations of T₁ and
   T₂ then A and B holds. This is guaranteed by
   [prediction_soundness] and [invariant_1].

 *)

Definition oracle_correlation_relation
  {L₁ L₂ : language} (OT : oracle_language L₁ L₂)
  (genv : oracle_genv OT) (state : oracle_state OT)
: trace_relation L₁ L₂ :=
  λ tr₁ tr₂,
    let '(genv₁, S₁) := first tr₁ in
    let '(genv₂, S₂) := first tr₂ in
    trace L₁ (genv₁, S₁) tr₁ ∧ trace L₂ (genv₂, S₂) tr₂
    ∧ genv_eq L₁ (left_genv OT genv) genv₁
    ∧ genv_eq L₂ (right_genv OT genv) genv₂
    ∧ state_eq L₁ (left_state  OT state) S₁
    ∧ state_eq L₂ (right_state OT state) S₂
    ∧ invariant OT genv state.

Definition correlation_relation
  {L1 L2 : language} (OT : oracle_language L1 L2)
: trace_relation L1 L2 :=
  λ tr₁ tr₂,
    ∃ genv os, oracle_correlation_relation OT genv os tr₁ tr₂.

(** An oracle realizes a relation between configurations of two languages. *)
CoFixpoint correlation_trace_from
  {L₁ L₂ : language} (OT : oracle_language L₁ L₂)
  (genv  : oracle_genv OT)
  (state : oracle_state OT)
  : coplist (configuration L₁ * configuration L₂) :=
  let c₁ := (left_genv  OT genv, left_state  OT state) in
  let c₂ := (right_genv OT genv, right_state OT state) in
  match oracle_step OT genv state with
  | Some next_state =>
    CCons _ (c₁, c₂) (correlation_trace_from OT genv next_state)
  | None =>
    (* Oracle is stuck *)
    CSingle _ (c₁, c₂)
  end.

Definition left_trace {L1 L2 : language} 
           (cotrace : coplist (configuration L1 * configuration L2)) 
:= Common.map fst cotrace.

Definition right_trace {L1 L2 : language} 
           (cotrace : coplist (configuration L1 * configuration L2)) 
:= Common.map snd cotrace.

(** If two traces are related by an oracle, then the oracle realizes a
    relation between the configurations of these two traces. *)
Theorem oracle_relates_configurations_of_programs
  {L₁ L₂ : language} (OT : oracle_language L₁ L₂) :
  ∀ (genv : oracle_genv OT) (state : oracle_state OT) tr₁ tr₂,
    oracle_correlation_relation OT genv state tr₁ tr₂ →
    let correlation_trace := correlation_trace_from OT genv state in
    let subtr₁ := left_trace correlation_trace in
    let subtr₂ := right_trace correlation_trace in
    included_subtrace subtr₁ tr₁ ∧ included_subtrace subtr₂ tr₂.
Proof.
  intros genv state tr₁ tr₂ H.
  unfold oracle_correlation_relation in H.
  destruct (first tr₁) as [genv₁ S₁] eqn:Hc₁, (first tr₂) as [genv₂ S₂] eqn:Hc₂.
  destruct H as [Htr₁ [Htr₂ [Hgenv₁ [Hgenv₂ [HS₁ [HS₂ Hinv]]]]]].
  split.
  - clear genv₂ Hgenv₂ S₂ HS₂ Htr₂ Hc₂ tr₂.
    revert genv₁ S₁ tr₁ Hgenv₁ HS₁ Hc₁ Htr₁ ;
    revert state Hinv.
    cofix H.
    intros state Hinv
           genv₁ S₁ tr₁ Hgenv₁ HS₁ Hc₁ Htr₁.
    rewrite <- (coplist_id_force_eq _ (correlation_trace_from _ _ _)).
    simpl.
    unfold left_trace.
    rewrite <- (coplist_id_force_eq _ (map fst _)).
    destruct (oracle_step OT genv state) as [state' | ] eqn:Hstate'.
    + destruct (prediction_soundness L₁ L₂ OT genv state state' Hinv Hstate')
        as [n₁ [_ [H₁ _]]].
      apply (opt_state_eq_trans _ _ _
                                (nsteps_eq _ _ _ _ n₁ (genv_eq_sym _ _ _ Hgenv₁)
                                           (state_eq_sym _ _ _ HS₁))) in H₁.
      pose proof (invariant_1 L₁ L₂ OT genv state state' Hinv Hstate') as Hinv'.
      pose proof (λ S₁ tr₁, H _ Hinv' genv₁ S₁ tr₁ Hgenv₁) as H'.
      clear H Hinv'. rename H' into H.
      simpl. destruct n₁ as [ | n₁ ].
      { (* stutter *)
        unfold nsteps in H₁. simpl in H₁. rewrite <- state_eq__opt_state_eq in H₁.
        apply (Stutter _ (map fst (correlation_trace_from OT genv state'))) ;
          auto.
        + simpl. rewrite Hc₁. split ; auto.
        + apply (λ H', H S₁ tr₁ H' Hc₁ Htr₁).
          eauto using state_eq_trans, state_eq_sym. }
      inversion Htr₁ as [genv₁_ S₁_ HS₁' | genv₁_ S₁_ S₁' tr₁' HS₁' Htr₁' ] ;
        subst genv₁_ S₁_ tr₁.
      { unfold nsteps in H₁. simpl in H₁.
        rewrite HS₁' in H₁. simpl in H₁.
        destruct H₁ as [[_ HFalse] | [A [B [HFalse _]]]] ; discriminate HFalse. }
      apply SameFirst.
      { simpl. split ; auto. }
      intros tr₁'' Htr₁''. simpl in Htr₁''. inversion Htr₁''. clear Htr₁''. subst tr₁''.
      exists tr₁'. split ; auto.
      unfold nsteps in H₁. simpl in H₁. rewrite HS₁' in H₁. simpl in H₁.
      destruct n₁ as [ | n₁ ].
      { simpl in H₁. rewrite <- state_eq__opt_state_eq in H₁.
        apply (H S₁' tr₁') ; auto using state_eq_sym.
        apply trace_first ; auto. }
      destruct (nsteps_skipn L₁ tr₁' (S n₁) genv₁ _ _ Htr₁' H₁)
        as [tr₁'' [S₁'' [Htr₁'' [H'tr₁'' HS₁'']]]].
      apply (IgnoreFirstSN n₁) with tr₁'' ; auto.
      apply (H S₁'') ; auto using state_eq_sym.
      apply trace_first. auto.
    + destruct (prediction_completeness L₁ L₂ OT genv state Hinv Hstate')
        as [H₁ _].
      simpl. apply SameFirst ; auto ; try discriminate.
      unfold configuration_eq. simpl. rewrite Hc₁. simpl.
      split ; auto using genv_eq_refl.
  - clear genv₁ Hgenv₁ S₁ HS₁ Htr₁ Hc₁ tr₁.
    revert genv₂ S₂ tr₂ Hgenv₂ HS₂ Hc₂ Htr₂ ;
    revert state Hinv.
    cofix H.
    intros state Hinv
           genv₂ S₂ tr₂ Hgenv₂ HS₂ Hc₂ Htr₂.
    rewrite <- (coplist_id_force_eq _ (correlation_trace_from _ _ _)).
    simpl.
    unfold right_trace.
    rewrite <- (coplist_id_force_eq _ (map snd _)).
    destruct (oracle_step OT genv state) as [state' | ] eqn:Hstate'.
    + destruct (prediction_soundness L₁ L₂ OT genv state state' Hinv Hstate')
        as [_ [n₂ [_ [H₁ _]]]].
      apply (opt_state_eq_trans _ _ _
                                (nsteps_eq _ _ _ _ n₂ (genv_eq_sym _ _ _ Hgenv₂)
                                           (state_eq_sym _ _ _ HS₂))) in H₁.
      pose proof (invariant_1 L₁ L₂ OT genv state state' Hinv Hstate') as Hinv'.
      pose proof (λ S₂ tr₂, H _ Hinv' genv₂ S₂ tr₂ Hgenv₂) as H'.
      clear H Hinv'. rename H' into H.
      simpl. destruct n₂ as [ | n₂ ].
      { (* stutter *)
        unfold nsteps in H₁. simpl in H₁. rewrite <- state_eq__opt_state_eq in H₁.
        apply (Stutter _ (map snd (correlation_trace_from OT genv state'))) ;
          auto.
        + simpl. rewrite Hc₂. split ; auto.
        + apply (λ H', H S₂ tr₂ H' Hc₂ Htr₂).
          eauto using state_eq_trans, state_eq_sym. }
      inversion Htr₂ as [genv₂_ S₂_ HS₂' | genv₂_ S₂_ S₂' tr₂' HS₂' Htr₂' ] ;
        subst genv₂_ S₂_ tr₂.
      { unfold nsteps in H₁. simpl in H₁.
        rewrite HS₂' in H₁. simpl in H₁.
        destruct H₁ as [[_ HFalse] | [A [B [HFalse _]]]] ; discriminate HFalse. }
      apply SameFirst.
      { simpl. split ; auto. }
      intros tr₂'' Htr₂''. simpl in Htr₂''. inversion Htr₂''. clear Htr₂''. subst tr₂''.
      exists tr₂'. split ; auto.
      unfold nsteps in H₁. simpl in H₁. rewrite HS₂' in H₁. simpl in H₁.
      destruct n₂ as [ | n₂ ].
      { simpl in H₁. rewrite <- state_eq__opt_state_eq in H₁.
        apply (H S₂' tr₂') ; auto using state_eq_sym.
        apply trace_first ; auto. }
      destruct (nsteps_skipn L₂ tr₂' (S n₂) genv₂ _ _ Htr₂' H₁)
        as [tr₂'' [S₂'' [Htr₂'' [H'tr₂'' HS₂'']]]].
      apply (IgnoreFirstSN n₂) with tr₂'' ; auto.
      apply (H S₂'') ; auto using state_eq_sym.
      apply trace_first. auto.
    + unfold right_trace. rewrite <- (coplist_id_force_eq _ (map snd _)).
      simpl. subst.
      apply SameFirst ; auto ; try discriminate.
      unfold configuration_eq. simpl. rewrite Hc₂. simpl.
      split ; auto using genv_eq_refl.
Qed.

(**
    The definition [oracle_correlation_relation] defined previously only depends on
    the initial states of those traces, and exploiting them may rely on oracle-specific
    knowledge.

    Fortunately, every pair of traces related by an oracle are also related by a slightly
    more general γ-correlation trace relation that abstracts away a few of the oracle
    details.
*)

Definition γ_from_oracle {L₁ L₂ : language}
           (OT : oracle_language L₁ L₂) (genv : oracle_genv OT)
  : configuration L₁ → configuration L₂ → Prop :=
  λ c₁ c₂,
  ∃ os,
      configuration_eq c₁ (left_genv  OT genv, left_state  OT os)
      ∧ configuration_eq c₂ (right_genv OT genv, right_state OT os)
      ∧ invariant OT genv os.

Theorem oracle_gamma_correlated
        {L₁ L₂ : language} (OT : oracle_language L₁ L₂) :
  ∀ (genv : oracle_genv OT) (state : oracle_state OT) tr₁ tr₂,
    oracle_correlation_relation OT genv state tr₁ tr₂ →
    gamma_correlated (γ_from_oracle OT genv) tr₁ tr₂.
Proof.
  intros genv state tr₁ tr₂ H.
  unfold oracle_correlation_relation in H.
  unfold trace' in *.
  destruct (first tr₁) as [genv₁ S₁], (first tr₂) as [genv₂ S₂].
  destruct H as [Htr₁ [Htr₂ [Hgenv₁ [Hgenv₂ [HS₁ [HS₂ Hinv]]]]]].
  revert genv state genv₁ genv₂ S₁ S₂ tr₁ tr₂ Hgenv₁ Hgenv₂ HS₁ HS₂ Htr₁ Htr₂ Hinv.
  cofix H.
  intros genv state genv₁ genv₂ S₁ S₂ tr₁ tr₂ Hgenv₁ Hgenv₂ HS₁ HS₂ Htr₁ Htr₂ Hinv.
  destruct (oracle_step OT genv state) as [state' | ] eqn:Hstate'.
  - destruct (prediction_soundness L₁ L₂ OT genv state state' Hinv Hstate')
      as [n₁ [n₂ [H₁ [H₂ Hprod]]]].
    apply (opt_state_eq_trans _ _ _
                              (nsteps_eq _ _ _ _ n₁ (genv_eq_sym _ _ _ Hgenv₁)
                                         (state_eq_sym _ _ _ HS₁))) in H₁.
    apply (opt_state_eq_trans _ _ _
                              (nsteps_eq _ _ _ _ n₂ (genv_eq_sym _ _ _ Hgenv₂)
                                         (state_eq_sym _ _ _ HS₂))) in H₂.
    destruct (nsteps_skipn _ _ _ _ _ _ Htr₁ H₁)
      as [tr₁' [S₁'' [Htr₁' [H'tr₁' HS₁'']]]].
    destruct (nsteps_skipn _ _ _ _ _ _ Htr₂ H₂)
      as [tr₂' [S₂'' [Htr₂' [H'tr₂' HS₂'']]]].
    apply (StepCorrelated _ tr₁ tr₂ n₁ n₂ tr₁' tr₂') ; auto.
    + unfold γ_from_oracle.
      exists state. unfold configuration_eq.
      simpl. rewrite (trace_first _ _ Htr₁), (trace_first _ _ Htr₂). simpl.
      repeat (split ; auto using genv_eq_sym, state_eq_sym).
    + apply (H genv state' genv₁ genv₂ S₁'' S₂'') ;
        eauto using state_eq_sym, invariant_1.
  - clear H.
    destruct (prediction_completeness L₁ L₂ OT genv state Hinv Hstate') as [H₁ H₂].
    rewrite <- (step_eq' _ genv₁ _ S₁) in H₁ ; auto using state_eq_sym, genv_eq_sym.
    rewrite <- (step_eq' _ genv₂ _ S₂) in H₂ ; auto using state_eq_sym, genv_eq_sym.
    inversion Htr₁ as [genv₁_ S₁_ HS₁' | genv₁_ S₁_ S₁' tr₁' HS₁' Htr₁' ] ;
      subst genv₁_ S₁_ tr₁ ; try (rewrite H₁ in HS₁' ; discriminate HS₁').
    inversion Htr₂ as [genv₂_ S₂_ HS₂' | genv₂_ S₂_ S₂' tr₂' HS₂' Htr₂' ] ;
      subst genv₂_ S₂_ tr₂ ; try (rewrite H₂ in HS₂' ; discriminate HS₂').
    apply TerminatedCorrelated.
    unfold γ_from_oracle. exists state.
    repeat (split ; auto using state_eq_sym, genv_eq_sym).
Qed.


(* Oracle equality is defined co-inductively *)
CoInductive oracle_eq {L₁ L₂ : language}
            (OT₁ OT₂ : oracle_language L₁ L₂)
            (genv₁ : oracle_genv OT₁) (genv₂ : oracle_genv OT₂) :
  oracle_state OT₁ → oracle_state OT₂ → Prop :=
| OracleStuckEquality :
    ∀ os₁ os₂,
      state_eq L₁ (left_state  OT₁ os₁) (left_state  OT₂ os₂) →
      state_eq L₂ (right_state OT₁ os₁) (right_state OT₂ os₂) →
      oracle_step OT₁ genv₁ os₁ = None →
      oracle_step OT₂ genv₂ os₂ = None →
      oracle_eq OT₁ OT₂ genv₁ genv₂ os₁ os₂
| OracleStepEquality :
    ∀ os₁ os₂ os₁' os₂',
      state_eq L₁ (left_state  OT₁ os₁) (left_state  OT₂ os₂) →
      state_eq L₂ (right_state OT₁ os₁) (right_state OT₂ os₂) →
      oracle_step OT₁ genv₁ os₁ = Some os₁' →
      oracle_step OT₂ genv₂ os₂ = Some os₂' →
      oracle_eq OT₁ OT₂ genv₁ genv₂ os₁' os₂' →
      oracle_eq OT₁ OT₂ genv₁ genv₂ os₁ os₂.

Lemma oracle_eq_refl:
  ∀ {L₁ L₂ : language} {OT : oracle_language L₁ L₂}
    (genv : oracle_genv OT) (os : oracle_state OT),
    oracle_eq OT OT genv genv os os.
Proof.
  intros L₁ L₂ OT genv.
  cofix.
  intros os.
  destruct (oracle_step _ genv os) as [os' | ] eqn:H.
  - apply (OracleStepEquality OT OT genv genv os os os' os') ;
      auto using state_eq_refl.
  - apply OracleStuckEquality ;
      auto using state_eq_refl.
Qed.

Lemma oracle_eq_sym:
  ∀ {L₁ L₂ : language}
    {OT₁ OT₂ : oracle_language L₁ L₂}
    (genv₁ : oracle_genv OT₁) (genv₂ : oracle_genv OT₂)
    (os₁ : oracle_state OT₁) (os₂ : oracle_state OT₂),
    oracle_eq OT₁ OT₂ genv₁ genv₂ os₁ os₂ →
    oracle_eq OT₂ OT₁ genv₂ genv₁ os₂ os₁.
Proof.
  intros L₁ L₂ OT₁ OT₂ genv₁ genv₂.
  cofix.
  intros os₁ os₂ H.
  destruct H as [os₁ os₂ Heq₁ Heq₂ H₁ H₂ | os₁ os₂ os₁' os₂' Heq₁ Heq₂ H₁ H₂ H].
  - apply OracleStuckEquality ;
      auto using state_eq_sym.
  - apply (OracleStepEquality OT₂ OT₁ genv₂ genv₁ os₂ os₁ os₂' os₁') ;
      auto using state_eq_sym.
Qed.

Lemma oracle_eq_trans:
  ∀ {L₁ L₂ : language}
    {OT₁ OT₂ OT₃ : oracle_language L₁ L₂}
    (genv₁ : oracle_genv OT₁) (genv₂ : oracle_genv OT₂) (genv₃ : oracle_genv OT₃)
    (os₁ : oracle_state OT₁) (os₂ : oracle_state OT₂) (os₃ : oracle_state OT₃),
    oracle_eq OT₁ OT₂ genv₁ genv₂ os₁ os₂ →
    oracle_eq OT₂ OT₃ genv₂ genv₃ os₂ os₃ →
    oracle_eq OT₁ OT₃ genv₁ genv₃ os₁ os₃.
Proof.
  intros L₁ L₂ OT₁ OT₂ OT₃ genv₁ genv₂ genv₃.
  cofix.
  intros os₁ os₂ os₃ H₁ H₂.
  destruct H₁ as [os₁ os₂ Heq₁ Heq₂ H₁ H₁' | os₁ os₂ os₁' os₂' Heq₁ Heq₂ H₁ H₁' H].
  - destruct H₂ as [os₂ os₃ Heq₂' Heq₃ H₂ H₂' | os₂ os₃ os₂' os₃' Heq₂' Heq₃ H₂ H₂' H'].
    + apply OracleStuckEquality ;
        eauto using state_eq_trans.
    + rewrite H₁' in H₂. discriminate H₂.
  - destruct H₂ as [os₂ os₃ Heq₂' Heq₃ H₂ H₂' | os₂ os₃ os₂'' os₃' Heq₂' Heq₃ H₂ H₂' H'].
    + rewrite H₁' in H₂. discriminate H₂.
    + apply (OracleStepEquality OT₁ OT₃ genv₁ genv₃ os₁ os₃ os₁' os₃') ;
        eauto using state_eq_trans.
      apply (oracle_eq_trans os₁' os₂'' os₃') ;
        eauto using state_eq_trans.
      rewrite H₁' in H₂. inversion H₂ as [H'₂]. subst os₂'.
      auto.
Qed.

(* Helper function to repeatedly call an oracle *)
Definition oracle_nsteps {L₁ L₂} {OT : oracle_language L₁ L₂}
         (genv : oracle_genv OT)
         (n : nat) (os : oracle_state OT) :
  option (oracle_state OT) := napply (oracle_step OT genv) n os.

Definition oracle_nsteps' {L₁ L₂} {OT : oracle_language L₁ L₂}
         (genv : oracle_genv OT)
         (n : nat)
         (os : oracle_state OT) :
  oracle_state OT * nat := napply' (oracle_step OT genv) n os.

Lemma oracle_nsteps_left:
  ∀ {L₁ L₂} {OT : oracle_language L₁ L₂} genv n os os',
    invariant OT genv os →
    oracle_nsteps genv n os = Some os' →
    ∃ n',
      opt_state_eq (nsteps (left_genv OT genv) n' (left_state OT os))
                   (Some (left_state OT os')).
Proof.
  intros L₁ L₂ OT genv n.
  induction n as [ | n IH ] ;
    intros os os' Hinv Hos'.
  - exists 0. simpl.
    simpl in Hos'. inversion Hos'. clear Hos'. subst os'.
    apply opt_state_eq_refl.
  - unfold oracle_nsteps in Hos'. simpl in Hos'.
    destruct (oracle_step OT genv os) as [os'' | ] eqn:Hos'' ;
      try discriminate Hos'.
    simpl in Hos'.
    destruct (prediction_soundness L₁ L₂ OT genv os os'' Hinv Hos'')
      as [n₁ [n₂ [H₁ _]]].
    pose proof (invariant_1 L₁ L₂ OT genv os os'' Hinv Hos'') as Hinv'.
    destruct (IH os'' os' Hinv' Hos') as [n' H'].
    exists (n₁ + n').
    simpl. simpl in H₁.
    rewrite nsteps_plus_0.
    destruct H₁ as [[H₁ H₁'] | [S' [S'' [HS' [HS'' HS'S'']]]]] ;
      try discriminate H₁'.
    rewrite HS'. simpl. inversion HS''. clear HS''. subst S''.
    apply opt_state_eq_trans
      with (nsteps (left_genv OT genv) n' (left_state OT os'')) ; auto.
    apply nsteps_eq ; auto using genv_eq_refl.
Qed.

Lemma oracle_nsteps_right:
  ∀ {L₁ L₂} {OT : oracle_language L₁ L₂} genv n os os',
    invariant OT genv os →
    oracle_nsteps genv n os = Some os' →
    ∃ n',
      opt_state_eq (nsteps (right_genv OT genv) n' (right_state OT os))
                   (Some (right_state OT os')).
Proof.
  intros L₁ L₂ OT genv n.
  induction n as [ | n IH ] ;
    intros os os' Hinv Hos'.
  - exists 0. simpl.
    simpl in Hos'. inversion Hos'. clear Hos'. subst os'.
    apply opt_state_eq_refl.
  - unfold oracle_nsteps in Hos'. simpl in Hos'.
    destruct (oracle_step OT genv os) as [os'' | ] eqn:Hos'' ;
      try discriminate Hos'.
    simpl in Hos'.
    destruct (prediction_soundness L₁ L₂ OT genv os os'' Hinv Hos'')
      as [n₁ [n₂ [_ [H₂ _]]]].
    pose proof (invariant_1 L₁ L₂ OT genv os os'' Hinv Hos'') as Hinv'.
    destruct (IH os'' os' Hinv' Hos') as [n' H'].
    exists (n₂ + n').
    simpl. simpl in H₂.
    rewrite nsteps_plus_0.
    destruct H₂ as [[H₁ H₁'] | [S' [S'' [HS' [HS'' HS'S'']]]]] ;
      try discriminate H₁'.
    rewrite HS'. simpl. inversion HS''. clear HS''. subst S''.
    apply opt_state_eq_trans
      with (nsteps (right_genv OT genv) n' (right_state OT os'')) ; auto.
    apply nsteps_eq ; auto using genv_eq_refl.
Qed.

Require Import Omega.

(**

    Given a pair of programs related by an oracle, if neither those programs diverges,
    the oracle reaches those two programs' final states in a finite number of iterations.

*)

Lemma oracle_nsteps_termination:
  ∀ {L₁ L₂} {OT : oracle_language L₁ L₂}
    (genv : oracle_genv OT)
    (n n' n₁ n₂ : nat) (os os' : oracle_state OT)
    S₁'' S₂'',
    let S₁ := left_state  OT os in
    let S₂ := right_state OT os in
    invariant OT genv os →
    opt_state_eq (nsteps (left_genv  OT genv) n₁ S₁) (Some S₁'') →
    opt_state_eq (nsteps (right_genv OT genv) n₂ S₂) (Some S₂'') →
    step L₁ (left_genv  OT genv) S₁'' = None →
    step L₂ (right_genv OT genv) S₂'' = None →
    oracle_nsteps' genv n os = (os', n') →
    n ≥ n₁ + n₂ →
    let S₁' := left_state  OT os' in
    let S₂' := right_state OT os' in
    state_eq L₁ S₁' S₁'' ∧ state_eq L₂ S₂' S₂'' ∧ invariant OT genv os'.
Proof.
  intros L₁ L₂ OT genv n.
  induction n as [ | n IH ] ;
    intros n' n₁ n₂ os os' S₁'' S₂''
           S₁ S₂ Hinv HS₁'' HS₂'' H'S₁'' H'S₂'' Hos' Hn S₁' S₂'.
  - simpl in Hos'. inversion Hos'. clear Hos'. subst os'.
    destruct n₁ as [ | n₁ ], n₂ as [ | n₂ ] ; try omega.
    clear Hn.
    unfold nsteps in HS₁'', HS₂''. simpl in HS₁'', HS₂''.
    rewrite <- state_eq__opt_state_eq in HS₁''.
    rewrite <- state_eq__opt_state_eq in HS₂''.
    subst S₁ S₂ S₁' S₂'.
    repeat split ; auto.
  - unfold oracle_nsteps' in Hos'. simpl in Hos'.
    destruct (oracle_step OT genv os) as [os'' | ] eqn:Hos''.
    + destruct (prediction_soundness _ _ OT genv os os'' Hinv Hos'')
      as [n₁' [n₂' [H₁ [H₂ Hn₁'n₂']]]].
      destruct (opt_state_eq_Some_r__state_eq _ _ HS₁'') as [S₁'_ [HS₁'_ H'S₁'_]].
      destruct (opt_state_eq_Some_r__state_eq _ _ HS₂'') as [S₂'_ [HS₂'_ H'S₂'_]].
      clear HS₁'' HS₂''.
      (* The idea here is that an oracle that hasn't reached termination yet
         is forced to make progress *)
      apply (opt_state_eq_trans (nsteps (left_genv OT genv) n₁' S₁) _ _) in H₁ ;
        [ | apply nsteps_eq ; simpl ;
            eauto using state_eq_sym, genv_eq_refl, state_eq_refl ].
      apply (opt_state_eq_trans (nsteps (right_genv OT genv) n₂' S₂) _ _) in H₂ ;
        [ | apply nsteps_eq ; simpl ;
            eauto using state_eq_sym, genv_eq_refl, state_eq_refl ].
      rewrite <- (step_eq' _ _ _ _ _ (genv_eq_refl L₁ _) H'S₁'_) in H'S₁''.
      rewrite <- (step_eq' _ _ _ _ _ (genv_eq_refl L₂ _) H'S₂'_) in H'S₂''.
      destruct (nsteps_split_helper _ _ _ _ _ _ HS₁'_ H'S₁'' H₁)
        as [ Hn' | [[S₁i [Hn' HS₁i]] | [Hn' HNone]]].
      * subst n₁'.
        rewrite HS₁'_ in H₁.
        apply opt_state_eq_sym in H₁.
        destruct (opt_state_eq_Some_r__state_eq _ _ H₁) as [S₁i [HS₁i H'S₁i]].
        inversion HS₁i. clear H₁ HS₁i. subst S₁i.
        destruct (nsteps_split_helper _ _ _ _ _ _ HS₂'_ H'S₂'' H₂)
          as [ Hn'' | [[S₂i [Hn'' HS₂i]] | [Hn'' HNone]]].
        { subst n₂'.
          rewrite HS₂'_ in H₂.
          apply opt_state_eq_sym in H₂.
          destruct (opt_state_eq_Some_r__state_eq _ _ H₂) as [S₂i [HS₂i H'S₂i]].
          inversion HS₂i. clear HS₂i. subst S₂i.
          apply (IH n' 0 0 os'' os' S₁'' S₂'') ; eauto with arith.
          + eauto using invariant_1.
          + simpl. apply state_eq__opt_state_eq in H'S₁i.
            apply opt_state_eq_trans with (Some S₁'_) ; auto.
            rewrite <- state_eq__opt_state_eq. auto.
          + simpl. apply state_eq__opt_state_eq in H'S₂i.
            apply opt_state_eq_trans with (Some S₂'_) ; auto.
            rewrite <- state_eq__opt_state_eq.
            auto.
          + rewrite <- (step_eq' _ _ _ _ _ (genv_eq_refl L₁ _) H'S₁'_) ; auto.
          + rewrite <- (step_eq' _ _ _ _ _ (genv_eq_refl L₂ _) H'S₂'_) ; auto. }
        { apply (IH n' 0 (n₂ - n₂') os'' os' S₁'' S₂'') ; auto.
          + eauto using invariant_1.
          + simpl. apply state_eq__opt_state_eq in H'S₁i.
            apply opt_state_eq_trans with (Some S₁'_) ; auto.
            rewrite <- state_eq__opt_state_eq. simpl.
            auto.
          + apply opt_state_eq_trans
              with (nsteps (right_genv OT genv) (n₂' + (n₂ - n₂')) S₂).
            { apply opt_state_eq_sym, nsteps_plus_1'. auto. }
            replace (n₂' + (n₂ - n₂')) with n₂ ; [ | auto with arith ].
            rewrite HS₂'_. rewrite <- state_eq__opt_state_eq.
            simpl. auto.
          + rewrite <- (step_eq' _ _ _ _ _ (genv_eq_refl L₁ _) H'S₁'_) ; auto.
          + rewrite <- (step_eq' _ _ _ _ _ (genv_eq_refl L₂ _) H'S₂'_) ; auto.
          + omega. }
        { exfalso.
          rewrite HNone in H₂.
          destruct H₂ as [[_ HFalse] | [S [_ [HFalse _]]]] ;
            discriminate HFalse. }
      * destruct (nsteps_split_helper _ _ _ _ _ _ HS₂'_ H'S₂'' H₂)
          as [ Hn'' | [[S₂i [Hn'' HS₂i]] | [Hn'' HNone]]].
        { subst n₂'.
          rewrite HS₂'_ in H₂.
          apply opt_state_eq_sym in H₂.
          destruct (opt_state_eq_Some_r__state_eq _ _ H₂) as [S₂i [HS₂i H'S₂i]].
          apply (IH n' (n₁ - n₁') 0 os'' os' S₁'' S₂'') ; auto with arith.
          + eauto using invariant_1.
          + apply opt_state_eq_trans
              with (nsteps (left_genv OT genv) (n₁' + (n₁ - n₁')) S₁).
            { apply opt_state_eq_sym, nsteps_plus_1'. auto. }
            replace (n₁' + (n₁ - n₁')) with n₁ ; [ | auto with arith ].
            rewrite HS₁'_. rewrite <- state_eq__opt_state_eq.
            simpl. auto.
          + simpl. apply state_eq__opt_state_eq in H'S₂i.
            apply opt_state_eq_trans with (Some S₂'_) ; auto.
            rewrite <- state_eq__opt_state_eq. simpl. auto.
          + rewrite <- (step_eq' _ _ _ _ _ (genv_eq_refl L₁ _) H'S₁'_) ; auto.
          + rewrite <- (step_eq' _ _ _ _ _ (genv_eq_refl L₂ _) H'S₂'_) ; auto.
          + omega. }
        { apply (IH n' (n₁ - n₁') (n₂ - n₂') os'' os' S₁'' S₂'') ; auto.
          + eauto using invariant_1.
          + apply opt_state_eq_trans
              with (nsteps (left_genv OT genv) (n₁' + (n₁ - n₁')) S₁).
            { apply opt_state_eq_sym, nsteps_plus_1'. auto. }
            replace (n₁' + (n₁ - n₁')) with n₁ ; [ | auto with arith ].
            rewrite HS₁'_. rewrite <- state_eq__opt_state_eq.
            simpl. auto.
          + apply opt_state_eq_trans
            with (nsteps (right_genv OT genv) (n₂' + (n₂ - n₂')) S₂).
            { apply opt_state_eq_sym, nsteps_plus_1'. auto. }
            replace (n₂' + (n₂ - n₂')) with n₂ ; [ | auto with arith ].
            rewrite HS₂'_. rewrite <- state_eq__opt_state_eq.
            simpl. auto.
          + rewrite <- (step_eq' _ _ _ _ _ (genv_eq_refl L₁ _) H'S₁'_) ; auto.
          + rewrite <- (step_eq' _ _ _ _ _ (genv_eq_refl L₂ _) H'S₂'_) ; auto.
          + omega. }
        { exfalso.
          rewrite HNone in H₂.
          destruct H₂ as [[_ HFalse] | [S [_ [HFalse _]]]] ;
            discriminate HFalse. }
      * exfalso. rewrite HNone in H₁.
        destruct H₁ as [[_ HFalse] | [S [_ [HFalse _]]]] ;
          discriminate HFalse.
    + (* Both programs have terminated… *)
      subst S₁ S₂.
      inversion Hos'. clear Hos'. subst n' os'.
      destruct (prediction_completeness L₁ L₂ OT genv os Hinv Hos'') as
          [Hfinal₁ Hfinal₂].
      destruct n₁ ; unfold nsteps in HS₁'' ; simpl in HS₁'' ;
        try rewrite Hfinal₁ in HS₁'' ;
        try (destruct HS₁'' as [ [_ HFalse] | [foo [_ [HFalse _]]]] ; discriminate HFalse).
      destruct n₂ ; unfold nsteps in HS₂'' ; simpl in HS₂'' ;
        try rewrite Hfinal₂ in HS₂'' ;
        try (destruct HS₂'' as [ [_ HFalse] | [foo [_ [HFalse _]]]] ; discriminate HFalse).
      apply state_eq__opt_state_eq in HS₁''.
      apply state_eq__opt_state_eq in HS₂''.
      auto.
Qed.

Lemma oracle_nsteps_termination'':
  ∀ {L₁ L₂} {OT : oracle_language L₁ L₂}
    (genv : oracle_genv OT)
    (os : oracle_state OT),
    let S₁ := left_state  OT os in
    let S₂ := right_state OT os in
    invariant OT genv os →
    terminates (left_genv  OT genv) S₁ →
    terminates (right_genv OT genv) S₂ →
    ∃ n os',
      oracle_nsteps genv n os = Some os'
      ∧ invariant OT genv os'
      ∧ let S₁' := left_state  OT os' in
        let S₂' := right_state OT os' in
          step L₁ (left_genv  OT genv) S₁' = None
        ∧ step L₂ (right_genv OT genv) S₂' = None.
Proof.
  intros L₁ L₂ OT genv os S₁ S₂ Hinv H'S₁ H'S₂.
  destruct H'S₁ as [n₁ [S₁' [HS₁' H'S₁']]].
  destruct H'S₂ as [n₂ [S₂' [HS₂' H'S₂']]].
  destruct (oracle_nsteps' genv (n₁ + n₂) os) as [os' n'] eqn:Hsteps.
  pose proof (oracle_nsteps_termination genv (n₁ + n₂) n' n₁ n₂ os os' S₁' S₂' Hinv) as H.
  subst S₁ S₂.
  rewrite HS₁', HS₂' in H.
  destruct H as [HS₁''S₁ [HS₂''S₂ Hinv']] ; eauto using opt_state_eq_refl.
  unfold oracle_nsteps' in *.
  apply (napply'_prefix (oracle_step OT genv)) in Hsteps.
  apply napply_napply'_0_iff in Hsteps.
  exists (n₁ + n₂ - n'), os'.
  split ; auto. split ; auto. split ; auto.
  - rewrite (step_eq' _ _ _ _ _ (genv_eq_refl _ _) HS₁''S₁). auto.
  - rewrite (step_eq' _ _ _ _ _ (genv_eq_refl _ _) HS₂''S₂). auto.
Qed.
