(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
 *)

Require Import Common.
Require Import ProgrammingLanguage.
Require Import OracleLanguage.

Parameters L₁ L₂ : language.

Definition _state : Type := state L₁ * state L₂.
Definition _genv : Type := genv_type L₁ * genv_type L₂.

Definition _step (genv : _genv) (S : _state) : option _state :=
  let '(S₁, S₂) := S in
  match step L₁ (fst genv) S₁, step L₂ (snd genv) S₂ with
  | None, None => None
  | Some S₁', Some S₂' => Some (S₁', S₂')
  | Some S₁', None => Some (S₁', S₂)
  | None, Some S₂' => Some (S₁, S₂')
  end.

Lemma _prediction_soundness:
  ∀ genv os os',
    True →
    _step genv os = Some os' →
    ∃ n₁ n₂,
      opt_state_eq (nsteps (fst genv) n₁ (fst os))
                     (Some (fst os'))
      ∧ opt_state_eq (nsteps (snd genv) n₂ (snd os))
                     (Some (snd os'))
      ∧ n₁ + n₂ > 0.
Proof.
  intros [genv₁ genv₂] [S₁ S₂] [S₁' S₂'] _ Hos'.
  simpl. unfold _step in Hos'. simpl in Hos'.
  destruct (step L₁ genv₁ S₁) as [S₁'_ | ] eqn:H₁,
           (step L₂ genv₂ S₂) as [S₂'_ | ] eqn:H₂ ;
    inversion Hos' ; clear Hos' ;
      (subst S₁'_ || subst S₁') ;
      (subst S₂'_ || subst S₂').
  - exists 1, 1. unfold nsteps. simpl. rewrite H₁, H₂.
    simpl. split ; auto using opt_state_eq_refl.
  - exists 1, 0. unfold nsteps. simpl. rewrite H₁.
    simpl. split ; auto using opt_state_eq_refl.
  - exists 0, 1. unfold nsteps. simpl. rewrite H₂.
    simpl. split ; auto using opt_state_eq_refl.
Qed.

Lemma _prediction_completeness:
  ∀ genv os,
    True →
    _step genv os = None →
    step L₁ (fst genv) (fst os) = None
    ∧ step L₂ (snd genv) (snd os) = None.
Proof.
  intros [genv₁ genv₂] [S₁ S₂] _ H.
  simpl.
  unfold _step in H. simpl in H.
  destruct (step L₁ genv₁ S₁), (step L₂ genv₂ S₂) ; auto ; discriminate H.
Qed.

Definition universal_oracle : oracle_language L₁ L₂ :=
  {|
    oracle_state := _state;
    oracle_genv := _genv;
    oracle_step := _step;
    left_state := fst;
    right_state := snd;
    left_genv := fst;
    right_genv := snd;
    invariant := λ _ _, True;
    invariant_1 := λ genv os os' _ _, I;
    prediction_soundness := _prediction_soundness;
    prediction_completeness := _prediction_completeness
  |}.
