(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale
  Copyright © 2017 Yann Régis-Gianas

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
 *)

(** Basic definitions about deterministic programming languages.

    We specify a record signature whose inhabitants define a
    deterministic programming language.

    This signature requires a type [state] for the dynamically
    changing part of the configuration while [genv_type] is the static
    part of the configuration (typically containing the source code
    and runtime constants and primitives).

    This signature also requires that the semantics of the programming
    language is described by a small-step reduction function [step].
    This reduction function is a partial function that may rewrite a
    dynamic state into a new dynamic state.

    For practical reasons, the dynamic state must be equipped with a
    decidable equality.
 *)

Require Import Common.

Record language : Type :=
  {
    (* Semantics *)
    return_type : Type;
    state : Type;
    genv_type : Type;

    (* Transition relation *)
    step : genv_type → state → option state;
    (* Whether a state is final, and if so, its return value *)
    state_final : state → option return_type;
    (* There is no transition step from a final state *)
    state_final_no_step : ∀ genv s v, state_final s = Some v → step genv s = None;

    (* Syntax *)
    prog : Type;
    initial_state : prog → genv_type * state;

    (* Decidable equality *)
    genv_eq : genv_type → genv_type → Prop;
    (* TODO: there ought to be a Coq-ish way to define that *)
    genv_eq_refl : ∀ genv, genv_eq genv genv;
    genv_eq_sym : ∀ genv₁ genv₂, genv_eq genv₁ genv₂ → genv_eq genv₂ genv₁;
    genv_eq_trans : ∀ genv₁ genv₂ genv₃,
        genv_eq genv₁ genv₂ →
        genv_eq genv₂ genv₃ →
        genv_eq genv₁ genv₃;
    genv_eq_dec : ∀ genv genv',
        {genv_eq genv genv'} + {¬ genv_eq genv genv'};

    state_eq : state → state → Prop;
    state_eq_refl : ∀ S, state_eq S S;
    state_eq_sym : ∀ S S', state_eq S S' → state_eq S' S;
    state_eq_trans : ∀ S₁ S₂ S₃,
        state_eq S₁ S₂ → state_eq S₂ S₃ → state_eq S₁ S₃;
    state_eq_dec : ∀ S S',
        {state_eq S S'} + {¬ state_eq S S'};

    state_final_eq : ∀ S S',
        state_eq S S' →
        state_final S = state_final S';
    step_eq : ∀ genv₁ genv₂ S₁ S₂ S₁' S₂',
        genv_eq genv₁ genv₂ →
        state_eq S₁ S₂ →
        step genv₁ S₁ = Some S₁' →
        step genv₂ S₂ = Some S₂' →
        state_eq S₁' S₂';
    step_eq' : ∀ genv₁ genv₂ S₁ S₂,
        genv_eq genv₁ genv₂ →
        state_eq S₁ S₂ →
        step genv₁ S₁ = None ↔ step genv₂ S₂ = None
  }.

(** We call a [configuration] the pair of the static global
    environment and the dynamic state of a program.
*)

Definition configuration (L : language) : Type := genv_type L * state L.

Definition initial_configuration {L : language} (p : prog L) :=
  initial_state L p.

Definition configuration_eq {L : language} (c₁ c₂ : configuration L) :=
  genv_eq L (fst c₁) (fst c₂) ∧ state_eq L (snd c₁) (snd c₂).

Definition opt_state_eq {L : language} (oS₁ oS₂ : option (state L)) : Prop :=
  (oS₁ = None ∧ oS₂ = None)
  ∨ ∃ S₁ S₂, oS₁ = Some S₁ ∧ oS₂ = Some S₂ ∧ state_eq L S₁ S₂.

Lemma opt_state_eq_dec:
  ∀ {L : language} (oS₁ oS₂ : option (state L)),
    { opt_state_eq oS₁ oS₂ } + { ¬ opt_state_eq oS₁ oS₂ }.
Proof.
  intros L [ S₁ | ] [S₂ | ].
  - destruct (state_eq_dec L S₁ S₂) as [H | H].
    + left. right. eauto.
    + right. intros [H' | H'].
      * destruct H' as [H' _]. discriminate H'.
      * destruct H' as [S₁_ [S₂_ [HS₁_ [HS₂_ H']]]].
        inversion HS₁_ as [H'S₁_].
        inversion HS₂_ as [H'S₂_].
        clear HS₁_ HS₂_.
        rewrite <- H'S₁_ in H'.
        rewrite <- H'S₂_ in H'.
        apply H. apply H'.
  - right. intros [[H H'] | [S₁_ [S₂ [H [H' H'']]]]].
    + discriminate H.
    + discriminate H'.
  - right. intros [[H H'] | [S₁ [S₂_ [H [H' H'']]]]].
    + discriminate H'.
    + discriminate H.
  - left. left. auto.
Qed.

Lemma state_eq__opt_state_eq:
  ∀ {L} (S₁ S₂ : state L),
    state_eq L S₁ S₂ <-> opt_state_eq (Some S₁) (Some S₂).
Proof.
  intros L S₁ S₂.
  split ; intro H.
  - right. exists S₁, S₂. repeat split ; auto.
  - destruct H as [[HFalse _] | H ] ; try discriminate HFalse.
    destruct H as [S₁_ [S₂_ [HS₁_ [HS₂_ H]]]].
    inversion HS₁_. inversion HS₂_. clear HS₁_ HS₂_. subst S₁_ S₂_.
    auto.
Qed.

Lemma opt_state_eq_Some_r__state_eq:
  ∀ {L} (S : state L) (oS : option (state L)),
    opt_state_eq oS (Some S) →
    ∃ S', oS = Some S' ∧ state_eq L S' S.
Proof.
  intros L S oS H.
  destruct H as [[_ HFalse] | H] ; try discriminate HFalse.
  destruct H as [S₁ [S₂ [H₁ [H₂ H]]]].
  inversion H₂. clear H₂. subst S₂.
  exists S₁. subst oS.
  split ; auto.
Qed.

Lemma opt_state_eq_refl:
  ∀ {L} (oS : option (state L)),
    opt_state_eq oS oS.
Proof.
  intros L [S | ].
  - right. exists S, S. auto using state_eq_refl.
  - left. auto.
Qed.

Lemma opt_state_eq_sym:
  ∀ {L} (oS₁ oS₂ : option (state L)),
    opt_state_eq oS₁ oS₂ → opt_state_eq oS₂ oS₁.
Proof.
  intros L oS₁ oS₂ H.
  destruct H as [[HS₁ HS₂] | [S₁ [S₂ [HS₁ [HS₂ H]]]]].
  - left. auto.
  - right. exists S₂, S₁.
    repeat split ; auto using state_eq_sym.
Qed.

Lemma opt_state_eq_trans:
  ∀ {L} (oS₁ oS₂ oS₃ : option (state L)),
    opt_state_eq oS₁ oS₂ → opt_state_eq oS₂ oS₃ → opt_state_eq oS₁ oS₃.
Proof.
  intros L oS₁ oS₂ oS₃ H₁ H₂.
  destruct H₁ as [[HS₁ HS₂] | [S₁ [S₂ [HS₁ [HS₂ H]]]]].
  - left.
    destruct H₂ as [[H'S₂ HS₃] | [S₂ [S₃ [H'S₂ [HS₃ H]]]]].
    + split ; auto.
    + rewrite H'S₂ in HS₂. discriminate HS₂.
  - right.
    destruct H₂ as [[H'S₂ HS₃] | [S'₂ [S₃ [H'S₂ [HS₃ H']]]]].
    + rewrite H'S₂ in HS₂. discriminate HS₂.
    + exists S₁, S₃. repeat split ; auto.
      rewrite H'S₂ in HS₂. inversion HS₂ as [H''S₂].
      clear HS₂. subst S'₂.
      eauto using state_eq_trans.
Qed.

(** We use [nsteps] to repeatedly call the reduction function [step].
    As [step] is a partial function, so is [nsteps].

    We also define a total function [nsteps'] that always return a
    state together with the number of reduction steps it could not
    perform.

    Hence, if [nsteps' n state = (state', 0)], the reduction
    function has been applied [n] times successfully ([nsteps_nsteps'_0_iff]).
    On the other hand, if [nsteps' n state = (state', S n')], [state'] is
    guaranteed to be a stuck state and was obtained by rewriting [state] [n - S n']
    times.
 *)

Definition nsteps {lang : language} (genv : genv_type lang)
           (n : nat) (S : state lang) : option (state lang) :=
  napply (step lang genv) n S.

Definition nsteps' {lang : language} (genv : genv_type lang)
           (n : nat) (S : state lang) : state lang * nat :=
  napply' (step lang genv) n S.

Lemma nsteps_eq:
  ∀ {L : language} (genv₁ genv₂ : genv_type L) (S₁ S₂ : state L) (n : nat),
    genv_eq L genv₁ genv₂ →
    state_eq L S₁ S₂ →
    opt_state_eq (nsteps genv₁ n S₁) (nsteps genv₂ n S₂).
Proof.
  intros L genv₁ genv₂ S₁ S₂ n. revert S₁ S₂.
  unfold nsteps.
  induction n as [ | n IH ].
  - intros S₁ S₂ Heq H.
    right. simpl. eauto.
  - intros S₁ S₂ Heq H. simpl.
    pose proof (step_eq' L genv₁ genv₂ S₁ S₂ Heq H) as HNone.
    destruct (step L genv₁ S₁) as [S₁' | ] eqn:HS₁'.
    + destruct (step L genv₂ S₂) as [S₂' | ] eqn:HS₂'.
      * clear HNone.
        simpl.
        apply IH ;
          eauto using step_eq.
      * exfalso. destruct HNone as [_ HNone].
        discriminate HNone. reflexivity.
    + destruct HNone as [HNone _].
      rewrite (HNone eq_refl).
      left. simpl. auto.
Qed.

Lemma nsteps_plus_0:
  ∀ {L : language} (genv : genv_type L) {n₁ n₂ : nat} (S₁ : state L),
    nsteps genv (n₁ + n₂) S₁ = nsteps genv n₁ S₁ >>= λ S₂, nsteps genv n₂ S₂.
Proof.
  intros L genv n₁.
  unfold nsteps.
  induction n₁ as [ | n₁ IH ] ;
    intros n₂ S₁.
  - simpl in *. reflexivity.
  - simpl.
    destruct (step L genv S₁) as [S₁' | ] ;
      auto.
    apply IH.
Qed.

Lemma nsteps_plus_1:
  ∀ {L : language} (genv : genv_type L) {n₁ n₂ : nat} (S₁ S₂ : state L),
    nsteps genv n₁ S₁ = Some S₂ → nsteps genv (n₁ + n₂) S₁ = nsteps genv n₂ S₂.
Proof.
  intros L genv n₁ n₂ S₁ S₂ H.
  rewrite nsteps_plus_0. rewrite H. reflexivity.
Qed.

Lemma nsteps_plus_1':
  ∀ {L : language} (genv : genv_type L) {n₁ n₂ : nat} (S₁ S₂ : state L),
    opt_state_eq (nsteps genv n₁ S₁) (Some S₂) →
    opt_state_eq (nsteps genv (n₁ + n₂) S₁) (nsteps genv n₂ S₂).
Proof.
  intros L genv n₁.
  intros n₂ S₁ S₂ H.
  rewrite nsteps_plus_0.
  destruct (opt_state_eq_Some_r__state_eq _ _ H) as [S₂_ [HS₂_ HS₂_S₂]].
  rewrite HS₂_. simpl. apply nsteps_eq ; auto using genv_eq_refl.
Qed.

Lemma nsteps_plus_2: (* TODO: better name *)
  ∀ {L} {genv : genv_type L} n n' S S',
    nsteps genv (n + n') S = Some S' →
    ∃ S₁,
      nsteps genv n S = Some S₁.
Proof.
  intros L genv n.
  unfold nsteps.
  induction n as [ | n IH ] ;
    intros n' S S' H.
  - simpl. exists S. auto.
  - simpl in H. simpl.
    destruct (step L genv S) as [S₁ | ] ;
      try discriminate H.
    simpl in H.
    simpl.
    eauto.
Qed.

(* A program may terminate because of a crash, or it may end in a correct final state *)
Definition terminates {L} genv S :=
  ∃ n S', nsteps genv n S = Some S' ∧ step L genv S' = None.

Require Import Omega.

Lemma nsteps_split_helper: (* TODO: better name *)
  ∀ {L} genv n n' S S' oS',
    nsteps genv n S = Some S' →
    step L genv S' = None →
    opt_state_eq (nsteps genv n' S) oS' →
    (n' = n)
    ∨ (∃ S', n > n' ∧ nsteps genv n' S = Some S')
    ∨ (n' > n ∧ nsteps genv n' S = None).
Proof.
  intros L genv n.
  unfold nsteps.
  induction n as [ | n IH ] ;
    intros n' S S' oS' H HS' H'.
  - destruct n' as [ | n' ].
    + left. auto.
    + right. right. split ; auto with arith.
      simpl in H. inversion H. clear H. subst S.
      simpl. rewrite HS'. reflexivity.
  - destruct (n' ?= Datatypes.S n) eqn:Hn'.
    + apply nat_compare_eq in Hn'. left. auto.
    + apply nat_compare_lt in Hn'. right. left.
      assert (∃ n'', Datatypes.S n = n' + n'') as H''.
      { clear IH S S' oS' H HS' H'.
        remember (S n) as m eqn:H. clear H. clear n.
        induction Hn' as [ | m' Hm' IH ].
        + exists 1. omega.
        + destruct IH as [n'' H]. exists (S n''). omega. }
      destruct H'' as [n'' Hn''].
      rewrite Hn'' in H.
      destruct (nsteps_plus_2 n' n'' _ _ H) as [S₁ H''].
      exists S₁. split ; auto.
    + apply nat_compare_gt in Hn'.
      right. right.
      assert (∃ n'', n' = Datatypes.S (Datatypes.S n) + n'') as H''.
      { clear IH S S' oS' H HS' H'.
        remember (S n) as m eqn:H. clear H. clear n.
        induction Hn' as [ | m' Hm' IH ] ; eauto.
        destruct IH as [n'' H]. exists (S n''). omega. }
      destruct H'' as [n'' H''].
      rewrite H''. split ; auto with arith.
      clear H'' H' Hn' oS' IH n'.
      remember (Datatypes.S n) as n'. clear Heqn' n.
      revert S H.
      induction n' as [ | n' IH ] ; intros S H.
      * simpl in H. inversion H. clear H. subst S.
        simpl.
        rewrite HS'.
        reflexivity.
      * simpl in H. simpl.
        destruct (step L genv S) as [S'' | ] eqn:HS'' ;
          try discriminate H.
        simpl in H.
        simpl in IH. simpl. apply IH.
        auto.
Qed.
