(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale
  Copyright © 2017 Yann Régis-Gianas

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
*)

(** Basic definitions about reduction traces.

    We introduce reduction traces for any deterministic programming
    language coinductively. Reachability and convergence are defined
    on top of this notion.

*)

Require Import ProgrammingLanguage.
Require Import Common.

(** While we will often abusively refer to potentially infinite non-empty
    lists of configurations as traces, actual reduction traces are
    characterized by the following [trace] predicate:
 *)

CoInductive trace (L : language) :
  configuration L → coplist (configuration L) → Prop :=
(* This does not distinguish between legitimate final states
   and crashed states, as those states can be further inspected. *)
| TFinal : ∀ genv state,
    step L genv state = None →
    trace L (genv, state) (CSingle _ (genv, state))
| TStep : ∀ genv state S' tr,
    step L genv state = Some S' →
    trace L (genv, S') tr →
    trace L (genv, state) (CCons _ (genv, state) tr).

(** We also provide a variant [trace'] of the [trace] predicate that
    hides the precise configuration the trace starts from:
*)

Definition trace' (L : language) (l : coplist (configuration L)) : Prop :=
  trace L (first l) l.

Lemma trace_first {L : language}:
  ∀ (c : configuration L) l, trace L c l → first l = c.
Proof.
  intros c l H.
  inversion H ; auto.
Qed.

Lemma trace_trace':
  ∀ L l, trace' L l <-> (∃ c, trace L c l).
Proof.
  intros L l. split.
  - intro H. unfold trace' in H. exists (first l). auto.
  - intros [c H].
    inversion H as [ genv state HNone | genv state oS' tr HoS' Htr ].
    + subst c l. unfold trace'. simpl. apply TFinal, HNone.
    + subst c l. unfold trace'. simpl. apply (TStep L genv state oS') ; auto.
Qed.

(** Converging and crashing traces are characterized by inductive predicates:
*)

Inductive nsteps_converging_trace {L : language}
  : nat → coplist (configuration L) → Prop :=
| FinallyConverges:
  ∀ c result,
  state_final L (snd c) = Some result →
  nsteps_converging_trace 0 (CSingle _ c)
| CConsConverges : ∀ c tr n,
  nsteps_converging_trace n tr →
  nsteps_converging_trace (S n) (CCons _ c tr).

Inductive nsteps_crashing_trace {L : language}
  : nat → coplist (configuration L) → Prop :=
| FinallyCrashes:
  ∀ c,
  step L (fst c) (snd c) = None →
  state_final L (snd c) = None →
  nsteps_crashing_trace 0 (CSingle _ c)
| CConsCrashes : ∀ c tr n,
  nsteps_crashing_trace n tr →
  nsteps_crashing_trace (S n) (CCons _ c tr).

CoFixpoint traceFrom {L : language} (c : configuration L) : coplist (configuration L) :=
  match step L (fst c) (snd c) with
  | None =>
    CSingle _ c
  | Some state' =>
    CCons _ c (traceFrom (fst c, state'))
  end.

Lemma traceFrom_trace:
  ∀ L (c : configuration L),
    trace L c (traceFrom c).
Proof.
  intro L.
  cofix H.
  intros [genv S].
  rewrite <- coplist_id_force_eq. simpl.
  destruct (step L genv S) as [S' | ] eqn:HS'.
  - apply (TStep L genv S S' _ HS'). apply H.
  - apply TFinal. apply HS'.
Qed.

Definition program_trace (L : language) (p : prog L) :=
  traceFrom (initial_configuration p).

Lemma nsteps_skipn:
  ∀ L tr n genv S S',
    trace L (genv, S) tr →
    opt_state_eq (nsteps genv n S) (Some S') →
    ∃ tr' S'',
      skipn n tr = Some tr'
      ∧ trace L (genv, S'') tr'
      ∧ state_eq L S'' S'.
Proof.
  intros L tr n genv.
  unfold nsteps.
  revert tr.
  induction n as [ | n IH ] ;
    intros tr S S' Htr H.
  - simpl in *.
    exists tr, S. repeat split ; auto.
    apply state_eq__opt_state_eq. auto.
  - inversion Htr as [genv_ S_ HS' | genv_ S_ S'' tr' HS'' Htr' ] ;
      subst genv_ S_ tr.
    + simpl in H. rewrite HS' in H. simpl in H.
      destruct H as [[_ HFalse] | [A [B [HFalse _]]]] ;
        discriminate HFalse.
    + simpl in H. rewrite HS'' in H. simpl in H. simpl.
      eauto.
Qed.

CoInductive included_subtrace {L : language}
  : coplist (configuration L) → coplist (configuration L) → Prop :=
| SameFirst:
    ∀ tr₁ tr₂,
    configuration_eq (first tr₁) (first tr₂) →
    (∀ tr₁',
        next tr₁ = Some tr₁' →
        ∃ tr₂', next tr₂ = Some tr₂' ∧ included_subtrace tr₁' tr₂') →
    included_subtrace tr₁ tr₂
| Stutter:
    ∀ tr₁ tr₁' tr₂,
      configuration_eq (first tr₁) (first tr₂) →
      next tr₁ = Some tr₁' →
      included_subtrace tr₁' tr₂ →
      included_subtrace tr₁ tr₂
| IgnoreFirstSN:
    ∀ n tr₁ tr₂ tr₂',
      skipn (S n) tr₂ = Some tr₂' →
      included_subtrace tr₁ tr₂' →
      included_subtrace tr₁ tr₂.